import numpy as np
import pandas as pd
from _typeshed import Incomplete
from enum import Enum, IntEnum as IntEnum
from iode import NA as NA
from iode.common import IodeFileType as IodeFileType
from iode.iode_cython import BinaryOperation as BinaryOperation, ExportFormats as ExportFormats, HighToLowType as HighToLowType, ImportFormats as ImportFormats, LowToHighMethod as LowToHighMethod, LowToHighType as LowToHighType, SimulationInitialization as SimulationInitialization, VarsMode as VarsMode
from iode.iode_database.abstract_database import IodeDatabase as IodeDatabase
from iode.time.period import Period as Period
from iode.time.sample import Sample as Sample
from iode.util import JUSTIFY as JUSTIFY, check_filepath as check_filepath, split_list as split_list, table2str as table2str
from pathlib import Path
from typing import Any, Self

Self = Any
StrEnum = Enum
Array: Incomplete
Array = Any

class VarPositionalIndexer:
    database: Incomplete
    def __init__(self, database) -> None: ...
    def __getitem__(self, index: int | tuple[int, int]): ...
    def __setitem__(self, index: int | tuple[int, int], value): ...

class Variables(IodeDatabase):
    '''
    IODE Variables database. 

    Attributes
    ----------
    filename: str
    description: str
    mode: int
    sample: Sample
    nb_periods: int
    periods: list(str)
    periods_as_float: list(float)

    Parameters
    ----------
    filepath: str, optional
        file containing the IODE variables to load.

    Returns
    -------
    Variables

    Examples
    --------
    >>> from iode import variables, SAMPLE_DATA_DIR
    >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")      # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
    Loading .../fun.var
    394 objects loaded 
    >>> len(variables)
    394
    >>> variables           # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
    Workspace: Variables
    nb variables: 394
    filename: ...\\tests\\data\\fun.var
    description: Modèle fun - Simulation 1
    sample: 1960Y1:2015Y1
    mode: LEVEL
    <BLANKLINE>
     name       1960Y1  1961Y1  1962Y1  1963Y1  1964Y1  1965Y1  ...     2009Y1  2010Y1  2011Y1  2012Y1  2013Y1  2014Y1  2015Y1
    ACAF            na      na      na      na      na      na  ...     -37.46  -37.83  -44.54  -55.56  -68.89  -83.34  -96.41
    ACAG            na      na      na      na      na      na  ...      27.23   28.25   29.28   30.32   31.37   32.42   33.47
    AOUC            na    0.25    0.25    0.26    0.28    0.29  ...       1.29    1.31    1.33    1.36    1.39    1.42    1.46
    AOUC_           na      na      na      na      na      na  ...       1.23    1.25    1.27    1.30    1.34    1.37    1.41
    AQC           0.22    0.22    0.22    0.23    0.24    0.25  ...       1.45    1.46    1.48    1.51    1.56    1.61    1.67
    ...            ...     ...     ...     ...     ...     ...  ...        ...     ...     ...     ...     ...     ...     ...
    ZJ              na      na      na      na      na      na  ...       1.49    1.51    1.53    1.56    1.59    1.63    1.67
    ZKF           0.80    0.81    0.82    0.81    0.83    0.82  ...       0.87    0.87    0.87    0.87    0.87    0.87    0.87
    ZKFO          1.00    1.00    1.00    1.00    1.00    1.00  ...       1.02    1.02    1.02    1.02    1.02    1.02    1.02
    ZX            0.00    0.00    0.00    0.00    0.00    0.00  ...       0.00    0.00    0.00    0.00    0.00    0.00    0.00
    ZZF_          0.69    0.69    0.69    0.69    0.69    0.69  ...       0.69    0.69    0.69    0.69    0.69    0.69    0.69
    <BLANKLINE>
    '''
    def __init__(self, filepath: str = None) -> None: ...
    @classmethod
    def get_instance(cls) -> Self: ...
    def copy(self, pattern: str = None) -> Self:
        '''
        Create a new Variables database in which each variable is a *copy* of the original variable 
        from the global Variables workspace. Any change made to the *copied database* (*subset*) will 
        not be applied to the global workspace. This can be useful for example if you want to 
        save previous values of variables before making a simulation.

        Parameters
        ----------
        pattern : str, optional
            If provided, the copied database will only contain the variables whose name matches the 
            provided pattern. By default (None), the copied database will contain all the variables 
            from the global Variables workspace. The pattern syntax is the same as the one used for the 
            `__getitem__` method. If the pattern is an empty string, the copied database will be 
            empty, creating a new *detached* database.
            Default to None.

        Returns
        -------
        Variables

        See Also
        --------
        Variables.new_detached
        
        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")      # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded 

        Variables subset

        >>> # without using copy(), any modification made on 
        >>> # the subset will also change the corresponding 
        >>> # global Variables workspace
        >>> var_subset = variables["A*"]
        >>> var_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> # a) add a variable
        >>> var_subset["A0"] = 0.0
        >>> "A0" in var_subset
        True
        >>> "A0" in variables
        True
        >>> variables["A0"]             # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name    1960Y1  1961Y1  1962Y1  ...  2013Y1  2014Y1  2015Y1
        A0        0.00    0.00    0.00  ...    0.00    0.00    0.00
        <BLANKLINE>

        >>> # b) modify a variable
        >>> var_subset["ACAF"] = 1.0
        >>> var_subset["ACAF"]          # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name    1960Y1  1961Y1  1962Y1  ...  2013Y1  2014Y1  2015Y1
        ACAF      1.00    1.00    1.00  ...    1.00    1.00    1.00
        <BLANKLINE>
        >>> variables["ACAF"]           # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name    1960Y1  1961Y1  1962Y1  ...  2013Y1  2014Y1  2015Y1
        ACAF      1.00    1.00    1.00  ...    1.00    1.00    1.00
        <BLANKLINE>
        >>> # c) delete a variable
        >>> del var_subset["ACAG"]
        >>> "ACAG" in var_subset
        False
        >>> "ACAG" in variables
        False

        Copied database subset

        >>> var_subset_copy = variables["B*"].copy()
        >>> var_subset_copy.names
        [\'BENEF\', \'BQY\', \'BRUGP\', \'BVY\']
        >>> # or equivalently
        >>> var_subset_copy = variables.copy("B*")
        >>> var_subset_copy.names
        [\'BENEF\', \'BQY\', \'BRUGP\', \'BVY\']
        >>> # by using copy(), any modification made on the copy subset 
        >>> # let the global workspace unchanged
        >>> # a) add a variable -> only added in the copied subset
        >>> var_subset_copy["B0"] = 0.0
        >>> "B0" in var_subset_copy
        True
        >>> "B0" in variables
        False
        >>> # b) modify a variable -> only modified in the copied subset
        >>> var_subset_copy["BENEF"] = 1.0
        >>> var_subset_copy["BENEF"]           # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name    1960Y1  1961Y1  1962Y1  ...  2013Y1  2014Y1  2015Y1
        BENEF     1.00    1.00    1.00  ...    1.00    1.00    1.00
        <BLANKLINE>
        >>> variables["BENEF"]           # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name    1960Y1  1961Y1  1962Y1  ...  2013Y1  2014Y1   2015Y1
        BENEF    11.66   13.61   12.21  ...   19.00  -32.20  -117.38
        <BLANKLINE>
        >>> # c) delete a variable -> only deleted in the copied subset
        >>> del var_subset_copy["BQY"]
        >>> "BQY" in var_subset_copy
        False
        >>> "BQY" in variables
        True

        New detached Variables database

        >>> # a new empty *detached* Variables database can be created by passing 
        >>> # an empty string to the copy() method 
        >>> var_detached = variables.copy("")
        >>> var_detached.names
        []
        >>> # or equivalently by using the new_detached() method 
        >>> var_detached = variables.new_detached()
        >>> var_detached.names
        []
        '''
    @property
    def i(self) -> VarPositionalIndexer:
        '''
        Allow to select the ith variable in the database.

        Examples
        --------
        >>> from iode import variables, SAMPLE_DATA_DIR
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> # get the value of the first period of the first variable
        >>> variables.i[0, 0]
        nan
        
        >>> # get the value of the last period of the last variable
        >>> variables.i[-1, -1]
        0.68840039
        >>> # update the value of the first period of the first variable
        >>> variables.i[0, 0] = 3.0
        >>> variables.i[0, 0]
        3.0
        >>> # update the value of the last period of the last variable
        >>> variables.i[-1, -1] = 3.0
        >>> variables.i[-1, -1]
        3.0
        '''
    def __getitem__(self, key) -> float | Self:
        '''
        Return the (subset of) variable(s) referenced by `key`.

        The `key` represents a selection on the Variables names and optionally on the periods.

        The selection on names can be: 
        
            - a single Variable name (e.g. "ACAF") 
            - a list of Variable names (e.g. "ACAF;ACAG;AOUC") 
            - a pattern (e.g. "A*;*_").
        
        If the selection on names represents a list of names or of sub-patterns, each name or sub-pattern must be 
        separated by a `separator` character which is either a whitespace ` `, or a comma `,`, or a semi-colon `;`, 
        or a tabulation `\\t`, or a newline `\\n`.

        A (sub-)`pattern` is a list of characters representing a group of object names. 
        It includes some special characters which have a special meaning:
        
            - `*` : any character sequence, even empty
            - `?` : any character (one and only one)
            - `@` : any alphanumerical char [A-Za-z0-9]
            - `&` : any non alphanumerical char
            - `|` : any alphanumeric character or none at the beginning and end of a string 
            - `!` : any non-alphanumeric character or none at the beginning and end of a string 
            - `\\` : escape the next character

        Note that the selection on names can contain references to IODE lists which are prefixed with the symbol `$`.

        The selection on periods can be:

           - the whole sample (e.g. None) 
           - a single period (e.g. "1990Y1") 
           - a range of periods (e.g. "1990Y1:2000Y1") 
           - a list of periods (e.g. ["1990Y1", "1995Y1", "2000Y1"]) 

        Parameters
        ----------
        key: str or list(str) or tuple(str, str) or tuple(str, list(str)) or tuple(str, str:str)
            The key is split into two parts: the selection on names and the selection on periods.
            The selection on names can be a single name, a list of names, or a pattern.
            The selection on periods (optional) can be a single period, a list of periods, or a range of periods.

        Returns
        -------
        float or Variables
            If `key` represents a single name and a single period, the returned value is a float.
            If `key` represents a list of non-contiguous periods, the returned object is a pandas DataFrame.
            Otherwise, a subset of the Variables workspace is returned.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables, NA
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> variables.sample
        Sample("1960Y1:2015Y1")

        >>> # -------- a) get one Variable --------
        >>> # get the variable values for the whole sample
        >>> variables["ACAF"]                       # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  ...  2014Y1   2015Y1
        ACAF            na      na  ...  -83.34   -96.41
        <BLANKLINE>
        >>> # get the variable value for a specific period
        >>> variables["ACAF", "1990Y1"]
        23.771
        >>> # get the variable values for range of periods (using a Python slice)
        >>> variables["ACAF", "1990Y1":"2000Y1"]    # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1990Y1:2000Y1
        mode: LEVEL
        <BLANKLINE>
        name        1990Y1  1991Y1  1992Y1  1993Y1  1994Y1  1995Y1  1996Y1  1997Y1  1998Y1  1999Y1  2000Y1
        ACAF         23.77   26.24   30.16   34.66    8.16  -13.13   32.17   39.94   29.65   13.53   10.05
        <BLANKLINE>
        >>> # same as above but with the colon \':\' inside the periods range string
        >>> variables["ACAF", "1990Y1:2000Y1"]      # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1990Y1:2000Y1
        mode: LEVEL
        <BLANKLINE>
        name        1990Y1  1991Y1  1992Y1  1993Y1  1994Y1  1995Y1  1996Y1  1997Y1  1998Y1  1999Y1  2000Y1
        ACAF         23.77   26.24   30.16   34.66    8.16  -13.13   32.17   39.94   29.65   13.53   10.05
        <BLANKLINE>

        >>> # b) -------- get the values for a single period --------
        >>> variables[:, "1990Y1"]                  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 394
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1990Y1:1990Y1
        mode: LEVEL
        <BLANKLINE>
         name       1990Y1
        ACAF         23.77
        ACAG        -28.17
        AOUC          1.00
        ...            ...
        ZKFO          1.00
        ZX            0.00
        ZZF_          0.69
        <BLANKLINE> 
        >>> # or equivalently:
        >>> variables["*", "1990Y1"]                # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 394
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1990Y1:1990Y1
        mode: LEVEL
        <BLANKLINE>
         name       1990Y1
        ACAF         23.77
        ACAG        -28.17
        AOUC          1.00
        AOUC_         0.94
        AQC           1.00
        ...            ...
        ZJ            1.09
        ZKF           0.81
        ZKFO          1.00
        ZX            0.00
        ZZF_          0.69
        <BLANKLINE>

        >>> # c) -------- get a subset of the Variables database using a pattern --------
        >>> vars_subset = variables["A*"]
        >>> vars_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> # get the variable values for a specific period -> Worspace subset
        >>> variables["A*", "1990Y1"]       # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1990Y1:1990Y1
        mode: LEVEL
        <BLANKLINE>
        name       1990Y1
        ACAF         23.77
        ACAG        -28.17
        AOUC          1.00
        AOUC_         0.94
        AQC           1.00
        <BLANKLINE>
        >>> # get the variable values for range of periods -> Workspace subset
        >>> variables["A*", "1990Y1:2000Y1"]      # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE 
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1990Y1:2000Y1
        mode: LEVEL
        <BLANKLINE>
         name       1990Y1  1991Y1  1992Y1  1993Y1  1994Y1  1995Y1  1996Y1  1997Y1  1998Y1  1999Y1  2000Y1
        ACAF         23.77   26.24   30.16   34.66    8.16  -13.13   32.17   39.94   29.65   13.53   10.05
        ACAG        -28.17  -30.93  -40.29  -43.16  -16.03  -41.85  -40.24  -32.93  -38.35  -39.86  -41.53
        AOUC          1.00    1.02    1.03    1.03    1.05    1.05    1.05    1.08    1.09    1.11    1.12
        AOUC_         0.94    0.96    0.97    0.98    0.99    1.00    1.00    1.03    1.04    1.08    1.10
        AQC           1.00    1.06    1.11    1.15    1.16    1.16    1.16    1.20    1.20    1.34    1.34
        <BLANKLINE>

        >>> # d) -------- get a subset of the Variables database using a list of names --------
        >>> vars_subset = variables[["ACAF", "ACAG", "AQC", "BQY", "BVY"]]
        >>> vars_subset.names
        [\'ACAF\', \'ACAG\', \'AQC\', \'BQY\', \'BVY\']
        >>> # get the variable values for a specific period -> pandas Series
        >>> variables[["ACAF", "ACAG", "AQC", "BQY", "BVY"], "1990Y1"]              # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1990Y1:1990Y1
        mode: LEVEL
        <BLANKLINE>
        name        1990Y1
        ACAF         23.77
        ACAG        -28.17
        AQC           1.00
        BQY         -34.10
        BVY         -34.10
        <BLANKLINE>
        >>> # get the variable values for a list of non-contiguous periods -> pandas DataFrame
        >>> variables[["ACAF", "ACAG", "AQC", "BQY", "BVY"], ["1990Y1", "1995Y1", "2000Y1"]]  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        time          1990Y1      1995Y1      2000Y1
        variables
        ACAF       23.771000  -13.130997   10.046611
        ACAG      -28.172186  -41.845993  -41.534787
        AQC         1.000000    1.161687    1.338603
        BQY       -34.099998   91.089355  119.955090
        BVY       -34.099997  105.500000  150.053352       
        >>> # get the variable values for range of periods -> Workspace subset
        >>> variables[["ACAF", "ACAG", "AQC", "BQY", "BVY"], "1990Y1:2000Y1"]      # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE 
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1990Y1:2000Y1
        mode: LEVEL
        <BLANKLINE>
        name        1990Y1  1991Y1  1992Y1  1993Y1  1994Y1  1995Y1  1996Y1  1997Y1  1998Y1  1999Y1  2000Y1
        ACAF         23.77   26.24   30.16   34.66    8.16  -13.13   32.17   39.94   29.65   13.53   10.05
        ACAG        -28.17  -30.93  -40.29  -43.16  -16.03  -41.85  -40.24  -32.93  -38.35  -39.86  -41.53
        AQC           1.00    1.06    1.11    1.15    1.16    1.16    1.16    1.20    1.20    1.34    1.34
        BQY         -34.10   -1.26  -13.75   52.16   66.63   91.09  104.68  113.52  116.19  117.91  119.96
        BVY         -34.10   -1.30  -14.70   58.10   75.90  105.50  123.20  135.62  140.74  144.86  150.05
        <BLANKLINE>
        '''
    def __setitem__(self, key, value) -> None:
        '''
        Update/add a (subset of) variable(s) referenced by `key` from/to the Variables database.

        The `key` represents a selection on the Variables names and optionally on the periods.

        The selection on names can be: 
        
            - a single Variable name (e.g. "ACAF") 
            - a list of Variable names (e.g. "ACAF;ACAG;AOUC") 
            - a pattern (e.g. "A*;*_").
        
        If the selection on names represents a list of names or of sub-patterns, each name or sub-pattern must be 
        separated by a `separator` character which is either a whitespace ` `, or a comma `,`, or a semi-colon `;`, 
        or a tabulation `\\t`, or a newline `\\n`.

        A (sub-)`pattern` is a list of characters representing a group of object names. 
        It includes some special characters which have a special meaning:
        
            - `*` : any character sequence, even empty
            - `?` : any character (one and only one)
            - `@` : any alphanumerical char [A-Za-z0-9]
            - `&` : any non alphanumerical char
            - `|` : any alphanumeric character or none at the beginning and end of a string 
            - `!` : any non-alphanumeric character or none at the beginning and end of a string 
            - `\\` : escape the next character

        Note that the selection on names can contain references to IODE lists which are prefixed with the symbol `$`.

        The selection on periods can be:

           - the whole sample (e.g. None) 
           - a single period (e.g. "1990Y1") 
           - a range of periods (e.g. "1990Y1:2000Y1") 
           - a list of periods (e.g. ["1990Y1", "1995Y1", "2000Y1"]) 

        Parameters
        ----------
        key: str or list(str) or tuple(str, str) or tuple(str, list(str)) or tuple(str, str:str)
            The key is split into two parts: the selection on names and the selection on periods.
            The selection on names can be a single name, a list of names, or a pattern.
            The selection on periods (optional) can be a single period, a list of periods, or a range of periods.

        value: str or int or float or dict(str, ...) or numpy array or pandas Series or pandas DataFrame or Variables
            If str, the value is interpreted as a LEC expression and is evaluated for each period.
            If int, the value is first converted to a float and then used for all periods.
            If float, the value is used for all periods.
            If dict, the keys represents the names of the variables to be modified. 
            If numpy array or pandas Series/DataFrame, there must be a value for each variable and period to be set.
            If Variables, names and periods must match.

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> import larray as la
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables, NA, Sample
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        
        >>> # a) -------- add one variable --------
        >>> # 1) same value for all periods
        >>> variables["A0"] = np.nan
        >>> variables["A0"]                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE 
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  ...  2014Y1   2015Y1
        A0              na      na  ...      na       na
        <BLANKLINE>
        >>> # or equivalently
        >>> variables["A0"] = NA
        >>> variables["A0"]                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  ...  2014Y1   2015Y1
        A0              na      na  ...      na       na
        <BLANKLINE>
        >>> # 2) LEC expression
        >>> variables["A1"] = "t + 10"
        >>> variables["A1"]                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  ...  2014Y1   2015Y1
        A1           10.00   11.00  ...   64.00    65.00
        <BLANKLINE>
        >>> # 3) list of values for each period
        >>> values = list(range(variables.nb_periods))
        >>> values[0] = NA
        >>> values[-1] = np.nan
        >>> variables["A2"] = values
        >>> variables["A2"]                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  1962Y1  ...  2013Y1  2014Y1  2015Y1
        A2              na    1.00    2.00  ...   53.00   54.00      na
        <BLANKLINE>
        >>> # 4) numpy ndarray
        >>> values = np.asarray(values)
        >>> variables["A3"] = values
        >>> variables["A3"]                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  1962Y1  ...  2013Y1  2014Y1  2015Y1
        A3              na    1.00    2.00  ...   53.00   54.00      na
        <BLANKLINE>
        >>> # 5) pandas Series
        >>> values = pd.Series(values, index=variables.periods_as_str)
        >>> variables["A4"] = values
        >>> variables["A4"]                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  1962Y1  ...  2013Y1  2014Y1  2015Y1
        A4              na    1.00    2.00  ...   53.00   54.00      na
        <BLANKLINE>
        >>> # 6) Variables object
        >>> variables["A5"] = variables["ACAF"]
        >>> variables["A5"]                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  ...  2013Y1  2014Y1  2015Y1
        A5              na      na  ...  -68.89  -83.34  -96.41
        <BLANKLINE> 

        >>> # b) -------- update one variable --------
        >>> # 1) set one value of a variable for a specific period
        >>> variables["ACAG", "1990Y1"]
        -28.1721855713507
        >>> variables["ACAG", "1990Y1"] = -28.2
        >>> variables["ACAG", "1990Y1"]
        -28.2

        >>> # 2) update all values of a variable
        >>> variables["ACAF"]                   # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  ...  2014Y1   2015Y1
        ACAF            na      na  ...  -83.34   -96.41
        <BLANKLINE>
        >>> # 2.1) same value for all periods
        >>> variables["ACAF"] = np.nan
        >>> variables["ACAF"]                   # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  ...  2014Y1   2015Y1
        ACAF            na      na  ...      na       na
        <BLANKLINE>
        >>> # or equivalently
        >>> variables["ACAF"] = NA
        >>> variables["ACAF"]                   # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  ...  2014Y1   2015Y1
        ACAF            na      na  ...      na       na
        <BLANKLINE>
        >>> # 2.2) LEC expression
        >>> variables["ACAF"] = "t + 10"
        >>> variables["ACAF"]                   # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  ...  2014Y1   2015Y1
        ACAF         10.00   11.00  ...   64.00    65.00
        <BLANKLINE>
        >>> # 2.3) list of values for each period
        >>> values = list(range(variables.nb_periods))
        >>> values[0] = NA
        >>> values[-1] = np.nan
        >>> variables["ACAF"] = values
        >>> variables["ACAF"]                   # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  1962Y1  ...  2013Y1  2014Y1  2015Y1
        ACAF            na    1.00    2.00  ...   53.00   54.00      na
        <BLANKLINE>
        >>> # 2.4) numpy array
        >>> values = np.asarray(values)
        >>> variables["ACAG"] = values
        >>> variables["ACAG"]                   # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  1962Y1  ...  2013Y1  2014Y1  2015Y1
        ACAG            na    1.00    2.00  ...   53.00   54.00      na
        <BLANKLINE>
        >>> # 2.5) pandas Series
        >>> values = pd.Series(values, index=variables.periods_as_str)
        >>> variables["AOUC"] = values
        >>> variables["AOUC"]                   # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  1962Y1  ...  2013Y1  2014Y1  2015Y1
        AOUC            na    1.00    2.00  ...   53.00   54.00      na
        <BLANKLINE>
        >>> # 2.6) Variables object
        >>> variables["AQC"]                    # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  1962Y1  ...  2013Y1  2014Y1  2015Y1
        AQC           0.22    0.22    0.22  ...    1.56    1.61    1.67
        <BLANKLINE>
        >>> variables["ACAF"] = variables["AQC"]
        >>> variables["ACAF"]                   # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  1962Y1  ...  2013Y1  2014Y1  2015Y1
        ACAF          0.22    0.22    0.22  ...    1.56    1.61    1.67
        <BLANKLINE>

        >>> # 3) set the values for range of (contiguous) periods 
        >>> # 3.1) variable[t:t+x] = same value for all periods
        >>> variables["ACAF", "1991Y1:1995Y1"] = 0.0
        >>> variables["ACAF", "1991Y1:1995Y1"]            # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE        
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
        name        1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF          0.00    0.00    0.00    0.00    0.00
        <BLANKLINE>
        >>> # 3.2) variable[t:t+x] = LEC expression
        >>> variables["ACAF", "1991Y1:1995Y1"] = "t + 10"
        >>> variables["ACAF", "1991Y1:1995Y1"]        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
        name        1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         41.00   42.00   43.00   44.00   45.00
        <BLANKLINE>
        >>> # 3.3) variable[t:t+x] = list of values for each period
        >>> values = [1.0, NA, 3.0, np.nan, 5.0]
        >>> variables["ACAF", "1991Y1:1995Y1"] = values
        >>> variables["ACAF", "1991Y1:1995Y1"]          # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
        name        1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF          1.00      na    3.00      na    5.00
        <BLANKLINE>
        >>> # 3.4) variable[t:t+x] = numpy array
        >>> values = np.asarray(values)
        >>> variables["ACAG", "1991Y1:1995Y1"] = values
        >>> variables["ACAG", "1991Y1:1995Y1"]          # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
        name        1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAG          1.00      na    3.00      na    5.00
        <BLANKLINE>
        >>> # 3.5) variable[t:t+x] = pandas Series
        >>> periods = Sample("1991Y1:1995Y1").periods
        >>> variables["AOUC", "1991Y1:1995Y1"] = pd.Series(values, index=periods)
        >>> variables["AOUC", "1991Y1:1995Y1"]          # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
        name        1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        AOUC          1.00      na    3.00      na    5.00
        <BLANKLINE>
        >>> # 3.6) variable[t:t+x] = Variables object
        >>> variables["AQC", "1991Y1:1995Y1"]           # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
        name        1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        AQC           1.06    1.11    1.15    1.16    1.16
        <BLANKLINE>
        >>> variables["ACAF", "1991Y1:1995Y1"] = variables["AQC", "1991Y1:1995Y1"]
        >>> variables["ACAF", "1991Y1:1995Y1"]        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
        name        1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF          1.06    1.11    1.15    1.16    1.16
        <BLANKLINE>
        
        >>> # 4) set the values for a list of non-contiguous periods
        >>> values = [1.0, 3.0, 5.0]
        >>> variables["ACAF", ["1991Y1", "1993Y1", "1995Y1"]] = values
        >>> variables["ACAF", ["1991Y1", "1993Y1", "1995Y1"]]        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        time
        1991Y1    1.0
        1993Y1    3.0
        1995Y1    5.0
        Name: ACAF, dtype: float64

        >>> # c) -------- update several variables at once --------
        >>> # 1) using a string
        >>> variables["ACAF, ACAG, AOUC", "1991Y1:1995Y1"] = "t + 1"
        >>> variables["ACAF, ACAG, AOUC", "1991Y1:1995Y1"]          # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 3
        filename: ...fun.var
        description: Modèle fun - Simulation 1                         
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
        name\t1991Y1\t1992Y1\t1993Y1\t1994Y1\t1995Y1
        ACAF\t 32.00\t 33.00\t 34.00\t 35.00\t 36.00
        ACAG\t 32.00\t 33.00\t 34.00\t 35.00\t 36.00
        AOUC\t 32.00\t 33.00\t 34.00\t 35.00\t 36.00
        <BLANKLINE>
        >>> # 2) using a dict of values
        >>> periods = ["1991Y1", "1992Y1", "1993Y1", "1994Y1", "1995Y1"]
        >>> values = {"ACAF": "ACAF * 1.05", 
        ...           "ACAG": [np.nan, -39.96, -42.88, -16.33, -41.16], 
        ...           "AOUC": pd.Series([1.023, np.nan, 1.046, np.nan, 1.064], index=periods)}
        >>> variables["ACAF, ACAG, AOUC", "1991Y1:1995Y1"] = values
        >>> variables["ACAF, ACAG, AOUC", "1991Y1:1995Y1"]          # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 3
        filename: ...fun.var
        description: Modèle fun - Simulation 1                         
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
        name\t1991Y1\t1992Y1\t1993Y1\t1994Y1\t1995Y1
        ACAF\t 33.60\t 34.65\t 35.70\t 36.75\t 37.80
        ACAG\t    na\t-39.96\t-42.88\t-16.33\t-41.16
        AOUC\t  1.02\t    na\t  1.05\t    na\t  1.06
        <BLANKLINE>
        >>> # 3) using a numpy array
        >>> data = [[28.89, 31.90, 36.66, 42.13, 9.92],
        ...         [np.nan, -39.96, -42.88, -16.33, -41.16], 
        ...         [1.023, np.nan, 1.046, np.nan, 1.064]]
        >>> data = np.asarray(data)
        >>> data                                                    # doctest: +NORMALIZE_WHITESPACE
        array([[ 28.89 ,  31.9  ,  36.66 ,  42.13 ,   9.92 ],
               [    nan, -39.96 , -42.88 , -16.33 , -41.16 ],
               [  1.023,     nan,   1.046,     nan,   1.064]])
        >>> variables["ACAF, ACAG, AOUC", "1991Y1:1995Y1"] = data
        >>> variables["ACAF, ACAG, AOUC", "1991Y1:1995Y1"]          # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 3
        filename: ...fun.var
        description: Modèle fun - Simulation 1                         
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
        name\t1991Y1\t1992Y1\t1993Y1\t1994Y1\t1995Y1
        ACAF\t 28.89\t 31.90\t 36.66\t 42.13\t  9.92
        ACAG\t    na\t-39.96\t-42.88\t-16.33\t-41.16
        AOUC\t  1.02\t    na\t  1.05\t    na\t  1.06
        <BLANKLINE>
        >>> # 4) using a pandas DataFrame
        >>> data += 2.0
        >>> df = pd.DataFrame(data, index=["ACAF", "ACAG", "AOUC"], columns=periods)
        >>> df                                                      # doctest: +NORMALIZE_WHITESPACE
        1991Y1\t1992Y1\t1993Y1\t1994Y1\t1995Y1
        ACAF\t30.890\t33.90\t38.660\t44.13\t11.920
        ACAG\tNaN\t-37.96\t-40.880\t-14.33\t-39.160
        AOUC\t3.023\tNaN\t3.046\tNaN\t3.064
        >>> variables["ACAF, ACAG, AOUC", "1991Y1:1995Y1"] = df
        >>> variables["ACAF, ACAG, AOUC", "1991Y1:1995Y1"]         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 3
        filename: ...fun.var
        description: Modèle fun - Simulation 1                         
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
        name\t1991Y1\t1992Y1\t1993Y1\t1994Y1\t1995Y1
        ACAF\t 30.89\t 33.90\t 38.66\t 44.13\t 11.92
        ACAG\t    na\t-37.96\t-40.88\t-14.33\t-39.16
        AOUC\t  3.02\t    na\t  3.05\t    na\t  3.06
        <BLANKLINE>
        >>> # 5) using an Array object (from the larray library)
        >>> data += 2.0
        >>> names_axis = la.Axis(name="names", labels=["ACAF", "ACAG", "AOUC"])
        >>> time_axis = la.Axis(name="time", labels=periods)
        >>> array = la.Array(data, axes=[names_axis, time_axis])
        >>> array                                                   # doctest: +NORMALIZE_WHITESPACE
        names\\time  1991Y1  1992Y1  1993Y1               1994Y1  1995Y1
              ACAF   32.89    35.9   40.66                46.13   13.92
              ACAG     nan  -35.96  -38.88  -12.329999999999998  -37.16
              AOUC   5.023     nan   5.046                  nan   5.064
        >>> variables["ACAF, ACAG, AOUC", "1991Y1:1995Y1"] = array
        >>> variables["ACAF, ACAG, AOUC", "1991Y1:1995Y1"]          # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 3
        filename: ...fun.var
        description: Modèle fun - Simulation 1                         
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
        name\t1991Y1\t1992Y1\t1993Y1\t1994Y1\t1995Y1
        ACAF\t 32.89\t 35.90\t 40.66\t 46.13\t 13.92
        ACAG\t    na\t-35.96\t-38.88\t-12.33\t-37.16
        AOUC\t  5.02\t    na\t  5.05\t    na\t  5.06
        <BLANKLINE>
        >>> # 6) using another variables database (subset)
        >>> variables_subset = variables["ACAF, ACAG, AOUC", "1991Y1:1995Y1"].copy()
        >>> variables_subset["ACAF"] = [1991, 1992, 1993, 1994, 1995]
        >>> variables_subset["ACAG"] = [1996, 1997, 1998, 1999, 2000]
        >>> variables_subset["AOUC"] = [2001, 2002, 2003, 2004, 2005]
        >>> variables["ACAF, ACAG, AOUC", "1991Y1:1995Y1"] = variables_subset
        >>> variables["ACAF, ACAG, AOUC", "1991Y1:1995Y1"]          # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 3
        filename: ...fun.var
        description: Modèle fun - Simulation 1                         
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
        name\t 1991Y1\t 1992Y1\t 1993Y1\t 1994Y1\t 1995Y1
        ACAF\t1991.00\t1992.00\t1993.00\t1994.00\t1995.00
        ACAG\t1996.00 1997.00 1998.00 1999.00 2000.00
        AOUC\t2001.00 2002.00 2003.00 2004.00 2005.00
        <BLANKLINE>

        >>> # d) -------- working on a subset (whole sample) --------
        >>> # reset variables database to initial state
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> # 1) get subset
        >>> vars_subset = variables["A*"]
        >>> vars_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> # 2) add a variable to the subset
        >>> vars_subset["A0"] = np.nan
        >>> vars_subset["A0"]              # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  ...  2014Y1   2015Y1
        A0              na      na  ...      na       na
        <BLANKLINE>
        >>> # --> new variable also appears in the global workspace
        >>> "A0" in variables
        True
        >>> variables["A0"]                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  ...  2014Y1   2015Y1
        A0              na      na  ...      na       na
        <BLANKLINE>
        >>> # 3) update a variable in the subset
        >>> vars_subset["A0"] = 0.0
        >>> vars_subset["A0"]              # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  ...  2014Y1   2015Y1
        A0            0.00    0.00  ...    0.00     0.00
        <BLANKLINE>
        >>> # --> variable is also updated in the global workspace
        >>> variables["A0"]                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  ...  2014Y1   2015Y1
        A0            0.00    0.00  ...    0.00     0.00
        <BLANKLINE>
        >>> # 4) delete a variable in the subset
        >>> del vars_subset["A0"]
        >>> "A0" in vars_subset
        False
        >>> # --> variable is also deleted in the global workspace
        >>> "A0" in variables
        False

        >>> # e) -------- working on a subset (names + periods) --------
        >>> # 1) get subset
        >>> vars_subset = variables["A*", "1991Y1:1995Y1"]
        >>> vars_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> vars_subset                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         26.24   30.16   34.66    8.16  -13.13
        ACAG        -30.93  -40.29  -43.16  -16.03  -41.85
        AOUC          1.02    1.03    1.03    1.05    1.05
        AOUC_         0.96    0.97    0.98    0.99    1.00
        AQC           1.06    1.11    1.15    1.16    1.16
        <BLANKLINE>
        >>> # 2) add a new variable in the subset -> Forbidden !
        >>> vars_subset["A0"] = 0.0         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
           ...
        RuntimeError: Cannot add the IODE variable \'A0\' when the subset does not cover the whole sample of the IODE Variables workspace.
        >>> # 3) update a variable in the subset
        >>> vars_subset["ACAF"] = 1.0
        >>> vars_subset["ACAF"]             # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
        name        1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF          1.00    1.00    1.00    1.00    1.00
        <BLANKLINE>
        >>> # --> variable is also updated in the global workspace
        >>> variables["ACAF", "1991Y1"]
        1.0
        >>> variables["ACAF", "1995Y1"]
        1.0
        >>> # 4) delete a variable in the subset -> Forbidden !
        >>> del vars_subset["ACAF"]           # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
          ...
        RuntimeError: Cannot delete variable(s) \'ACAF\' when the subset does not cover the whole sample of the IODE Variables workspace
        '''
    def __delitem__(self, key) -> None:
        '''
        Remove the (subset of) variable(s) referenced by `key` from the Variables database.

        Parameters
        ----------
        key: str or list(str)
            (list of) name(s) of the variable(s) to be removed.
            The list of names can be given as a string pattern (e.g. "A*;*_").

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded

        >>> # a) delete one variable
        >>> variables.get_names("A*")
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> del variables["ACAF"]
        >>> variables.get_names("A*")
        [\'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']

        >>> # b) delete several variables at once using a pattern
        >>> del variables["A*"]
        >>> variables.get_names("A*")
        []

        >>> # c) delete several variables at once using a list of names
        >>> variables.get_names("B*")
        [\'BENEF\', \'BQY\', \'BRUGP\', \'BVY\']
        >>> del variables[["BENEF", "BQY"]]
        >>> variables.get_names("B*")
        [\'BRUGP\', \'BVY\']

        >>> # delete one variable from a subset of the global workspace
        >>> vars_subset = variables["D*"]
        >>> vars_subset.names
        [\'DEBT\', \'DPU\', \'DPUF\', \'DPUG\', \'DPUGO\', \'DPUH\', \'DPUHO\', \'DPUU\', \'DTF\', \'DTFX\', \'DTH\', \'DTH1\', \'DTH1C\', \'DTHX\']
        >>> del vars_subset["DPUGO"]
        >>> vars_subset.names
        [\'DEBT\', \'DPU\', \'DPUF\', \'DPUG\', \'DPUH\', \'DPUHO\', \'DPUU\', \'DTF\', \'DTFX\', \'DTH\', \'DTH1\', \'DTH1C\', \'DTHX\']
        >>> # NOTE: the variable has also been deleted from the global workspace
        >>> "DPUGO" in variables
        False
        >>> variables.get_names("D*")
        [\'DEBT\', \'DPU\', \'DPUF\', \'DPUG\', \'DPUH\', \'DPUHO\', \'DPUU\', \'DTF\', \'DTFX\', \'DTH\', \'DTH1\', \'DTH1C\', \'DTHX\']
        >>> # WARNING: when deleting (one) variable(s), the period(s) cannot be specified
        >>> del variables["DPUG", "1990Y1:1995Y1"]          # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        RuntimeError: Cannot delete variable(s) \'DPUG\'.
        The syntax \'del variables[\'DPUG\']\' must be used instead of \'del variables[\'DPUG\', <periods>]\'
        
        >>> # subset over names and periods
        >>> vars_subset = variables["D*", "1990Y1:1995Y1"]
        >>> vars_subset.names
        [\'DEBT\', \'DPU\', \'DPUF\', \'DPUG\', \'DPUH\', \'DPUHO\', \'DPUU\', \'DTF\', \'DTFX\', \'DTH\', \'DTH1\', \'DTH1C\', \'DTHX\']
        >>> del vars_subset["DPUG"]                         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        RuntimeError: Cannot delete variable(s) \'DPUG\' when the subset does not cover the whole sample of the IODE Variables workspace
        '''
    def __binary_op__(self, other: int | float | np.ndarray | pd.Series | pd.DataFrame | Array | Self, op: BinaryOperation, copy_self: bool) -> Self: ...
    def __add__(self, other):
        '''
        Add `other` to the current (subset of) Variables object.

        Parameters
        ----------
        other: int, float, numpy ndarray, pandas Series, pandas DataFrame, larray Array or iode Variables
            If `other` is an int or a float, add the scalar to all values of the current (subset of) Variables object.
            If `other` is a numpy ndarray, the shape of the ndarray must be compatible with the current (subset of) 
            Variables object. Specifically, the number of rows must be equal to the number of variables and the number of 
            columns must be equal to the number of periods.
            If `other` is a pandas Series, it must represent either a single variable or a single period.
            If `other` is a pandas DataFrame, it must represent the same variables names and periods 
            as the current (subset of) Variables object. Specifically, the index of the DataFrame must be equal to the 
            variables names and the columns of the DataFrame must be equal to the periods.
            If `other` is an larray Array, its last axis must be equal to the periods and be named \'time\'. 
            If the Array has more than two axes, the first n-1 axes are combined to form the variables names. 
            The first (combined) axis must be equal to the variables names.
            If `other` is an iode Variables object, add the two Variables objects.
            `self` and `other` must share the same sample and represent the same set of variables names.
        
        Returns
        -------
        Variables

        Warnings
        --------
        Adding a numpy ndarray to a Variables object is not recommended as there is no compatibility check 
        between for the names and periods. The result is not guaranteed to be the one you expected. 
        This possibility is provided for speed reasons (when the database or the subset is large).

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> import larray as la
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables, NA, Sample
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> vars_subset = variables["A*", "1991Y1:1995Y1"]
        >>> vars_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> vars_subset                                         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         26.24   30.16   34.66    8.16  -13.13
        ACAG        -30.93  -40.29  -43.16  -16.03  -41.85
        AOUC          1.02    1.03    1.03    1.05    1.05
        AOUC_         0.96    0.97    0.98    0.99    1.00
        AQC           1.06    1.11    1.15    1.16    1.16
        <BLANKLINE>

        >>> # add a scalar to all values of a subset of a Variables object
        >>> new_vars_subset = vars_subset + 2.0
        >>> new_vars_subset                                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         28.24   32.16   36.66   10.16  -11.13
        ACAG        -28.93  -38.29  -41.16  -14.03  -39.85
        AOUC          3.02    3.03    3.03    3.05    3.05
        AOUC_         2.96    2.97    2.98    2.99    3.00
        AQC           3.06    3.11    3.15    3.16    3.16
        <BLANKLINE>

        >>> # add two (subsets of) a Variables object
        >>> new_vars_subset = vars_subset + vars_subset
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         52.48   60.32   69.32   16.32  -26.26
        ACAG        -61.87  -80.57  -86.32  -32.06  -83.69
        AOUC          2.05    2.06    2.06    2.09    2.10
        AOUC_         1.93    1.95    1.96    1.98    1.99
        AQC           2.13    2.22    2.31    2.31    2.32
        <BLANKLINE>

        >>> # add a pandas Series to a single variable
        >>> series = pd.Series([1.0, 2.0, 3.0, 4.0, 5.0], index=vars_subset.periods_as_str)
        >>> series                                              # doctest: +NORMALIZE_WHITESPACE
        1991Y1    1.0
        1992Y1    2.0
        1993Y1    3.0
        1994Y1    4.0
        1995Y1    5.0
        dtype: float64
        >>> updated_ACAF = vars_subset["ACAF"] + series
        >>> updated_ACAF                                        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         27.24   32.16   37.66   12.16   -8.13
        <BLANKLINE>
        
        >>> # add a pandas Series to a subset corresponding to a single period
        >>> series = pd.Series([1.0, 2.0, 3.0, 4.0, 5.0], index=vars_subset.names)
        >>> series                                          # doctest: +NORMALIZE_WHITESPACE
        ACAF     1.0
        ACAG     2.0
        AOUC     3.0
        AOUC_    4.0
        AQC      5.0
        dtype: float64
        >>> vars_subset_1995Y1 = vars_subset[:, "1995Y1"] + series
        >>> vars_subset_1995Y1                              # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1995Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1995Y1
        ACAF        -12.13
        ACAG        -39.85
        AOUC          4.05
        AOUC_         5.00
        AQC           6.16
        <BLANKLINE>

        >>> # add a pandas DataFrame to the subset of a Variables object  
        >>> data = np.array([[1.0, 2.0, 3.0, 4.0, 5.0], 
        ...                  [6.0, 7.0, 8.0, 9.0, 10.0], 
        ...                  [11.0, 12.0, 13.0, 14.0, 15.0], 
        ...                  [16.0, 17.0, 18.0, 19.0, 20.0], 
        ...                  [21.0, 22.0, 23.0, 24.0, 25.0]],)
        >>> df = pd.DataFrame(data, index=vars_subset.names, columns=vars_subset.periods_as_str) 
        >>> df                                              # doctest: +NORMALIZE_WHITESPACE
               1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF      1.0     2.0     3.0     4.0     5.0
        ACAG      6.0     7.0     8.0     9.0    10.0
        AOUC     11.0    12.0    13.0    14.0    15.0
        AOUC_    16.0    17.0    18.0    19.0    20.0
        AQC      21.0    22.0    23.0    24.0    25.0
        >>> new_vars_subset = vars_subset + df
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         27.24   32.16   37.66   12.16   -8.13
        ACAG        -24.93  -33.29  -35.16   -7.03  -31.85
        AOUC         12.02   13.03   14.03   15.05   16.05
        AOUC_        16.96   17.97   18.98   19.99   21.00
        AQC          22.06   23.11   24.15   25.16   26.16
        <BLANKLINE>

        >>> # add an larray Array to a subset of a Variables object
        >>> axis_names = la.Axis(name="names", labels=vars_subset.names)
        >>> axis_time = la.Axis(name="time", labels=vars_subset.periods_as_str)
        >>> array = la.Array(data, axes=(axis_names, axis_time))
        >>> array                                           # doctest: +NORMALIZE_WHITESPACE
        names\\time  1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
              ACAF     1.0     2.0     3.0     4.0     5.0
              ACAG     6.0     7.0     8.0     9.0    10.0
              AOUC    11.0    12.0    13.0    14.0    15.0
             AOUC_    16.0    17.0    18.0    19.0    20.0
               AQC    21.0    22.0    23.0    24.0    25.0
        >>> new_vars_subset = vars_subset + array
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         27.24   32.16   37.66   12.16   -8.13
        ACAG        -24.93  -33.29  -35.16   -7.03  -31.85
        AOUC         12.02   13.03   14.03   15.05   16.05
        AOUC_        16.96   17.97   18.98   19.99   21.00
        AQC          22.06   23.11   24.15   25.16   26.16
        <BLANKLINE>        

        >>> # WARNING: adding a numpy ndarray to a (subset of a) Variables object is not recommended 
        >>> #          as there is no compatibility check between for the names and periods.
        >>> #          The result is not guaranteed to be the one you expected.
        >>> #          This possibility is provided for speed reasons 
        >>> #          (when dealing with large subsets/databases).
        >>> # add a numpy 1D ndarray to a single variable
        >>> data = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        >>> updated_ACAF = vars_subset["ACAF"] + data
        >>> updated_ACAF                                    # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         27.24   32.16   37.66   12.16   -8.13
        <BLANKLINE>
        >>> # add a numpy 1D ndarray to the subset corresponding to a single period
        >>> vars_subset_1995Y1 = vars_subset[:, "1995Y1"] + data
        >>> vars_subset_1995Y1                              # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1995Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1995Y1
        ACAF        -12.13
        ACAG        -39.85
        AOUC          4.05
        AOUC_         5.00
        AQC           6.16
        <BLANKLINE>        
        >>> # add a numpy 2D ndarray to a (subset of a) Variables object
        >>> data = np.array([[1.0, 2.0, 3.0, 4.0, 5.0], 
        ...                  [6.0, 7.0, 8.0, 9.0, 10.0], 
        ...                  [11.0, 12.0, 13.0, 14.0, 15.0], 
        ...                  [16.0, 17.0, 18.0, 19.0, 20.0], 
        ...                  [21.0, 22.0, 23.0, 24.0, 25.0]])
        >>> new_vars_subset = vars_subset + data
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         27.24   32.16   37.66   12.16   -8.13
        ACAG        -24.93  -33.29  -35.16   -7.03  -31.85
        AOUC         12.02   13.03   14.03   15.05   16.05
        AOUC_        16.96   17.97   18.98   19.99   21.00
        AQC          22.06   23.11   24.15   25.16   26.16
        <BLANKLINE>
        '''
    def __radd__(self, other):
        '''
        Add `other` to the current (subset of) Variables object.

        Parameters
        ----------
        other: int, float, numpy ndarray, larray Array or iode Variables
            If `other` is an int or a float, add the scalar to all values of the current (subset of) Variables object.
            If `other` is a numpy ndarray, the shape of the ndarray must be compatible with the current (subset of) 
            Variables object. Specifically, the number of rows must be equal to the number of variables and the number of 
            columns must be equal to the number of periods.
            If `other` is an larray Array, its last axis must be equal to the periods and be named \'time\'. 
            If the Array has more than two axes, the first n-1 axes are combined to form the variables names. 
            The first (combined) axis must be equal to the variables names.
            If `other` is an iode Variables object, add the two Variables objects.
            `self` and `other` must share the same sample and represent the same set of variables names.
        
        Returns
        -------
        Variables

        Warnings
        --------
        Adding a numpy ndarray to a Variables object is not recommended as there is no compatibility check 
        between for the names and periods. The result is not guaranteed to be the one you expected. 
        This possibility is provided for speed reasons (when the database or the subset is large).

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> import larray as la
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables, NA, Sample
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> vars_subset = variables["A*", "1991Y1:1995Y1"]
        >>> vars_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> vars_subset                                         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         26.24   30.16   34.66    8.16  -13.13
        ACAG        -30.93  -40.29  -43.16  -16.03  -41.85
        AOUC          1.02    1.03    1.03    1.05    1.05
        AOUC_         0.96    0.97    0.98    0.99    1.00
        AQC           1.06    1.11    1.15    1.16    1.16
        <BLANKLINE>

        >>> # add a scalar to all values of a subset of a Variables object
        >>> new_vars_subset = 2.0 + vars_subset
        >>> new_vars_subset                                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         28.24   32.16   36.66   10.16  -11.13
        ACAG        -28.93  -38.29  -41.16  -14.03  -39.85
        AOUC          3.02    3.03    3.03    3.05    3.05
        AOUC_         2.96    2.97    2.98    2.99    3.00
        AQC           3.06    3.11    3.15    3.16    3.16
        <BLANKLINE>

        >>> # add two (subsets of) a Variables object
        >>> new_vars_subset = vars_subset + vars_subset
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         52.48   60.32   69.32   16.32  -26.26
        ACAG        -61.87  -80.57  -86.32  -32.06  -83.69
        AOUC          2.05    2.06    2.06    2.09    2.10
        AOUC_         1.93    1.95    1.96    1.98    1.99
        AQC           2.13    2.22    2.31    2.31    2.32
        <BLANKLINE>

        >>> # add an larray Array to a subset of a Variables object
        >>> data = np.array([[1.0, 2.0, 3.0, 4.0, 5.0], 
        ...                  [6.0, 7.0, 8.0, 9.0, 10.0], 
        ...                  [11.0, 12.0, 13.0, 14.0, 15.0], 
        ...                  [16.0, 17.0, 18.0, 19.0, 20.0], 
        ...                  [21.0, 22.0, 23.0, 24.0, 25.0]])
        >>> axis_names = la.Axis(name="names", labels=vars_subset.names)
        >>> axis_time = la.Axis(name="time", labels=vars_subset.periods_as_str)
        >>> array = la.Array(data, axes=(axis_names, axis_time))
        >>> array                                           # doctest: +NORMALIZE_WHITESPACE
        names\\time  1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
              ACAF     1.0     2.0     3.0     4.0     5.0
              ACAG     6.0     7.0     8.0     9.0    10.0
              AOUC    11.0    12.0    13.0    14.0    15.0
             AOUC_    16.0    17.0    18.0    19.0    20.0
               AQC    21.0    22.0    23.0    24.0    25.0
        >>> new_vars_subset = array + vars_subset
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         27.24   32.16   37.66   12.16   -8.13
        ACAG        -24.93  -33.29  -35.16   -7.03  -31.85
        AOUC         12.02   13.03   14.03   15.05   16.05
        AOUC_        16.96   17.97   18.98   19.99   21.00
        AQC          22.06   23.11   24.15   25.16   26.16
        <BLANKLINE>        

        >>> # WARNING: adding a numpy ndarray to a (subset of a) Variables object is not recommended 
        >>> #          as there is no compatibility check between for the names and periods.
        >>> #          The result is not guaranteed to be the one you expected.
        >>> #          This possibility is provided for speed reasons 
        >>> #          (when dealing with large subsets/databases).
        >>> # add a numpy 1D ndarray to a single variable
        >>> data = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        >>> updated_ACAF = data + vars_subset["ACAF"]
        >>> updated_ACAF                                    # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        array([ 27.240999 ,  32.159    ,  37.661999 ,  12.1610022,  -8.130997 ])
        >>> # add a numpy 1D ndarray to the subset corresponding to a single period
        >>> vars_subset_1995Y1 = data + vars_subset[:, "1995Y1"]
        >>> vars_subset_1995Y1                              # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        array([[-12.130997  , -11.130997  , -10.130997  ,  -9.130997  ,
                 -8.130997  ],
               [-40.845993  , -39.845993  , -38.845993  , -37.845993  ,
                -36.845993  ],
               [  2.0498914 ,   3.0498914 ,   4.0498914 ,   5.0498914 ,
                  6.0498914 ],
               [  1.99526324,   2.99526324,   3.99526324,   4.99526324,
                  5.99526324],
               [  2.1616869 ,   3.1616869 ,   4.1616869 ,   5.1616869 ,
                  6.1616869 ]])        
        >>> # add a numpy 2D ndarray to a (subset of a) Variables object
        >>> data = np.array([[1.0, 2.0, 3.0, 4.0, 5.0], 
        ...                  [6.0, 7.0, 8.0, 9.0, 10.0], 
        ...                  [11.0, 12.0, 13.0, 14.0, 15.0], 
        ...                  [16.0, 17.0, 18.0, 19.0, 20.0], 
        ...                  [21.0, 22.0, 23.0, 24.0, 25.0]])
        >>> new_vars_subset = data + vars_subset
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        array([[ 27.240999  ,  32.159     ,  37.661999  ,  12.1610022 ,
                 -8.130997  ],
               [-24.934     , -33.285999  , -35.157997  ,  -7.029003  ,
                -31.845993  ],
               [ 12.02443339,  13.0314501 ,  14.03091768,  15.04628419,
                 16.0498914 ],
               [ 16.96466659,  17.97403904,  18.97881286,  19.98955638,
                 20.99526324],
               [ 22.0628064 ,  23.1102825 ,  24.1532652 ,  25.1571276 ,
                 26.1616869 ]])
        '''
    def __iadd__(self, other):
        '''
        Add `other` to the current (subset of) Variables object.

        Parameters
        ----------
        other: int, float, numpy ndarray, pandas Series, pandas DataFrame, larray Array or iode Variables
            If `other` is an int or a float, add the scalar to all values of the current (subset of) Variables object.
            If `other` is a numpy ndarray, the shape of the ndarray must be compatible with the current (subset of) 
            Variables object. Specifically, the number of rows must be equal to the number of variables and the number of 
            columns must be equal to the number of periods.
            If `other` is a pandas Series, it must represent either a single variable or a single period.
            If `other` is a pandas DataFrame, it must represent the same variables names and periods 
            as the current (subset of) Variables object. Specifically, the index of the DataFrame must be equal to the 
            variables names and the columns of the DataFrame must be equal to the periods.
            If `other` is an larray Array, its last axis must be equal to the periods and be named \'time\'. 
            If the Array has more than two axes, the first n-1 axes are combined to form the variables names. 
            The first (combined) axis must be equal to the variables names.
            If `other` is an iode Variables object, add the two Variables objects.
            `self` and `other` must share the same sample and represent the same set of variables names.

        Warnings
        --------
        Adding a numpy ndarray to a Variables object is not recommended as there is no compatibility check 
        between for the names and periods. The result is not guaranteed to be the one you expected. 
        This possibility is provided for speed reasons (when the database or the subset is large).

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> import larray as la
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables, NA, Sample
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> vars_subset = variables["A*", "1991Y1:1995Y1"]
        >>> vars_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> vars_subset                                         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         26.24   30.16   34.66    8.16  -13.13
        ACAG        -30.93  -40.29  -43.16  -16.03  -41.85
        AOUC          1.02    1.03    1.03    1.05    1.05
        AOUC_         0.96    0.97    0.98    0.99    1.00
        AQC           1.06    1.11    1.15    1.16    1.16
        <BLANKLINE>

        >>> # add a scalar to all values of the current subset of a Variables object
        >>> vars_subset += 2.0
        >>> vars_subset                                         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         28.24   32.16   36.66   10.16  -11.13
        ACAG        -28.93  -38.29  -41.16  -14.03  -39.85
        AOUC          3.02    3.03    3.03    3.05    3.05
        AOUC_         2.96    2.97    2.98    2.99    3.00
        AQC           3.06    3.11    3.15    3.16    3.16
        <BLANKLINE>

        >>> # add a (subsets of) a Variables object to the current (subset of) Variables object
        >>> vars_subset += vars_subset
        >>> vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         56.48   64.32   73.32   20.32  -22.26
        ACAG        -57.87  -76.57  -82.32  -28.06  -79.69
        AOUC          6.05    6.06    6.06    6.09    6.10
        AOUC_         5.93    5.95    5.96    5.98    5.99
        AQC           6.13    6.22    6.31    6.31    6.32
        <BLANKLINE>

        >>> # add a pandas Series to a single variable
        >>> series = pd.Series([1.0, 2.0, 3.0, 4.0, 5.0], index=vars_subset.periods_as_str)
        >>> series                                              # doctest: +NORMALIZE_WHITESPACE
        1991Y1    1.0
        1992Y1    2.0
        1993Y1    3.0
        1994Y1    4.0
        1995Y1    5.0
        dtype: float64
        >>> vars_subset["ACAF"] += series
        >>> vars_subset["ACAF"]                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         57.48   66.32   76.32   24.32  -17.26
        <BLANKLINE>
        
        >>> # add a pandas Series to the subset corresponding to a single period
        >>> series = pd.Series([1.0, 2.0, 3.0, 4.0, 5.0], index=vars_subset.names)
        >>> series                                          # doctest: +NORMALIZE_WHITESPACE
        ACAF     1.0
        ACAG     2.0
        AOUC     3.0
        AOUC_    4.0
        AQC      5.0
        dtype: float64
        >>> vars_subset[:, "1995Y1"] += series
        >>> vars_subset[:, "1995Y1"]                        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1995Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1995Y1
        ACAF        -16.26
        ACAG        -77.69
        AOUC          9.10
        AOUC_         9.99
        AQC          11.32
        <BLANKLINE>

        >>> # add a pandas DataFrame to the current subset of the Variables object  
        >>> data = np.array([[1.0, 2.0, 3.0, 4.0, 5.0], 
        ...                  [6.0, 7.0, 8.0, 9.0, 10.0], 
        ...                  [11.0, 12.0, 13.0, 14.0, 15.0], 
        ...                  [16.0, 17.0, 18.0, 19.0, 20.0], 
        ...                  [21.0, 22.0, 23.0, 24.0, 25.0]],)
        >>> df = pd.DataFrame(data, index=vars_subset.names, columns=vars_subset.periods_as_str) 
        >>> df                                              # doctest: +NORMALIZE_WHITESPACE
               1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF      1.0     2.0     3.0     4.0     5.0
        ACAG      6.0     7.0     8.0     9.0    10.0
        AOUC     11.0    12.0    13.0    14.0    15.0
        AOUC_    16.0    17.0    18.0    19.0    20.0
        AQC      21.0    22.0    23.0    24.0    25.0
        >>> vars_subset += df
        >>> vars_subset                                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         58.48   68.32   79.32   28.32  -11.26
        ACAG        -51.87  -69.57  -74.32  -19.06  -67.69
        AOUC         17.05   18.06   19.06   20.09   24.10
        AOUC_        21.93   22.95   23.96   24.98   29.99
        AQC          27.13   28.22   29.31   30.31   36.32
        <BLANKLINE>

        >>> # add an larray Array to the current subset of the Variables object
        >>> axis_names = la.Axis(name="names", labels=vars_subset.names)
        >>> axis_time = la.Axis(name="time", labels=vars_subset.periods_as_str)
        >>> array = la.Array(data, axes=(axis_names, axis_time))
        >>> array                                           # doctest: +NORMALIZE_WHITESPACE
        names\\time  1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
              ACAF     1.0     2.0     3.0     4.0     5.0
              ACAG     6.0     7.0     8.0     9.0    10.0
              AOUC    11.0    12.0    13.0    14.0    15.0
             AOUC_    16.0    17.0    18.0    19.0    20.0
               AQC    21.0    22.0    23.0    24.0    25.0
        >>> vars_subset += array
        >>> vars_subset                                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         59.48   70.32   82.32   32.32   -6.26
        ACAG        -45.87  -62.57  -66.32  -10.06  -57.69
        AOUC         28.05   30.06   32.06   34.09   39.10
        AOUC_        37.93   39.95   41.96   43.98   49.99
        AQC          48.13   50.22   52.31   54.31   61.32
        <BLANKLINE>        

        >>> # WARNING: adding a numpy ndarray to a (subset of a) Variables object is not recommended 
        >>> #          as there is no compatibility check between for the names and periods.
        >>> #          The result is not guaranteed to be the one you expected.
        >>> #          This possibility is provided for speed reasons 
        >>> #          (when dealing with large subsets/databases).
        >>> # add a numpy 1D ndarray to a single variable
        >>> data = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        >>> vars_subset["ACAF"] += data
        >>> vars_subset["ACAF"]                             # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         60.48   72.32   85.32   36.32   -1.26
        <BLANKLINE>
        >>> # add a numpy 1D ndarray to the subset corresponding to a single period
        >>> vars_subset[:, "1995Y1"] += data
        >>> vars_subset[:, "1995Y1"]                        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1995Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1995Y1
        ACAF         -0.26
        ACAG        -55.69
        AOUC         42.10
        AOUC_        53.99
        AQC          66.32
        <BLANKLINE>        
        >>> # add a numpy 2D ndarray to the current (subset of the) Variables object
        >>> data = np.array([[1.0, 2.0, 3.0, 4.0, 5.0], 
        ...                  [6.0, 7.0, 8.0, 9.0, 10.0], 
        ...                  [11.0, 12.0, 13.0, 14.0, 15.0], 
        ...                  [16.0, 17.0, 18.0, 19.0, 20.0], 
        ...                  [21.0, 22.0, 23.0, 24.0, 25.0]])
        >>> vars_subset += data
        >>> vars_subset                                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         61.48   74.32   88.32   40.32    4.74
        ACAG        -39.87  -55.57  -58.32   -1.06  -45.69
        AOUC         39.05   42.06   45.06   48.09   57.10
        AOUC_        53.93   56.95   59.96   62.98   73.99
        AQC          69.13   72.22   75.31   78.31   91.32
        <BLANKLINE>
        '''
    def __sub__(self, other):
        '''
        Subtract `other` from the current (subset of) Variables object.

        Parameters
        ----------
        other: int, float, numpy ndarray, pandas Series, pandas DataFrame, larray Array or iode Variables
            If `other` is an int or a float, subtract the scalar from all values of the current (subset of) Variables object.
            If `other` is a numpy ndarray, the shape of the ndarray must be compatible with the current (subset of) 
            Variables object. Specifically, the number of rows must be equal to the number of variables and the number of 
            columns must be equal to the number of periods.
            If `other` is a pandas Series, it must represent either a single variable or a single period.
            If `other` is a pandas DataFrame, it must represent the same variables names and periods 
            as the current (subset of) Variables object. Specifically, the index of the DataFrame must be equal to the 
            variables names and the columns of the DataFrame must be equal to the periods.
            If `other` is an larray Array, its last axis must be equal to the periods and be named \'time\'. 
            If the Array has more than two axes, the first n-1 axes are combined to form the variables names. 
            The first (combined) axis must be equal to the variables names.
            If `other` is an iode Variables object, it must share the same sample and represent the same 
            set of variables names as `self`.
        
        Returns
        -------
        Variables

        Warnings
        --------
        Subtracting a numpy ndarray to a Variables object is not recommended as there is no compatibility check 
        between for the names and periods. The result is not guaranteed to be the one you expected. 
        This possibility is provided for speed reasons (when the database or the subset is large).

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> import larray as la
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables, NA, Sample
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> vars_subset = variables["A*", "1991Y1:1995Y1"]
        >>> vars_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> vars_subset                                         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         26.24   30.16   34.66    8.16  -13.13
        ACAG        -30.93  -40.29  -43.16  -16.03  -41.85
        AOUC          1.02    1.03    1.03    1.05    1.05
        AOUC_         0.96    0.97    0.98    0.99    1.00
        AQC           1.06    1.11    1.15    1.16    1.16
        <BLANKLINE>

        >>> # subtract a scalar from all values of a subset of a Variables object
        >>> new_vars_subset = vars_subset - 2.0
        >>> new_vars_subset                                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF     \t 24.24\t 28.16\t 32.66\t  6.16\t-15.13
        ACAG     \t-32.93\t-42.29\t-45.16\t-18.03\t-43.85
        AOUC    \t -0.98\t -0.97\t -0.97\t -0.95\t -0.95
        AOUC_   \t -1.04\t -1.03\t -1.02\t -1.01\t -1.00
        AQC  \t     -0.94\t -0.89\t -0.85\t -0.84\t -0.84
        <BLANKLINE>

        >>> # subtract (a subset of) a Variables object from another
        >>> new_vars_subset = vars_subset - vars_subset
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF \t      0.00\t  0.00\t  0.00\t  0.00\t  0.00
        ACAG \t      0.00\t  0.00\t  0.00\t  0.00\t  0.00
        AOUC \t      0.00\t  0.00\t  0.00\t  0.00\t  0.00
        AOUC_\t      0.00\t  0.00\t  0.00\t  0.00\t  0.00
        AQC  \t      0.00\t  0.00\t  0.00\t  0.00\t  0.00
        <BLANKLINE>

        >>> # subtract a pandas Series from a single variable
        >>> series = pd.Series([1.0, 2.0, 3.0, 4.0, 5.0], index=vars_subset.periods_as_str)
        >>> series                                              # doctest: +NORMALIZE_WHITESPACE
        1991Y1    1.0
        1992Y1    2.0
        1993Y1    3.0
        1994Y1    4.0
        1995Y1    5.0
        dtype: float64
        >>> updated_ACAF = vars_subset["ACAF"] - series
        >>> updated_ACAF                                        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF\t     25.24\t 28.16\t 31.66\t  4.16\t-18.13
        <BLANKLINE>
        
        >>> # subtract a pandas Series from a subset corresponding to a single period
        >>> series = pd.Series([1.0, 2.0, 3.0, 4.0, 5.0], index=vars_subset.names)
        >>> series                                          # doctest: +NORMALIZE_WHITESPACE
        ACAF     1.0
        ACAG     2.0
        AOUC     3.0
        AOUC_    4.0
        AQC      5.0
        dtype: float64
        >>> vars_subset_1995Y1 = vars_subset[:, "1995Y1"] - series
        >>> vars_subset_1995Y1                              # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1995Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1995Y1
        ACAF \t    -14.13
        ACAG     \t-43.85
        AOUC \t     -1.95
        AOUC_\t     -3.00
        AQC  \t     -3.84
        <BLANKLINE>

        >>> # subtract a pandas DataFrame from the subset of a Variables object  
        >>> data = np.array([[1.0, 2.0, 3.0, 4.0, 5.0], 
        ...                  [6.0, 7.0, 8.0, 9.0, 10.0], 
        ...                  [11.0, 12.0, 13.0, 14.0, 15.0], 
        ...                  [16.0, 17.0, 18.0, 19.0, 20.0], 
        ...                  [21.0, 22.0, 23.0, 24.0, 25.0]],)
        >>> df = pd.DataFrame(data, index=vars_subset.names, columns=vars_subset.periods_as_str) 
        >>> df                                              # doctest: +NORMALIZE_WHITESPACE
               1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF      1.0     2.0     3.0     4.0     5.0
        ACAG      6.0     7.0     8.0     9.0    10.0
        AOUC     11.0    12.0    13.0    14.0    15.0
        AOUC_    16.0    17.0    18.0    19.0    20.0
        AQC      21.0    22.0    23.0    24.0    25.0
        >>> new_vars_subset = vars_subset - df
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF    \t 25.24\t 28.16\t 31.66\t  4.16\t-18.13
        ACAG     \t-36.93\t-47.29\t-51.16\t-25.03\t-51.85
        AOUC     \t -9.98\t-10.97\t-11.97\t-12.95\t-13.95
        AOUC_    \t-15.04\t-16.03\t-17.02\t-18.01\t-19.00
        AQC     \t-19.94\t-20.89\t-21.85\t-22.84\t-23.84
        <BLANKLINE>

        >>> # subtract an larray Array from a subset of a Variables object
        >>> axis_names = la.Axis(name="names", labels=vars_subset.names)
        >>> axis_time = la.Axis(name="time", labels=vars_subset.periods_as_str)
        >>> array = la.Array(data, axes=(axis_names, axis_time))
        >>> array                                           # doctest: +NORMALIZE_WHITESPACE
        names\\time  1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
              ACAF     1.0     2.0     3.0     4.0     5.0
              ACAG     6.0     7.0     8.0     9.0    10.0
              AOUC    11.0    12.0    13.0    14.0    15.0
             AOUC_    16.0    17.0    18.0    19.0    20.0
               AQC    21.0    22.0    23.0    24.0    25.0
        >>> new_vars_subset = vars_subset - array
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF     \t 25.24\t 28.16\t 31.66\t  4.16\t-18.13
        ACAG     \t-36.93\t-47.29\t-51.16\t-25.03\t-51.85
        AOUC     \t -9.98\t-10.97\t-11.97\t-12.95\t-13.95
        AOUC_   \t-15.04\t-16.03\t-17.02\t-18.01\t-19.00
        AQC      \t-19.94\t-20.89\t-21.85\t-22.84\t-23.84
        <BLANKLINE>        

        >>> # WARNING: subtracting a numpy ndarray from a (subset of a) Variables object is not recommended 
        >>> #          as there is no compatibility check between for the names and periods.
        >>> #          The result is not guaranteed to be the one you expected.
        >>> #          This possibility is provided for speed reasons 
        >>> #          (when dealing with large subsets/databases).
        >>> # subtract a numpy 1D ndarray from a single variable
        >>> data = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        >>> updated_ACAF = vars_subset["ACAF"] - data
        >>> updated_ACAF                                    # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF\t     25.24\t 28.16\t 31.66\t  4.16\t-18.13
        <BLANKLINE>
        >>> # subtract a numpy 1D ndarray from the subset corresponding to a single period
        >>> vars_subset_1995Y1 = vars_subset[:, "1995Y1"] - data
        >>> vars_subset_1995Y1                              # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1995Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1995Y1
        ACAF \t    -14.13
        ACAG     \t-43.85
        AOUC     \t -1.95
        AOUC_   \t -3.00
        AQC      \t -3.84
        <BLANKLINE>        
        >>> # subtract a numpy 2D ndarray from a (subset of a) Variables object
        >>> data = np.array([[1.0, 2.0, 3.0, 4.0, 5.0], 
        ...                  [6.0, 7.0, 8.0, 9.0, 10.0], 
        ...                  [11.0, 12.0, 13.0, 14.0, 15.0], 
        ...                  [16.0, 17.0, 18.0, 19.0, 20.0], 
        ...                  [21.0, 22.0, 23.0, 24.0, 25.0]])
        >>> new_vars_subset = vars_subset - data
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF    \t 25.24\t 28.16\t 31.66\t  4.16\t-18.13
        ACAG    \t-36.93\t-47.29\t-51.16\t-25.03\t-51.85
        AOUC    \t -9.98\t-10.97\t-11.97\t-12.95\t-13.95
        AOUC_    \t-15.04\t-16.03\t-17.02\t-18.01\t-19.00
        AQC  \t    -19.94\t-20.89\t-21.85\t-22.84\t-23.84
        <BLANKLINE>
        '''
    def __rsub__(self, other):
        '''
        subtract `self` from `other`.

        Parameters
        ----------
        other: iode Variables
            `self` and `other` must share the same sample and represent 
            the same set of variables names.
        
        Returns
        -------
        Variables

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> import larray as la
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables, NA, Sample
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> vars_subset = variables["A*", "1991Y1:1995Y1"]
        >>> vars_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> vars_subset                                         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         26.24   30.16   34.66    8.16  -13.13
        ACAG        -30.93  -40.29  -43.16  -16.03  -41.85
        AOUC          1.02    1.03    1.03    1.05    1.05
        AOUC_         0.96    0.97    0.98    0.99    1.00
        AQC           1.06    1.11    1.15    1.16    1.16
        <BLANKLINE>

        >>> # subtract a (subset of) Variables object from another
        >>> new_vars_subset = vars_subset - vars_subset
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF \t      0.00\t  0.00\t  0.00\t  0.00\t  0.00
        ACAG     \t  0.00\t  0.00\t  0.00\t  0.00\t  0.00
        AOUC     \t  0.00\t  0.00\t  0.00\t  0.00\t  0.00
        AOUC_\t      0.00\t  0.00\t  0.00\t  0.00\t  0.00
        AQC  \t      0.00\t  0.00\t  0.00\t  0.00\t  0.00
        <BLANKLINE>
        '''
    def __isub__(self, other):
        '''
        subtract `other` from the current (subset of) Variables object.

        Parameters
        ----------
        other: int, float, numpy ndarray, pandas Series, pandas DataFrame, larray Array or iode Variables
            If `other` is an int or a float, subtract the scalar from all values of the current (subset of) Variables object.
            If `other` is a numpy ndarray, the shape of the ndarray must be compatible with the current (subset of) 
            Variables object. Specifically, the number of rows must be equal to the number of variables and the number of 
            columns must be equal to the number of periods.
            If `other` is a pandas Series, it must represent either a single variable or a single period.
            If `other` is a pandas DataFrame, it must represent the same variables names and periods 
            as the current (subset of) Variables object. Specifically, the index of the DataFrame must be equal to the 
            variables names and the columns of the DataFrame must be equal to the periods.
            If `other` is an larray Array, its last axis must be equal to the periods and be named \'time\'. 
            If the Array has more than two axes, the first n-1 axes are combined to form the variables names. 
            The first (combined) axis must be equal to the variables names.
            If `other` is an iode Variables object, it must share the same sample and represent the same 
            set of variables names as `self`.
        Warnings
        --------
        Subtracting a numpy ndarray to a Variables object is not recommended as there is no compatibility check 
        between for the names and periods. The result is not guaranteed to be the one you expected. 
        This possibility is provided for speed reasons (when the database or the subset is large).

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> import larray as la
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables, NA, Sample
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> vars_subset = variables["A*", "1991Y1:1995Y1"]
        >>> vars_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> vars_subset                                         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         26.24   30.16   34.66    8.16  -13.13
        ACAG        -30.93  -40.29  -43.16  -16.03  -41.85
        AOUC          1.02    1.03    1.03    1.05    1.05
        AOUC_         0.96    0.97    0.98    0.99    1.00
        AQC           1.06    1.11    1.15    1.16    1.16
        <BLANKLINE>

        >>> # subtract a scalar from all values of the current subset of a Variables object
        >>> vars_subset -= 2.0
        >>> vars_subset                                         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF     \t 24.24\t 28.16\t 32.66\t  6.16\t-15.13
        ACAG     \t-32.93\t-42.29\t-45.16\t-18.03\t-43.85
        AOUC    \t -0.98\t -0.97\t -0.97\t -0.95\t -0.95
        AOUC_   \t -1.04\t -1.03\t -1.02\t -1.01\t -1.00
        AQC  \t     -0.94\t -0.89\t -0.85\t -0.84\t -0.84
        <BLANKLINE>

        >>> # subtract a (subsets of) a Variables object from the current one
        >>> vars_subset -= vars_subset
        >>> vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF \t      0.00\t  0.00\t  0.00\t  0.00\t  0.00
        ACAG \t      0.00\t  0.00\t  0.00\t  0.00\t  0.00
        AOUC \t      0.00\t  0.00\t  0.00\t  0.00\t  0.00
        AOUC_\t      0.00\t  0.00\t  0.00\t  0.00\t  0.00
        AQC  \t      0.00\t  0.00\t  0.00\t  0.00\t  0.00
        <BLANKLINE>

        >>> # subtract a pandas Series from a single variable
        >>> series = pd.Series([1.0, 2.0, 3.0, 4.0, 5.0], index=vars_subset.periods_as_str)
        >>> series                                              # doctest: +NORMALIZE_WHITESPACE
        1991Y1    1.0
        1992Y1    2.0
        1993Y1    3.0
        1994Y1    4.0
        1995Y1    5.0
        dtype: float64
        >>> vars_subset["ACAF"] -= series
        >>> vars_subset["ACAF"]                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
        name        1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         -1.00   -2.00   -3.00   -4.00   -5.00
        <BLANKLINE>
        
        >>> # subtract a pandas Series from the subset corresponding to a single period
        >>> series = pd.Series([1.0, 2.0, 3.0, 4.0, 5.0], index=vars_subset.names)
        >>> series                                          # doctest: +NORMALIZE_WHITESPACE
        ACAF     1.0
        ACAG     2.0
        AOUC     3.0
        AOUC_    4.0
        AQC      5.0
        dtype: float64
        >>> vars_subset[:, "1995Y1"] -= series
        >>> vars_subset[:, "1995Y1"]                        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1995Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1995Y1
        ACAF         -6.00
        ACAG         -2.00
        AOUC         -3.00
        AOUC_        -4.00
        AQC          -5.00
        <BLANKLINE>

        >>> # subtract a pandas DataFrame from the current subset of the Variables object  
        >>> data = np.array([[1.0, 2.0, 3.0, 4.0, 5.0], 
        ...                  [6.0, 7.0, 8.0, 9.0, 10.0], 
        ...                  [11.0, 12.0, 13.0, 14.0, 15.0], 
        ...                  [16.0, 17.0, 18.0, 19.0, 20.0], 
        ...                  [21.0, 22.0, 23.0, 24.0, 25.0]],)
        >>> df = pd.DataFrame(data, index=vars_subset.names, columns=vars_subset.periods_as_str) 
        >>> df                                              # doctest: +NORMALIZE_WHITESPACE
               1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF      1.0     2.0     3.0     4.0     5.0
        ACAG      6.0     7.0     8.0     9.0    10.0
        AOUC     11.0    12.0    13.0    14.0    15.0
        AOUC_    16.0    17.0    18.0    19.0    20.0
        AQC      21.0    22.0    23.0    24.0    25.0
        >>> vars_subset -= df
        >>> vars_subset                                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         -2.00   -4.00   -6.00   -8.00  -11.00
        ACAG         -6.00   -7.00   -8.00   -9.00  -12.00
        AOUC        -11.00  -12.00  -13.00  -14.00  -18.00
        AOUC_       -16.00  -17.00  -18.00  -19.00  -24.00
        AQC         -21.00  -22.00  -23.00  -24.00  -30.00
        <BLANKLINE>

        >>> # subtract an larray Array from the current subset of the Variables object
        >>> axis_names = la.Axis(name="names", labels=vars_subset.names)
        >>> axis_time = la.Axis(name="time", labels=vars_subset.periods_as_str)
        >>> array = la.Array(data, axes=(axis_names, axis_time))
        >>> array                                           # doctest: +NORMALIZE_WHITESPACE
        names\\time  1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
              ACAF     1.0     2.0     3.0     4.0     5.0
              ACAG     6.0     7.0     8.0     9.0    10.0
              AOUC    11.0    12.0    13.0    14.0    15.0
             AOUC_    16.0    17.0    18.0    19.0    20.0
               AQC    21.0    22.0    23.0    24.0    25.0
        >>> vars_subset -= array
        >>> vars_subset                                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         -3.00   -6.00   -9.00  -12.00  -16.00
        ACAG        -12.00  -14.00  -16.00  -18.00  -22.00
        AOUC        -22.00  -24.00  -26.00  -28.00  -33.00
        AOUC_       -32.00  -34.00  -36.00  -38.00  -44.00
        AQC         -42.00  -44.00  -46.00  -48.00  -55.00
        <BLANKLINE>   

        >>> # WARNING: subtracting a numpy ndarray to a (subset of a) Variables object is not recommended 
        >>> #          as there is no compatibility check between for the names and periods.
        >>> #          The result is not guaranteed to be the one you expected.
        >>> #          This possibility is provided for speed reasons 
        >>> #          (when dealing with large subsets/databases).
        >>> # subtract a numpy 1D ndarray from a single variable
        >>> data = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        >>> vars_subset["ACAF"] -= data
        >>> vars_subset["ACAF"]                             # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
        name        1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         -4.00   -8.00  -12.00  -16.00  -21.00
        <BLANKLINE>
        >>> # subtract a numpy 1D ndarray from the subset corresponding to a single period
        >>> vars_subset[:, "1995Y1"] -= data
        >>> vars_subset[:, "1995Y1"]                        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1995Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1995Y1
        ACAF        -22.00
        ACAG        -24.00
        AOUC        -36.00
        AOUC_       -48.00
        AQC         -60.00
        <BLANKLINE>        
        >>> # subtract a numpy 2D ndarray from the current (subset of the) Variables object
        >>> data = np.array([[1.0, 2.0, 3.0, 4.0, 5.0], 
        ...                  [6.0, 7.0, 8.0, 9.0, 10.0], 
        ...                  [11.0, 12.0, 13.0, 14.0, 15.0], 
        ...                  [16.0, 17.0, 18.0, 19.0, 20.0], 
        ...                  [21.0, 22.0, 23.0, 24.0, 25.0]])
        >>> vars_subset -= data
        >>> vars_subset                                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         -5.00  -10.00  -15.00  -20.00  -27.00
        ACAG        -18.00  -21.00  -24.00  -27.00  -34.00
        AOUC        -33.00  -36.00  -39.00  -42.00  -51.00
        AOUC_       -48.00  -51.00  -54.00  -57.00  -68.00
        AQC         -63.00  -66.00  -69.00  -72.00  -85.00
        <BLANKLINE>
        '''
    def __mul__(self, other):
        '''
        multiply the current (subset of) Variables object by `other`.

        Parameters
        ----------
        other: int, float, numpy ndarray, pandas Series, pandas DataFrame, larray Array or iode Variables
            If `other` is an int or a float, multiply all values of the current (subset of) Variables object the scalar.
            If `other` is a numpy ndarray, the shape of the ndarray must be compatible with the current (subset of) 
            Variables object. Specifically, the number of rows must be equal to the number of variables and the number of 
            columns must be equal to the number of periods.
            If `other` is a pandas Series, it must represent either a single variable or a single period.
            If `other` is a pandas DataFrame, it must represent the same variables names and periods 
            as the current (subset of) Variables object. Specifically, the index of the DataFrame must be equal to the 
            variables names and the columns of the DataFrame must be equal to the periods.
            If `other` is an larray Array, its last axis must be equal to the periods and be named \'time\'. 
            If the Array has more than two axes, the first n-1 axes are combined to form the variables names. 
            The first (combined) axis must be equal to the variables names.
            If `other` is an iode Variables object, it must share the same sample and represent the same 
            set of variables names as `self`.        
        Returns
        -------
        Variables

        Warnings
        --------
        Multiplying a numpy ndarray to a Variables object is not recommended as there is no compatibility check 
        between for the names and periods. The result is not guaranteed to be the one you expected. 
        This possibility is provided for speed reasons (when the database or the subset is large).

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> import larray as la
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables, NA, Sample
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> vars_subset = variables["A*", "1991Y1:1995Y1"]
        >>> vars_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> vars_subset                                         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         26.24   30.16   34.66    8.16  -13.13
        ACAG        -30.93  -40.29  -43.16  -16.03  -41.85
        AOUC          1.02    1.03    1.03    1.05    1.05
        AOUC_         0.96    0.97    0.98    0.99    1.00
        AQC           1.06    1.11    1.15    1.16    1.16
        <BLANKLINE>

        >>> # multiply all values of a subset of a Variables object by a scalar
        >>> new_vars_subset = vars_subset * 2.0
        >>> new_vars_subset                                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF    \t 52.48\t 60.32\t 69.32\t 16.32\t-26.26
        ACAG     \t-61.87\t-80.57\t-86.32\t-32.06\t-83.69
        AOUC \t      2.05\t  2.06\t  2.06\t  2.09\t  2.10
        AOUC_\t      1.93\t  1.95\t  1.96\t  1.98\t  1.99
        AQC  \t      2.13\t  2.22\t  2.31\t  2.31\t  2.32
        <BLANKLINE>

        >>> # multiply two (subsets of) a Variables object
        >>> new_vars_subset = vars_subset * vars_subset
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1   1992Y1   1993Y1  1994Y1   1995Y1
        ACAF     \t688.59\t 909.57\t 1201.45   66.60   172.42
        ACAG     \t956.91\t1622.96\t 1862.61  256.93  1751.09
        AOUC \t      1.05\t   1.06\t    1.06\t1.09\t 1.10
        AOUC_\t      0.93\t   0.95\t    0.96\t0.98\t 0.99
        AQC  \t      1.13\t   1.23\t    1.33\t1.34\t 1.35
        <BLANKLINE>

        >>> # multiply a single variable by a pandas Series
        >>> series = pd.Series([1.0, 2.0, 3.0, 4.0, 5.0], index=vars_subset.periods_as_str)
        >>> series                                              # doctest: +NORMALIZE_WHITESPACE
        1991Y1    1.0
        1992Y1    2.0
        1993Y1    3.0
        1994Y1    4.0
        1995Y1    5.0
        dtype: float64
        >>> updated_ACAF = vars_subset["ACAF"] * series
        >>> updated_ACAF                                        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF\t     26.24\t 60.32\t103.99\t 32.64\t-65.65
        <BLANKLINE>
        
        >>> # multiply a subset corresponding to a single period by a pandas Series
        >>> series = pd.Series([1.0, 2.0, 3.0, 4.0, 5.0], index=vars_subset.names)
        >>> series                                          # doctest: +NORMALIZE_WHITESPACE
        ACAF     1.0
        ACAG     2.0
        AOUC     3.0
        AOUC_    4.0
        AQC      5.0
        dtype: float64
        >>> vars_subset_1995Y1 = vars_subset[:, "1995Y1"] * series
        >>> vars_subset_1995Y1                              # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1995Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name    1995Y1
        ACAF \t -13.13
        ACAG \t -83.69
        AOUC \t   3.15
        AOUC_\t   3.98
        AQC  \t   5.81
        <BLANKLINE>

        >>> # multiply a subset of a Variables object by a pandas DataFrame  
        >>> data = np.array([[1.0, 2.0, 3.0, 4.0, 5.0], 
        ...                  [6.0, 7.0, 8.0, 9.0, 10.0], 
        ...                  [11.0, 12.0, 13.0, 14.0, 15.0], 
        ...                  [16.0, 17.0, 18.0, 19.0, 20.0], 
        ...                  [21.0, 22.0, 23.0, 24.0, 25.0]],)
        >>> df = pd.DataFrame(data, index=vars_subset.names, columns=vars_subset.periods_as_str) 
        >>> df                                              # doctest: +NORMALIZE_WHITESPACE
               1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF      1.0     2.0     3.0     4.0     5.0
        ACAG      6.0     7.0     8.0     9.0    10.0
        AOUC     11.0    12.0    13.0    14.0    15.0
        AOUC_    16.0    17.0    18.0    19.0    20.0
        AQC      21.0    22.0    23.0    24.0    25.0
        >>> new_vars_subset = vars_subset * df
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name        1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF          26.24   60.32  103.99   32.64  -65.65
        ACAG        -185.60 -282.00 -345.26 -144.26 -418.46
        AOUC          11.27   12.38   13.40   14.65   15.75
        AOUC_         15.43   16.56   17.62   18.80   19.91
        AQC           22.32   24.43   26.53   27.77   29.04
        <BLANKLINE>

        >>> # multiply a subset of a Variables object by an larray Array
        >>> axis_names = la.Axis(name="names", labels=vars_subset.names)
        >>> axis_time = la.Axis(name="time", labels=vars_subset.periods_as_str)
        >>> array = la.Array(data, axes=(axis_names, axis_time))
        >>> array                                           # doctest: +NORMALIZE_WHITESPACE
        names\\time  1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
              ACAF     1.0     2.0     3.0     4.0     5.0
              ACAG     6.0     7.0     8.0     9.0    10.0
              AOUC    11.0    12.0    13.0    14.0    15.0
             AOUC_    16.0    17.0    18.0    19.0    20.0
               AQC    21.0    22.0    23.0    24.0    25.0
        >>> new_vars_subset = vars_subset * array
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name        1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF          26.24   60.32  103.99   32.64  -65.65
        ACAG        -185.60 -282.00 -345.26 -144.26 -418.46
        AOUC          11.27   12.38   13.40   14.65   15.75
        AOUC_         15.43   16.56   17.62   18.80   19.91
        AQC           22.32   24.43   26.53   27.77   29.04
        <BLANKLINE>     

        >>> # WARNING: multiplying a numpy ndarray to a (subset of a) Variables object is not recommended 
        >>> #          as there is no compatibility check between for the names and periods.
        >>> #          The result is not guaranteed to be the one you expected.
        >>> #          This possibility is provided for speed reasons 
        >>> #          (when dealing with large subsets/databases).
        >>> # multiply a single variable by a numpy 1D ndarray
        >>> data = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        >>> updated_ACAF = vars_subset["ACAF"] * data
        >>> updated_ACAF                                    # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF\t     26.24\t 60.32\t103.99\t 32.64\t-65.65
        <BLANKLINE>
        >>> # multiply the subset corresponding to a single period by a numpy 1D ndarray
        >>> vars_subset_1995Y1 = vars_subset[:, "1995Y1"] * data
        >>> vars_subset_1995Y1                              # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1995Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name   1995Y1
        ACAF \t-13.13
        ACAG \t-83.69
        AOUC \t  3.15
        AOUC_\t  3.98
        AQC  \t  5.81
        <BLANKLINE>       
        >>> # multiply a (subset of a) Variables object by a numpy 2D ndarray
        >>> data = np.array([[1.0, 2.0, 3.0, 4.0, 5.0], 
        ...                  [6.0, 7.0, 8.0, 9.0, 10.0], 
        ...                  [11.0, 12.0, 13.0, 14.0, 15.0], 
        ...                  [16.0, 17.0, 18.0, 19.0, 20.0], 
        ...                  [21.0, 22.0, 23.0, 24.0, 25.0]])
        >>> new_vars_subset = vars_subset * data
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name        1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF          26.24   60.32  103.99   32.64  -65.65
        ACAG        -185.60 -282.00 -345.26 -144.26 -418.46
        AOUC          11.27   12.38   13.40   14.65   15.75
        AOUC_         15.43   16.56   17.62   18.80   19.91
        AQC           22.32   24.43   26.53   27.77   29.04
        <BLANKLINE>
        '''
    def __rmul__(self, other):
        '''
        multiply `other` by the current (subset of) Variables object.

        Parameters
        ----------
        other: int, float, numpy ndarray, larray Array or iode Variables
            If `other` is an int or a float, multiply all values of the current (subset of) Variables object by the scalar.
            If `other` is a numpy ndarray, the shape of the ndarray must be compatible with the current (subset of) 
            Variables object. Specifically, the number of rows must be equal to the number of variables and the number of 
            columns must be equal to the number of periods.
            If `other` is an larray Array, its last axis must be equal to the periods and be named \'time\'. 
            If the Array has more than two axes, the first n-1 axes are combined to form the variables names. 
            The first (combined) axis must be equal to the variables names.
            If `other` is an iode Variables object, it must share the same sample and represent the same 
            set of variables names as `self`.        
        Returns
        -------
        Variables

        Warnings
        --------
        Multiplying a numpy ndarray to a Variables object is not recommended as there is no compatibility check 
        between for the names and periods. The result is not guaranteed to be the one you expected. 
        This possibility is provided for speed reasons (when the database or the subset is large).

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> import larray as la
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables, NA, Sample
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> vars_subset = variables["A*", "1991Y1:1995Y1"]
        >>> vars_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> vars_subset                                         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         26.24   30.16   34.66    8.16  -13.13
        ACAG        -30.93  -40.29  -43.16  -16.03  -41.85
        AOUC          1.02    1.03    1.03    1.05    1.05
        AOUC_         0.96    0.97    0.98    0.99    1.00
        AQC           1.06    1.11    1.15    1.16    1.16
        <BLANKLINE>

        >>> # multiply all values of a subset of a Variables object by a scalar
        >>> new_vars_subset = 2.0 * vars_subset
        >>> new_vars_subset                                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF    \t 52.48\t 60.32\t 69.32\t 16.32\t-26.26
        ACAG     \t-61.87\t-80.57\t-86.32\t-32.06\t-83.69
        AOUC \t      2.05\t  2.06\t  2.06\t  2.09\t  2.10
        AOUC_\t      1.93\t  1.95\t  1.96\t  1.98\t  1.99
        AQC  \t      2.13\t  2.22\t  2.31\t  2.31\t  2.32
        <BLANKLINE>

        >>> # multiply two (subsets of) a Variables object
        >>> new_vars_subset = vars_subset * vars_subset
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF     \t688.59\t 909.57\t 1201.45   66.60   172.42
        ACAG     \t956.91\t1622.96\t 1862.61  256.93  1751.09
        AOUC \t      1.05\t   1.06\t    1.06\t1.09\t 1.10
        AOUC_\t      0.93\t   0.95\t    0.96\t0.98\t 0.99
        AQC  \t      1.13\t   1.23\t    1.33\t1.34\t 1.35
        <BLANKLINE>

        >>> # multiply a subset of a Variables object by an larray Array
        >>> data = np.array([[1.0, 2.0, 3.0, 4.0, 5.0], 
        ...                  [6.0, 7.0, 8.0, 9.0, 10.0], 
        ...                  [11.0, 12.0, 13.0, 14.0, 15.0], 
        ...                  [16.0, 17.0, 18.0, 19.0, 20.0], 
        ...                  [21.0, 22.0, 23.0, 24.0, 25.0]])
        >>> axis_names = la.Axis(name="names", labels=vars_subset.names)
        >>> axis_time = la.Axis(name="time", labels=vars_subset.periods_as_str)
        >>> array = la.Array(data, axes=(axis_names, axis_time))
        >>> array                                           # doctest: +NORMALIZE_WHITESPACE
        names\\time  1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
              ACAF     1.0     2.0     3.0     4.0     5.0
              ACAG     6.0     7.0     8.0     9.0    10.0
              AOUC    11.0    12.0    13.0    14.0    15.0
             AOUC_    16.0    17.0    18.0    19.0    20.0
               AQC    21.0    22.0    23.0    24.0    25.0
        >>> new_vars_subset = array * vars_subset
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name        1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF          26.24   60.32  103.99   32.64  -65.65
        ACAG        -185.60 -282.00 -345.26 -144.26 -418.46
        AOUC          11.27   12.38   13.40   14.65   15.75
        AOUC_         15.43   16.56   17.62   18.80   19.91
        AQC           22.32   24.43   26.53   27.77   29.04
        <BLANKLINE>   

        >>> # WARNING: multiplying a numpy ndarray to a (subset of a) Variables object is not recommended 
        >>> #          as there is no compatibility check between for the names and periods.
        >>> #          The result is not guaranteed to be the one you expected.
        >>> #          This possibility is provided for speed reasons 
        >>> #          (when dealing with large subsets/databases).
        >>> # multiply a single variable by a numpy 1D ndarray
        >>> data = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        >>> result = data + vars_subset["ACAF"]
        >>> result                                      # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        array([ 27.240999 ,  32.159    ,  37.661999 ,  12.1610022,  -8.130997 ])
        >>> # multiply the subset corresponding to a single period by a numpy 1D ndarray
        >>> vars_subset_1995Y1 = data * vars_subset[:, "1995Y1"]
        >>> vars_subset_1995Y1                              # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        array([[ -13.130997  ,  -26.261994  ,  -39.392991  ,  -52.523988  ,
                -65.654985  ],
               [ -41.845993  ,  -83.691986  , -125.537979  , -167.383972  ,
                -209.229965  ],
               [   1.0498914 ,    2.0997828 ,    3.1496742 ,    4.1995656 ,
                   5.24945701],
               [   0.99526324,    1.99052648,    2.98578972,    3.98105296,
                   4.9763162 ],
               [   1.1616869 ,    2.3233738 ,    3.4850607 ,    4.6467476 ,
                   5.8084345 ]])       
        >>> # multiply a (subset of a) Variables object by a numpy 2D ndarray
        >>> data = np.array([[1.0, 2.0, 3.0, 4.0, 5.0], 
        ...                  [6.0, 7.0, 8.0, 9.0, 10.0], 
        ...                  [11.0, 12.0, 13.0, 14.0, 15.0], 
        ...                  [16.0, 17.0, 18.0, 19.0, 20.0], 
        ...                  [21.0, 22.0, 23.0, 24.0, 25.0]])
        >>> result = data * vars_subset
        >>> result                                  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        array([[  26.240999  ,   60.318     ,  103.985997  ,   32.6440088 ,
                  -65.654985  ],
                [-185.604     , -282.001993  , -345.263976  , -144.261027  ,
                 -418.45993   ],
                [  11.26876733,   12.37740123,   13.40192989,   14.6479787 ,
                   15.74837102],
                [  15.4346655 ,   16.55866374,   17.61863151,   18.80157119,
                   19.90526479],
                [  22.3189344 ,   24.426215  ,   26.5250996 ,   27.7710624 ,
                   29.0421725 ]])
        '''
    def __imul__(self, other):
        '''
        multiply the current (subset of) Variables object by `other`.

        Parameters
        ----------
        other: int, float, numpy ndarray, pandas Series, pandas DataFrame, larray Array or iode Variables
            If `other` is an int or a float, multiply all values of the current (subset of) Variables object by the scalar.
            If `other` is a numpy ndarray, the shape of the ndarray must be compatible with the current (subset of) 
            Variables object. Specifically, the number of rows must be equal to the number of variables and the number of 
            columns must be equal to the number of periods.
            If `other` is a pandas Series, it must represent either a single variable or a single period.
            If `other` is a pandas DataFrame, it must represent the same variables names and periods 
            as the current (subset of) Variables object. Specifically, the index of the DataFrame must be equal to the 
            variables names and the columns of the DataFrame must be equal to the periods.
            If `other` is an larray Array, its last axis must be equal to the periods and be named \'time\'. 
            If the Array has more than two axes, the first n-1 axes are combined to form the variables names. 
            The first (combined) axis must be equal to the variables names.
            If `other` is an iode Variables object, it must share the same sample and represent the same 
            set of variables names as `self`.

        Warnings
        --------
        Multiplying a numpy ndarray to a Variables object is not recommended as there is no compatibility check 
        between for the names and periods. The result is not guaranteed to be the one you expected. 
        This possibility is provided for speed reasons (when the database or the subset is large).

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> import larray as la
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables, NA, Sample
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> vars_subset = variables["A*", "1991Y1:1995Y1"]
        >>> vars_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> vars_subset                                         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         26.24   30.16   34.66    8.16  -13.13
        ACAG        -30.93  -40.29  -43.16  -16.03  -41.85
        AOUC          1.02    1.03    1.03    1.05    1.05
        AOUC_         0.96    0.97    0.98    0.99    1.00
        AQC           1.06    1.11    1.15    1.16    1.16
        <BLANKLINE>

        >>> # multiply all values of a subset of a Variables object by a scalar
        >>> vars_subset *= 2.0
        >>> vars_subset                                         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF    \t 52.48\t 60.32\t 69.32\t 16.32\t-26.26
        ACAG     \t-61.87\t-80.57\t-86.32\t-32.06\t-83.69
        AOUC \t      2.05\t  2.06\t  2.06\t  2.09\t  2.10
        AOUC_\t      1.93\t  1.95\t  1.96\t  1.98\t  1.99
        AQC  \t      2.13\t  2.22\t  2.31\t  2.31\t  2.32
        <BLANKLINE>

        >>> # multiply two (subsets of) a Variables object
        >>> vars_subset *= vars_subset
        >>> vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name        1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF        2754.36 3638.26 4805.82  266.41  689.69
        ACAG        3827.65 6491.85 7450.45 1027.72 7004.35
        AOUC           4.20    4.26    4.25    4.38    4.41
        AOUC_          3.72    3.80    3.83    3.92    3.96
        AQC            4.52    4.93    5.32    5.36    5.40
        <BLANKLINE>

        >>> # multiply a single variable by a pandas Series
        >>> series = pd.Series([1.0, 2.0, 3.0, 4.0, 5.0], index=vars_subset.periods_as_str)
        >>> series                                              # doctest: +NORMALIZE_WHITESPACE
        1991Y1    1.0
        1992Y1    2.0
        1993Y1    3.0
        1994Y1    4.0
        1995Y1    5.0
        dtype: float64
        >>> vars_subset["ACAF"] *= series
        >>> vars_subset["ACAF"]                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
        name         1991Y1   1992Y1    1993Y1   1994Y1   1995Y1
        ACAF        2754.36  7276.52  14417.45  1065.63  3448.46
        <BLANKLINE>
        
        >>> # multiply a subset corresponding to a single period by a pandas Series
        >>> series = pd.Series([1.0, 2.0, 3.0, 4.0, 5.0], index=vars_subset.names)
        >>> series                                          # doctest: +NORMALIZE_WHITESPACE
        ACAF     1.0
        ACAG     2.0
        AOUC     3.0
        AOUC_    4.0
        AQC      5.0
        dtype: float64
        >>> vars_subset[:, "1995Y1"] *= series
        >>> vars_subset[:, "1995Y1"]                        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1995Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name         1995Y1
        ACAF         3448.46
        ACAG        14008.70
        AOUC           13.23
        AOUC_          15.85
        AQC            26.99
        <BLANKLINE>

        >>> # multiply a subset of a Variables object by a pandas DataFrame  
        >>> data = np.array([[1.0, 2.0, 3.0, 4.0, 5.0], 
        ...                  [6.0, 7.0, 8.0, 9.0, 10.0], 
        ...                  [11.0, 12.0, 13.0, 14.0, 15.0], 
        ...                  [16.0, 17.0, 18.0, 19.0, 20.0], 
        ...                  [21.0, 22.0, 23.0, 24.0, 25.0]],)
        >>> df = pd.DataFrame(data, index=vars_subset.names, columns=vars_subset.periods_as_str) 
        >>> df                                              # doctest: +NORMALIZE_WHITESPACE
               1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF      1.0     2.0     3.0     4.0     5.0
        ACAG      6.0     7.0     8.0     9.0    10.0
        AOUC     11.0    12.0    13.0    14.0    15.0
        AOUC_    16.0    17.0    18.0    19.0    20.0
        AQC      21.0    22.0    23.0    24.0    25.0
        >>> vars_subset *= df
        >>> vars_subset                                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name         1991Y1          1992Y1          1993Y1         1994Y1    1995Y1
        ACAF         2754.36        14553.04        43252.35        4262.53   17242.31
        ACAG        22965.90        45442.93        59603.61        9249.44  140086.97
        AOUC           46.18           51.07           55.27          61.30     198.41
        AOUC_          59.56           64.52           68.98          74.42     316.98
        AQC            94.88          108.48          122.36         128.54     674.76
        <BLANKLINE>

        >>> # multiply a subset of a Variables object by an larray Array
        >>> axis_names = la.Axis(name="names", labels=vars_subset.names)
        >>> axis_time = la.Axis(name="time", labels=vars_subset.periods_as_str)
        >>> array = la.Array(data, axes=(axis_names, axis_time))
        >>> array                                           # doctest: +NORMALIZE_WHITESPACE
        names\\time  1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
              ACAF     1.0     2.0     3.0     4.0     5.0
              ACAG     6.0     7.0     8.0     9.0    10.0
              AOUC    11.0    12.0    13.0    14.0    15.0
             AOUC_    16.0    17.0    18.0    19.0    20.0
               AQC    21.0    22.0    23.0    24.0    25.0
        >>> vars_subset *= array
        >>> vars_subset                                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name          1991Y1          1992Y1          1993Y1         1994Y1            1995Y1
        ACAF          2754.36        29106.09       129757.05       17050.10          86211.54
        ACAG        137795.38       318100.50       476828.85       83244.98        1400869.70
        AOUC           507.94          612.80          718.45         858.25           2976.13
        AOUC_          952.92         1096.76         1241.66        1414.00           6339.51
        AQC           1992.54         2386.56         2814.32        3084.93          16868.96
        <BLANKLINE>    

        >>> # WARNING: multiplying a numpy ndarray to a (subset of a) Variables object is not recommended 
        >>> #          as there is no compatibility check between for the names and periods.
        >>> #          The result is not guaranteed to be the one you expected.
        >>> #          This possibility is provided for speed reasons 
        >>> #          (when dealing with large subsets/databases).
        >>> # multiply a single variable by a numpy 1D ndarray
        >>> data = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        >>> vars_subset["ACAF"] *= data
        >>> vars_subset["ACAF"]                             # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
        name         1991Y1    1992Y1     1993Y1    1994Y1     1995Y1
        ACAF        2754.36  58212.18  389271.15  68200.40  431057.71
        <BLANKLINE>
        >>> # multiply the subset corresponding to a single period by a numpy 1D ndarray
        >>> vars_subset[:, "1995Y1"] *= data
        >>> vars_subset[:, "1995Y1"]                        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1995Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name           1995Y1
        ACAF         431057.71
        ACAG        2801739.41
        AOUC           8928.40
        AOUC_         25358.05
        AQC           84344.78
        <BLANKLINE>      
        >>> # multiply a (subset of a) Variables object by a numpy 2D ndarray
        >>> data = np.array([[1.0, 2.0, 3.0, 4.0, 5.0], 
        ...                  [6.0, 7.0, 8.0, 9.0, 10.0], 
        ...                  [11.0, 12.0, 13.0, 14.0, 15.0], 
        ...                  [16.0, 17.0, 18.0, 19.0, 20.0], 
        ...                  [21.0, 22.0, 23.0, 24.0, 25.0]])
        >>> vars_subset *= data
        >>> vars_subset                                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name          1991Y1          1992Y1          1993Y1          1994Y1            1995Y1
        ACAF          2754.36        116424.36      1167813.46      272801.62        2155288.53
        ACAG        826772.28       2226703.47      3814630.82      749204.78       28017394.08
        AOUC          5587.35          7353.60         9339.81       12015.54         133926.04
        AOUC_        15246.65         18644.88        22349.96       26865.93         507161.04
        AQC          41843.33         52504.32        64729.44       74038.26        2108619.46
        <BLANKLINE>
        '''
    def __truediv__(self, other):
        '''
        divide the current (subset of) Variables object by `other`.

        Parameters
        ----------
        other: int, float, numpy ndarray, pandas Series, pandas DataFrame, larray Array or iode Variables
            If `other` is an int or a float, divide all values of the current (subset of) Variables object by the scalar.
            If `other` is a numpy ndarray, the shape of the ndarray must be compatible with the current (subset of) 
            Variables object. Specifically, the number of rows must be equal to the number of variables and the number of 
            columns must be equal to the number of periods.
            If `other` is a pandas Series, it must represent either a single variable or a single period.
            If `other` is a pandas DataFrame, it must represent the same variables names and periods 
            as the current (subset of) Variables object. Specifically, the index of the DataFrame must be equal to the 
            variables names and the columns of the DataFrame must be equal to the periods.
            If `other` is an larray Array, its last axis must be equal to the periods and be named \'time\'. 
            If the Array has more than two axes, the first n-1 axes are combined to form the variables names. 
            The first (combined) axis must be equal to the variables names.
            If `other` is an iode Variables object, it must share the same sample and represent the same 
            set of variables names as `self`.
                
        Returns
        -------
        Variables

        Warnings
        --------
        Dividing a numpy ndarray to a Variables object is not recommended as there is no compatibility check 
        between for the names and periods. The result is not guaranteed to be the one you expected. 
        This possibility is provided for speed reasons (when the database or the subset is large).

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> import larray as la
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables, NA, Sample
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> vars_subset = variables["A*", "1991Y1:1995Y1"]
        >>> vars_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> vars_subset                                         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         26.24   30.16   34.66    8.16  -13.13
        ACAG        -30.93  -40.29  -43.16  -16.03  -41.85
        AOUC          1.02    1.03    1.03    1.05    1.05
        AOUC_         0.96    0.97    0.98    0.99    1.00
        AQC           1.06    1.11    1.15    1.16    1.16
        <BLANKLINE>

        >>> # divide all values of a subset of a Variables object by a scalar
        >>> new_vars_subset = vars_subset / 2.0
        >>> new_vars_subset                                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name\t1991Y1\t1992Y1\t1993Y1\t1994Y1\t1995Y1
        ACAF \t 13.12\t 15.08\t 17.33\t  4.08\t -6.57
        ACAG \t-15.47\t-20.14\t-21.58\t -8.01\t-20.92
        AOUC \t  0.51\t  0.52\t  0.52\t  0.52\t  0.52
        AOUC_\t  0.48\t  0.49\t  0.49\t  0.49\t  0.50
        AQC  \t  0.53\t  0.56\t  0.58\t  0.58\t  0.58
        <BLANKLINE>

        >>> # divide (a subset of) a Variables object by another
        >>> new_vars_subset = vars_subset / vars_subset
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name\t1991Y1\t1992Y1\t1993Y1\t1994Y1\t1995Y1
        ACAF \t  1.00\t  1.00\t  1.00\t  1.00\t  1.00
        ACAG \t  1.00\t  1.00\t  1.00\t  1.00\t  1.00
        AOUC \t  1.00\t  1.00\t  1.00\t  1.00\t  1.00
        AOUC_\t  1.00\t  1.00\t  1.00\t  1.00\t  1.00
        AQC  \t  1.00\t  1.00\t  1.00\t  1.00\t  1.00
        <BLANKLINE>

        >>> # divide a a single variable by a pandas Series
        >>> series = pd.Series([1.0, 2.0, 3.0, 4.0, 5.0], index=vars_subset.periods_as_str)
        >>> series                                              # doctest: +NORMALIZE_WHITESPACE
        1991Y1    1.0
        1992Y1    2.0
        1993Y1    3.0
        1994Y1    4.0
        1995Y1    5.0
        dtype: float64
        >>> updated_ACAF = vars_subset["ACAF"] / series
        >>> updated_ACAF                                        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
        name\t1991Y1\t1992Y1\t1993Y1\t1994Y1\t1995Y1
        ACAF\t 26.24\t 15.08\t 11.55\t  2.04\t -2.63
        <BLANKLINE>
        
        >>> # divide a pandas Series to the subset corresponding to a single period
        >>> series = pd.Series([1.0, 2.0, 3.0, 4.0, 5.0], index=vars_subset.names)
        >>> series                                          # doctest: +NORMALIZE_WHITESPACE
        ACAF     1.0
        ACAG     2.0
        AOUC     3.0
        AOUC_    4.0
        AQC      5.0
        dtype: float64
        >>> vars_subset_1995Y1 = vars_subset[:, "1995Y1"] / series
        >>> vars_subset_1995Y1                              # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1995Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name\t1995Y1
        ACAF \t-13.13
        ACAG \t-20.92
        AOUC \t  0.35
        AOUC_\t  0.25
        AQC  \t  0.23
        <BLANKLINE>

        >>> # divide a (subset of a) Variables object by a pandas DataFrame  
        >>> data = np.array([[1.0, 2.0, 3.0, 4.0, 5.0], 
        ...                  [6.0, 7.0, 8.0, 9.0, 10.0], 
        ...                  [11.0, 12.0, 13.0, 14.0, 15.0], 
        ...                  [16.0, 17.0, 18.0, 19.0, 20.0], 
        ...                  [21.0, 22.0, 23.0, 24.0, 25.0]],)
        >>> df = pd.DataFrame(data, index=vars_subset.names, columns=vars_subset.periods_as_str) 
        >>> df                                              # doctest: +NORMALIZE_WHITESPACE
               1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF      1.0     2.0     3.0     4.0     5.0
        ACAG      6.0     7.0     8.0     9.0    10.0
        AOUC     11.0    12.0    13.0    14.0    15.0
        AOUC_    16.0    17.0    18.0    19.0    20.0
        AQC      21.0    22.0    23.0    24.0    25.0
        >>> new_vars_subset = vars_subset / df
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name\t1991Y1\t1992Y1\t1993Y1\t1994Y1\t1995Y1
        ACAF \t 26.24\t 15.08\t 11.55\t  2.04\t -2.63
        ACAG \t -5.16\t -5.76\t -5.39\t -1.78\t -4.18
        AOUC \t  0.09\t  0.09\t  0.08\t  0.07\t  0.07
        AOUC_\t  0.06\t  0.06\t  0.05\t  0.05\t  0.05
        AQC  \t  0.05\t  0.05\t  0.05\t  0.05\t  0.05
        <BLANKLINE>

        >>> # divide (a subset of) a Variables object by an larray Array
        >>> axis_names = la.Axis(name="names", labels=vars_subset.names)
        >>> axis_time = la.Axis(name="time", labels=vars_subset.periods_as_str)
        >>> array = la.Array(data, axes=(axis_names, axis_time))
        >>> array                                           # doctest: +NORMALIZE_WHITESPACE
        names\\time  1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
              ACAF     1.0     2.0     3.0     4.0     5.0
              ACAG     6.0     7.0     8.0     9.0    10.0
              AOUC    11.0    12.0    13.0    14.0    15.0
             AOUC_    16.0    17.0    18.0    19.0    20.0
               AQC    21.0    22.0    23.0    24.0    25.0
        >>> new_vars_subset = vars_subset / array
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name\t1991Y1\t1992Y1\t1993Y1\t1994Y1\t1995Y1
        ACAF \t 26.24\t 15.08\t 11.55\t  2.04\t -2.63
        ACAG \t -5.16\t -5.76\t -5.39\t -1.78\t -4.18
        AOUC \t  0.09\t  0.09\t  0.08\t  0.07\t  0.07
        AOUC_\t  0.06\t  0.06\t  0.05\t  0.05\t  0.05
        AQC  \t  0.05\t  0.05\t  0.05\t  0.05\t  0.05
        <BLANKLINE>        

        >>> # WARNING: dividing a numpy ndarray to a (subset of a) Variables object is not recommended 
        >>> #          as there is no compatibility check between for the names and periods.
        >>> #          The result is not guaranteed to be the one you expected.
        >>> #          This possibility is provided for speed reasons 
        >>> #          (when dealing with large subsets/databases).
        >>> # divide a single variable by a numpy 1D ndarray
        >>> data = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        >>> updated_ACAF = vars_subset["ACAF"] / data
        >>> updated_ACAF                                    # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
        name\t1991Y1\t1992Y1\t1993Y1\t1994Y1\t1995Y1
        ACAF\t 26.24\t 15.08\t 11.55\t  2.04\t -2.63
        <BLANKLINE>
        >>> # divide the subset corresponding to a single period by a numpy 1D ndarray
        >>> vars_subset_1995Y1 = vars_subset[:, "1995Y1"] / data
        >>> vars_subset_1995Y1                              # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1995Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name\t1995Y1
        ACAF \t-13.13
        ACAG \t-20.92
        AOUC \t  0.35
        AOUC_\t  0.25
        AQC  \t  0.23
        <BLANKLINE>
        >>> # divide a (subset of a) Variables object by a numpy 2D ndarray
        >>> data = np.array([[1.0, 2.0, 3.0, 4.0, 5.0], 
        ...                  [6.0, 7.0, 8.0, 9.0, 10.0], 
        ...                  [11.0, 12.0, 13.0, 14.0, 15.0], 
        ...                  [16.0, 17.0, 18.0, 19.0, 20.0], 
        ...                  [21.0, 22.0, 23.0, 24.0, 25.0]])
        >>> new_vars_subset = vars_subset / data
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name\t1991Y1\t1992Y1\t1993Y1\t1994Y1\t1995Y1
        ACAF \t 26.24\t 15.08\t 11.55\t  2.04\t -2.63
        ACAG \t -5.16\t -5.76\t -5.39\t -1.78\t -4.18
        AOUC \t  0.09\t  0.09\t  0.08\t  0.07\t  0.07
        AOUC_\t  0.06\t  0.06\t  0.05\t  0.05\t  0.05
        AQC  \t  0.05\t  0.05\t  0.05\t  0.05\t  0.05
        <BLANKLINE>
        '''
    def __rtruediv__(self, other):
        '''
        divide `other` by the current (subset of) Variables object.

        Parameters
        ----------
        other: iode Variables
            `self` and `other` must share the same sample and represent 
            the same set of variables names.
        
        Returns
        -------
        Variables

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> import larray as la
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables, NA, Sample
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> vars_subset = variables["A*", "1991Y1:1995Y1"]
        >>> vars_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> vars_subset                                         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         26.24   30.16   34.66    8.16  -13.13
        ACAG        -30.93  -40.29  -43.16  -16.03  -41.85
        AOUC          1.02    1.03    1.03    1.05    1.05
        AOUC_         0.96    0.97    0.98    0.99    1.00
        AQC           1.06    1.11    1.15    1.16    1.16
        <BLANKLINE>

        >>> # divide (a subset of) a Variables object by another
        >>> new_vars_subset = vars_subset / vars_subset
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name\t1991Y1\t1992Y1\t1993Y1\t1994Y1\t1995Y1
        ACAF \t  1.00\t  1.00\t  1.00\t  1.00\t  1.00
        ACAG \t  1.00\t  1.00\t  1.00\t  1.00\t  1.00
        AOUC \t  1.00\t  1.00\t  1.00\t  1.00\t  1.00
        AOUC_\t  1.00\t  1.00\t  1.00\t  1.00\t  1.00
        AQC  \t  1.00\t  1.00\t  1.00\t  1.00\t  1.00
        <BLANKLINE>
        '''
    def __itruediv__(self, other):
        '''
        divide the current (subset of) Variables object by `other`.

        Parameters
        ----------
        other: int, float, numpy ndarray, pandas Series, pandas DataFrame, larray Array or iode Variables
            If `other` is an int or a float, divide all values of the current (subset of) Variables object by the scalar.
            If `other` is a numpy ndarray, the shape of the ndarray must be compatible with the current (subset of) 
            Variables object. Specifically, the number of rows must be equal to the number of variables and the number of 
            columns must be equal to the number of periods.
            If `other` is a pandas Series, it must represent either a single variable or a single period.
            If `other` is a pandas DataFrame, it must represent the same variables names and periods 
            as the current (subset of) Variables object. Specifically, the index of the DataFrame must be equal to the 
            variables names and the columns of the DataFrame must be equal to the periods.
            If `other` is an larray Array, its last axis must be equal to the periods and be named \'time\'. 
            If the Array has more than two axes, the first n-1 axes are combined to form the variables names. 
            The first (combined) axis must be equal to the variables names.
            If `other` is an iode Variables object, it must share the same sample and represent the same 
            set of variables names as `self`.
        
        Warnings
        --------
        Dividing a numpy ndarray to a Variables object is not recommended as there is no compatibility check 
        between for the names and periods. The result is not guaranteed to be the one you expected. 
        This possibility is provided for speed reasons (when the database or the subset is large).

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> import larray as la
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables, NA, Sample
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> vars_subset = variables["A*", "1991Y1:1995Y1"]
        >>> vars_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> vars_subset                                         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         26.24   30.16   34.66    8.16  -13.13
        ACAG        -30.93  -40.29  -43.16  -16.03  -41.85
        AOUC          1.02    1.03    1.03    1.05    1.05
        AOUC_         0.96    0.97    0.98    0.99    1.00
        AQC           1.06    1.11    1.15    1.16    1.16
        <BLANKLINE>

        >>> # divide all values of a subset of a Variables object by a scalar
        >>> vars_subset /= 2.0
        >>> vars_subset                                         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name\t1991Y1\t1992Y1\t1993Y1\t1994Y1\t1995Y1
        ACAF \t 13.12\t 15.08\t 17.33\t  4.08\t -6.57
        ACAG \t-15.47\t-20.14\t-21.58\t -8.01\t-20.92
        AOUC \t  0.51\t  0.52\t  0.52\t  0.52\t  0.52
        AOUC_\t  0.48\t  0.49\t  0.49\t  0.49\t  0.50
        AQC  \t  0.53\t  0.56\t  0.58\t  0.58\t  0.58
        <BLANKLINE>

        >>> # divide (a subset of) a Variables object by another
        >>> vars_subset /= vars_subset
        >>> vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name\t1991Y1\t1992Y1\t1993Y1\t1994Y1\t1995Y1
        ACAF \t  1.00\t  1.00\t  1.00\t  1.00\t  1.00
        ACAG \t  1.00\t  1.00\t  1.00\t  1.00\t  1.00
        AOUC \t  1.00\t  1.00\t  1.00\t  1.00\t  1.00
        AOUC_\t  1.00\t  1.00\t  1.00\t  1.00\t  1.00
        AQC  \t  1.00\t  1.00\t  1.00\t  1.00\t  1.00
        <BLANKLINE>

        >>> # divide a a single variable by a pandas Series
        >>> series = pd.Series([1.0, 0.5, 0.25, 0.2, 0.1], index=vars_subset.periods_as_str)
        >>> series                                              # doctest: +NORMALIZE_WHITESPACE
        1991Y1    1.00
        1992Y1    0.50
        1993Y1    0.25
        1994Y1    0.20
        1995Y1    0.10
        dtype: float64
        >>> vars_subset["ACAF"] /= series
        >>> vars_subset["ACAF"]                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
        name        1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF          1.00    2.00    4.00    5.00   10.00
        <BLANKLINE>
        
        >>> # divide a pandas Series to the subset corresponding to a single period
        >>> series = pd.Series([1.0, 0.5, 0.25, 0.2, 0.1], index=vars_subset.names)
        >>> series                                          # doctest: +NORMALIZE_WHITESPACE
        ACAF     1.00
        ACAG     0.50
        AOUC     0.25
        AOUC_    0.20
        AQC      0.10
        dtype: float64
        >>> vars_subset[:, "1995Y1"] /= series
        >>> vars_subset[:, "1995Y1"]                        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1995Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1995Y1
        ACAF         10.00
        ACAG          2.00
        AOUC          4.00
        AOUC_         5.00
        AQC          10.00
        <BLANKLINE>

        >>> # divide a (subset of a) Variables object by a pandas DataFrame  
        >>> data = np.array([[0.1, 0.2, 0.25, 0.5, 1.0], 
        ...                  [1.0, 0.5, 0.25, 0.2, 0.1], 
        ...                  [1.0, 0.5, 0.25, 0.2, 0.1], 
        ...                  [1.0, 0.5, 0.25, 0.2, 0.1], 
        ...                  [1.0, 0.5, 0.25, 0.2, 0.1]])
        >>> df = pd.DataFrame(data, index=vars_subset.names, columns=vars_subset.periods_as_str) 
        >>> df                                              # doctest: +NORMALIZE_WHITESPACE
               1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF      0.1     0.2    0.25     0.5     1.0
        ACAG      1.0     0.5    0.25     0.2     0.1
        AOUC      1.0     0.5    0.25     0.2     0.1
        AOUC_     1.0     0.5    0.25     0.2     0.1
        AQC       1.0     0.5    0.25     0.2     0.1
        >>> vars_subset /= df
        >>> vars_subset                                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         10.00   10.00   16.00   10.00   10.00
        ACAG          1.00    2.00    4.00    5.00   20.00
        AOUC          1.00    2.00    4.00    5.00   40.00
        AOUC_         1.00    2.00    4.00    5.00   50.00
        AQC           1.00    2.00    4.00    5.00  100.00
        <BLANKLINE>

        >>> # divide (a subset of) a Variables object by an larray Array
        >>> axis_names = la.Axis(name="names", labels=vars_subset.names)
        >>> axis_time = la.Axis(name="time", labels=vars_subset.periods_as_str)
        >>> array = la.Array(data, axes=(axis_names, axis_time))
        >>> array                                           # doctest: +NORMALIZE_WHITESPACE
        names\\time  1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
            ACAF       0.1     0.2    0.25     0.5     1.0
            ACAG       1.0     0.5    0.25     0.2     0.1
            AOUC       1.0     0.5    0.25     0.2     0.1
            AOUC_      1.0     0.5    0.25     0.2     0.1
            AQC        1.0     0.5    0.25     0.2     0.1
        >>> vars_subset /= array
        >>> vars_subset                                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1   1995Y1
        ACAF        100.00   50.00   64.00   20.00    10.00
        ACAG          1.00    4.00   16.00   25.00   200.00
        AOUC          1.00    4.00   16.00   25.00   400.00
        AOUC_         1.00    4.00   16.00   25.00   500.00
        AQC           1.00    4.00   16.00   25.00  1000.00
        <BLANKLINE>      

        >>> # WARNING: dividing a numpy ndarray to a (subset of a) Variables object is not recommended 
        >>> #          as there is no compatibility check between for the names and periods.
        >>> #          The result is not guaranteed to be the one you expected.
        >>> #          This possibility is provided for speed reasons 
        >>> #          (when dealing with large subsets/databases).
        >>> # divide a single variable by a numpy 1D ndarray
        >>> data = np.array([1.0, 0.5, 1.0, 0.5, 1.0])
        >>> vars_subset["ACAF"] /= data
        >>> vars_subset["ACAF"]                             # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
        name        1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF        100.00  100.00   64.00   40.00   10.00
        <BLANKLINE>
        >>> # divide the subset corresponding to a single period by a numpy 1D ndarray
        >>> vars_subset[:, "1995Y1"] /= data
        >>> vars_subset[:, "1995Y1"]                        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1995Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name        1995Y1
        ACAF          10.00
        ACAG         400.00
        AOUC         400.00
        AOUC_       1000.00
        AQC         1000.00
        <BLANKLINE>
        >>> # divide a (subset of a) Variables object by a numpy 2D ndarray
        >>> data = np.array([[0.1, 0.2, 0.25, 0.5, 1.0], 
        ...                  [1.0, 0.5, 0.25, 0.2, 0.1], 
        ...                  [1.0, 0.5, 0.25, 0.2, 0.1], 
        ...                  [1.0, 0.5, 0.25, 0.2, 0.1], 
        ...                  [1.0, 0.5, 0.25, 0.2, 0.1]])
        >>> vars_subset /= data
        >>> vars_subset                                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name        1991Y1  1992Y1  1993Y1  1994Y1    1995Y1
        ACAF        1000.00  500.00  256.00   80.00     10.00
        ACAG           1.00    8.00   64.00  125.00   4000.00
        AOUC           1.00    8.00   64.00  125.00   4000.00
        AOUC_          1.00    8.00   64.00  125.00  10000.00
        AQC            1.00    8.00   64.00  125.00  10000.00
        <BLANKLINE>
        '''
    def __pow__(self, other):
        '''
        Compute the expression :math:`self^{other}` ( self \\*\\* other ).

        Parameters
        ----------
        other: int, float, numpy ndarray, pandas Series, pandas DataFrame, larray Array or iode Variables
            If `other` is an int or a float, compute \'value \\*\\* other\' for all values of the current (subset of) 
            Variables object.
            If `other` is a numpy ndarray, the shape of the ndarray must be compatible with the current (subset of) 
            Variables object. Specifically, the number of rows must be equal to the number of variables and the number of 
            columns must be equal to the number of periods.
            If `other` is a pandas Series, it must represent either a single variable or a single period.
            If `other` is a pandas DataFrame, it must represent the same variables names and periods 
            as the current (subset of) Variables object. Specifically, the index of the DataFrame must be equal to the 
            variables names and the columns of the DataFrame must be equal to the periods.
            If `other` is an larray Array, its last axis must be equal to the periods and be named \'time\'. 
            If the Array has more than two axes, the first n-1 axes are combined to form the variables names. 
            The first (combined) axis must be equal to the variables names.
            If `other` is an iode Variables object, it must share the same sample and represent the same 
            set of variables names as `self`.
                
        Returns
        -------
        Variables

        Warnings
        --------
        Using a numpy ndarray is not recommended as there is no compatibility check 
        between for the names and periods. The result is not guaranteed to be the one you expected. 
        This possibility is provided for speed reasons (when the database or the subset is large).

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> import larray as la
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables, NA, Sample
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> vars_subset = variables["A*", "1991Y1:1995Y1"]
        >>> vars_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> vars_subset                                         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF         26.24   30.16   34.66    8.16  -13.13
        ACAG        -30.93  -40.29  -43.16  -16.03  -41.85
        AOUC          1.02    1.03    1.03    1.05    1.05
        AOUC_         0.96    0.97    0.98    0.99    1.00
        AQC           1.06    1.11    1.15    1.16    1.16
        <BLANKLINE>

        >>> # compute \'value ** other\' for all values of the current 
        >>> # (subset of) Variables object.
        >>> new_vars_subset = vars_subset ** 2
        >>> new_vars_subset                                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1   1992Y1   1993Y1  1994Y1   1995Y1
        ACAF \t    688.59\t 909.57\t 1201.45   66.60   172.42
        ACAG     \t956.91\t1622.96\t 1862.61  256.93  1751.09
        AOUC \t      1.05\t   1.06\t    1.06\t1.09\t 1.10
        AOUC_    \t  0.93\t   0.95\t    0.96\t0.98\t 0.99
        AQC  \t      1.13\t   1.23\t    1.33\t1.34\t 1.35
        <BLANKLINE>

        >>> # compute \'V[name, period] ** W[name, period]\' for each name and period
        >>> # for all names and periods
        >>> other = vars_subset.copy()
        >>> other = 2.0
        >>> new_vars_subset = vars_subset ** other
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1   1992Y1   1993Y1  1994Y1   1995Y1
        ACAF \t    688.59\t 909.57\t 1201.45   66.60   172.42
        ACAG     \t956.91\t1622.96\t 1862.61  256.93  1751.09
        AOUC \t      1.05\t   1.06\t    1.06\t1.09\t 1.10
        AOUC_    \t  0.93\t   0.95\t    0.96\t0.98\t 0.99
        AQC  \t      1.13\t   1.23\t    1.33\t1.34\t 1.35
        <BLANKLINE>

        >>> # compute \'iode_var[period] ** series[period]\' for each period
        >>> series = pd.Series([1.0, 2.0, 0.5, 1./4., 2.0], index=vars_subset.periods_as_str)
        >>> series                                              # doctest: +NORMALIZE_WHITESPACE
        1991Y1    1.00
        1992Y1    2.00
        1993Y1    0.50
        1994Y1    0.25
        1995Y1    2.00
        dtype: float64
        >>> updated_ACAF = vars_subset["ACAF"] ** series
        >>> updated_ACAF                                        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF\t     26.24\t909.57\t  5.89\t  1.69\t172.42
        <BLANKLINE>
        
        >>> # compute \'single_period_subset[name] ** series[name]\' for each name
        >>> series = pd.Series([1.0, 2.0, 0.5, 1./4., 2.0], index=vars_subset.names)
        >>> series                                          # doctest: +NORMALIZE_WHITESPACE
        ACAF     1.00
        ACAG     2.00
        AOUC     0.50
        AOUC_    0.25
        AQC      2.00
        dtype: float64
        >>> vars_subset_1995Y1 = vars_subset[:, "1995Y1"] ** series
        >>> vars_subset_1995Y1                              # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1995Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name\t 1995Y1
        ACAF \t -13.13
        ACAG \t1751.09
        AOUC \t   1.02
        AOUC_\t   1.00
        AQC  \t   1.35
        <BLANKLINE>

        >>> # compute \'iode_var[name, period] ** df[name, period]\' for each name and period
        >>> data = np.array([[1.0, 2.0, 0.5, 1./4., 2.0],
        ...                  [2.0, -1.0, 2.0, -1.0, 2.0], 
        ...                  [1./4., 2.0, 1.0, 2.0, 0.5],
        ...                  [0.5, 1./4., 2.0, 1.0, 2.0],
        ...                  [2.0, 0.5, 1./4., 2.0, 1.0]])
        >>> df = pd.DataFrame(data, index=vars_subset.names, columns=vars_subset.periods_as_str) 
        >>> df                                              # doctest: +NORMALIZE_WHITESPACE
              1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF\t1.00\t2.00\t0.50\t0.25\t 2.0
        ACAG\t2.00   -1.00\t2.00   -1.00\t 2.0
        AOUC\t0.25\t2.00\t1.00\t2.00\t 0.5
        AOUC_\t0.50\t0.25\t2.00\t1.00\t 2.0
        AQC\t    2.00\t0.50\t0.25\t2.00\t 1.0
        >>> new_vars_subset = vars_subset ** df
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1   1993Y1  1994Y1   1995Y1
        ACAF \t     26.24\t909.57\t   5.89\t   1.69\t  172.42
        ACAG     \t956.91\t -0.02\t1862.61\t  -0.06\t 1751.09
        AOUC \t      1.01\t  1.06\t   1.03\t   1.09\t    1.02
        AOUC_\t      0.98\t  0.99\t   0.96\t   0.99\t    0.99
        AQC  \t      1.13\t  1.05\t   1.04\t   1.34\t    1.16
        <BLANKLINE>

        >>> # compute \'iode_var[name, period] ** array[name, period]\' for each name and period
        >>> axis_names = la.Axis(name="names", labels=vars_subset.names)
        >>> axis_time = la.Axis(name="time", labels=vars_subset.periods_as_str)
        >>> array = la.Array(data, axes=(axis_names, axis_time))
        >>> array                                           # doctest: +NORMALIZE_WHITESPACE
        names\\time  1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
              ACAF     1.0     2.0     0.5    0.25     2.0
              ACAG     2.0    -1.0     2.0    -1.0     2.0
              AOUC    0.25     2.0     1.0     2.0     0.5
             AOUC_     0.5    0.25     2.0     1.0     2.0
               AQC     2.0     0.5    0.25     2.0     1.0
        >>> new_vars_subset = vars_subset ** array
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1   1993Y1  1994Y1  1995Y1
        ACAF \t     26.24\t909.57\t   5.89\t  1.69\t 172.42
        ACAG \t    956.91\t -0.02\t1862.61\t -0.06\t1751.09
        AOUC \t      1.01\t  1.06\t   1.03\t  1.09\t   1.02
        AOUC_\t      0.98\t  0.99\t   0.96\t  0.99\t   0.99
        AQC  \t      1.13\t  1.05\t   1.04\t  1.34\t   1.16
        <BLANKLINE>        

        >>> # WARNING: using a numpy ndarray to a (subset of a) Variables object is not recommended 
        >>> #          as there is no compatibility check between for the names and periods.
        >>> #          The result is not guaranteed to be the one you expected.
        >>> #          This possibility is provided for speed reasons 
        >>> #          (when dealing with large subsets/databases).
        >>> # compute \'iode_var[period] ** data[t]\' for each period
        >>> data = np.array([1.0, 2.0, 0.5, 1./4., 2.0])
        >>> updated_ACAF = vars_subset["ACAF"] ** data
        >>> updated_ACAF                                    # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1  1993Y1  1994Y1  1995Y1
        ACAF\t     26.24\t909.57\t  5.89\t  1.69\t172.42
        <BLANKLINE>
        >>> # compute \'single_period_subset[name] ** data[i]\' for each name
        >>> vars_subset_1995Y1 = vars_subset[:, "1995Y1"] ** data
        >>> vars_subset_1995Y1                              # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1995Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1995Y1
        ACAF \t    -13.13
        ACAG       1751.09
        AOUC \t      1.02
        AOUC_\t      1.00
        AQC  \t      1.35
        <BLANKLINE>        
        >>> # compute \'iode_var[name, period] ** data[i, t]\' for each name and period
        >>> data = np.array([[1.0, 2.0, 0.5, 1./4., 2.0],
        ...                  [2.0, -1.0, 2.0, -1.0, 2.0], 
        ...                  [1./4., 2.0, 1.0, 2.0, 0.5],
        ...                  [0.5, 1./4., 2.0, 1.0, 2.0],
        ...                  [2.0, 0.5, 1./4., 2.0, 1.0]])
        >>> new_vars_subset = vars_subset ** data
        >>> new_vars_subset                                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1991Y1:1995Y1
        mode: LEVEL
        <BLANKLINE>
         name       1991Y1  1992Y1   1993Y1  1994Y1   1995Y1
        ACAF    \t 26.24\t909.57\t   5.89\t   1.69\t  172.42
        ACAG    \t956.91\t -0.02\t1862.61\t  -0.06\t 1751.09
        AOUC \t      1.01\t  1.06\t   1.03\t   1.09\t    1.02
        AOUC_\t      0.98\t  0.99\t   0.96\t   0.99\t    0.99
        AQC  \t      1.13\t  1.05\t   1.04\t   1.34\t    1.16
        <BLANKLINE>
        '''
    def from_numpy(self, data: np.ndarray, vars_names: str | list[str] = None, first_period: str | Period = None, last_period: str | Period = None):
        '''
        Copy the numpy ndarray `array` into the IODE Variables database.
        A row of the ndarray represents a variable.
        A column of the ndarray represents a period.

        Parameters
        ----------
        data: numpy ndarray
            Numpy ndarray containing the variables values to copy into the 
            IODE Variables database. If the ndarray is a 1D array, either 
            `var_names` must represent a single variable or `first_period` 
            must be equal to `last_period`.
        vars_names: str or list of str, optional
            Names of the variables to copy into the IODE Variables database.
            Default to all variables names found in the present database.
        first_period: str or Period, optional
            First period of the values to copy into the IODE Variables database. 
            Default to the first period of the present database.
        last_period: str or Period, optional
            Last period of the values to copy into the IODE Variables database. 
            Default to the last period of the present database.

        Warnings
        --------
        IODE and pandas don\'t use the same constant to represent NaN values.
        When loading a pandas DataFrame into the Variables database, the pandas 
        NaN values (:math:`nan`) are converted to IODE NaN values (:math:`NA`).

        See Also
        --------
        Variables.from_frame
        Variables.from_array

        Examples
        --------
        >>> from iode import variables, SAMPLE_DATA_DIR, Sample
        >>> import numpy as np
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded  
        >>> # define the list of Variables to override, the first and last periods to copy
        >>> vars_names = variables.get_names("A*")
        >>> vars_names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> first_period = "2000Y1"
        >>> last_period = "2010Y1"
        >>> sample = Sample(first_period, last_period)
        >>> nb_periods = sample.nb_periods
        >>> nb_periods
        11
        >>> # save original values to restore them later
        >>> original_values = variables["A*", "2000Y1:2010Y1"].to_numpy()
        >>> # create the numpy ndarray containing the values to copy into the Variables database
        >>> data = np.zeros((len(vars_names), nb_periods), dtype=float)
        >>> for i in range(len(vars_names)):
        ...     for j in range(nb_periods):
        ...         data[i, j] = i * nb_periods + j
        >>> data
        array([[  0.,   1.,   2.,   3.,   4.,   5.,   6.,   7.,   8.,   9.,  10.],
               [ 11.,  12.,  13.,  14.,  15.,  16.,  17.,  18.,  19.,  20.,  21.],
               [ 22.,  23.,  24.,  25.,  26.,  27.,  28.,  29.,  30.,  31.,  32.],
               [ 33.,  34.,  35.,  36.,  37.,  38.,  39.,  40.,  41.,  42.,  43.],
               [ 44.,  45.,  46.,  47.,  48.,  49.,  50.,  51.,  52.,  53.,  54.]])
        
        >>> variables["A*", "2000Y1:2010Y1"]                # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 2000Y1:2010Y1
        mode: LEVEL
        <BLANKLINE>
         name       2000Y1  2001Y1  2002Y1  2003Y1  2004Y1  2005Y1  2006Y1  2007Y1  2008Y1  2009Y1  2010Y1
        ACAF         10.05    2.87   -0.93   -6.09  -14.58  -26.54  -28.99  -33.38  -38.41  -37.46  -37.83
        ACAG        -41.53   18.94   19.98   21.02   22.07   23.11   24.13   25.16   26.19   27.23   28.25
        AOUC          1.12    1.14    1.16    1.17    1.17    1.18    1.20    1.22    1.26    1.29    1.31
        AOUC_         1.10    1.14    1.15    1.16    1.15    1.16    1.19    1.20    1.21    1.23    1.25
        AQC           1.34    1.38    1.41    1.42    1.40    1.40    1.40    1.41    1.43    1.45    1.46
        <BLANKLINE>
        >>> # copy the numpy ndarray into the Variables database (overriding the existing values)
        >>> variables.from_numpy(data, vars_names, first_period, last_period)
        >>> variables["A*", "2000Y1:2010Y1"]                # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 2000Y1:2010Y1
        mode: LEVEL
        <BLANKLINE>
         name       2000Y1  2001Y1  2002Y1  2003Y1  2004Y1  2005Y1  2006Y1  2007Y1  2008Y1  2009Y1  2010Y1
        ACAF          0.00    1.00    2.00    3.00    4.00    5.00    6.00    7.00    8.00    9.00   10.00
        ACAG         11.00   12.00   13.00   14.00   15.00   16.00   17.00   18.00   19.00   20.00   21.00
        AOUC         22.00   23.00   24.00   25.00   26.00   27.00   28.00   29.00   30.00   31.00   32.00
        AOUC_        33.00   34.00   35.00   36.00   37.00   38.00   39.00   40.00   41.00   42.00   43.00
        AQC          44.00   45.00   46.00   47.00   48.00   49.00   50.00   51.00   52.00   53.00   54.00
        <BLANKLINE>

        >>> # if a subset represents all values to be updated, the values for the arguments 
        >>> # vars_names, first_period and last_period can be omitted
        >>> vars_subset = variables["A*", "2000Y1:2010Y1"]
        >>> vars_subset.from_numpy(original_values)
        >>> vars_subset                # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 2000Y1:2010Y1
        mode: LEVEL
        <BLANKLINE>
         name       2000Y1  2001Y1  2002Y1  2003Y1  2004Y1  2005Y1  2006Y1  2007Y1  2008Y1  2009Y1  2010Y1
        ACAF         10.05    2.87   -0.93   -6.09  -14.58  -26.54  -28.99  -33.38  -38.41  -37.46  -37.83
        ACAG        -41.53   18.94   19.98   21.02   22.07   23.11   24.13   25.16   26.19   27.23   28.25
        AOUC          1.12    1.14    1.16    1.17    1.17    1.18    1.20    1.22    1.26    1.29    1.31
        AOUC_         1.10    1.14    1.15    1.16    1.15    1.16    1.19    1.20    1.21    1.23    1.25
        AQC           1.34    1.38    1.41    1.42    1.40    1.40    1.40    1.41    1.43    1.45    1.46
        <BLANKLINE>
        '''
    def to_numpy(self) -> np.ndarray:
        '''
        Create a Numpy ndarray from the current Variables database.

        Returns
        -------
        np.ndarray
            Numpy ndarray containing the variables values of the current Variables database or subset.

        Examples
        --------
        >>> from iode import variables, SAMPLE_DATA_DIR
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> len(variables)
        394
        >>> variables.sample
        Sample("1960Y1:2015Y1")
        >>> variables.nb_periods
        56

        >>> # export the whole Variables workspace to a numpy ndarray (394 variables x 56 periods)
        >>> data = variables.to_numpy()
        >>> data.shape
        (394, 56)
        >>> data[5, 40]
        442.26441085858613
        >>> variables.i[5, 40]
        442.26441085858613

        >>> # export a subset of names
        >>> vars_subset = variables["A*"]
        >>> vars_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> vars_subset                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
         name       1960Y1  1961Y1  1962Y1  1963Y1  ...  2013Y1  2014Y1  2015Y1
        ACAF            na      na      na      na  ...  -68.89  -83.34  -96.41
        ACAG            na      na      na      na  ...   31.37   32.42   33.47
        AOUC            na    0.25    0.25    0.26  ...    1.39    1.42    1.46
        AOUC_           na      na      na      na  ...    1.34    1.37    1.41
        AQC           0.22    0.22    0.22    0.23  ...    1.56    1.61    1.67
        <BLANKLINE>
        >>> data = vars_subset.to_numpy()
        >>> data.shape
        (5, 56)
        >>> data                        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        array([[            nan,          nan,          nan,          nan,
                                          ...
                   -55.55928982, -68.89465432, -83.34062511, -96.41041983],
                  [         nan,          nan,          nan,          nan,
                                          ...
                    30.32396115,  31.37013881,  32.42029883,  33.46960134],
                  [         nan,   0.24783192,   0.25456766,   0.26379573,
                                          ...
                     1.35553983,   1.38777697,   1.42371396,   1.46086261],
                  [         nan,          nan,          nan,          nan,
                                          ...
                     1.30459041,   1.33808573,   1.37301015,   1.4075568 ],
                  [  0.21753037,   0.21544869,   0.22228125,   0.22953896,
                                          ...
                     1.51366598,   1.55803879,   1.61318117,   1.67429058]])

        >>> # export a subset of names and periods
        >>> vars_subset = variables["A*", "2000Y1:2010Y1"]
        >>> vars_subset                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 2000Y1:2010Y1
        mode: LEVEL
        <BLANKLINE>
         name       2000Y1  2001Y1  2002Y1  2003Y1  ...  2007Y1  2008Y1  2009Y1  2010Y1
        ACAF         10.05    2.87   -0.93   -6.09  ...  -33.38  -38.41  -37.46  -37.83
        ACAG        -41.53   18.94   19.98   21.02  ...   25.16   26.19   27.23   28.25
        AOUC          1.12    1.14    1.16    1.17  ...    1.22    1.26    1.29    1.31
        AOUC_         1.10    1.14    1.15    1.16  ...    1.20    1.21    1.23    1.25
        AQC           1.34    1.38    1.41    1.42  ...    1.41    1.43    1.45    1.46
        <BLANKLINE>
        >>> data = vars_subset.to_numpy()
        >>> data.shape
        (5, 11)
        >>> data                        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        array([[ 10.04661079,   2.86792274,  -0.92921251,  -6.09156499,
                -14.58209446, -26.53878957, -28.98728798, -33.37842578,
                -38.40951778, -37.46350964, -37.82742883],
               [-41.53478657,  18.93980114,  19.98081488,  21.02050218,
                 22.06647552,  23.10796216,  24.12963715,  25.16090905,
                 26.19211148,  27.22995512,  28.25392898],
               [  1.11623762,   1.14047639,   1.15716928,   1.17048954,
                  1.16767464,   1.1815207 ,   1.19946163,   1.21933288,
                  1.26280574,   1.28713178,   1.3071099 ],
               [  1.1019572 ,   1.13624426,   1.15021519,   1.16082895,
                  1.14802147,   1.16412337,   1.18589708,   1.19516611,
                  1.21383423,   1.23185399,   1.25016433],
               [  1.33860286,   1.37918825,   1.40881647,   1.41970458,
                  1.40065206,   1.39697298,   1.39806354,   1.40791334,
                  1.42564488,   1.44633167,   1.46286837]])
        '''
    def __array__(self, dtype: Incomplete | None = None):
        '''
        Return a Numpy ndarray from the current Variables database.

        Returns
        -------
        np.ndarray
            Numpy ndarray containing the variables values of the current Variables database or subset.

        Examples
        --------
        >>> import numpy as np
        >>> from iode import variables, SAMPLE_DATA_DIR
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> len(variables)
        394
        >>> variables.sample
        Sample("1960Y1:2015Y1")
        >>> variables.nb_periods
        56

        >>> # export the whole Variables workspace to a numpy ndarray (394 variables x 56 periods)
        >>> data = np.asarray(variables)
        >>> data.shape
        (394, 56)
        >>> data[5, 40]
        442.26441085858613
        >>> variables.i[5, 40]
        442.26441085858613

        >>> # export a subset of names
        >>> vars_subset = variables["A*"]
        >>> vars_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> vars_subset                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
         name       1960Y1  1961Y1  1962Y1  1963Y1  ...  2013Y1  2014Y1  2015Y1
        ACAF            na      na      na      na  ...  -68.89  -83.34  -96.41
        ACAG            na      na      na      na  ...   31.37   32.42   33.47
        AOUC            na    0.25    0.25    0.26  ...    1.39    1.42    1.46
        AOUC_           na      na      na      na  ...    1.34    1.37    1.41
        AQC           0.22    0.22    0.22    0.23  ...    1.56    1.61    1.67
        <BLANKLINE>
        >>> data = np.asarray(vars_subset)
        >>> data.shape
        (5, 56)
        >>> data                        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        array([[            nan,          nan,          nan,          nan,
                                          ...
                   -55.55928982, -68.89465432, -83.34062511, -96.41041983],
                  [         nan,          nan,          nan,          nan,
                                          ...
                    30.32396115,  31.37013881,  32.42029883,  33.46960134],
                  [         nan,   0.24783192,   0.25456766,   0.26379573,
                                          ...
                     1.35553983,   1.38777697,   1.42371396,   1.46086261],
                  [         nan,          nan,          nan,          nan,
                                          ...
                     1.30459041,   1.33808573,   1.37301015,   1.4075568 ],
                  [  0.21753037,   0.21544869,   0.22228125,   0.22953896,
                                          ...
                     1.51366598,   1.55803879,   1.61318117,   1.67429058]])

        >>> # export a subset of names and periods
        >>> vars_subset = variables["A*", "2000Y1:2010Y1"]
        >>> vars_subset                 # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 5
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 2000Y1:2010Y1
        mode: LEVEL
        <BLANKLINE>
         name       2000Y1  2001Y1  2002Y1  2003Y1  ...  2007Y1  2008Y1  2009Y1  2010Y1
        ACAF         10.05    2.87   -0.93   -6.09  ...  -33.38  -38.41  -37.46  -37.83
        ACAG        -41.53   18.94   19.98   21.02  ...   25.16   26.19   27.23   28.25
        AOUC          1.12    1.14    1.16    1.17  ...    1.22    1.26    1.29    1.31
        AOUC_         1.10    1.14    1.15    1.16  ...    1.20    1.21    1.23    1.25
        AQC           1.34    1.38    1.41    1.42  ...    1.41    1.43    1.45    1.46
        <BLANKLINE>
        >>> data = np.asarray(vars_subset)
        >>> data.shape
        (5, 11)
        >>> data                        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        array([[ 10.04661079,   2.86792274,  -0.92921251,  -6.09156499,
                -14.58209446, -26.53878957, -28.98728798, -33.37842578,
                -38.40951778, -37.46350964, -37.82742883],
               [-41.53478657,  18.93980114,  19.98081488,  21.02050218,
                 22.06647552,  23.10796216,  24.12963715,  25.16090905,
                 26.19211148,  27.22995512,  28.25392898],
               [  1.11623762,   1.14047639,   1.15716928,   1.17048954,
                  1.16767464,   1.1815207 ,   1.19946163,   1.21933288,
                  1.26280574,   1.28713178,   1.3071099 ],
               [  1.1019572 ,   1.13624426,   1.15021519,   1.16082895,
                  1.14802147,   1.16412337,   1.18589708,   1.19516611,
                  1.21383423,   1.23185399,   1.25016433],
               [  1.33860286,   1.37918825,   1.40881647,   1.41970458,
                  1.40065206,   1.39697298,   1.39806354,   1.40791334,
                  1.42564488,   1.44633167,   1.46286837]])
        '''
    def from_frame(self, df: pd.DataFrame):
        '''
        Copy the pandas DataFrame `df` into the IODE Variables database.
        The variable names to copy are deduced from the index of the DataFrame.
        The column names must match the sub-periods of the current Variables sample.

        Parameters
        ----------
        df: DataFrame
            pandas DataFrame containing the variables to copy into the IODE Variables database.

        Notes
        -----
        The index of the passed DataFrame is sorted in alphabetical order before 
        copying to IODE Variables database.

        Warnings
        --------
        IODE and pandas don\'t use the same constant to represent NaN values.
        When loading a pandas DataFrame into the Variables database, the pandas 
        NaN values (:math:`nan`) are converted to IODE NaN values (:math:`NA`).

        See Also
        --------
        Variables.to_frame

        Examples
        --------
        >>> from iode import variables
        >>> import numpy as np
        >>> import pandas as pd
        >>> variables.clear()
        >>> len(variables)
        0

        >>> # create the pandas DataFrame
        >>> vars_names = [f"{region}_{code}" for region in ["VLA", "WAL", "BXL"] for code in ["00", "01", "02"]]
        >>> periods_list = [f"{i}Y1" for i in range(1960, 1971)]
        >>> nb_periods = len(periods_list)
        >>> data = np.zeros((len(vars_names), nb_periods), dtype=float)
        >>> for i in range(len(vars_names)):
        ...     for j in range(nb_periods):
        ...         data[i, j] = i * nb_periods + j
        >>> df = pd.DataFrame(index=vars_names, columns=periods_list, data=data)
        >>> # display the dataframe
        >>> df          # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                1960Y1  1961Y1  1962Y1  1963Y1  ...  1967Y1  1968Y1  1969Y1  1970Y1
        VLA_00     0.0     1.0     2.0     3.0  ...     7.0     8.0     9.0    10.0
        VLA_01    11.0    12.0    13.0    14.0  ...    18.0    19.0    20.0    21.0
        VLA_02    22.0    23.0    24.0    25.0  ...    29.0    30.0    31.0    32.0
        WAL_00    33.0    34.0    35.0    36.0  ...    40.0    41.0    42.0    43.0
        WAL_01    44.0    45.0    46.0    47.0  ...    51.0    52.0    53.0    54.0
        WAL_02    55.0    56.0    57.0    58.0  ...    62.0    63.0    64.0    65.0
        BXL_00    66.0    67.0    68.0    69.0  ...    73.0    74.0    75.0    76.0
        BXL_01    77.0    78.0    79.0    80.0  ...    84.0    85.0    86.0    87.0
        BXL_02    88.0    89.0    90.0    91.0  ...    95.0    96.0    97.0    98.0
        <BLANKLINE>
        [9 rows x 11 columns]

        >>> # load into the IODE Variables database
        >>> variables.from_frame(df)
        >>> len(variables)
        9
        >>> variables.names
        [\'BXL_00\', \'BXL_01\', \'BXL_02\', \'VLA_00\', \'VLA_01\', \'VLA_02\', \'WAL_00\', \'WAL_01\', \'WAL_02\']
        >>> variables.sample
        Sample("1960Y1:1970Y1")
        >>> variables["VLA_00"]         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 1960Y1:1970Y1
        mode: LEVEL
        <BLANKLINE>
         name       1960Y1  1961Y1  1962Y1  ...  1968Y1  1969Y1  1970Y1
        VLA_00        0.00    1.00    2.00  ...    8.00    9.00   10.00
        <BLANKLINE>
        >>> variables["BXL_02"]         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 1960Y1:1970Y1
        mode: LEVEL
        <BLANKLINE>
         name       1960Y1  1961Y1  1962Y1  ...  1968Y1  1969Y1  1970Y1
        BXL_02       88.00   89.00   90.00  ...   96.00   97.00   98.00
        <BLANKLINE>
        '''
    def to_frame(self, vars_axis_name: str = 'names', time_axis_name: str = 'time', periods_as_type: str | type = ...) -> pd.DataFrame:
        '''
        Create a pandas DataFrame from the current Variables database.
        The index of the returned DataFrame is build from the Variables names 
        and the columns from the periods.

        Parameters
        ----------
        vars_axis_name: str, optional
            Name of the axis representing the Variables names.
            Defaults to \'names\'.
        time_axis_name: str, optional
            Name of the axis representing the periods.
            Defaults to \'time\'.
        periods_as_type: str or type, optional
            Type in which periods are converted. 
            Possible values are: str, int or float.
            Defaults to str.

        Warnings
        --------
        IODE and pandas don\'t use the same constant to represent NaN values.
        When exporting IODE variables as a pandas DataFrame, the IODE NaN values 
        (:math:`NA`) are converted to pandas NaN values (:math:`nan`).

        See Also
        --------
        Variables.from_frame

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables
        >>> import pandas as pd
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded

        >>> # Export the IODE Variables database as a pandas DataFrame
        >>> variables                       # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 394
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
         name       1960Y1  1961Y1  1962Y1  1963Y1  ...  2013Y1  2014Y1  2015Y1
        ACAF            na      na      na      na  ...  -68.89  -83.34  -96.41
        ACAG            na      na      na      na  ...   31.37   32.42   33.47
        AOUC            na    0.25    0.25    0.26  ...    1.39    1.42    1.46
        AOUC_           na      na      na      na  ...    1.34    1.37    1.41
        AQC           0.22    0.22    0.22    0.23  ...    1.56    1.61    1.67
        ...            ...     ...     ...     ...  ...     ...     ...     ...
        ZJ              na      na      na      na  ...    1.59    1.63    1.67
        ZKF           0.80    0.81    0.82    0.81  ...    0.87    0.87    0.87
        ZKFO          1.00    1.00    1.00    1.00  ...    1.02    1.02    1.02
        ZX            0.00    0.00    0.00    0.00  ...    0.00    0.00    0.00
        ZZF_          0.69    0.69    0.69    0.69  ...    0.69    0.69    0.69
        <BLANKLINE>
        >>> df = variables.to_frame()
        >>> df.shape
        (394, 56)
        >>> df.index.to_list()              # doctest: +ELLIPSIS
        [\'ACAF\', \'ACAG\', \'AOUC\', ..., \'ZKFO\', \'ZX\', \'ZZF_\']
        >>> df.columns.to_list()            # doctest: +ELLIPSIS
        [\'1960Y1\', \'1961Y1\', ..., \'2014Y1\', \'2015Y1\']
        >>> df                              # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        time     1960Y1    1961Y1    1962Y1  ...     2013Y1     2014Y1     2015Y1
        names                                ...
        ACAF        NaN       NaN       NaN  ... -68.894654 -83.340625 -96.410420
        ACAG        NaN       NaN       NaN  ...  31.370139  32.420299  33.469601
        AOUC        NaN  0.247832  0.254568  ...   1.387777   1.423714   1.460863
        AOUC_       NaN       NaN       NaN  ...   1.338086   1.373010   1.407557
        AQC    0.217530  0.215449  0.222281  ...   1.558039   1.613181   1.674291
        ...         ...       ...       ...  ...        ...        ...        ...
        ZJ          NaN       NaN       NaN  ...   1.591981   1.630309   1.667971
        ZKF    0.802574  0.812873  0.819252  ...   0.874883   0.874351   0.873593
        ZKFO   1.000000  1.000000  1.000000  ...   1.015990   1.015990   1.015990
        ZX     0.000000  0.000000  0.000000  ...   0.000000   0.000000   0.000000
        ZZF_   0.688400  0.688400  0.688400  ...   0.688400   0.688400   0.688400
        <BLANKLINE>
        [394 rows x 56 columns]

        >>> # Export a subset of the IODE Variables database as a pandas DataFrame
        >>> vars_subset = variables["A*;*_", "2000Y1:2010Y1"]
        >>> vars_subset                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 33
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 2000Y1:2010Y1
        mode: LEVEL
        <BLANKLINE>
         name        2000Y1  2001Y1  2002Y1  2003Y1 ...  2007Y1   2008Y1   2009Y1   2010Y1
        ACAF          10.05    2.87   -0.93   -6.09 ...  -33.38   -38.41   -37.46   -37.83
        ACAG         -41.53   18.94   19.98   21.02 ...   25.16    26.19    27.23    28.25
        AOUC           1.12    1.14    1.16    1.17 ...    1.22     1.26     1.29     1.31
        AOUC_          1.10    1.14    1.15    1.16 ...    1.20     1.21     1.23     1.25
        AQC            1.34    1.38    1.41    1.42 ...    1.41     1.43     1.45     1.46
        ...             ...     ...     ...     ... ...     ...      ...      ...      ...
        WCF_        3716.45 3863.90 3999.57 4147.95 ... 4665.91  4916.65  5042.74  5170.60
        WIND_       1000.14 1035.22 1070.93 1102.91 ... 1178.12  1231.49  1268.86  1301.03
        WNF_        2334.76 2427.49 2512.87 2606.28 ... 2932.23  3089.99  3169.32  3249.75
        YDH_        7276.61 7635.91 7958.39 8331.07 ... 9685.61 10228.84 10630.74 10995.83
        ZZF_           0.69    0.69    0.69    0.69 ...    0.69     0.69     0.69     0.69
        <BLANKLINE>
        >>> df = vars_subset.to_frame()
        >>> df.shape
        (33, 11)
        >>> df.index.to_list()              # doctest: +ELLIPSIS
        [\'ACAF\', \'ACAG\', \'AOUC\', ..., \'WNF_\', \'YDH_\', \'ZZF_\']
        >>> df.columns.to_list()            # doctest: +ELLIPSIS
        [\'2000Y1\', \'2001Y1\', \'2002Y1\', ..., \'2008Y1\', \'2009Y1\', \'2010Y1\']
        >>> df                              # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        time           2000Y1       2001Y1  ...        2009Y1        2010Y1
        names                               ...
        ACAF        10.046611     2.867923  ...    -37.463510    -37.827429
        ACAG       -41.534787    18.939801  ...     27.229955     28.253929
        AOUC         1.116238     1.140476  ...      1.287132      1.307110
        AOUC_        1.101957     1.136244  ...      1.231854      1.250164
        AQC          1.338603     1.379188  ...      1.446332      1.462868
        ...               ...          ...  ...           ...           ...
        WCF_      3716.447509  3863.897550  ...   5042.743118   5170.600010
        WIND_     1000.144577  1035.218800  ...   1268.861647   1301.025126
        WNF_      2334.763628  2427.492334  ...   3169.316544   3249.751702
        YDH_      7276.607740  7635.905667  ...  10630.736896  10995.831393
        ZZF_         0.688400     0.688400  ...      0.688400      0.688400
        <BLANKLINE>
        [33 rows x 11 columns]

        >>> # Export a subset of the IODE Variables database with periods as int
        >>> df = variables["A*;*_", "2000Y1:2010Y1"].to_frame(periods_as_type=int) 
        >>> df.columns.to_list()            # doctest: +ELLIPSIS
        [2000, 2001, 2002, 2003, ..., 2007, 2008, 2009, 2010]
        >>> df                              # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        time             2000         2001  ...          2009          2010
        names                               ...
        ACAF        10.046611     2.867923  ...    -37.463510    -37.827429
        ACAG       -41.534787    18.939801  ...     27.229955     28.253929
        AOUC         1.116238     1.140476  ...      1.287132      1.307110
        AOUC_        1.101957     1.136244  ...      1.231854      1.250164
        AQC          1.338603     1.379188  ...      1.446332      1.462868
        ...               ...          ...  ...           ...           ...
        WCF_      3716.447509  3863.897550  ...   5042.743118   5170.600010
        WIND_     1000.144577  1035.218800  ...   1268.861647   1301.025126
        WNF_      2334.763628  2427.492334  ...   3169.316544   3249.751702
        YDH_      7276.607740  7635.905667  ...  10630.736896  10995.831393
        ZZF_         0.688400     0.688400  ...      0.688400      0.688400
        <BLANKLINE>
        [33 rows x 11 columns]
        '''
    def from_array(self, array: Array, time_axis_name: str = 'time', sep: str = '_'):
        '''
        Copies the Array array into the IODE Variables database.
        
        If the passed array has more than two dimensions, the non-time axes are grouped 
        (with \'sep\' as separator) to form the Variables names.

        Parameters
        ----------
        array: Array
            Input Array object.
        time_axis_name: str, optional
            Name of the time axis in array.
            Assumed to be \'time\' by default.
        sep: str, optional
            If the \'array\' as more than two axes, the separator 
            \'sep\' is used to group labels of the non-time axes.

        Warnings
        --------
        IODE and LArray don\'t use the same constant to represent NaN values.
        When loading a LArray Array into the Variables database, the LArray 
        NaN values (:math:`nan`) are converted to IODE NaN values (:math:`NA`).

        Examples
        --------
        >>> from iode import variables
        >>> import larray as la
        >>> import numpy as np
        >>> variables.clear()
        >>> len(variables)
        0

        >>> regions_axis = la.Axis("region=VLA,WAL,BXL")
        >>> code_axis = la.Axis("code=00..02")
        >>> periods_axis = la.Axis("time=1960Y1..1970Y1")
        >>> array = la.ndtest((regions_axis, code_axis, periods_axis), dtype=float)
        >>> array
        region  code\\time  1960Y1  1961Y1  1962Y1  ...  1968Y1  1969Y1  1970Y1
           VLA         00     0.0     1.0     2.0  ...     8.0     9.0    10.0
           VLA         01    11.0    12.0    13.0  ...    19.0    20.0    21.0
           VLA         02    22.0    23.0    24.0  ...    30.0    31.0    32.0
           WAL         00    33.0    34.0    35.0  ...    41.0    42.0    43.0
           WAL         01    44.0    45.0    46.0  ...    52.0    53.0    54.0
           WAL         02    55.0    56.0    57.0  ...    63.0    64.0    65.0
           BXL         00    66.0    67.0    68.0  ...    74.0    75.0    76.0
           BXL         01    77.0    78.0    79.0  ...    85.0    86.0    87.0
           BXL         02    88.0    89.0    90.0  ...    96.0    97.0    98.0
        
        >>> # load the IODE Variables from the Array object
        >>> variables.from_array(array)
        >>> len(variables)
        9
        >>> variables.names
        [\'BXL_00\', \'BXL_01\', \'BXL_02\', \'VLA_00\', \'VLA_01\', \'VLA_02\', \'WAL_00\', \'WAL_01\', \'WAL_02\']
        >>> variables.sample
        Sample("1960Y1:1970Y1")
        >>> variables["VLA_00"]         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 1960Y1:1970Y1
        mode: LEVEL
        <BLANKLINE>
         name       1960Y1  1961Y1  1962Y1  ...  1968Y1  1969Y1  1970Y1
        VLA_00        0.00    1.00    2.00  ...    8.00    9.00   10.00
        <BLANKLINE>
        >>> variables["BXL_02"]         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 1960Y1:1970Y1
        mode: LEVEL
        <BLANKLINE>
         name       1960Y1  1961Y1  1962Y1  ...  1968Y1  1969Y1  1970Y1
        BXL_02       88.00   89.00   90.00  ...   96.00   97.00   98.00
        <BLANKLINE>
        '''
    def to_array(self, vars_axis_name: str = 'names', time_axis_name: str = 'time', periods_as_type: str | type = ...) -> Array:
        '''
        Creates an Array from the current IODE Variables database.
        
        Parameters
        ----------
        vars_axis_name: str, optional
            Name of the axis representing the Variables names.
            Defaults to \'names\'.
        time_axis_name: str, optional
            Name of the axis representing the periods.
            Defaults to \'time\'.
        periods_as_type: str or type, optional
            Type in which periods are converted. 
            Possible values are: str, int or float.
            Defaults to str.

        Warnings
        --------
        IODE and LArray don\'t use the same constant to represent NaN values.
        When exporting IODE variables as a LArray Array, the IODE NaN values 
        (:math:`NA`) are converted to LArray NaN values (:math:`nan`).

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded

        >>> # Export the IODE Variables database as an (larray) Array object
        >>> variables                       # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 394
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
         name       1960Y1  1961Y1  1962Y1  1963Y1  ...  2013Y1  2014Y1   2015Y1
        ACAF            na      na      na      na  ...  -68.89  -83.34   -96.41
        ACAG            na      na      na      na  ...   31.37   32.42    33.47
        AOUC            na    0.25    0.25    0.26  ...    1.39    1.42     1.46
        AOUC_           na      na      na      na  ...    1.34    1.37     1.41
        AQC           0.22    0.22    0.22    0.23  ...    1.56    1.61     1.67
        ...            ...     ...     ...     ...  ...     ...     ...      ...
        ZJ              na      na      na      na  ...    1.59    1.63     1.67
        ZKF           0.80    0.81    0.82    0.81  ...    0.87    0.87     0.87
        ZKFO          1.00    1.00    1.00    1.00  ...    1.02    1.02     1.02
        ZX            0.00    0.00    0.00    0.00  ...    0.00    0.00     0.00
        ZZF_          0.69    0.69    0.69    0.69  ...    0.69    0.69     0.69
        <BLANKLINE>
        >>> array = variables.to_array()
        >>> array.shape
        (394, 56)
        >>> array.axes.info
        394 x 56
         names [394]: \'ACAF\' \'ACAG\' \'AOUC\' ... \'ZKFO\' \'ZX\' \'ZZF_\'
         time [56]: \'1960Y1\' \'1961Y1\' \'1962Y1\' ... \'2013Y1\' \'2014Y1\' \'2015Y1\'
        >>> array                           # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        names\\time      1960Y1  ...              2015Y1
               ACAF         nan  ...  -96.41041982848331
               ACAG         nan  ...   33.46960134488098
               AOUC         nan  ...  1.4608626117037322
              AOUC_         nan  ...  1.4075567973393608
                AQC  0.21753037  ...  1.6742905757021305
                ...         ...  ...                 ...
                 ZJ         nan  ...  1.6679707618363606
                ZKF  0.80257398  ...  0.8735925955073036
               ZKFO         1.0  ...           1.0159901
                 ZX         0.0  ...                 0.0
               ZZF_  0.68840039  ...          0.68840039

        >>> # Export a subset of the IODE Variables database as an (larray) Array object
        >>> vars_subset = variables["A*;*_", "2000Y1:2010Y1"]
        >>> vars_subset                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 33
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 2000Y1:2010Y1
        mode: LEVEL
        <BLANKLINE>
         name        2000Y1  2001Y1  2002Y1  2003Y1 ...  2007Y1   2008Y1   2009Y1   2010Y1
        ACAF          10.05    2.87   -0.93   -6.09 ...  -33.38   -38.41   -37.46   -37.83
        ACAG         -41.53   18.94   19.98   21.02 ...   25.16    26.19    27.23    28.25
        AOUC           1.12    1.14    1.16    1.17 ...    1.22     1.26     1.29     1.31
        AOUC_          1.10    1.14    1.15    1.16 ...    1.20     1.21     1.23     1.25
        AQC            1.34    1.38    1.41    1.42 ...    1.41     1.43     1.45     1.46
        ...             ...     ...     ...     ... ...     ...      ...      ...      ...
        WCF_        3716.45 3863.90 3999.57 4147.95 ... 4665.91  4916.65  5042.74  5170.60
        WIND_       1000.14 1035.22 1070.93 1102.91 ... 1178.12  1231.49  1268.86  1301.03
        WNF_        2334.76 2427.49 2512.87 2606.28 ... 2932.23  3089.99  3169.32  3249.75
        YDH_        7276.61 7635.91 7958.39 8331.07 ... 9685.61 10228.84 10630.74 10995.83
        ZZF_           0.69    0.69    0.69    0.69 ...    0.69     0.69     0.69     0.69
        <BLANKLINE>
        >>> array = vars_subset.to_array()
        >>> array.shape
        (33, 11)
        >>> array.axes.info
        33 x 11
         names [33]: \'ACAF\' \'ACAG\' \'AOUC\' ... \'WNF_\' \'YDH_\' \'ZZF_\'
         time [11]: \'2000Y1\' \'2001Y1\' \'2002Y1\' ... \'2008Y1\' \'2009Y1\' \'2010Y1\'
        >>> array                           # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        names\\time               2000Y1  ...               2010Y1
               ACAF   10.046610792200543  ...   -37.82742883229439
               ACAG   -41.53478656734795  ...   28.253928978210485
               AOUC   1.1162376230972206  ...   1.3071099004906368
              AOUC_     1.10195719812178  ...   1.2501643331956398
                AQC   1.3386028553645442  ...   1.4628683697450802
                ...                 ...  ...                 ...    
               WCF_   3716.4475089520292  ...    5170.600010384268
              WIND_   1000.1445769794319  ...    1301.025126372868
               WNF_   2334.7636275081923  ...   3249.7517024908175
               YDH_    7276.607740221424  ...   10995.831392939246
               ZZF_           0.68840039  ...           0.68840039

        >>> # Export a subset of the IODE Variables database as an (larray) Array object
        >>> # with periods as int
        >>> array = variables["A*;*_", "2000Y1:2010Y1"].to_array(periods_as_type=int)
        >>> array.axes.info
        33 x 11
         names [33]: \'ACAF\' \'ACAG\' \'AOUC\' ... \'WNF_\' \'YDH_\' \'ZZF_\'
         time [11]: 2000 2001 2002 ... 2008 2009 2010
        >>> array                           # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        names\\time                 2000  ...                 2010
              ACAF   10.046610792200543  ...   -37.82742883229439
              ACAG   -41.53478656734795  ...   28.253928978210485
              AOUC   1.1162376230972206  ...   1.3071099004906368
             AOUC_     1.10195719812178  ...   1.2501643331956398
               AQC   1.3386028553645442  ...   1.4628683697450802
               ...                 ...  ...                 ...    
              WCF_   3716.4475089520292  ...    5170.600010384268
             WIND_   1000.1445769794319  ...    1301.025126372868
              WNF_   2334.7636275081923  ...   3249.7517024908175
              YDH_    7276.607740221424  ...   10995.831392939246
              ZZF_           0.68840039  ...           0.68840039
        '''
    @property
    def mode(self) -> str:
        '''
        Current display mode for the IODE Variables values.
        The default mode is \'LEVEL\'.

        Parameters
        ----------
        value: VarsMode or str
            New mode value.
            Possible modes are LEVEL, DIFF, GROWTH_RATE, Y0Y_DIFF or Y0Y_GROWTH_RATE.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR, variables, VarsMode
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> variables.mode
        \'LEVEL\'
        >>> variables["ACAF", "1990Y1"]
        23.771
        
        >>> variables.mode = VarsMode.GROWTH_RATE
        >>> variables.mode
        \'GROWTH_RATE\'
        >>> variables["ACAF", "1990Y1"]
        38.48528176607737
        >>> variables.mode = "level"
        >>> variables.mode
        \'LEVEL\'
        >>> variables["ACAF", "1990Y1"]
        23.771
        '''
    @mode.setter
    def mode(self, value: VarsMode | str): ...
    @property
    def first_period(self) -> Period:
        """
        First period of the current Variables database.
        """
    @property
    def last_period(self) -> Period:
        """
        Last period of the current Variables database.
        """
    @property
    def sample(self) -> Sample:
        '''
        Current (or new) sample of the IODE Variables database.

        If a new sample is given, two cases are possible:
          
          - the sample is shorter than the current sample: the data beyond the new sample is destroyed,
          - the sample is longer than the current sample: the value `NA` (not available) is set for the added periods.

        Parameters
        ----------
        value: str or tuple(str, str)
            New sample as string \'start_period:last_period\' or as a tuple \'start_period\', \'last_period\'.
        
        Warnings
        --------
        Changing the sample on a subset of the Variables workspace is not allowed unless the 
        :py:meth:`~iode.Variables.copy` method has been used (in that case, any change made on the subset 
        will not be reflected in the global Variables workspace).

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables
        >>> variables.clear()
        >>> variables.sample
        None

        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        
        >>> variables.sample
        Sample("1960Y1:2015Y1")
        >>> variables[\'ACAF\']                                   # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  ...  2014Y1   2015Y1
        ACAF            na      na  ...  -83.34   -96.41
        <BLANKLINE>

        >>> # -- update sample by passing a string
        >>> # case 1: new sample is shorter than the current sample
        >>> #         the data beyond the new sample is destroyed
        >>> variables.sample = \'1970Y1:2010Y1\'
        >>> variables.sample
        Sample("1970Y1:2010Y1")
        >>> variables[\'ACAF\']                                   # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1970Y1:2010Y1
        mode: LEVEL
        <BLANKLINE>
        name        1970Y1  1971Y1  ...  2009Y1   2010Y1
        ACAF          1.21    5.20  ...  -37.46   -37.83
        <BLANKLINE>
        >>> # case 2: new sample is longer than the current sample
        >>> #         the value NA (not available) is set for the added periods
        >>> variables.sample = \'1968Y1:2012Y1\'
        >>> variables.sample
        Sample("1968Y1:2012Y1")
        >>> variables[\'ACAF\']                                   # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1968Y1:2012Y1
        mode: LEVEL
        <BLANKLINE>
        name        1968Y1  1969Y1  1970Y1  ...  2010Y1 2011Y1   2012Y1
        ACAF            na      na    1.21  ...  -37.83     na       na
        <BLANKLINE>
        
        >>> # -- start period is optional -> the start period is kept as it is
        >>> variables.sample = \':2010Y1\'
        >>> variables.sample
        Sample("1968Y1:2010Y1")
        >>> variables[\'ACAF\']                                   # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1968Y1:2010Y1
        mode: LEVEL
        <BLANKLINE>
        name        1968Y1  1969Y1  1970Y1  ...  2009Y1   2010Y1
        ACAF            na      na    1.21  ...  -37.46   -37.83
        <BLANKLINE>
        >>> # -- end period is optional -> the end period is kept as it is
        >>> variables.sample = \'1970Y1:\'
        >>> variables.sample
        Sample("1970Y1:2010Y1")
        >>> variables[\'ACAF\']                                   # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1970Y1:2010Y1
        mode: LEVEL
        <BLANKLINE>
        name        1970Y1  1971Y1  ...  2009Y1   2010Y1
        ACAF          1.21    5.20  ...  -37.46   -37.83
        <BLANKLINE>

        >>> # update sample by passing a start period and 
        >>> # an end period separated by a comma
        >>> variables.sample = \'1968Y1\', \'2012Y1\'
        >>> variables.sample
        Sample("1968Y1:2012Y1")
        >>> variables[\'ACAF\']                                   # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1968Y1:2012Y1
        mode: LEVEL
        <BLANKLINE>
        name        1968Y1  1969Y1  1970Y1  ...  2010Y1 2011Y1   2012Y1
        ACAF            na      na    1.21  ...  -37.83     na       na
        <BLANKLINE>
        '''
    @sample.setter
    def sample(self, value: str | tuple[str | Period, str | Period]): ...
    @property
    def nb_periods(self) -> int:
        '''
        Return the number of periods from the current Variables sample.

        Returns
        -------
        int

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> variables.sample
        Sample("1960Y1:2015Y1")
        >>> variables.nb_periods
        56
        '''
    @property
    def periods(self) -> list[Period]:
        '''
        Return the list of periods from the current Variables sample.

        Returns
        -------
        list(Period)

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> variables.sample
        Sample("1960Y1:2015Y1")
        >>> variables.periods       # doctest: +ELLIPSIS 
        [Period("1960Y1"), Period("1961Y1"), ..., Period("2014Y1"), Period("2015Y1")]
        '''
    @property
    def periods_as_str(self) -> list[str]:
        '''
        Return the list of periods as string from the current Variables sample.

        Returns
        -------
        list(str)

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> variables.sample
        Sample("1960Y1:2015Y1")
        >>> variables.periods_as_str          # doctest: +ELLIPSIS 
        [\'1960Y1\', \'1961Y1\', ..., \'2014Y1\', \'2015Y1\']
        '''
    @property
    def periods_as_float(self) -> list[float]:
        '''
        Return the list of periods as float from the current Variables sample.

        Returns
        -------
        list(str)

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> variables.sample
        Sample("1960Y1:2015Y1")
        >>> variables.periods_as_float          # doctest: +ELLIPSIS 
        [1960.0, 1961.0, ..., 2014.0, 2015.0]
        '''
    @property
    def threshold(self) -> float:
        '''
        Threshold under which the difference between 2 variables are considered equal.
        
        The comparison test is:
        .. math::

           \\begin{cases} 
           \\frac{x_1 - x_2}{x_1} < \\epsilon, & \\text{if } x_1 \\neq 0 \\\\\n           |x_2| < \\epsilon, & \\text{otherwise}
           \\end{cases}

        where :math:`x_1` and :math:`x_2` are the values of the variables to compare 
        and :math:`\\epsilon` is the threshold value.

        Parameters
        ----------
        value: float
        \tNew threshold value.
            Default value is 1e-7.

        See Also
        --------
        iode.Variables.compare

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> variables.threshold
        1e-07
        >>> variables.threshold = 1e-5
        >>> variables.threshold
        1e-05
        '''
    @threshold.setter
    def threshold(self, value: float): ...
    @property
    def df(self) -> pd.DataFrame:
        '''
        Create a pandas DataFrame from the current Variables database.
        The index of the returned DataFrame is build from the Variables names 
        and the columns from the periods.

        Warnings
        --------
        IODE and pandas don\'t use the same constant to represent NaN values.
        When exporting IODE variables as a pandas DataFrame, the IODE NaN values 
        (:math:`NA`) are converted to pandas NaN values (:math:`nan`).

        See Also
        --------
        Variables.to_frame

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables
        >>> import pandas as pd
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> len(variables)
        394
        >>> variables.sample
        Sample("1960Y1:2015Y1")
        >>> variables.nb_periods
        56

        >>> # Export the IODE Variables database as a pandas DataFrame
        >>> df = variables.df
        >>> df.shape
        (394, 56)
        >>> df.index.to_list()              # doctest: +ELLIPSIS
        [\'ACAF\', \'ACAG\', \'AOUC\', ..., \'ZKFO\', \'ZX\', \'ZZF_\']
        >>> df.columns.to_list()            # doctest: +ELLIPSIS
        [\'1960Y1\', \'1961Y1\', ..., \'2014Y1\', \'2015Y1\']
        >>> variables["AOUC"]               # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  ...  2014Y1   2015Y1
        AOUC            na    0.25  ...    1.42     1.46
        <BLANKLINE>
        >>> df.loc["AOUC"]                  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        time
        1960Y1         NaN
        1961Y1    0.247832
        ...
        2014Y1    1.423714
        2015Y1    1.460863
        Name: AOUC, dtype: float64    
        >>> variables["ZKFO"]               # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  ...  2014Y1   2015Y1
        ZKFO          1.00    1.00  ...    1.02     1.02
        <BLANKLINE>
        >>> df.loc["ZKFO"]                  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        time
        1960Y1    1.00000
        1961Y1    1.00000
        ...
        2014Y1    1.01599
        2015Y1    1.01599
        Name: ZKFO, dtype: float64

        >>> # Export a subset of the IODE Variables database as a pandas DataFrame
        >>> df = variables["A*;*_"].df
        >>> df.shape
        (33, 56)
        >>> df.index.to_list()              # doctest: +ELLIPSIS
        [\'ACAF\', \'ACAG\', \'AOUC\', ..., \'WNF_\', \'YDH_\', \'ZZF_\']
        >>> df.columns.to_list()            # doctest: +ELLIPSIS
        [\'1960Y1\', \'1961Y1\', ..., \'2014Y1\', \'2015Y1\']
        >>> variables["AOUC"]               # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  ...  2014Y1   2015Y1
        AOUC            na    0.25  ...    1.42     1.46
        <BLANKLINE>
        >>> df.loc["AOUC"]                  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        time
        1960Y1         NaN
        1961Y1    0.247832
        ...
        2014Y1    1.423714
        2015Y1    1.460863
        Name: AOUC, dtype: float64    
        >>> variables["ZZF_"]               # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ...fun.var
        description: Modèle fun - Simulation 1
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
        name        1960Y1  1961Y1  ...  2014Y1   2015Y1
        ZZF_          0.69    0.69  ...    0.69     0.69
        <BLANKLINE>
        >>> df.loc["ZZF_"]                  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        time
        1960Y1    0.6884
        1961Y1    0.6884
        ...
        2014Y1    0.6884
        2015Y1    0.6884
        Name: ZZF_, dtype: float64 
        '''
    def periods_subset(self, from_period: str | Period = None, to_period: str | Period = None, as_float: bool = False) -> list[str | float]:
        '''
        Return a subset of the periods from the current Variables sample.

        Parameters
        ----------
        from_period: str or Period, optional
            first period of the subset of periods.
            Defaults to the first period of the current Variables sample.
        to_period: str or Period, optional
            last period of the subset of periods.
            Defaults to the last period of the current Variables sample.
        as_float: bool, optional
            whether or not to return periods as float.
            Defaults to False.

        Returns
        -------
        list(str) or list(float)

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> variables.sample
        Sample("1960Y1:2015Y1")
        >>> variables.periods_subset("1990Y1", "2000Y1")                    # doctest: +ELLIPSIS
        [\'1990Y1\', \'1991Y1\', ..., \'1999Y1\', \'2000Y1\']
        >>> variables.periods_subset("1990Y1", "2000Y1", as_float=True)     # doctest: +ELLIPSIS
        [1990.0, 1991.0, ..., 1999.0, 2000.0]
        '''
    def copy_from(self, input_files: str | list[str], from_period: str | Period = None, to_period: str | Period = None, names: str | list[str] = '*'):
        """
        Copy (a subset of) variables from the input file(s) 'input_files' into the current database.

        Parameters
        ----------
        input_file: str or list(str)
            file(s) from which the copied variables are read.
        from_period: str or Period, optional
            start period for copying the variables values.
        to_period: str or Period, optional
            end period for copying the variables values.
        names: str or list(str), optional
            list of variables to copy from the input file(s).
            Defaults to load all variables from the input file(s). 
        """
    def low_to_high(self, type_of_series: LowToHighType | str, method: LowToHighMethod | str, filepath: str | Path, var_list: str | list[str]):
        '''
        Build series with higher periodicity for *stock data* (Unemployment, Debt, ...) or 
        *flow data* (GNP, Deficit, ...).

        The list of specified series (variables) from the input file are loaded into the current 
        Variables database and the periodicity of these series (variables) is modified simultaneously. 
        The new periodicity is the one currently defined in the current Variables database.

        The loaded series are added to or replace those (for existing names) in the current 
        Variables database.

        This procedure exists for the following cases:

            - annual to monthly 
            - annual to quarterly 
            - quarterly to monthly 
        
        Two types of series are available, one for stocks (STOCK), the other for flows (FLOW).

        Three interpolation methods are available:

            - linear (LINEAR): A[1980Q{1,2,3,4}] = A[1979Y1] + i * (A[1980Y1] - A[1979Y1])/4 i = 1,2,3,4 
            - cubic splines (CUBIC_SPLINES): cubic interpolation 
            - step (STEP) : A[1980Q{1,2,3,4}] = A[1980Y1] 

        Parameters
        ----------
        type_of_series : LowToHighType or str
            Two types of series are considered: \'stock\' and \'flow\':
                - STOCK: stock data (Unemployment, Debt, ...)
                - FLOW: flow data (GNP, Deficit, ...)

        method : LowToHighMethod or str
            Method to use for transformation. Three methods can be used:
                - LINEAR (\'L\'): Linear interpolation
                - CUBIC_SPLINES (\'C\'): Cubic Spliness
                - STEP (\'S\'): Step

        filepath : str or Path
            Filepath to the source data file.

        var_list : str or list(str)
            List of variables to include in the transformation.

        Returns
        -------
        None

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR, variables, LowToHighType, LowToHighMethod
        >>> variables.clear()
        >>> # define a yearly sample
        >>> variables.sample = "2010Q1:2020Q4"
        >>> # input filepath
        >>> filepath = f"{SAMPLE_DATA_DIR}/fun.var"
        
        Linear interpolation / stock
        
        >>> # "stock" -> the result is a linear interpolation of the 2 surrounding source values.
        >>> variables.low_to_high(LowToHighType.STOCK, LowToHighMethod.LINEAR, filepath, ["ACAF", "ACAG"])
        >>> variables["ACAF", "2014Q1":"2014Q4"]    # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2014Q1:2014Q4
        mode: LEVEL
        <BLANKLINE>
        name        2014Q1  2014Q2  2014Q3  2014Q4
        ACAF        -72.51  -76.12  -79.73  -83.34
        <BLANKLINE>
        >>> variables["ACAG", "2014Q1":"2014Q4"]    # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2014Q1:2014Q4
        mode: LEVEL
        <BLANKLINE>
        name        2014Q1  2014Q2  2014Q3  2014Q4
        ACAG         31.63   31.90   32.16   32.42
        <BLANKLINE>

        Linear interpolation / flow
        
        >>> # "flow" -> the result is the source value divided by the nb of sub-periods. 
        >>> variables.low_to_high(LowToHighType.FLOW, LowToHighMethod.LINEAR, filepath, ["ACAF", "ACAG"])
        >>> variables["ACAF", "2014Q1":"2014Q4"]    # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2014Q1:2014Q4
        mode: LEVEL
        <BLANKLINE>
        name        2014Q1  2014Q2  2014Q3  2014Q4
        ACAF        -20.84  -20.84  -20.84  -20.84
        <BLANKLINE>
        >>> variables["ACAG", "2014Q1":"2014Q4"]    # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2014Q1:2014Q4
        mode: LEVEL
        <BLANKLINE>
        name        2014Q1  2014Q2  2014Q3  2014Q4
        ACAG          8.11    8.11    8.11    8.11
        <BLANKLINE>

        Cubic splines / stock
        
        >>> variables.low_to_high(LowToHighType.STOCK, LowToHighMethod.CUBIC_SPLINES, filepath, ["ACAF", "ACAG"])
        >>> variables["ACAF", "2012Q1":"2012Q4"]    # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2012Q1:2012Q4
        mode: LEVEL
        <BLANKLINE>
        name        2012Q1  2012Q2  2012Q3  2012Q4
        ACAF        -47.30  -50.05  -52.81  -55.56
        <BLANKLINE>
        >>> variables["ACAG", "2012Q1":"2012Q4"]    # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2012Q1:2012Q4
        mode: LEVEL
        <BLANKLINE>
        name        2012Q1  2012Q2  2012Q3  2012Q4
        ACAG         29.54   29.80   30.06   30.32
        <BLANKLINE>

        Cubic splines / flow
        
        >>> variables.low_to_high(LowToHighType.FLOW, LowToHighMethod.CUBIC_SPLINES, filepath, ["ACAF", "ACAG"])
        >>> variables["ACAF", "2012Q1":"2012Q4"]    # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2012Q1:2012Q4
        mode: LEVEL
        <BLANKLINE>
        name        2012Q1  2012Q2  2012Q3  2012Q4
        ACAF        -12.75  -13.44  -14.27  -15.10
        <BLANKLINE>
        >>> variables["ACAG", "2012Q1":"2012Q4"]    # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2012Q1:2012Q4
        mode: LEVEL
        <BLANKLINE>
        name        2012Q1  2012Q2  2012Q3  2012Q4
        ACAG          7.48    7.55    7.61    7.68
        <BLANKLINE>

        Step / stock
        
        >>> # "stock" -> the result has the same value as the source
        >>> variables.low_to_high(LowToHighType.STOCK, LowToHighMethod.STEP, filepath, ["ACAF", "ACAG"])
        >>> variables["ACAF", "2014Q1":"2014Q4"]    # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2014Q1:2014Q4
        mode: LEVEL
        <BLANKLINE>
        name        2014Q1  2014Q2  2014Q3  2014Q4
        ACAF        -83.34  -83.34  -83.34  -83.34
        <BLANKLINE>
        >>> variables["ACAG", "2014Q1":"2014Q4"]    # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2014Q1:2014Q4
        mode: LEVEL
        <BLANKLINE>
        name        2014Q1  2014Q2  2014Q3  2014Q4
        ACAG         32.42   32.42   32.42   32.42
        <BLANKLINE>

        Step / flow
        
        >>> # "flow" -> the result is the source value plus a portion of 
        >>> # the difference between the 2 surrounding values in the source
        >>> variables.low_to_high(LowToHighType.FLOW, LowToHighMethod.STEP, filepath, ["ACAF", "ACAG"])
        >>> variables["ACAF", "2014Q1":"2014Q4"]    # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2014Q1:2014Q4
        mode: LEVEL
        <BLANKLINE>
        name        2014Q1  2014Q2  2014Q3  2014Q4
        ACAF        -20.84  -20.84  -20.84  -20.84
        <BLANKLINE>
        >>> variables["ACAG", "2014Q1":"2014Q4"]    # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2014Q1:2014Q4
        mode: LEVEL
        <BLANKLINE>
        name        2014Q1  2014Q2  2014Q3  2014Q4
        ACAG          8.11    8.11    8.11    8.11
        <BLANKLINE>
        '''
    def high_to_low(self, type_of_series: HighToLowType | str, filepath: str | Path, var_list: str | list[str]):
        '''
        Build series of lower periodicity by (summing the / taking the average of the / taking the 
        last observation of) sub-periods.

        The list of specified series (variables) from the input file are loaded into the current 
        Variables database and the periodicity of these series (variables) is modified simultaneously. 
        The new periodicity is the one currently defined in the current Variables database.

        The loaded series are added to or replace those (for existing names) in the current 
        Variables database.

        This procedure exists for the following cases:

            - monthly to annual 
              (annual observation = sum of 12 months / average of 12 months / December value) 
            - quarterly to annual 
              (annual observation = sum of 4 quarters / average of 4 quarters / last quarter value) 
            - monthly to quarterly 
              (quarterly observation = sum of 3 months / average of 3 months / value for the last month of the quarter) 

        Three types of series are available:

            - SIM: addition of sub-periods data
            - MEAN: average of sub-periods data
            - LAST: last observation

        In the case of a non-existent value (:math:`NA`) for one of the sub-periods, the result is :math:`NA`.

        Parameters
        ----------
        type_of_series : HighToLowType or str
            Three types of series are available: 
                - SUM : addition of sub-period data
                - MEAN : average of sub-period data
                - LAST : last observation

        filepath : str or Path
            Filepath to the source data file.

        var_list : str or list(str)
            List of variables to include in the transformation.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR, variables, HighToLowType
        >>> variables.clear()
        >>> # define a yearly sample
        >>> variables.sample = "2000Y1:2020Y1"
        >>> # input filepath
        >>> filepath = f"{SAMPLE_DATA_DIR}/fun_q.var"
        
        Last Obs in year
        
        >>> variables.high_to_low(HighToLowType.LAST, filepath, ["ACAF", "ACAG"])
        >>> variables["ACAF", "2010Y1":"2014Y1"]    # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2010Y1:2014Y1
        mode: LEVEL
        <BLANKLINE>
        name        2010Y1  2011Y1  2012Y1  2013Y1  2014Y1
        ACAF        -37.83  -44.54  -55.56  -68.89  -83.34
        <BLANKLINE>
        >>> variables["ACAG", "2010Y1":"2014Y1"]    # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2010Y1:2014Y1
        mode: LEVEL
        <BLANKLINE>
        name        2010Y1  2011Y1  2012Y1  2013Y1  2014Y1
        ACAG          7.06    7.32    7.58    7.84    8.11
        <BLANKLINE>

        Mean of year
        
        >>> variables.high_to_low(HighToLowType.MEAN, filepath, ["ACAF", "ACAG"])
        >>> variables["ACAF", "2010Y1":"2014Y1"]    # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2010Y1:2014Y1
        mode: LEVEL
        <BLANKLINE>
        name        2010Y1  2011Y1  2012Y1  2013Y1  2014Y1
        ACAF        -37.83  -44.54  -55.56  -68.89  -83.34
        <BLANKLINE>
        >>> variables["ACAG", "2010Y1":"2014Y1"]    # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2010Y1:2014Y1
        mode: LEVEL
        <BLANKLINE>
        name        2010Y1  2011Y1  2012Y1  2013Y1  2014Y1
        ACAG          7.06    7.32    7.58    7.84    8.11
        <BLANKLINE>

        Sum
        
        >>> variables.high_to_low(HighToLowType.SUM, filepath, ["ACAF", "ACAG"])
        >>> variables["ACAF", "2010Y1":"2014Y1"]    # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2010Y1:2014Y1
        mode: LEVEL
        <BLANKLINE>
        name         2010Y1  2011Y1  2012Y1  2013Y1  2014Y1
        ACAF        -151.31 -178.18 -222.24 -275.58 -333.36
        <BLANKLINE>
        >>> variables["ACAG", "2010Y1":"2014Y1"]    # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2010Y1:2014Y1
        mode: LEVEL
        <BLANKLINE>
        name        2010Y1  2011Y1  2012Y1  2013Y1  2014Y1
        ACAG         28.25   29.28   30.32   31.37   32.42
        <BLANKLINE>
        '''
    def extrapolate(self, method: SimulationInitialization | str, from_period: str | Period = None, to_period: str | Period = None, variables_list: str | list[str] = None):
        '''
        Extrapolate variables using one the method described below, based on previous periods.

        The possible methods are as follows:

          - :math:`Y := Y[-1], if Y null or NA` (TM1) : each null or NA endogen at the start takes the value of 
            the previous period,
          - :math:`Y := Y[-1], always` (TM1_A) : each endogen takes the value of the previous period at the start,
          - :math:`Y := extrapolation, if Y null or NA` (EXTRA) : each null or NA endogen takes as value a linear 
            extrapolation of the two previous periods,
          - :math:`Y := extrapolation, always` (EXTRA_A) : each endogen takes as its value a linear extrapolation of 
            the two preceding periods, whether or not it is zero at the start,
          - :math:`Y unchanged` (ASIS): endogenous values are not initialized. They retain their value whether or 
            not they are zero,
          - :math:`Y := Y[-1], if Y = NA` (TM1_NA): each NA value takes the value of the previous period,
          - :math:`Y := extrapolation, if Y = NA` (extra_na): each NA value takes the value of a linear extrapolation of 
            the two previous periods.

        The sample over which the calculation is to be performed must be supplied. 
        This sample must be a sub-sample of the global sample defined in the current Variables database.

        In addition, the list of variables to be extrapolated can be specified. 
        If this list is not given, all variables will be extrapolated.

        Parameters
        ----------
        method: SimulationInitialization or str
            initialization method. Possible values are:
              - TM1: :math:`Y := Y[-1], if Y null or NA`
              - TM1_A: :math:`Y := Y[-1], always`
              - EXTRA: :math:`Y := extrapolation, if Y null or NA`
              - EXTRA_A: :math:`Y := extrapolation, always`
              - ASIS: :math:`Y unchanged`
              - TM1_NA: :math:`Y := Y[-1], if Y = NA`
              - EXTRA_NA: :math:`Y := extrapolation, if Y = NA`
        from_period: str or Period, optional
            starting period to extrapolate variables.
            Defaults to the starting period if the current sample.
        to_period: str or Period, optional
            ending period to extrapolate variables.
            Defaults to the ending period if the current sample.
        variables_list: str or list(str), optional
            list of variables to extrapolate.
            Defaults to None (all variables).

        Examples
        --------
        >>> from iode import variables, NA, SimulationInitialization
        >>> variables.clear()
        >>> variables.sample = "2000Y1:2020Y1"

        >>> def reset_ACAF():
        ...     variables["ACAF"] = "t"
        ...     variables["ACAF", ["2005Y1", "2007Y1"]] = NA
        >>> # create ACAF
        >>> reset_ACAF()
        >>> variables["ACAF", :"2010Y1"]                # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2000Y1:2010Y1
        mode: LEVEL
        <BLANKLINE>
        name        2000Y1  2001Y1  2002Y1  2003Y1  2004Y1  2005Y1  2006Y1  2007Y1  2008Y1  2009Y1  2010Y1
        ACAF          0.00    1.00    2.00    3.00    4.00      na    6.00      na    8.00    9.00   10.00
        <BLANKLINE>

        >>> # "TM1" (Y := Y[-1], if Y null or NA)
        >>> reset_ACAF()
        >>> variables.extrapolate(SimulationInitialization.TM1, "2005Y1", "2010Y1")
        >>> variables["ACAF", "2003Y1":"2009Y1"]        # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2003Y1:2009Y1
        mode: LEVEL
        <BLANKLINE>
        name        2003Y1  2004Y1  2005Y1  2006Y1  2007Y1  2008Y1  2009Y1
        ACAF          3.00    4.00    4.00    6.00    6.00    8.00    9.00
        <BLANKLINE>

        >>> # "TM1_A" (Y := Y[-1], always)
        >>> reset_ACAF()
        >>> variables.extrapolate(SimulationInitialization.TM1_A, "2005Y1", "2010Y1")
        >>> variables["ACAF", "2003Y1":"2009Y1"]        # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2003Y1:2009Y1
        mode: LEVEL
        <BLANKLINE>
        name        2003Y1  2004Y1  2005Y1  2006Y1  2007Y1  2008Y1  2009Y1
        ACAF          3.00    4.00    4.00    4.00    4.00    4.00    4.00
        <BLANKLINE>

        >>> # "EXTRA" (Y := extrapolation, if Y null or NA)
        >>> reset_ACAF()
        >>> variables.extrapolate(SimulationInitialization.EXTRA, "2005Y1", "2010Y1")
        >>> variables["ACAF", "2003Y1":"2009Y1"]        # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2003Y1:2009Y1
        mode: LEVEL
        <BLANKLINE>
        name        2003Y1  2004Y1  2005Y1  2006Y1  2007Y1  2008Y1  2009Y1
        ACAF          3.00    4.00    5.00    6.00    7.00    8.00    9.00
        <BLANKLINE>

        >>> # "EXTRA_A" (Y := extrapolation, always)
        >>> reset_ACAF()
        >>> variables.extrapolate(SimulationInitialization.EXTRA_A, "2005Y1", "2010Y1")
        >>> variables["ACAF", "2003Y1":"2009Y1"]        # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2003Y1:2009Y1
        mode: LEVEL
        <BLANKLINE>
        name        2003Y1  2004Y1  2005Y1  2006Y1  2007Y1  2008Y1  2009Y1
        ACAF          3.00    4.00    5.00    6.00    7.00    8.00    9.00
        <BLANKLINE>

        >>> # "ASIS" (Y unchanged)
        >>> reset_ACAF()
        >>> variables.extrapolate(SimulationInitialization.ASIS, "2005Y1", "2010Y1")
        >>> variables["ACAF", "2003Y1":"2009Y1"]        # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2003Y1:2009Y1
        mode: LEVEL
        <BLANKLINE>
        name        2003Y1  2004Y1  2005Y1  2006Y1  2007Y1  2008Y1  2009Y1
        ACAF          3.00    4.00      na    6.00      na    8.00    9.00
        <BLANKLINE>

        >>> # "TM1_NA" (Y := Y[-1], if Y = NA)
        >>> reset_ACAF()
        >>> variables.extrapolate(SimulationInitialization.TM1_NA, "2005Y1", "2010Y1")
        >>> variables["ACAF", "2003Y1":"2009Y1"]        # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2003Y1:2009Y1
        mode: LEVEL
        <BLANKLINE>
        name        2003Y1  2004Y1  2005Y1  2006Y1  2007Y1  2008Y1  2009Y1
        ACAF          3.00    4.00    4.00    6.00    6.00    8.00    9.00
        <BLANKLINE>

        >>> # "EXTRA_NA" (Y := extrapolation, if Y = NA)
        >>> reset_ACAF()
        >>> variables.extrapolate(SimulationInitialization.EXTRA_NA, "2005Y1", "2010Y1")
        >>> variables["ACAF", "2003Y1":"2009Y1"]        # doctest: +NORMALIZE_WHITESPACE
        Workspace: Variables
        nb variables: 1
        filename: ws
        sample: 2003Y1:2009Y1
        mode: LEVEL
        <BLANKLINE>
        name        2003Y1  2004Y1  2005Y1  2006Y1  2007Y1  2008Y1  2009Y1
        ACAF          3.00    4.00    5.00    6.00    7.00    8.00    9.00
        <BLANKLINE>
        '''
    def seasonal_adjustment(self, input_file: str, eps_test: float = 5.0, series: str | list[str] = None):
        """
        Eliminate seasonal variations in monthly series (= variables).

        The method constructs deseasonalized series using the *Census XI method*, as well as cyclical 
        and stochastic trend components. Thus, In addition to the deseasonalized series (named after 
        the series in the input file), the result is two other series:

            - the series containing the cyclical trend component named : `_C<name>` where *name* is the original name
            - the series containing the stochastic component called : `_I<name>` where *name* is the original name

        Note that the deseasonalized series is the product of the other two.

        To do this, the list of *series* (= variables) are loaded from the specified file *input_file* 
        into the current IODE Variables database and simultaneously modifies the selected *series* if necessary.

        The new *series* are added to or replace (for existing names) those in the current IODE Variables database.

        In the case of non-existent value (NA) for one of the periods, the result is NA for the whole year.

        Parameters
        ----------
        input_file: str or Path
            Filepath to the input file.
        eps_test: float, optional
            Criterion verifying whether a seasonal influence is present in a series
            Default to 5.0.
        series: str or list(str), optional
            list of series on which to apply the seasonal adjustment.
            Defaults to None (all variables).

        See Also
        --------
        Variables.trend_correction
        """
    def trend_correction(self, input_file: str, lambda_: float, series: str | list[str] = None, log: bool = False):
        """
        Implementation of the *Hodrick-Prescott method* for trend series (= variables) construction. 
        The principle is the same as for deseasonalization: series read from a file are imported and transformed simultaneously.

        If the `log` flag is set, the series are log-transformed before calculation and exp-transformed after calculation. 
        The series values must therefore be strictly positive.

        Parameters
        ----------
        input_file: str
            filepath to the input file.
        lambda_: float
            Lambda parameter of the Hodrick-Prescott method.
        series: str or list(str), optional
            list of series on which to apply the trend correction.
            Defaults to None (all variables).
        log: bool, optional
            Whether or not the series are log-transformed before calculation and exp-transformed 
            after calculation. The series values must therefore be strictly positive.
            Defaults to False.

        See Also
        --------
        Variables.seasonal_adjustment
        """
    @classmethod
    def convert_file(cls, input_file: str | Path, input_format: str | ImportFormats, save_file: str | Path, rule_file: str | Path, from_period: str | Period, to_period: str | Period, debug_file: str | Path = None):
        '''
        Convert an external file representing IODE variables to an IODE variables file (.var). 
        The possible formats for the input file are:
          
          - `Ascii`: IODE-specific Ascii format for objects 
          - `Rotated Ascii`: Ascii format for variables with series in columns 
          - `DIF`: DIF format (Data Interchange Format) 
          - `DIF` Belgostat: (old) exchange format specific to Belgostat 
          - `NIS`: National Institute of Statistics Ascii format (old) 
          - `GEM`: Ascii format of Chronos software 
          - `PRN-Aremos`: Ascii format from Aremos software 
          - `TXT Belgostat`: (old) Belgostat-specific exchange format 

        The rule file is a simple text file contains the rules for: 
        
          - selecting the objects to be imported 
          - determining the objects names.

        Each rule consists of two fields:

          - the selection pattern, containing a description of the names concerned by the rule. 
            This mask is defined in the same way as the :py:meth:`~iode.Comments.search` method.
          - the transcoding algorithm for the names, which can contain : 
            - `+` : indicates that the character must be included in the name 
            - `-` : indicates that the character should be skipped 
            - any other character: included in the name 
        
        Example:

            B* C+-+          -> transforms B1234 into CB2, BCDEF into CBE, etc 
            *X ++++++++++    -> keeps names ending in X unchanged
            * ++++++++++     -> keeps all names unchanged

        Parameters
        ----------
        input_file : str or Path
            The path to the input file to be converted. 
        input_format : str or ImportFormats
            The format of the input file. Possible formats are ASCII, ROT_ASCII (Rotated Ascii), 
            DIF, BISTEL, NIS, GEM, PRN, TXT (TXT Belgostat).
        save_file : str or Path
            The path to the output file where the IODE variables will be saved.
        rule_file : str or Path
            The path to the rule file that defines the selection and transcoding rules. 
        from_period : str or Period
            The first period of the series to be imported. 
        to_period : str or Period
            The last period of the series to be imported.
        debug_file : str or Path, optional
            The path to the debug file where the debug information will be saved. 
            If not provided, the debug information will be printed to the console.

        Examples
        --------
        >>> from pathlib import Path
        >>> from iode import SAMPLE_DATA_DIR, variables, ImportFormats
        >>> output_dir = getfixture(\'tmp_path\')

        >>> input_file = f"{SAMPLE_DATA_DIR}/fun_xode.av.ref"
        >>> input_format = ImportFormats.ASCII
        >>> save_file = str(output_dir / "imported_var.var")
        >>> rule_file = f"{SAMPLE_DATA_DIR}/rules.txt"
        >>> debug_file = str(output_dir / "debug.log")

        >>> # print rules
        >>> with open(rule_file, "r") as f:         # doctest: +NORMALIZE_WHITESPACE
        ...     print(f.read())
        ...
        AC*  KK_--+++++++++++++
        *U   UU_++++++++++++++++
        >>> # get list of variables with a name starting with \'AC\' 
        >>> # and ending with \'U\' from the input file
        >>> with open(input_file, "r") as f:         # doctest: +NORMALIZE_WHITESPACE, +ELLIPSIS
        ...     for line in f:
        ...         name = line.split(" ")[0]
        ...         if name.startswith("AC") or name.endswith("U"):
        ...             print(line.strip())
        ...
        ACAF na na ... -83.3406251108009 -96.4104198284833
        ACAG na na ... 32.4202988291984 33.469601344881
        CGU 69.354416 70.728317 ... 2652.457356636 2800.12343205764 
        DPU 56.285999 58.596001 ... 1794.28676968594 1879.1395597413
        DPUU 56.285999 58.596001 ... 1795.98222768555 1880.91519686508
        IFU na na ... 1895.9196231884 1952.4775760035
        IHU na na ... 855.342842036469 904.6210534989
        MU na na ... 0.278260325684654 0.278260325684654
        NAWRU na na ... 0.139645850151953 0.139645850151953
        WBU 256.177 268.75299 ... 8525.33576585068 8986.56510007165
        >>> # import variables from input_file to save_file
        >>> # using the rules defined in rule_file
        >>> variables.convert_file(input_file, input_format, save_file, rule_file, "2000Y1", "2010Y1", debug_file)
        Reading object 1 : KK_AF
        Reading object 2 : KK_AG
        Reading object 3 : UU_CGU
        Reading object 4 : UU_DPU
        Reading object 5 : UU_DPUU
        Reading object 6 : UU_IFU
        Reading object 7 : UU_IHU
        Reading object 8 : UU_MU
        Reading object 9 : UU_NAWRU
        Reading object 10 : UU_WBU
        10 objects saved
        >>> # check content of the saved file
        >>> variables.load(save_file)        # doctest: +ELLIPSIS
        Loading ...\\imported_var.var
        10 objects loaded
        >>> variables                        # doctest: +NORMALIZE_WHITESPACE, +ELLIPSIS
        Workspace: Variables
        nb variables: 10
        filename: ...\\imported_var.var
        sample: 1960Y1:2015Y1
        mode: LEVEL
        <BLANKLINE>
          name      1960Y1  1961Y1  1962Y1  1963Y1  1964Y1  ...      2010Y1  2011Y1  2012Y1  2013Y1  2014Y1  2015Y1 
        KK_AF           na      na      na      na      na  ...      -37.83  -44.54  -55.56  -68.89  -83.34  -96.41 
        KK_AG           na      na      na      na      na  ...       28.25   29.28   30.32   31.37   32.42   33.47 
        UU_CGU       69.35   70.73   77.93   88.62   95.53  ...     2173.77 2268.92 2381.03 2510.32 2652.46 2800.12 
        UU_DPU       56.29   58.60   61.96   67.21   74.28  ...     1531.90 1584.26 1642.98 1712.98 1794.29 1879.14 
        UU_DPUU      56.29   58.60   61.96   67.21   74.28  ...     1533.35 1585.75 1644.53 1714.60 1795.98 1880.92 
        UU_IFU          na      na      na      na      na  ...     1566.97 1591.35 1676.27 1802.71 1895.92 1952.48 
        UU_IHU          na      na      na      na      na  ...      726.94  755.66  784.87  815.30  855.34  904.62 
        UU_MU           na      na      na      na      na  ...        0.28    0.28    0.28    0.28    0.28    0.28 
        UU_NAWRU        na      na      na      na      na  ...        0.14    0.14    0.14    0.14    0.14    0.14 
        UU_WBU      256.18  268.75  295.29  324.89  366.34  ...     7072.79 7328.34 7664.35 8073.34 8525.34 8986.57 
        <BLANKLINE>
        >>> # content of the debug file
        >>> with open(debug_file, "r") as f:         # doctest: +NORMALIZE_WHITESPACE
        ...     for line in f:
        ...         print(line.strip())
        ...
        ACAF -> KK_AF       (Rule KK_--+++++++++++++)
        ACAG -> KK_AG       (Rule KK_--+++++++++++++)
        CGU -> UU_CGU       (Rule UU_++++++++++++++++)
        DPU -> UU_DPU       (Rule UU_++++++++++++++++)
        DPUU -> UU_DPUU     (Rule UU_++++++++++++++++)
        IFU -> UU_IFU       (Rule UU_++++++++++++++++)
        IHU -> UU_IHU       (Rule UU_++++++++++++++++)
        MU -> UU_MU         (Rule UU_++++++++++++++++)
        NAWRU -> UU_NAWRU   (Rule UU_++++++++++++++++)
        WBU -> UU_WBU       (Rule UU_++++++++++++++++)
        '''
    @classmethod
    def export_as_file(cls, variables_file: str | Path, rule_file: str | Path, save_file: str | Path, export_format: str | ExportFormats, from_period: str | Period, to_period: str | Period, comments_file: str | Path, nan_value: str = '#N/A', separator: str = ';', debug_file: str | Path = None):
        '''
        Convert an IODE Variables file to a format used by some other programs. 
        The possible output formats are:
          
          - `CSV` 
          - `RCSV` 
          - `DIF` 
          - `WKS` 
          - `TSP` 

        If an IODE Comments file is passed, comments of the same name will be associated with variables 
        in the result file when the output format allows it.

        The rule file is a simple text file contains the rules for: 
        
          - selecting the variables (and comments) to be exported 
          - determining the new variables names in the saved file.

        Each rule consists of two fields:

          - the selection pattern, containing a description of the names concerned by the rule. 
            This mask is defined in the same way as the :py:meth:`~iode.Comments.search` method.
          - the transcoding algorithm for the names, which can contain : 
            - `+` : indicates that the character must be included in the name 
            - `-` : indicates that the character should be skipped 
            - any other character: included in the name 
        
        Example:

            B* C+-+          -> transforms B1234 into CB2, BCDEF into CBE, etc 
            *X ++++++++++    -> keeps names ending in X unchanged
            * ++++++++++     -> keeps all names unchanged

        Parameters
        ----------
        variables_file : str or Path
            The path to the input Variables file to be converted. 
        rule_file : str or Path
            The path to the rule file that defines the selection and transcoding rules. 
        save_file : str or Path
            The path to the output file where the IODE variables will be saved.
        export_format : str or ExportFormats
            The format of the output file. Possible formats are CSV, RCSV (rotated CSV), 
            DIF, WKS, TSP.
        from_period : str or Period, optional
            The first period of the series to be exported. 
        to_period : str or Period, optional
            The last period of the series to be exported.        
        comments_file : str or Path, optional
            The path to the input Comments file. 
        nan_value: str, optional
            The value to be used for missing data. 
            Only used for `CSV` and `RCSV` (rotated CSV).
            Default is "#N/A". 
        separator: str, optional
            The character to be used as separator. 
            Only used for `CSV` and `RCSV` (rotated CSV).
            Default is ";".
        debug_file : str or Path, optional
            The path to the debug file where the debug information will be saved. 
            If not provided, the debug information will be printed to the console.

        Examples
        --------
        >>> from pathlib import Path
        >>> from iode import SAMPLE_DATA_DIR, comments, variables, ExportFormats
        >>> output_dir = getfixture(\'tmp_path\')

        >>> variables_file = f"{SAMPLE_DATA_DIR}/fun.av"
        >>> comments_file = f"{SAMPLE_DATA_DIR}/fun.ac"
        >>> rule_file = f"{SAMPLE_DATA_DIR}/rules.txt"
        >>> from_period = "2000Y1"
        >>> to_period = "2010Y1"
        
        >>> comments.load(comments_file)            # doctest: +NORMALIZE_WHITESPACE, +ELLIPSIS
        Loading .../fun.ac
        Reading object 1 : ACAF
        Reading object 2 : ACAG
        ...
        Reading object 316 : ZX
        Reading object 317 : ZZ_
        317 objects loaded 
        >>> variables.load(variables_file)          # doctest: +NORMALIZE_WHITESPACE, +ELLIPSIS
        Loading .../fun.av
        Reading object 1 : ACAF
        Reading object 2 : ACAG
        ...
        Reading object 393 : ZX
        Reading object 394 : ZZF_
        394 objects loaded

        >>> # print file containing rules
        >>> with open(rule_file, "r") as f:         # doctest: +NORMALIZE_WHITESPACE
        ...     print(f.read())
        ...
        AC*  KK_--+++++++++++++
        *U   UU_++++++++++++++++
        >>> # get list of variables with a name starting with \'AC\' 
        >>> # and ending with \'U\'
        >>> variables.get_names("AC*;*U")
        [\'ACAF\', \'ACAG\', \'CGU\', \'DPU\', \'DPUU\', \'IFU\', \'IHU\', \'MU\', \'NAWRU\', \'WBU\']
        >>> # get list of comments with a name starting with \'AC\' 
        >>> # and ending with \'U\'
        >>> comments.get_names("AC*;*U")
        [\'ACAF\', \'ACAG\', \'DPU\', \'DPUU\', \'IFU\', \'IHU\', \'WBU\']

        >>> # export variables to CSV
        >>> export_format = ExportFormats.CSV
        >>> save_file = str(output_dir / "exported_var.csv")
        >>> debug_file = str(output_dir / "debug_csv.log")
        >>> variables.export_as_file(variables_file, rule_file, save_file, export_format, 
        ...                          from_period, to_period, comments_file, debug_file=debug_file)      # doctest: +NORMALIZE_WHITESPACE, +ELLIPSIS
        Loading ...\\fun.av
        Reading object 1 : ACAF
        Reading object 2 : ACAG
        ...
        Reading object 393 : ZX
        Reading object 394 : ZZF_
        394 objects loaded
        Loading ...\\fun.ac
        Reading object 1 : ACAF
        Reading object 2 : ACAG
        ...
        Reading object 316 : ZX
        Reading object 317 : ZZ_
        317 objects loaded     
        >>> # check content of the saved file
        >>> # note: no comment found for variables CGU, MU and NAWRU
        >>> with open(save_file, "r") as f:         # doctest: +NORMALIZE_WHITESPACE, +ELLIPSIS
        ...     print(f.read())
        ...
        code;comment;2000Y1;2001Y1;2002Y1;2003Y1;2004Y1;2005Y1;2006Y1;2007Y1;2008Y1;2009Y1;2010Y1;
        KK_AF; Ondernemingen: ontvangen kapitaaloverdrachten.; 10.046611;...;-37.827429;
        KK_AG; Totale overheid: netto ontvangen kapitaaloverdrachten.; -41.534787;...;28.253929;
        UU_CGU; ; 1383.2586;...;2173.7682;
        UU_DPU; Nominale afschrijvingen op de kapitaalvoorraad.; 953.60012;...;1531.9025;
        UU_DPUU; Nominale afschrijvingen op de kapitaalvoorraad (aangepast:  inkomensoptiek).; 954.5012;...;1533.35;
        UU_IFU; Bruto kapitaalvorming: ondernemingen.; 1076.1795;...;1566.9738;
        UU_IHU; Bruto kapitaalvorming: gezinnen.; 471.00145;...;726.93744;
        UU_MU; ; 0.42001992;...;0.27826033;
        UU_NAWRU; ; 0.14141811;...;0.13964585;
        UU_WBU; Totale loonmassa (inclusief werkgeversbijdragen).; 4922.5664;...;7072.7855;
        >>> # content of the debug file
        >>> with open(debug_file, "r") as f:         # doctest: +NORMALIZE_WHITESPACE
        ...     print(f.read())
        ...
        ACAF -> KK_AF       (Rule KK_--+++++++++++++)
        ACAG -> KK_AG       (Rule KK_--+++++++++++++)
        CGU -> UU_CGU       (Rule UU_++++++++++++++++)
        DPU -> UU_DPU       (Rule UU_++++++++++++++++)
        DPUU -> UU_DPUU     (Rule UU_++++++++++++++++)
        IFU -> UU_IFU       (Rule UU_++++++++++++++++)
        IHU -> UU_IHU       (Rule UU_++++++++++++++++)
        MU -> UU_MU         (Rule UU_++++++++++++++++)
        NAWRU -> UU_NAWRU   (Rule UU_++++++++++++++++)
        WBU -> UU_WBU       (Rule UU_++++++++++++++++)

        >>> # export variables to rotated CSV
        >>> export_format = ExportFormats.RCSV
        >>> save_file = str(output_dir / "exported_var.rcsv")
        >>> debug_file = str(output_dir / "debug_rcsv.log")
        >>> variables.export_as_file(variables_file, rule_file, save_file, export_format, 
        ...                          from_period, to_period, comments_file, debug_file=debug_file)      # doctest: +NORMALIZE_WHITESPACE, +ELLIPSIS
        Loading ...\\fun.av
        Reading object 1 : ACAF
        Reading object 2 : ACAG
        ...
        Reading object 393 : ZX
        Reading object 394 : ZZF_
        394 objects loaded
        Loading ...\\fun.ac
        Reading object 1 : ACAF
        Reading object 2 : ACAG
        ...
        Reading object 316 : ZX
        Reading object 317 : ZZ_
        317 objects loaded 
        >>> # check content of the saved file
        >>> # warning: the comments file is not used for rotated CSV
        >>> with open(save_file, "r") as f:         # doctest: +NORMALIZE_WHITESPACE, +ELLIPSIS
        ...     print(f.read())
        ...
        ; KK_AF; KK_AG; UU_CGU; UU_DPU; UU_DPUU; UU_IFU; UU_IHU; UU_MU; UU_NAWRU; UU_WBU;
        2000Y1; 10.046611; -41.534787; ... 0.42001992; 0.14141811; 4922.5664;
        2001Y1; 2.8679227; 18.939801; ... 0.40711156; 0.14138538; 5138.9458;
        2002Y1; -0.92921251; 19.980815; ... 0.39212964; 0.14125761; 5341.3233;
        2003Y1; -6.091565; 21.020502; ... 0.37923534; 0.14106277; 5556.2476;
        2004Y1; -14.582094; 22.066476; ... 0.36772624; 0.14083541; 5696.1652;
        2005Y1; -26.53879; 23.107962; ... 0.35617242; 0.14059196; 5814.7965;
        2006Y1; -28.987288; 24.129637; ... 0.34370718; 0.14034559; 6015.8951;
        2007Y1; -33.378426; 25.160909; ... 0.32978662; 0.1401158; 6295.5108;
        2008Y1; -38.409518; 26.192111; ... 0.31416594; 0.13991636; 6650.3069;
        2009Y1; -37.46351; 27.229955; ... 0.29691377; 0.13975922; 6861.5824;
        2010Y1; -37.827429; 28.253929; ... 0.27826033; 0.13964585; 7072.7855;

        >>> # export variables to TSP
        >>> export_format = ExportFormats.TSP
        >>> save_file = str(output_dir / "exported_var.tsp")
        >>> debug_file = str(output_dir / "debug_tsp.log")
        >>> variables.export_as_file(variables_file, rule_file, save_file, export_format, 
        ...                          from_period, to_period, comments_file, debug_file=debug_file)      # doctest: +NORMALIZE_WHITESPACE, +ELLIPSIS
        Loading ...\\fun.av
        Reading object 1 : ACAF
        Reading object 2 : ACAG
        ...
        Reading object 393 : ZX
        Reading object 394 : ZZF_
        394 objects loaded
        Loading ...\\fun.ac
        Reading object 1 : ACAF
        Reading object 2 : ACAG
        ...
        Reading object 316 : ZX
        Reading object 317 : ZZ_
        317 objects loaded 
        >>> # check content of the saved file
        >>> with open(save_file, "r") as f:         # doctest: +NORMALIZE_WHITESPACE, +ELLIPSIS
        ...     print(f.read())
        ...
        FREQ A;
        SMPL 2000 2010 ;
        LOAD KK_AF ;
        <BLANKLINE>
        ? Ondernemingen: ontvangen kapitaaloverdrachten.
        10.046611 2.8679227 -0.92921251 -6.091565 -14.582094 -26.53879 -28.987288
        -33.378426 -38.409518 -37.46351 -37.827429
        ;
        LOAD KK_AG ;
        <BLANKLINE>
        ? Totale overheid: netto ontvangen kapitaaloverdrachten.
        -41.534787 18.939801 19.980815 21.020502 22.066476 23.107962 24.129637
        25.160909 26.192111 27.229955 28.253929
        ;
        LOAD UU_CGU ;
        <BLANKLINE>
        1383.2586 1463.8679 1539.7198 1615.7986 1672.9449 1723.928 1787.6062 1873.3473
        1987.2756 2083.4747 2173.7682
        ;
        LOAD UU_DPU ;
        <BLANKLINE>
        ? Nominale afschrijvingen op de kapitaalvoorraad.
        953.60012 1007.4142 1056.9933 1101.323 1143.7441 1195.2886 1252.6131 1316.011
        1402.6103 1471.9715 1531.9025
        ;
        LOAD UU_DPUU ;
        <BLANKLINE>
        ? Nominale afschrijvingen op de kapitaalvoorraad (aangepast:
        ? inkomensoptiek).
        954.5012 1008.3662 1057.9921 1102.3636 1144.8248 1196.418 1253.7967 1317.2545
        1403.9357 1473.3624 1533.35
        ;
        LOAD UU_IFU ;
        <BLANKLINE>
        ? Bruto kapitaalvorming: ondernemingen.
        1076.1795 1136.1372 1150.8846 1197.4509 1277.7354 1399.6986 1490.8074 1562.3181
        1617.3543 1602.6872 1566.9738
        ;
        LOAD UU_IHU ;
        <BLANKLINE>
        ? Bruto kapitaalvorming: gezinnen.
        471.00145 486.53108 514.2291 552.55878 600.62646 634.63051 655.61486 671.66908
        665.98197 697.87295 726.93744
        ;
        LOAD UU_MU ;
        <BLANKLINE>
        0.42001992 0.40711156 0.39212964 0.37923534 0.36772624 0.35617242 0.34370718
        0.32978662 0.31416594 0.29691377 0.27826033
        ;
        LOAD UU_NAWRU ;
        <BLANKLINE>
        0.14141811 0.14138538 0.14125761 0.14106277 0.14083541 0.14059196 0.14034559
        0.1401158 0.13991636 0.13975922 0.13964585
        ;
        LOAD UU_WBU ;
        <BLANKLINE>
        ? Totale loonmassa (inclusief werkgeversbijdragen).
        4922.5664 5138.9458 5341.3233 5556.2476 5696.1652 5814.7965 6015.8951 6295.5108
        6650.3069 6861.5824 7072.7855
        ;
        '''
    def print_to_file(self, filepath: str | Path, names: str | list[str] = None, format: str = None):
        '''
        Print the list variables defined by `names` to the file `filepath` using the format `format`.

        Argument `format` must be in the list:
        - \'H\' (HTML file)
        - \'M\' (MIF file)
        - \'R\' (RTF file)
        - \'C\' (CSV file)

        If argument `format` is null (default), the *A2M* format will be used to print the output.

        If the filename does not contain an extension, it is automatically added based on 
        the value of `format`.

        If `names` is a string, it is considered as a *pattern* and the function will print 
        all variables matching the pattern. The following characters in *pattern* have a 
        special meaning:
        
            - `*` : any character sequence, even empty
            - `?` : any character (one and only one)
            - `@` : any alphanumerical char [A-Za-z0-9]
            - `&` : any non alphanumerical char
            - `|` : any alphanumeric character or none at the beginning and end of a string 
            - `!` : any non-alphanumeric character or none at the beginning and end of a string 
            - `\\` : escape the next character

        If `names` is None, print all variables of the (subset of the) current database.

        Parameters
        ----------
        filepath: str or Path
            path to the file to print.
            If the filename does not contain an extension, it is automatically 
            added based on the value of the format argument.
        names: str or list of str, optional
            pattern or list of names of the variables to print.
            If None, print all variables of the (subset of the) current database.
            Defaults to None.
        format: str, optional
            format of the output file. Possible values are: \'H\' (HTML file), 
            \'M\' (MIF file), \'R\' (RTF file) or \'C\' (CSV file).
            Defaults to None meaning that the variables will be dumped in the *A2M* format.

        Examples
        --------
        >>> from iode import variables, SAMPLE_DATA_DIR
        >>> output_dir = getfixture(\'tmp_path\')
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")         # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> variables.print_to_file(output_dir / "variables.csv", ["ACAF", "ACAG"])       # doctest: +ELLIPSIS
        Printing IODE objects definition to file \'...variables.csv\'...
        Printing ACAF ...
        Printing ACAG ...
        Print done
        >>> with open(output_dir / "variables.csv") as f:                     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        ...     print(f.read())
        ...
        <BLANKLINE>
        "Name","1960Y1","1961Y1","1962Y1","1963Y1",...,"2012Y1","2013Y1","2014Y1","2015Y1",
        <BLANKLINE>
        "ACAF","#N/A","#N/A","#N/A","#N/A",...,"-55.55929","-68.894654","-83.340625","-96.41042",
        "ACAG","#N/A","#N/A","#N/A","#N/A",...,"30.323961","31.370139","32.420299","33.469601",
        <BLANKLINE>
        '''
    def __hash__(self) -> int:
        '''
        Return a hash value for the current Variables database.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR, variables
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")        # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        Loading .../fun.var
        394 objects loaded
        >>> len(variables)
        394
        >>> original_hash = hash(variables)
        
        >>> # rename 1 variable
        >>> variables.rename("ACAF", "ACAF_")
        >>> original_hash == hash(variables)
        False
        >>> variables.rename("ACAF_", "ACAF")  # revert the change
        >>> original_hash == hash(variables)
        True

        >>> # modify one variable
        >>> original_variable = variables["ACAF"].copy()
        >>> variables["ACAF"] = 0.0
        >>> original_hash == hash(variables)
        False
        >>> variables["ACAF"] = original_variable  # revert the change
        >>> original_hash == hash(variables)
        True

        >>> # delete a variable
        >>> original_variable = variables["ACAF"].copy()
        >>> del variables["ACAF"]
        >>> original_hash == hash(variables)
        False
        >>> variables["ACAF"] = original_variable
        >>> original_hash == hash(variables)
        True

        >>> # add a variable
        >>> variables["NEW"] = 0.0
        >>> original_hash == hash(variables)
        False
        >>> del variables["NEW"]
        >>> original_hash == hash(variables)
        True
        '''

variables: Variables
