import enum
import typing
from _typeshed import Incomplete
from iode.util import JUSTIFY as JUSTIFY, check_filepath as check_filepath, join_lines as join_lines, split_list as split_list, table2str as table2str
from larray.core.array import Array
from pandas.core.frame import DataFrame
from pandas.core.series import Series
from pathlib import Path
from typing import Any, Callable, ClassVar, Self, overload

EQ_TEST_NAMES: tuple
ESTIMATION_EPS: float
ESTIMATION_MAXIT: int
IODE_DEFAULT_DATABASE_FILENAME: str
NA: float
NAN_REP: str
PERIODICITY_LIST: list
__test__: dict
__version__: str
comments: Comments
equations: Equations
identities: Identities
installation_dir: str
lists: Lists
scalars: Scalars
tables: Tables
variables: Variables

class AdjustmentMethod(enum.IntEnum):
    __new__: ClassVar[Callable] = ...
    ERROR_CORRECTION: ClassVar[AdjustmentMethod] = ...
    PARTIAL: ClassVar[AdjustmentMethod] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[int]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    def __format__(self, *args, **kwargs) -> str: ...

class Comments(_AbstractDatabase):
    __pyx_vtable__: ClassVar[PyCapsule] = ...
    i: PositionalIndexer
    series: Series
    def __init__(self, *args, **kwargs) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""
    @overload
    def from_series(self, s: Series) -> Any:
        '''from_series(self, s: Series)

                Copy the pandas Series `s` into the IODE Comments database.
                The comment names to copy are deduced from the index of the Series.

                Parameters
                ----------
                s: Series
                    pandas Series containing the comments to copy into the IODE Comments database.

                Notes
                -----
                The index of the passed Series is sorted in alphabetical order before 
                copying to IODE Comments database.

                See Also
                --------
                Comments.to_series

                Examples
                --------
                >>> from iode import comments
                >>> import pandas as pd
                >>> comments.clear()
                >>> len(comments)
                0

                >>> # create the pandas Series
                >>> names = ["A0", "A1", "B0", "B1", "C0", "C1"]
                >>> data = ["A zero", "A one", "B zero", "B one", "C zero", "C one"]
                >>> s = pd.Series(data=data, index=names, dtype=str, name="Comments")
                >>> # display the pandas series
                >>> s          # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                A0     A zero
                A1     A one
                B0    B zero
                B1     B one
                C0    C zero
                C1     C one
                Name: Comments, dtype: object

                >>> # load into the IODE Comments database
                >>> comments.from_series(s)
                >>> len(comments)
                6
        
                >>> comments.names             # doctest: +ELLIPSIS
                [\'A0\', \'A1\', \'B0\', \'B1\', \'C0\', \'C1\']
                >>> comments["B0"]
                \'B zero\'
                >>> comments["C1"]
                \'C one\'
        '''
    @overload
    def from_series(cls, s) -> Any:
        '''from_series(self, s: Series)

                Copy the pandas Series `s` into the IODE Comments database.
                The comment names to copy are deduced from the index of the Series.

                Parameters
                ----------
                s: Series
                    pandas Series containing the comments to copy into the IODE Comments database.

                Notes
                -----
                The index of the passed Series is sorted in alphabetical order before 
                copying to IODE Comments database.

                See Also
                --------
                Comments.to_series

                Examples
                --------
                >>> from iode import comments
                >>> import pandas as pd
                >>> comments.clear()
                >>> len(comments)
                0

                >>> # create the pandas Series
                >>> names = ["A0", "A1", "B0", "B1", "C0", "C1"]
                >>> data = ["A zero", "A one", "B zero", "B one", "C zero", "C one"]
                >>> s = pd.Series(data=data, index=names, dtype=str, name="Comments")
                >>> # display the pandas series
                >>> s          # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                A0     A zero
                A1     A one
                B0    B zero
                B1     B one
                C0    C zero
                C1     C one
                Name: Comments, dtype: object

                >>> # load into the IODE Comments database
                >>> comments.from_series(s)
                >>> len(comments)
                6
        
                >>> comments.names             # doctest: +ELLIPSIS
                [\'A0\', \'A1\', \'B0\', \'B1\', \'C0\', \'C1\']
                >>> comments["B0"]
                \'B zero\'
                >>> comments["C1"]
                \'C one\'
        '''
    @overload
    def to_series(self) -> Series:
        '''to_series(self) -> Series

                Create a pandas Series from the current Comments database.
                The index of the returned Series is build from the Comments names.

                See Also
                --------
                Comments.from_series

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, comments
                >>> import pandas as pd
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> len(comments)
                317

                >>> # Export the IODE Comments database as a pandas Series
                >>> s = comments.to_series()
                >>> len(s)
                317

                >>> s.index.to_list()               # doctest: +ELLIPSIS
                [\'ACAF\', \'ACAG\', \'AOUC\', \'AQC\', ..., \'ZJ\', \'ZKF\', \'ZX\', \'ZZ_\']
                >>> comments["ACAF"]                # doctest: +NORMALIZE_WHITESPACE
                \'Ondernemingen: ontvangen kapitaaloverdrachten.\' 
                >>> s["ACAF"]                       # doctest: +NORMALIZE_WHITESPACE
                \'Ondernemingen: ontvangen kapitaaloverdrachten.\' 
                >>> comments["ZZ_"]                 # doctest: +NORMALIZE_WHITESPACE
                \'Marktsector (ondernemingen en zelfstandigen): loonquote\\n(gemiddelde 1954-94).\'
                >>> s["ZZ_"]                        # doctest: +NORMALIZE_WHITESPACE
                \'Marktsector (ondernemingen en zelfstandigen): loonquote\\n(gemiddelde 1954-94).\'

                >>> # Export a subset of the IODE Comments database as a pandas Series
                >>> s = comments["A*;*_"].to_series()
                >>> len(s)
                34

                >>> s.index.to_list()               # doctest: +ELLIPSIS
                [\'ACAF\', \'ACAG\', \'AOUC\', \'AQC\', ..., \'WIND_\', \'WNF_\', \'YDH_\', \'ZZ_\']
                >>> comments["ACAF"]                # doctest: +NORMALIZE_WHITESPACE
                \'Ondernemingen: ontvangen kapitaaloverdrachten.\' 
                >>> s["ACAF"]                       # doctest: +NORMALIZE_WHITESPACE
                \'Ondernemingen: ontvangen kapitaaloverdrachten.\' 
                >>> comments["ZZ_"]                 # doctest: +NORMALIZE_WHITESPACE
                \'Marktsector (ondernemingen en zelfstandigen): loonquote\\n(gemiddelde 1954-94).\'
                >>> s["ZZ_"]                        # doctest: +NORMALIZE_WHITESPACE
                \'Marktsector (ondernemingen en zelfstandigen): loonquote\\n(gemiddelde 1954-94).\'
        '''
    @overload
    def to_series(cls) -> Any:
        '''to_series(self) -> Series

                Create a pandas Series from the current Comments database.
                The index of the returned Series is build from the Comments names.

                See Also
                --------
                Comments.from_series

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, comments
                >>> import pandas as pd
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> len(comments)
                317

                >>> # Export the IODE Comments database as a pandas Series
                >>> s = comments.to_series()
                >>> len(s)
                317

                >>> s.index.to_list()               # doctest: +ELLIPSIS
                [\'ACAF\', \'ACAG\', \'AOUC\', \'AQC\', ..., \'ZJ\', \'ZKF\', \'ZX\', \'ZZ_\']
                >>> comments["ACAF"]                # doctest: +NORMALIZE_WHITESPACE
                \'Ondernemingen: ontvangen kapitaaloverdrachten.\' 
                >>> s["ACAF"]                       # doctest: +NORMALIZE_WHITESPACE
                \'Ondernemingen: ontvangen kapitaaloverdrachten.\' 
                >>> comments["ZZ_"]                 # doctest: +NORMALIZE_WHITESPACE
                \'Marktsector (ondernemingen en zelfstandigen): loonquote\\n(gemiddelde 1954-94).\'
                >>> s["ZZ_"]                        # doctest: +NORMALIZE_WHITESPACE
                \'Marktsector (ondernemingen en zelfstandigen): loonquote\\n(gemiddelde 1954-94).\'

                >>> # Export a subset of the IODE Comments database as a pandas Series
                >>> s = comments["A*;*_"].to_series()
                >>> len(s)
                34

                >>> s.index.to_list()               # doctest: +ELLIPSIS
                [\'ACAF\', \'ACAG\', \'AOUC\', \'AQC\', ..., \'WIND_\', \'WNF_\', \'YDH_\', \'ZZ_\']
                >>> comments["ACAF"]                # doctest: +NORMALIZE_WHITESPACE
                \'Ondernemingen: ontvangen kapitaaloverdrachten.\' 
                >>> s["ACAF"]                       # doctest: +NORMALIZE_WHITESPACE
                \'Ondernemingen: ontvangen kapitaaloverdrachten.\' 
                >>> comments["ZZ_"]                 # doctest: +NORMALIZE_WHITESPACE
                \'Marktsector (ondernemingen en zelfstandigen): loonquote\\n(gemiddelde 1954-94).\'
                >>> s["ZZ_"]                        # doctest: +NORMALIZE_WHITESPACE
                \'Marktsector (ondernemingen en zelfstandigen): loonquote\\n(gemiddelde 1954-94).\'
        '''
    @overload
    def to_series(cls) -> Any:
        '''to_series(self) -> Series

                Create a pandas Series from the current Comments database.
                The index of the returned Series is build from the Comments names.

                See Also
                --------
                Comments.from_series

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, comments
                >>> import pandas as pd
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> len(comments)
                317

                >>> # Export the IODE Comments database as a pandas Series
                >>> s = comments.to_series()
                >>> len(s)
                317

                >>> s.index.to_list()               # doctest: +ELLIPSIS
                [\'ACAF\', \'ACAG\', \'AOUC\', \'AQC\', ..., \'ZJ\', \'ZKF\', \'ZX\', \'ZZ_\']
                >>> comments["ACAF"]                # doctest: +NORMALIZE_WHITESPACE
                \'Ondernemingen: ontvangen kapitaaloverdrachten.\' 
                >>> s["ACAF"]                       # doctest: +NORMALIZE_WHITESPACE
                \'Ondernemingen: ontvangen kapitaaloverdrachten.\' 
                >>> comments["ZZ_"]                 # doctest: +NORMALIZE_WHITESPACE
                \'Marktsector (ondernemingen en zelfstandigen): loonquote\\n(gemiddelde 1954-94).\'
                >>> s["ZZ_"]                        # doctest: +NORMALIZE_WHITESPACE
                \'Marktsector (ondernemingen en zelfstandigen): loonquote\\n(gemiddelde 1954-94).\'

                >>> # Export a subset of the IODE Comments database as a pandas Series
                >>> s = comments["A*;*_"].to_series()
                >>> len(s)
                34

                >>> s.index.to_list()               # doctest: +ELLIPSIS
                [\'ACAF\', \'ACAG\', \'AOUC\', \'AQC\', ..., \'WIND_\', \'WNF_\', \'YDH_\', \'ZZ_\']
                >>> comments["ACAF"]                # doctest: +NORMALIZE_WHITESPACE
                \'Ondernemingen: ontvangen kapitaaloverdrachten.\' 
                >>> s["ACAF"]                       # doctest: +NORMALIZE_WHITESPACE
                \'Ondernemingen: ontvangen kapitaaloverdrachten.\' 
                >>> comments["ZZ_"]                 # doctest: +NORMALIZE_WHITESPACE
                \'Marktsector (ondernemingen en zelfstandigen): loonquote\\n(gemiddelde 1954-94).\'
                >>> s["ZZ_"]                        # doctest: +NORMALIZE_WHITESPACE
                \'Marktsector (ondernemingen en zelfstandigen): loonquote\\n(gemiddelde 1954-94).\'
        '''
    def __delitem__(self, other) -> None:
        '''
        Remove the (subset of) comment(s) referenced by `key` from the Comments database.

        Parameters
        ----------
        key: str or list(str)
            (list of) name(s) of the comment(s) to be removed.
            The list of names can be given as a string pattern (e.g. "A*;*_").

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import comments
        >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")

        >>> # a) delete one comment
        >>> comments.get_names("A*")
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AQC\']
        >>> del comments["ACAF"]
        >>> comments.get_names("A*")
        [\'ACAG\', \'AOUC\', \'AQC\']

        >>> # b) delete several comments at once using a pattern
        >>> del comments["A*"]
        >>> comments.get_names("A*")
        []

        >>> # c) delete several comments at once using a list of names
        >>> comments.get_names("B*")
        [\'BENEF\', \'BENEF_\', \'BQY\', \'BVY\']
        >>> del comments[["BENEF", "BQY"]]
        >>> comments.get_names("B*")
        [\'BENEF_\', \'BVY\']

        >>> # delete one comment from a subset of the global database
        >>> comments_subset = comments["D*"]
        >>> comments_subset.names
        [\'DPU\', \'DPUF\', \'DPUG\', \'DPUGO\', \'DPUH\', \'DPUU\', \'DTF\', \'DTFX\', \'DTH\', \'DTH1\', \'DTH1C\', \'DTHX\']
        >>> del comments_subset["DPUGO"]
        >>> comments_subset.names
        [\'DPU\', \'DPUF\', \'DPUG\', \'DPUH\', \'DPUU\', \'DTF\', \'DTFX\', \'DTH\', \'DTH1\', \'DTH1C\', \'DTHX\']
        >>> # NOTE: the comment has also been deleted from the global database
        >>> "DPUGO" in comments
        False
        >>> comments.get_names("D*")
        [\'DPU\', \'DPUF\', \'DPUG\', \'DPUH\', \'DPUU\', \'DTF\', \'DTFX\', \'DTH\', \'DTH1\', \'DTH1C\', \'DTHX\']
        '''
    def __getitem__(self, index):
        '''
        Return the (subset of) comment(s) referenced by `key`.

        The `key` can represent a single object name (e.g. "ACAF") or a list of object names ("ACAF;ACAG;AOUC") 
        or a pattern (e.g. "A*") or a list of sub-patterns (e.g. "A*;*_").
        
        If the `key` represents a list of object names or of sub-patterns, each name or sub-pattern is separated 
        by a `separator` character which is either a whitespace ` `, or a comma `,`, or a semi-colon `;`, or a 
        tabulation `\\t`, or a newline `\\n`.

        A (sub-)`pattern` is a list of characters representing a group of object names. 
        It includes some special characters which have a special meaning:
        
            - `*` : any character sequence, even empty
            - `?` : any character (one and only one)
            - `@` : any alphanumerical char [A-Za-z0-9]
            - `&` : any non alphanumerical char
            - `|` : any alphanumeric character or none at the beginning and end of a string 
            - `!` : any non-alphanumeric character or none at the beginning and end of a string 
            - `\\` : escape the next character

        Note that the `key` can contain references to IODE lists which are prefixed with the symbol `$`.

        Parameters
        ----------
        key: str or list(str)
            (the list of) name(s) of the comment(s) to get.
            The list of objects to get can be specified by a pattern or by a list of sub-patterns (e.g. "A*;*_").

        Returns
        -------
        Single comment or a subset of the database.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import comments
        >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")

        >>> # a) get one Comment
        >>> comments["ACAF"]
        \'Ondernemingen: ontvangen kapitaaloverdrachten.\'

        >>> # b) get a subset of the Comments database using a pattern
        >>> comments_subset = comments["A*"]
        >>> comments_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AQC\']
        
        >>> # c) get a subset of the Comments database using a list of names
        >>> comments_subset = comments[["ACAF", "AOUC", "BQY", "BVY"]]
        >>> comments_subset.names
        [\'ACAF\', \'AOUC\', \'BQY\', \'BVY\']
        '''
    def __hash__(self) -> int:
        '''
        Return a hash value for the current Comments database.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR, comments
        >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
        >>> len(comments)
        317
        >>> original_hash = hash(comments)
        
        >>> # rename 1 comment
        >>> comments.rename("ACAF", "ACAF_")
        >>> original_hash == hash(comments)
        False
        >>> comments.rename("ACAF_", "ACAF")  # revert the change
        >>> original_hash == hash(comments)
        True

        >>> # modify one comment
        >>> original_comment = comments["ACAF"]
        >>> comments["ACAF"] = "modified comment"
        >>> original_hash == hash(comments)
        False
        >>> comments["ACAF"] = original_comment  # revert the change
        >>> original_hash == hash(comments)
        True

        >>> # delete a comment
        >>> original_comment = comments["ACAF"]
        >>> del comments["ACAF"]
        >>> original_hash == hash(comments)
        False
        >>> comments["ACAF"] = original_comment
        >>> original_hash == hash(comments)
        True

        >>> # add a comment
        >>> comments["NEW"] = "new comment"
        >>> original_hash == hash(comments)
        False
        >>> del comments["NEW"]
        >>> original_hash == hash(comments)
        True
        '''
    def __reduce__(self):
        """__reduce_cython__(self)"""
    def __setitem__(self, index, object) -> None:
        '''
        Update/add a (subset of) comment(s) referenced by `key` from/to the Comments database.

        The `key` can represent a single object name (e.g. "ACAF") or a list of object names ("ACAF;ACAG;AOUC") 
        or a pattern (e.g. "A*") or a list of sub-patterns (e.g. "A*;*_").
        
        If the `key` represents a list of object names or of sub-patterns, each name or sub-pattern is separated 
        by a `separator` character which is either a whitespace ` `, or a comma `,`, or a semi-colon `;`, or a 
        tabulation `\\t`, or a newline `\\n`.

        A (sub-)`pattern` is a list of characters representing a group of object names. 
        It includes some special characters which have a special meaning:
        
            - `*` : any character sequence, even empty
            - `?` : any character (one and only one)
            - `@` : any alphanumerical char [A-Za-z0-9]
            - `&` : any non alphanumerical char
            - `|` : any alphanumeric character or none at the beginning and end of a string 
            - `!` : any non-alphanumeric character or none at the beginning and end of a string 
            - `\\` : escape the next character

        Note that the `key` can contain references to IODE lists which are prefixed with the symbol `$`.

        Parameters
        ----------
        key: str or list(str)
            (the list of) name(s) of the comment(s) to update/add.
            The list of objects to update/add can be specified by a pattern or by a list of sub-patterns 
            (e.g. "A*;*_").
        value: str or list(str)
            (new) comment value(s).

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import comments
        >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
        
        >>> # a) add one comment
        >>> comments["BDY"] = "Difference net incomes (YN - YK)"
        >>> comments["BDY"]
        \'Difference net incomes (YN - YK)\'

        >>> # b) update one comment
        >>> comments["ACAF"]
        \'Ondernemingen: ontvangen kapitaaloverdrachten.\'
        >>> comments["ACAF"] = "New Value"
        >>> comments["ACAF"]
        \'New Value\'

        >>> # c) working on a subset
        >>> # 1) get subset
        >>> comments_subset = comments["A*"]
        >>> comments_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AQC\']
        >>> # 2) add a comment to the subset 
        >>> comments_subset["A0"] = "New Comment"
        >>> comments_subset["A0"]
        \'New Comment\'
        >>> # --> new comment also appears in the global workspace
        >>> "A0" in comments
        True
        >>> comments["A0"]
        \'New Comment\'
        >>> # 3) update a comment in the subset
        >>> comments_subset["A0"] = "Updated Comment"
        >>> comments_subset["A0"]
        \'Updated Comment\'
        >>> # --> comment is also updated in the global workspace
        >>> comments["A0"]
        \'Updated Comment\'
        '''

class ComputedTable:
    __pyx_vtable__: ClassVar[PyCapsule] = ...
    columns: list[str]
    files: list[str]
    lines: list[str]
    nb_columns: int
    nb_decimals: int
    nb_files: int
    nb_lines: int
    nb_operations_between_files: int
    nb_periods: int
    sample: Sample
    title: str
    def __init__(self, *args, **kwargs) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""
    def is_editable(self, row: int | str, column: int | str) -> bool:
        '''is_editable(self, row: Union[int, str], column: Union[int, str]) -> bool

                Check if a cell in the computed table is editable.

                A cell is **not** editable if:

                  1. the corresponding column contains on operation on periods or files
                  2. the corresponding column does not refer to the current workspace
                  3. the corresponding LEC expression from the original table starts with 0+
                  4. the corresponding LEC expression from the original table does not refer to at least one variable

                Parameters
                ----------
                row : int or str
                    The row passed as position (int) or as name (str).
                column : int or str
                    The column passed as position (int) or as name (str).

                Returns
                -------
                bool: True if the cell is editable, False otherwise.

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import Table, comments, variables
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")

                >>> title = "Example Table"
                >>> lecs = ["Q_F", "Q_I", "Q_F/Q_I", "ln(Q_I+Q_F)", "KNFF", "KNFF[-1]", "3+ln(10)", "0+KNFF"]
                >>> table = Table(2, title, lecs)
                >>> table                       # doctest: +NORMALIZE_WHITESPACE
                DIVIS | 1             |
                TITLE |       "Example Table"
                ----- | ---------------------------
                CELL  |               |     "#S"
                ----- | ---------------------------
                CELL  | "Q_F"         |         Q_F
                CELL  | "Q_I"         |         Q_I
                CELL  | "Q_F/Q_I"     |     Q_F/Q_I
                CELL  | "ln(Q_I+Q_F)" | ln(Q_I+Q_F)
                CELL  | "KNFF"        |        KNFF
                CELL  | "KNFF[-1]"    |    KNFF[-1]
                CELL  | "3+ln(10)"    |    3+ln(10)
                CELL  | "0+KNFF"      |      0+KNFF
                <BLANKLINE>
                nb lines: 12
                nb columns: 2
                language: \'ENGLISH\'
                gridx: \'MAJOR\'
                gridy: \'MAJOR\'
                graph_axis: \'VALUES\'
                graph_alignment: \'LEFT\'
                <BLANKLINE>

                >>> # compute table
                >>> computed_table = table.compute("(2010;2010/2009)[1;2]:2", extra_files=f"{SAMPLE_DATA_DIR}/ref.av")
                >>> computed_table                      # doctest: +NORMALIZE_WHITESPACE
                 line title \\ period[file]  | 2010[1]  | 2010[2]  | 2010/2009[1] | 2010/2009[2] | 2011[1]  | 2011[2]  | 2011/2010[1] | 2011/2010[2]
                ------------------------------------------------------------------------------------------------------------------------------------
                Q_F                         |  5842.74 |  5725.89 |         1.90 |         1.90 |  5930.75 |  5812.13 |         1.51 |         1.51
                Q_I                         |  1093.37 |  1071.50 |         0.89 |         0.89 |  1114.60 |  1092.30 |         1.94 |         1.94
                Q_F/Q_I                     |     5.34 |     5.34 |         1.00 |         1.00 |     5.32 |     5.32 |        -0.43 |        -0.43
                ln(Q_I+Q_F)                 |     8.84 |     8.82 |         0.20 |         0.20 |     8.86 |     8.84 |         0.18 |         0.18
                KNFF                        | 11525.01 | 11294.51 |         2.05 |         2.05 | 11736.78 | 11502.05 |         1.84 |         1.84
                KNFF[-1]                    | 11293.85 | 11067.97 |         2.82 |         2.82 | 11525.01 | 11294.51 |         2.05 |         2.05
                3+ln(10)                    |     5.30 |     5.30 |         0.00 |         0.00 |     5.30 |     5.30 |         0.00 |         0.00
                0+KNFF                      | 11525.01 | 11294.51 |         2.05 |         2.05 | 11736.78 | 11502.05 |         1.84 |         1.84
                <BLANKLINE>

                >>> computed_table.is_editable("Q_F", "2011[1]")
                True
                >>> # 1. cell is not editable if the corresponding column contains on operation on periods or files
                >>> computed_table.is_editable("Q_F", "2010/2009[1]")
                False
                >>> # 2. cell is not editable if the corresponding column does not refer to the current workspace
                >>> computed_table.is_editable("Q_F", "2011[2]")
                False
                >>> # 3. cell is not editable if the corresponding LEC expression from the original table starts with 0+
                >>> computed_table.is_editable("0+KNFF", "2011[1]")
                False
                >>> # 4. cell is not editable if the corresponding LEC expression from the original table 
                >>> #    does not refer to at least one variable
                >>> computed_table.is_editable("3+ln(10)", "2011[1]")
                False
        '''
    def print_to_file(self, destination_file: str | Path, format: str = ...) -> Any:
        '''print_to_file(self, destination_file: Union[str, Path], format: str = None)

                Print the present computed table to a file.

                Argument `format` must be in the list:
                - \'H\' (HTML file)
                - \'M\' (MIF file)
                - \'R\' (RTF file)
                - \'C\' (CSV file)
                - \'D\' (DUMMY file)

                If argument `format` is null (default), the A2M format will be used
                to print the output.

                If the filename does not contain an extension, it is automatically
                added based on the format.

                Parameters
                ----------
                destination_file: str or Path 
                    The destination file for printing
                format: str, optional
                    The format of the output file. Deduced from the extension if not provided.
                    If destination_file has no extension and format is None, the A2M format is used.
            
                Examples
                --------
                >>> from pathlib import Path
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import Table, tables, variables
                >>> tables.load(f"{SAMPLE_DATA_DIR}/fun.tbl")
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")

                >>> computed_table = tables["C8_1"].compute("(2010;2010/2009):5")
                >>> computed_table              # doctest: +NORMALIZE_WHITESPACE, +ELLIPSIS
                   line title \\ period[file]     |    10    | 10/09 |    11    | 11/10 | ... |    14    | 14/13
                -------------------------------------------------------------------------...--------------------
                Output potentiel                 |  6936.11 |  1.74 |  7045.34 |  1.57 | ... |  7460.12 |  2.16
                Stock de capital                 | 11293.85 |  2.82 | 11525.01 |  2.05 | ... | 12263.95 |  2.41
                Intensité de capital             |     0.39 | -2.17 |     0.38 | -2.05 | ... |     0.36 | -1.90
                Productivité totale des facteurs |     1.10 |  1.00 |     1.11 |  1.00 | ... |     1.14 |  1.00
                <BLANKLINE>
                >>> computed_table.print_to_file(tmp_dir / "computed_table_2_periods.csv")
                >>> with open(tmp_dir / "computed_table_2_periods.csv", "r") as f:    # doctest: +NORMALIZE_WHITESPACE, +ELLIPSIS
                ...     print(f.read())
                "Déterminants de l\'output potentiel"
                <BLANKLINE>
                " ","10","10/09","11","11/10","12","12/11","13","13/12","14","14/13",
                <BLANKLINE>
                "Output potentiel","6936.11","1.74",...,"7460.12","2.16",       
                "Stock de capital","11293.85","2.82",...,"12263.95","2.41",  
                "Intensité de capital","0.39","-2.17",...,"0.36","-1.90",
                "Productivité totale des facteurs","1.10","1.00",...,"1.14","1.00",      
                <BLANKLINE>

                >>> extra_files = Path(SAMPLE_DATA_DIR) / "ref.av"
                >>> computed_table = tables["C8_1"].compute("2010[1-2]:5", extra_files=extra_files)
                >>> computed_table              # doctest: +NORMALIZE_WHITESPACE
                   line title \\ period[file]     | 10[1-2] | 11[1-2] | 12[1-2] | 13[1-2] | 14[1-2]
                -----------------------------------------------------------------------------------
                Output potentiel                 |  138.72 |  140.91 |  143.23 |  146.05 |  149.20
                Stock de capital                 |  225.88 |  230.50 |  234.74 |  239.51 |  245.28
                Intensité de capital             |    0.01 |    0.01 |    0.01 |    0.01 |    0.01
                Productivité totale des facteurs |    0.02 |    0.02 |    0.02 |    0.02 |    0.02
                <BLANKLINE>
                >>> computed_table.print_to_file(tmp_dir / "computed_table_2_files.csv")    
                >>> with open(tmp_dir / "computed_table_2_files.csv", "r") as f:
                ...     print(f.read())
                "Déterminants de l\'output potentiel"
                <BLANKLINE>
                " ","10[1-2]","11[1-2]","12[1-2]","13[1-2]","14[1-2]",
                <BLANKLINE>
                "Output potentiel","138.72","140.91","143.23","146.05","149.20",
                "Stock de capital","225.88","230.50","234.74","239.51","245.28",
                "Intensité de capital","0.01","0.01","0.01","0.01","0.01",
                "Productivité totale des facteurs","0.02","0.02","0.02","0.02","0.02",
                <BLANKLINE>
        '''
    @overload
    def to_array(self) -> Array:
        '''to_array(self) -> Array

                Convert the computed table to a larray Array.

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import Table, tables, variables
                >>> tables.load(f"{SAMPLE_DATA_DIR}/fun.tbl")
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")

                >>> # simple time series (current workspace) - 6 observations
                >>> computed_table = tables["C8_1"].compute("2000:6")
                >>> computed_table              # doctest: +NORMALIZE_WHITESPACE
                   line title \\ period[file]     |    00   |    01   |    02   |    03   |    04   |    05
                ---------------------------------------------------------------------------------------------
                Output potentiel                 | 5495.21 | 5627.86 | 5748.78 | 5857.95 | 5966.20 | 6103.63
                Stock de capital                 | 8083.55 | 8359.89 | 8647.94 | 8910.34 | 9175.81 | 9468.89
                Intensité de capital             |    0.50 |    0.49 |    0.48 |    0.46 |    0.45 |    0.43
                Productivité totale des facteurs |    0.99 |    1.00 |    1.01 |    1.02 |    1.03 |    1.04
                <BLANKLINE>
                >>> array = computed_table.to_array()
                >>> array                       # doctest: +NORMALIZE_WHITESPACE
                name\\period[file]                  00  ...                  05
                 Output potentiel   5495.212781863819  ...   6103.631843893574
                 Stock de capital   8083.551747773469  ...   9468.886506037787
                     Intensité de  0.5031658911658933  ...  0.4349138159428717
                          capital
                     Productivité   0.993773299183911  ...  1.0444657249633247
                       totale des
                         facteurs
        '''
    @overload
    def to_array(cls) -> Any:
        '''to_array(self) -> Array

                Convert the computed table to a larray Array.

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import Table, tables, variables
                >>> tables.load(f"{SAMPLE_DATA_DIR}/fun.tbl")
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")

                >>> # simple time series (current workspace) - 6 observations
                >>> computed_table = tables["C8_1"].compute("2000:6")
                >>> computed_table              # doctest: +NORMALIZE_WHITESPACE
                   line title \\ period[file]     |    00   |    01   |    02   |    03   |    04   |    05
                ---------------------------------------------------------------------------------------------
                Output potentiel                 | 5495.21 | 5627.86 | 5748.78 | 5857.95 | 5966.20 | 6103.63
                Stock de capital                 | 8083.55 | 8359.89 | 8647.94 | 8910.34 | 9175.81 | 9468.89
                Intensité de capital             |    0.50 |    0.49 |    0.48 |    0.46 |    0.45 |    0.43
                Productivité totale des facteurs |    0.99 |    1.00 |    1.01 |    1.02 |    1.03 |    1.04
                <BLANKLINE>
                >>> array = computed_table.to_array()
                >>> array                       # doctest: +NORMALIZE_WHITESPACE
                name\\period[file]                  00  ...                  05
                 Output potentiel   5495.212781863819  ...   6103.631843893574
                 Stock de capital   8083.551747773469  ...   9468.886506037787
                     Intensité de  0.5031658911658933  ...  0.4349138159428717
                          capital
                     Productivité   0.993773299183911  ...  1.0444657249633247
                       totale des
                         facteurs
        '''
    @overload
    def to_frame(self) -> DataFrame:
        '''to_frame(self) -> DataFrame

                Convert the computed table to a pandas DataFrame.

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import Table, tables, variables
                >>> tables.load(f"{SAMPLE_DATA_DIR}/fun.tbl")
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")

                >>> # simple time series (current workspace) - 6 observations
                >>> computed_table = tables["C8_1"].compute("2000:6")
                >>> computed_table              # doctest: +NORMALIZE_WHITESPACE
                   line title \\ period[file]     |    00   |    01   |    02   |    03   |    04   |    05
                ---------------------------------------------------------------------------------------------
                Output potentiel                 | 5495.21 | 5627.86 | 5748.78 | 5857.95 | 5966.20 | 6103.63
                Stock de capital                 | 8083.55 | 8359.89 | 8647.94 | 8910.34 | 9175.81 | 9468.89
                Intensité de capital             |    0.50 |    0.49 |    0.48 |    0.46 |    0.45 |    0.43
                Productivité totale des facteurs |    0.99 |    1.00 |    1.01 |    1.02 |    1.03 |    1.04
                <BLANKLINE>
                >>> df = computed_table.to_frame()
                >>> df                          # doctest: +NORMALIZE_WHITESPACE
                period[file]                               00  ...           05
                name                                           ...
                Output potentiel                  5495.212782  ...  6103.631844
                Stock de capital                  8083.551748  ...  9468.886506
                Intensité de capital                 0.503166  ...     0.434914
                Productivité totale des facteurs     0.993773  ...     1.044466
                <BLANKLINE>
                [4 rows x 6 columns]
        '''
    @overload
    def to_frame(cls) -> Any:
        '''to_frame(self) -> DataFrame

                Convert the computed table to a pandas DataFrame.

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import Table, tables, variables
                >>> tables.load(f"{SAMPLE_DATA_DIR}/fun.tbl")
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")

                >>> # simple time series (current workspace) - 6 observations
                >>> computed_table = tables["C8_1"].compute("2000:6")
                >>> computed_table              # doctest: +NORMALIZE_WHITESPACE
                   line title \\ period[file]     |    00   |    01   |    02   |    03   |    04   |    05
                ---------------------------------------------------------------------------------------------
                Output potentiel                 | 5495.21 | 5627.86 | 5748.78 | 5857.95 | 5966.20 | 6103.63
                Stock de capital                 | 8083.55 | 8359.89 | 8647.94 | 8910.34 | 9175.81 | 9468.89
                Intensité de capital             |    0.50 |    0.49 |    0.48 |    0.46 |    0.45 |    0.43
                Productivité totale des facteurs |    0.99 |    1.00 |    1.01 |    1.02 |    1.03 |    1.04
                <BLANKLINE>
                >>> df = computed_table.to_frame()
                >>> df                          # doctest: +NORMALIZE_WHITESPACE
                period[file]                               00  ...           05
                name                                           ...
                Output potentiel                  5495.212782  ...  6103.631844
                Stock de capital                  8083.551748  ...  9468.886506
                Intensité de capital                 0.503166  ...     0.434914
                Productivité totale des facteurs     0.993773  ...     1.044466
                <BLANKLINE>
                [4 rows x 6 columns]
        '''
    def __del__(self, *args, **kwargs) -> None: ...
    def __delitem__(self, other) -> None:
        """Delete self[key]."""
    def __getitem__(self, index):
        '''
        Get the value of a cell of the computed table.

        Parameters
        ----------
        key: tuple(int or str, int or str) or int or str
            Row and column of the cell to access. 
            Row and column can be passed either as position (int) or as name (str).

        Returns
        -------
        float: 
            The value of a given cell.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import Table, comments, variables
        >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")

        >>> title = "Example Table"
        >>> lecs = ["Q_F", "Q_I", "Q_F/Q_I", "ln(Q_I+Q_F)", "KNFF", "KNFF[-1]", "3+ln(10)", "0+KNFF"]
        >>> table = Table(2, title, lecs)
        >>> table                       # doctest: +NORMALIZE_WHITESPACE
        DIVIS | 1             |
        TITLE |       "Example Table"
        ----- | ---------------------------
        CELL  |               |     "#S"
        ----- | ---------------------------
        CELL  | "Q_F"         |         Q_F
        CELL  | "Q_I"         |         Q_I
        CELL  | "Q_F/Q_I"     |     Q_F/Q_I
        CELL  | "ln(Q_I+Q_F)" | ln(Q_I+Q_F)
        CELL  | "KNFF"        |        KNFF
        CELL  | "KNFF[-1]"    |    KNFF[-1]
        CELL  | "3+ln(10)"    |    3+ln(10)
        CELL  | "0+KNFF"      |      0+KNFF
        <BLANKLINE>
        nb lines: 12
        nb columns: 2
        language: \'ENGLISH\'
        gridx: \'MAJOR\'
        gridy: \'MAJOR\'
        graph_axis: \'VALUES\'
        graph_alignment: \'LEFT\'
        <BLANKLINE>

        >>> # compute table
        >>> computed_table = table.compute("(2010;2010/2009)[1;2]:2", extra_files=f"{SAMPLE_DATA_DIR}/ref.av")
        >>> computed_table                      # doctest: +NORMALIZE_WHITESPACE
         line title \\ period[file]  | 2010[1]  | 2010[2]  | 2010/2009[1] | 2010/2009[2] | 2011[1]  | 2011[2]  | 2011/2010[1] | 2011/2010[2]
        ------------------------------------------------------------------------------------------------------------------------------------
        Q_F                         |  5842.74 |  5725.89 |         1.90 |         1.90 |  5930.75 |  5812.13 |         1.51 |         1.51
        Q_I                         |  1093.37 |  1071.50 |         0.89 |         0.89 |  1114.60 |  1092.30 |         1.94 |         1.94
        Q_F/Q_I                     |     5.34 |     5.34 |         1.00 |         1.00 |     5.32 |     5.32 |        -0.43 |        -0.43
        ln(Q_I+Q_F)                 |     8.84 |     8.82 |         0.20 |         0.20 |     8.86 |     8.84 |         0.18 |         0.18
        KNFF                        | 11525.01 | 11294.51 |         2.05 |         2.05 | 11736.78 | 11502.05 |         1.84 |         1.84
        KNFF[-1]                    | 11293.85 | 11067.97 |         2.82 |         2.82 | 11525.01 | 11294.51 |         2.05 |         2.05
        3+ln(10)                    |     5.30 |     5.30 |         0.00 |         0.00 |     5.30 |     5.30 |         0.00 |         0.00
        0+KNFF                      | 11525.01 | 11294.51 |         2.05 |         2.05 | 11736.78 | 11502.05 |         1.84 |         1.84
        <BLANKLINE>

        >>> # get cell value by position
        >>> computed_table[2, 2]
        0.9975986300775119
        >>> # get cell value by labels
        >>> computed_table["Q_F/Q_I", "2010/2009[1]"]
        0.9975986300775119
        '''
    def __reduce__(self):
        """__reduce_cython__(self)"""
    def __setitem__(self, index, object) -> None:
        '''
        Assign a new value to a cell. 
        May modify the values of other cells.

        A cell is **not** editable if:

          1. the corresponding column contains on operation on periods or files
          2. the corresponding column does not refer to the current workspace
          3. the corresponding LEC expression from the original table starts with 0+
          4. the corresponding LEC expression from the original table does not refer to at least one variable

        Parameters
        ----------
        key: tuple(int or str, int or str) 
            Row and column to access the cell. 
            Row and column can be passed either as position (int) or as name (str).
        value: float 
            New value of the cell.

        Notes
        -----
        When the corresponding LEC expression involves more than one variable, only 
        the value of the first variable is modified.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import Table, comments, variables
        >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")

        >>> title = "Example Table"
        >>> lecs = ["Q_F", "Q_I", "Q_F/Q_I", "ln(Q_I+Q_F)", "KNFF", "KNFF[-1]", "3+ln(10)", "0+KNFF"]
        >>> table = Table(2, title, lecs)
        >>> table                       # doctest: +NORMALIZE_WHITESPACE
        DIVIS | 1             |
        TITLE |       "Example Table"
        ----- | ---------------------------
        CELL  |               |     "#S"
        ----- | ---------------------------
        CELL  | "Q_F"         |         Q_F
        CELL  | "Q_I"         |         Q_I
        CELL  | "Q_F/Q_I"     |     Q_F/Q_I
        CELL  | "ln(Q_I+Q_F)" | ln(Q_I+Q_F)
        CELL  | "KNFF"        |        KNFF
        CELL  | "KNFF[-1]"    |    KNFF[-1]
        CELL  | "3+ln(10)"    |    3+ln(10)
        CELL  | "0+KNFF"      |      0+KNFF
        <BLANKLINE>
        nb lines: 12
        nb columns: 2
        language: \'ENGLISH\'
        gridx: \'MAJOR\'
        gridy: \'MAJOR\'
        graph_axis: \'VALUES\'
        graph_alignment: \'LEFT\'
        <BLANKLINE>

        >>> # compute table
        >>> computed_table = table.compute("(2010;2010/2009)[1;2]:2", extra_files=f"{SAMPLE_DATA_DIR}/ref.av")
        >>> computed_table                      # doctest: +NORMALIZE_WHITESPACE
         line title \\ period[file]  | 2010[1]  | 2010[2]  | 2010/2009[1] | 2010/2009[2] | 2011[1]  | 2011[2]  | 2011/2010[1] | 2011/2010[2]
        ------------------------------------------------------------------------------------------------------------------------------------
        Q_F                         |  5842.74 |  5725.89 |         1.90 |         1.90 |  5930.75 |  5812.13 |         1.51 |         1.51
        Q_I                         |  1093.37 |  1071.50 |         0.89 |         0.89 |  1114.60 |  1092.30 |         1.94 |         1.94
        Q_F/Q_I                     |     5.34 |     5.34 |         1.00 |         1.00 |     5.32 |     5.32 |        -0.43 |        -0.43
        ln(Q_I+Q_F)                 |     8.84 |     8.82 |         0.20 |         0.20 |     8.86 |     8.84 |         0.18 |         0.18
        KNFF                        | 11525.01 | 11294.51 |         2.05 |         2.05 | 11736.78 | 11502.05 |         1.84 |         1.84
        KNFF[-1]                    | 11293.85 | 11067.97 |         2.82 |         2.82 | 11525.01 | 11294.51 |         2.05 |         2.05
        3+ln(10)                    |     5.30 |     5.30 |         0.00 |         0.00 |     5.30 |     5.30 |         0.00 |         0.00
        0+KNFF                      | 11525.01 | 11294.51 |         2.05 |         2.05 | 11736.78 | 11502.05 |         1.84 |         1.84
        <BLANKLINE>

        >>> # set cell value by position
        >>> computed_table[1, 4] = 1114.0
        >>> computed_table[1, 4]
        1114.0
        >>> # set cell value by labels
        >>> computed_table["Q_I", "2011[1]"] = 1115.0
        >>> computed_table["Q_I", "2011[1]"]
        1115.0

        >>> # warning: when the corresponding LEC expression involves more than one variable, 
        >>> #          only the value of the first variable is modified.
        >>> # variables values before:
        >>> variables["Q_F", "2011Y1"]
        5930.747852857564
        >>> variables["Q_I", "2011Y1"]
        1115.0
        >>> # update cell of the computed table
        >>> computed_table["Q_F/Q_I", "2011[1]"] = 5.28
        >>> # variables values after:
        >>> #   Q_F/Q_I = new_value 
        >>> #   Q_F = new_value * Q_I 
        >>> variables["Q_F", "2011Y1"]
        5887.20023393631
        >>> variables["Q_I", "2011Y1"]
        1115.0
        '''

class CorrelationMatrix:
    __pyx_vtable__: ClassVar[PyCapsule] = ...
    names: Incomplete
    shape: Incomplete
    def __init__(self, *args, **kwargs) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""
    def name(self, index: int) -> str:
        """name(self, index: int) -> str"""
    def __delitem__(self, other) -> None:
        """Delete self[key]."""
    def __getitem__(self, index):
        """Return self[key]."""
    def __len__(self) -> int:
        """Return len(self)."""
    def __reduce__(self):
        """__reduce_cython__(self)"""
    def __setitem__(self, index, object) -> None:
        """Set self[key] to value."""

class EditAndEstimateEquations:
    block: str
    correlation_matrix: CorrelationMatrix
    current_equation: Equation
    equations_db: Equations
    equations_list: Incomplete
    instruments: str | list[str]
    is_done: Incomplete
    method: str
    next_equation: Equation
    sample: Sample
    scalars_db: Scalars
    def __init__(self, *args, **kwargs) -> None:
        """Create and return a new object.  See help(type) for accurate signature."""
    def copy_eq_tests_values(self) -> Any:
        """copy_eq_tests_values(self)

                Extract and save the resulting values for the tests for each equations in the 
                local estimation Equations database.

                See :meth:`EditAndEstimateEquations.estimate` for the examples.

                Notes
                -----
                Equivalent to the ODE_blk_cur_atests() function from o_est.c from the old GUI (developed by Jean-Marc Paul)
        """
    @classmethod
    def estimate(cls, *args, **kwargs):
        '''estimate(self)

                Estimate the current block of equations (which is not necessarily all the equations 
                in the local Equations database \'kdb_eqs\').

                Notes
                -----
                Equivalent to ODE_blk_est() from o_est.c from the old GUI (developed by Jean-Marc Paul).

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, equations, scalars, variables, Equation
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> equations.clear()
                >>> scalars.clear()

                >>> equations["ACAF"] = Equation("ACAF", "(ACAF/VAF[-1]) := acaf1 + acaf2 * GOSF[-1] + acaf4 * (TIME=1995)")
                >>> equations["DPUH"] = Equation("DPUH", "dln(DPUH/DPUHO) := dpuh_1 + dpuh_2 * dln(IHU/PI5) + dln(PC)")

                >>> from iode.iode_python import EditAndEstimateEquations
                >>> estimation = EditAndEstimateEquations("1980Y1", "1996Y1")
                >>> estimation.block = "ACAF;DPUH", "ACAF"
                >>> estimation.block
                \'ACAF;DPUH\'

                >>> scalars_est = estimation.scalars_db
                >>> scalars_est.names
                []
                >>> estimation.update_scalars()
                >>> scalars_est.names
                [\'acaf1\', \'acaf2\', \'acaf4\', \'dpuh_1\', \'dpuh_2\']
                >>> scalars_est["acaf1"]
                Scalar(0.9, 1, na)
                >>> scalars_est["dpuh_1"]
                Scalar(0.9, 1, na)

                >>> equations_est = estimation.equations_db
                >>> equations_est["ACAF"]               # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'ACAF\',
                         lec = \'(ACAF/VAF[-1]) := acaf1 + acaf2 * GOSF[-1] + acaf4 * (TIME=1995)\',
                         method = \'LSQ\',
                         from_period = \'1960Y1\',
                         to_period = \'2015Y1\')
                >>> equations_est["DPUH"]               # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'DPUH\',
                         lec = \'dln(DPUH/DPUHO) := dpuh_1 + dpuh_2 * dln(IHU/PI5) + dln(PC)\',
                         method = \'LSQ\',
                         from_period = \'1960Y1\',
                         to_period = \'2015Y1\')

                >>> estimation.is_done
                False
                >>> estimation.estimate()
                >>> estimation.is_done
                True

                >>> scalars_est["acaf1"]
                Scalar(0.0157705, 1, 0.00136079)
                >>> scalars_est["dpuh_1"]
                Scalar(0.0109855, 1, 0.00481857)

                >>> corr_matrix = estimation.correlation_matrix
                >>> corr_matrix.names
                [\'acaf1\', \'acaf2\', \'acaf4\', \'dpuh_1\', \'dpuh_2\']
                >>> corr_matrix.shape
                (5, 5)
                >>> corr_matrix         # doctest: +NORMALIZE_WHITESPACE
                           |      acaf1       acaf2       acaf4      dpuh_1      dpuh_2
                ------------------------------------------------------------------------
                     acaf1 |          1   -0.935266    0.200167   0.0448324  -0.0372903
                     acaf2 |  -0.935266           1   -0.300833  -0.0016619   0.0395814
                     acaf4 |   0.200167   -0.300833           1  0.00037477  -0.00892588
                    dpuh_1 |  0.0448324  -0.0016619  0.00037477           1  -0.0419869
                    dpuh_2 | -0.0372903   0.0395814  -0.00892588  -0.0419869           1
                <BLANKLINE>

                >>> estimation.copy_eq_tests_values()
                >>> equations_est["ACAF"]             # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'ACAF\',
                        lec = \'(ACAF/VAF[-1]) := acaf1 + acaf2 * GOSF[-1] + acaf4 * (TIME=1995)\',
                        method = \'LSQ\',
                        from_period = \'1960Y1\',
                        to_period = \'2015Y1\',
                        tests = {corr = 1,
                                 dw = 2.33007,
                                 fstat = 32.2851,
                                 loglik = 83.8101,
                                 meany = 0.00818467,
                                 r2 = 0.821815,
                                 r2adj = 0.79636,
                                 ssres = 5.19787e-05,
                                 stderr = 0.00192685,
                                 stderrp = 23.5422,
                                 stdev = 0.0042699})
                >>> equations_est["DPUH"]             # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'DPUH\',
                        lec = \'dln(DPUH/DPUHO) := dpuh_1 + dpuh_2 * dln(IHU/PI5) + dln(PC)\',
                        method = \'LSQ\',
                        from_period = \'1960Y1\',
                        to_period = \'2015Y1\',
                        tests = {corr = 0.126096,
                                 dw = 3.15593,
                                 fstat = 3.51611,
                                 loglik = 43.5743,
                                 meany = 0.0505132,
                                 r2 = 0.189895,
                                 r2adj = 0.135888,
                                 ssres = 0.00591031,
                                 stderr = 0.01985,
                                 stderrp = 39.2966,
                                 stdev = 0.0213538})

                >>> # global Equations and Scalars databases are left intact by the method estimate()
                >>> equations["ACAF"]                   # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'ACAF\',
                         lec = \'(ACAF/VAF[-1]) := acaf1 + acaf2 * GOSF[-1] + acaf4 * (TIME=1995)\',
                         method = \'LSQ\',
                         from_period = \'1960Y1\',
                         to_period = \'2015Y1\')
                >>> equations["DPUH"]                   # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'DPUH\',
                         lec = \'dln(DPUH/DPUHO) := dpuh_1 + dpuh_2 * dln(IHU/PI5) + dln(PC)\',
                         method = \'LSQ\',
                         from_period = \'1960Y1\',
                         to_period = \'2015Y1\')
                >>> scalars["acaf1"]
                Scalar(0.9, 1, na)
                >>> scalars["dpuh_1"]
                Scalar(0.9, 1, na)

                >>> # save results in the global databases
                >>> new_eqs = estimation.save()
                >>> new_eqs
                []
                >>> equations["ACAF"]             # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                Equation(endogenous = \'ACAF\',
                         lec = \'(ACAF/VAF[-1]) := acaf1 + acaf2 * GOSF[-1] + acaf4 * (TIME=1995)\',
                         method = \'LSQ\',
                         from_period = \'1980Y1\',
                         to_period = \'1996Y1\',
                         block = \'ACAF;DPUH\',
                         tests = {corr = 1,
                                  dw = 2.33007,
                                  fstat = 32.2851,
                                  loglik = 83.8101,
                                  meany = 0.00818467,
                                  r2 = 0.821815,
                                  r2adj = 0.79636,
                                  ssres = 5.19787e-05,
                                  stderr = 0.00192685,
                                  stderrp = 23.5422,
                                  stdev = 0.0042699},
                         date = ...)
                >>> equations["DPUH"]             # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                Equation(endogenous = \'DPUH\',
                         lec = \'dln(DPUH/DPUHO) := dpuh_1 + dpuh_2 * dln(IHU/PI5) + dln(PC)\',
                         method = \'LSQ\',
                         from_period = \'1980Y1\',
                         to_period = \'1996Y1\',
                         block = \'ACAF;DPUH\',
                         tests = {corr = 0.126096,
                                  dw = 3.15593,
                                  fstat = 3.51611,
                                  loglik = 43.5743,
                                  meany = 0.0505132,
                                  r2 = 0.189895,
                                  r2adj = 0.135888,
                                  ssres = 0.00591031,
                                  stderr = 0.01985,
                                  stderrp = 39.2966,
                                  stdev = 0.0213538},
                         date = ...)
                >>> scalars["acaf1"]
                Scalar(0.0157705, 1, 0.00136079)
                >>> scalars["dpuh_1"]
                Scalar(0.0109855, 1, 0.00481857)
        '''
    def get_fitted_values(self, name: str) -> list[float]:
        '''get_fitted_values(self, name: str) -> List[float]

                Fitted values for estimation (saved as the Variable `_CALC`)

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, equations, scalars, variables, Equation
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> equations.clear()
                >>> scalars.clear()

                >>> equations["ACAF"] = Equation("ACAF", "(ACAF/VAF[-1]) := acaf1 + acaf2 * GOSF[-1] + acaf4 * (TIME=1995)")
                >>> equations["DPUH"] = Equation("DPUH", "dln(DPUH/DPUHO) := dpuh_1 + dpuh_2 * dln(IHU/PI5) + dln(PC)")

                >>> from iode.iode_python import EditAndEstimateEquations
                >>> estimation = EditAndEstimateEquations("1980Y1", "1996Y1")
                >>> estimation.block = "ACAF;DPUH", "ACAF"
                >>> estimation.estimate()

                >>> estimation.get_observed_values("ACAF")      # doctest: +ELLIPSIS
                [0.011412041862997465, 0.016028202180810566, ..., -0.002985052229901132, 0.00695696806902914]
                >>> estimation.get_fitted_values("ACAF")        # doctest: +ELLIPSIS
                [0.01256212379902523, 0.01249107519078205, ..., -0.002985052229901133, 0.004490336374473415]
                >>> estimation.get_residual_values("ACAF")      # doctest: +ELLIPSIS
                [-0.0011500819360277647, 0.0035371269900285156, ..., 8.673617379884035e-19, 0.002466631694555725]

                >>> estimation.get_observed_values("DPUH")      # doctest: +ELLIPSIS
                [0.06044527980207867, 0.08768972383253629, ..., 0.0424313077256923, 0.0064336499579307135]
                >>> estimation.get_fitted_values("DPUH")        # doctest: +ELLIPSIS
                [0.07361898875462985, 0.06423949088326213, ..., 0.028792670295126173, 0.032048802201331064]
                >>> estimation.get_residual_values("DPUH")      # doctest: +ELLIPSIS
                [-0.013173708952551183, 0.02345023294927416, ..., 0.013638637430566126, -0.02561515224340035]
        '''
    def get_observed_values(self, name: str) -> list[float]:
        '''get_observed_values(self, name: str) -> List[float]

                Observed values (saved as the Variable `_YOBS`).

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, equations, scalars, variables, Equation
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> equations.clear()
                >>> scalars.clear()

                >>> equations["ACAF"] = Equation("ACAF", "(ACAF/VAF[-1]) := acaf1 + acaf2 * GOSF[-1] + acaf4 * (TIME=1995)")
                >>> equations["DPUH"] = Equation("DPUH", "dln(DPUH/DPUHO) := dpuh_1 + dpuh_2 * dln(IHU/PI5) + dln(PC)")

                >>> from iode.iode_python import EditAndEstimateEquations
                >>> estimation = EditAndEstimateEquations("1980Y1", "1996Y1")
                >>> estimation.block = "ACAF;DPUH", "ACAF"
                >>> estimation.estimate()

                >>> estimation.get_observed_values("ACAF")      # doctest: +ELLIPSIS
                [0.011412041862997465, 0.016028202180810566, ..., -0.002985052229901132, 0.00695696806902914]
                >>> estimation.get_fitted_values("ACAF")        # doctest: +ELLIPSIS
                [0.01256212379902523, 0.01249107519078205, ..., -0.002985052229901133, 0.004490336374473415]
                >>> estimation.get_residual_values("ACAF")      # doctest: +ELLIPSIS
                [-0.0011500819360277647, 0.0035371269900285156, ..., 8.673617379884035e-19, 0.002466631694555725]

                >>> estimation.get_observed_values("DPUH")      # doctest: +ELLIPSIS
                [0.06044527980207867, 0.08768972383253629, ..., 0.0424313077256923, 0.0064336499579307135]
                >>> estimation.get_fitted_values("DPUH")        # doctest: +ELLIPSIS
                [0.07361898875462985, 0.06423949088326213, ..., 0.028792670295126173, 0.032048802201331064]
                >>> estimation.get_residual_values("DPUH")      # doctest: +ELLIPSIS
                [-0.013173708952551183, 0.02345023294927416, ..., 0.013638637430566126, -0.02561515224340035]
        '''
    def get_residual_values(self, name: str) -> list[float]:
        '''get_residual_values(self, name: str) -> List[float]

                Difference between the observations and the estimate values (saved as the Variable `YRES`).

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, equations, scalars, variables, Equation
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> equations.clear()
                >>> scalars.clear()

                >>> equations["ACAF"] = Equation("ACAF", "(ACAF/VAF[-1]) := acaf1 + acaf2 * GOSF[-1] + acaf4 * (TIME=1995)")
                >>> equations["DPUH"] = Equation("DPUH", "dln(DPUH/DPUHO) := dpuh_1 + dpuh_2 * dln(IHU/PI5) + dln(PC)")

                >>> from iode.iode_python import EditAndEstimateEquations
                >>> estimation = EditAndEstimateEquations("1980Y1", "1996Y1")
                >>> estimation.block = "ACAF;DPUH", "ACAF"
                >>> estimation.estimate()

                >>> estimation.get_observed_values("ACAF")      # doctest: +ELLIPSIS
                [0.011412041862997465, 0.016028202180810566, ..., -0.002985052229901132, 0.00695696806902914]
                >>> estimation.get_fitted_values("ACAF")        # doctest: +ELLIPSIS
                [0.01256212379902523, 0.01249107519078205, ..., -0.002985052229901133, 0.004490336374473415]
                >>> estimation.get_residual_values("ACAF")      # doctest: +ELLIPSIS
                [-0.0011500819360277647, 0.0035371269900285156, ..., 8.673617379884035e-19, 0.002466631694555725]

                >>> estimation.get_observed_values("DPUH")      # doctest: +ELLIPSIS
                [0.06044527980207867, 0.08768972383253629, ..., 0.0424313077256923, 0.0064336499579307135]
                >>> estimation.get_fitted_values("DPUH")        # doctest: +ELLIPSIS
                [0.07361898875462985, 0.06423949088326213, ..., 0.028792670295126173, 0.032048802201331064]
                >>> estimation.get_residual_values("DPUH")      # doctest: +ELLIPSIS
                [-0.013173708952551183, 0.02345023294927416, ..., 0.013638637430566126, -0.02561515224340035]
        '''
    def save(self, from_period: str | list[str] = ..., to_period: str | list[str] = ...) -> list[str]:
        """save(self, from_period: Union[str, List[str]] = None, to_period: Union[str, List[str]] = None) -> List[str]

                - copy the equations referenced in the vector 'v_equations' from the local Equations database 
                  to the global Equations database,
                - if estimation -> create/update the scalars containing the results of the estimated equation(s),
                - merge the local Scalars database into the global Scalars database,
                - return the list of names of the new equations.

                See :meth:`EditAndEstimateEquations.estimate` for the examples.
        
                Parameters
                ----------
                from_period: str or Period, optional
                    Starting period to copy if no estimation has been done
                to_period: str or Period, optional
                    Ending period to copy if no estimation has been done
        
                Returns
                -------
                list(str)
                    List of the names of the new equations

                Notes
                -----
                Equivalent to the ODE_blk_save() function from o_est.c from the old GUI (developed by Jean-Marc Paul).
        """
    def update_current_equation(self, lec: str, comment: str) -> Any:
        """update_current_equation(self, lec: str, comment: str)

                Update the LEC and comment of the current equation
        
                Notes
                -----
                Equivalent to the ODE_blk_save_cur() function from o_est.c from the old GUI (developed by Jean-Marc Paul)
        """
    @overload
    def update_scalars(self) -> Any:
        """update_scalars(self)

                Update the local estimation Scalars database 'kdb_scl'.
        
                What the C++ method update_scalars() does:

                    - Step 1: for each equation in the local estimation Equations database, get the list if corresponding scalars
                    - Step 2: remove duplicated scalar names
                    - Step 3: for each scalar name, check if it is already present in the local database 'kdb_scl':
        \t                  - no  -> check if in the global Scalars database
                                       - yes -> copy scalars from the global database to 'kdb_scl'. 
        \t                           - no  -> add a new scalar with value = 0.0 and relax = 1.0 to 'kdb_scl'.
        \t                  - yes -> does nothing.
                    - Step 4: remove the scalars associated with equations which are not in the present block to estimate
        
                See :meth:`EditAndEstimateEquations.estimate` for the examples.

                Notes
                -----
                Equivalent to the ODE_blk_coef() function from o_est.c from the old GUI (developed by Jean-Marc Paul)
        """
    @overload
    def update_scalars(cls) -> Any:
        """update_scalars(self)

                Update the local estimation Scalars database 'kdb_scl'.
        
                What the C++ method update_scalars() does:

                    - Step 1: for each equation in the local estimation Equations database, get the list if corresponding scalars
                    - Step 2: remove duplicated scalar names
                    - Step 3: for each scalar name, check if it is already present in the local database 'kdb_scl':
        \t                  - no  -> check if in the global Scalars database
                                       - yes -> copy scalars from the global database to 'kdb_scl'. 
        \t                           - no  -> add a new scalar with value = 0.0 and relax = 1.0 to 'kdb_scl'.
        \t                  - yes -> does nothing.
                    - Step 4: remove the scalars associated with equations which are not in the present block to estimate
        
                See :meth:`EditAndEstimateEquations.estimate` for the examples.

                Notes
                -----
                Equivalent to the ODE_blk_coef() function from o_est.c from the old GUI (developed by Jean-Marc Paul)
        """
    def __reduce__(self):
        """__reduce_cython__(self)"""

class EqMethod(enum.IntEnum):
    __new__: ClassVar[Callable] = ...
    GLS: ClassVar[EqMethod] = ...
    INSTRUMENTAL: ClassVar[EqMethod] = ...
    LSQ: ClassVar[EqMethod] = ...
    MAX_LIKELIHOOD: ClassVar[EqMethod] = ...
    ZELLNER: ClassVar[EqMethod] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[int]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    def __format__(self, *args, **kwargs) -> str: ...

class EqTest(enum.IntEnum):
    __new__: ClassVar[Callable] = ...
    CORR: ClassVar[EqTest] = ...
    DW: ClassVar[EqTest] = ...
    FSTAT: ClassVar[EqTest] = ...
    LOGLIK: ClassVar[EqTest] = ...
    MEANY: ClassVar[EqTest] = ...
    R2: ClassVar[EqTest] = ...
    R2ADJ: ClassVar[EqTest] = ...
    SSRES: ClassVar[EqTest] = ...
    STDERR: ClassVar[EqTest] = ...
    STDERRP: ClassVar[EqTest] = ...
    STDEV: ClassVar[EqTest] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[int]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    def __format__(self, *args, **kwargs) -> str: ...

class Equation:
    __pyx_vtable__: ClassVar[PyCapsule] = ...
    block: Incomplete
    coefficients: list[str]
    comment: Incomplete
    date: Incomplete
    endogenous: Incomplete
    instruments: str | list[str]
    lec: str
    method: str
    sample: Sample
    tests: Incomplete
    variables: list[str]
    def __init__(self, endogenous=..., 
lec=..., 
method=..., 
from_period=..., 
to_period=...) -> Any:
        """Initialize self.  See help(type(self)) for accurate signature."""
    def estimate(self, from_period: str | Period = ..., to_period: str | Period = ...) -> Any:
        '''estimate(self, from_period: Union[str, Period] = None, to_period: Union[str, Period] = None)

                Estimate the present equation.

                At the end of the estimation process, certain variables and scalars are automatically created 
                if the process has converged. These variables and scalars can be used for computational purposes and, 
                as they are part of the global workspace, can be saved for future use.

                The tests resulting from the last estimation are saved as scalars. The same applies to residuals, 
                left-hand and right-hand members of equations.

                Saved tests (as scalars) have the following names:

                    - `e0_n` : number of sample periods 
                    - `e0_k` : number of estimated coefficients 
                    - `e0_stdev` : std dev of residuals 
                    - `e0_meany` : mean of Y 
                    - `e0_ssres` : sum of squares of residuals 
                    - `e0_stderr` : std error 
                    - `e0_stderrp` : std error percent (in %) 
                    - `e0_fstat` : F-Stat 
                    - `e0_r2` : R square 
                    - `e0_r2adj` : adjusted R-squared 
                    - `e0_dw` : Durbin-Watson 
                    - `e0_loglik` : Log Likelihood 

                Calculated series are saved in special variables:

                    - `_YCALC0` : right-hand side of the equation
                    - `_YOBS0` : left-hand side of the equation
                    - `_YRES0` : residuals of the equation

                Outside the estimation sample, the series values are :math:`NA`.

                Parameters
                ----------
                from_period : str or Period, optional
                    The starting period of the execution range. 
                    Defaults to the starting period of the current Variables sample.
                to_period : str or Period, optional
                    The ending period of the execution range.
                    Defaults to the ending period of the current Variables sample.

                Warnings
                --------
                If the present equation belongs to a block, you must use the :meth:`Equations.estimate` method instead.

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, equations, scalars, variables
                >>> equations.load(f"{SAMPLE_DATA_DIR}/fun.eqs")
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")

                >>> eq_ACAF = equations["ACAF"]
                >>> eq_ACAF.lec
                \'(ACAF/VAF[-1]) :=acaf1+acaf2*GOSF[-1]+\\nacaf4*(TIME=1995)\'

                >>> # create scalars
                >>> scalars["acaf1"] = 0., 1.
                >>> scalars["acaf2"] = 0., 1.
                >>> scalars["acaf4"] = 0., 1.

                >>> # estimate the ACQF equation
                >>> eq_ACAF.estimate("1980Y1", "1996Y1")
                >>> scalars["acaf1"]
                Scalar(0.0157705, 1, 0.00136949)
                >>> scalars["acaf2"]
                Scalar(-7.96505e-06, 1, 1.48247e-06)
                >>> scalars["acaf4"]
                Scalar(-0.0085027, 1, 0.00208257)
        '''
    def get_formated_date(self, format: str = ...) -> str:
        '''get_formated_date(self, format: str = u\'dd-mm-yyyy\') -> str

                Return the date of last estimation in a given format.

                Returns
                -------
                str

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, equations
                >>> equations.load(f"{SAMPLE_DATA_DIR}/fun.eqs")
                >>> # date as default format "dd-mm-yyyy"
                >>> equations["ACAF"].date
                \'12-06-1998\'
                >>> # date with specific format
                >>> equations["ACAF"].get_formated_date("dd/mm/yyyy")
                \'12/06/1998\'
        '''
    @overload
    def split_equation(self) -> tuple[str, str]:
        '''split_equation(self) -> Tuple[str, str]

                Split an equation into its left-hand side and its right-hand side.

                Returns
                -------
                tuple(str, str)

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import Equation
                >>> eq_ACAF = Equation("ACAF", "(ACAF / VAF[-1]) := acaf1 + acaf2 * GOSF[-1] + acaf4 * (TIME=1995)")
                >>> lhs, rhs = eq_ACAF.split_equation()
                >>> # left-hand side
                >>> lhs
                \'(ACAF / VAF[-1])\'
                >>> # right-hand side
                >>> rhs
                \'acaf1 + acaf2 * GOSF[-1] + acaf4 * (TIME=1995)\'
        '''
    @overload
    def split_equation(cls) -> Any:
        '''split_equation(self) -> Tuple[str, str]

                Split an equation into its left-hand side and its right-hand side.

                Returns
                -------
                tuple(str, str)

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import Equation
                >>> eq_ACAF = Equation("ACAF", "(ACAF / VAF[-1]) := acaf1 + acaf2 * GOSF[-1] + acaf4 * (TIME=1995)")
                >>> lhs, rhs = eq_ACAF.split_equation()
                >>> # left-hand side
                >>> lhs
                \'(ACAF / VAF[-1])\'
                >>> # right-hand side
                >>> rhs
                \'acaf1 + acaf2 * GOSF[-1] + acaf4 * (TIME=1995)\'
        '''
    def __copy__(self) -> Equation:
        '''__copy__(self) -> Equation

                Return a copy of the current Equation.

                Examples
                --------
                >>> import copy
                >>> from iode import SAMPLE_DATA_DIR, equations, variables
                >>> equations.load(f"{SAMPLE_DATA_DIR}/fun.eqs")
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> equations["ACAF"]           # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'ACAF\',
                        lec = \'(ACAF/VAF[-1]) :=acaf1+acaf2*GOSF[-1]+\\nacaf4*(TIME=1995)\',
                        method = \'LSQ\',
                        from_period = \'1980Y1\',
                        to_period = \'1996Y1\',
                        block = \'ACAF\',
                        tests = {corr = 1,
                                 dw = 2.32935,
                                 fstat = 32.2732,
                                 loglik = 83.8075,
                                 meany = 0.00818467,
                                 r2 = 0.821761,
                                 r2adj = 0.796299,
                                 ssres = 5.19945e-05,
                                 stderr = 0.00192715,
                                 stderrp = 23.5458,
                                 stdev = 0.0042699},
                        date = \'12-06-1998\')
                >>> copied_eq = copy.copy(equations["ACAF"])
                >>> copied_eq                   # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'ACAF\',
                        lec = \'(ACAF/VAF[-1]) :=acaf1+acaf2*GOSF[-1]+\\nacaf4*(TIME=1995)\',
                        method = \'LSQ\',
                        from_period = \'1980Y1\',
                        to_period = \'1996Y1\',
                        block = \'ACAF\',
                        tests = {corr = 1,
                                 dw = 2.32935,
                                 fstat = 32.2732,
                                 loglik = 83.8075,
                                 meany = 0.00818467,
                                 r2 = 0.821761,
                                 r2adj = 0.796299,
                                 ssres = 5.19945e-05,
                                 stderr = 0.00192715,
                                 stderrp = 23.5458,
                                 stdev = 0.0042699},
                        date = \'12-06-1998\')
        '''
    def __eq__(self, other: object) -> bool:
        """Return self==value."""
    def __ge__(self, other: object) -> bool:
        """Return self>=value."""
    def __gt__(self, other: object) -> bool:
        """Return self>value."""
    def __hash__(self) -> int:
        """Return hash(self)."""
    def __le__(self, other: object) -> bool:
        """Return self<=value."""
    def __lt__(self, other: object) -> bool:
        """Return self<value."""
    def __ne__(self, other: object) -> bool:
        """Return self!=value."""
    def __reduce__(self):
        """__reduce_cython__(self)"""

class Equations(_AbstractDatabase):
    __pyx_vtable__: ClassVar[PyCapsule] = ...
    df: DataFrame
    i: PositionalIndexer
    def __init__(self, *args, **kwargs) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""
    def estimate(self, from_period: str | Period = ..., to_period: str | Period = ..., list_eqs: str | list[str] = ...) -> Any:
        '''estimate(self, from_period: Union[str, Period] = None, to_period: Union[str, Period] = None, list_eqs: Union[str, List[str]] = None)

                Estimate an equation or a block of equations.

                At the end of the estimation process, certain variables and scalars are automatically created 
                if the process has converged. These variables and scalars can be used for computational purposes and, 
                as they are part of the global workspace, can be saved for future use.

                The tests resulting from the last estimation are saved as scalars. The same applies to residuals, 
                left-hand and right-hand members of equations.

                Saved tests (as scalars) have the following names (`e<ith>_*` for the <ith> equation of the block):

                    - `e<ith>_n`  : number of sample periods 
                    - `e<ith>_k`  : number of estimated coefficients 
                    - `e<ith>_stdev` : std dev of residuals 
                    - `e<ith>_meany` : mean of Y 
                    - `e<ith>_ssres` : sum of squares of residuals 
                    - `e<ith>_stderr` : std error 
                    - `e<ith>_stderrp` : std error percent (in %) 
                    - `e<ith>_fstat` : F-Stat 
                    - `e<ith>_r2`  : R square 
                    - `e<ith>_r2adj`  : adjusted R-squared 
                    - `e<ith>_dw` : Durbin-Watson 
                    - `e<ith>_loglik`  : Log Likelihood 

                Calculated series are saved in special variables:

                    - `_YCALC<i>` : right-hand side of the <ith> equation in the block
                    - `_YOBS<i>` : left-hand side of the <ith> equation in the block 
                    - `_YRES<i>` : residuals of the <ith> equation of the block 

                Outside the estimation sample, the series values are :math:`NA`.

                Parameters
                ----------
                from_period : str or Period, optional
                    The starting period of the execution range. 
                    Defaults to the starting period of the current Variables sample.
                to_period : str or Period, optional
                    The ending period of the execution range.
                    Defaults to the ending period of the current Variables sample.
                list_eqs: str or list(str), optional
                    List of equations to be estimated. 
                    If not provided, all equations of the present Equations database will be estimated.
                    Default to None (all equations).

                Warnings
                --------
                If some equations to be estimated belongs to a same block, the *block* (:meth:`Equation.block`), 
                *method* (:meth:`Equation.method`), and *instruments* (\':meth:`Equation.instruments`) of each of 
                them must have been updated before.

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, equations, scalars, variables
                >>> equations.load(f"{SAMPLE_DATA_DIR}/fun.eqs")
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")

                >>> equations["ACAF"].lec
                \'(ACAF/VAF[-1]) :=acaf1+acaf2*GOSF[-1]+\\nacaf4*(TIME=1995)\'
                >>> equations["DPUH"].lec
                \'dln (DPUH/DPUHO):=dpuh_1+dpuh_2*dln(IHU/PI5)+dln PC\'

                >>> # create scalars
                >>> scalars["acaf1"] = 0., 1.
                >>> scalars["acaf2"] = 0., 1.
                >>> scalars["acaf4"] = 0., 1.
                >>> scalars["dpuh_1"] = 0., 1.
                >>> scalars["dpuh_2"] = 0., 1.

                >>> # estimate an equation
                >>> equations.estimate("1980Y1", "1996Y1", "ACAF")
                >>> # or equivalenty
                >>> equations["ACAF"].estimate("1980Y1", "1996Y1")
                >>> scalars["acaf1"]
                Scalar(0.0157705, 1, 0.00136949)
                >>> scalars["acaf2"]
                Scalar(-7.96505e-06, 1, 1.48247e-06)
                >>> scalars["acaf4"]
                Scalar(-0.0085027, 1, 0.00208257)
                >>> scalars["dpuh_1"]
                Scalar(0, 1, na)
                >>> scalars["dpuh_2"]
                Scalar(0, 1, na)

                >>> # estimate a block
                >>> # prepare equations (same block and method)
                >>> block = "ACAF;DPUH"
                >>> for name in block.split(";"):
                ...     equations[name] = {"block": block, "method": "LSQ"}
                >>> equations.estimate("1980Y1", "1996Y1", block)
                >>> scalars["acaf1"]
                Scalar(0.0157705, 1, 0.00136079)
                >>> scalars["acaf2"]
                Scalar(-7.96505e-06, 1, 1.47188e-06)
                >>> scalars["acaf4"]
                Scalar(-0.0085027, 1, 0.00206603)
                >>> scalars["dpuh_1"]
                Scalar(0.0109855, 1, 0.00481857)
                >>> scalars["dpuh_2"]
                Scalar(0.0574893, 1, 0.0368951)
        '''
    @overload
    def from_frame(self, df: DataFrame) -> Any:
        '''from_frame(self, df: DataFrame)

                Copy the pandas DataFrame `df` into the IODE Equations database.
                The equation names to copy are deduced from the index of the DataFrame.

                Parameters
                ----------
                df: DataFrame
                    pandas DataFrame containing the equations to copy into the Equations database.
                    The passed DataFrame must at least contains:
                        - an index containing the names of the equations
                        - a column labeled \'lec\' containing the LEC expressions of the equations.
                    Other possible columns are \'method\', \'sample\', \'comment\', \'instruments\' and \'block\'
                    (see :class:`Equation`).

                Notes
                -----
                The index of the passed DataFrame is sorted in alphabetical order before 
                copying to IODE Equations database.

                See Also
                --------
                Equations.from_series
                Equations.to_frame

                Examples
                --------
                >>> from iode import equations, variables
                >>> import pandas as pd
                >>> import numpy as np
                >>> variables.clear()
                >>> variables.sample = "1960Y1:2015Y1"
                >>> equations.clear()
                >>> len(equations)
                0

                >>> # create the pandas DataFrame
                >>> columns=["lec", "method", "sample", "corr", "fstat", "r2", "date"]
                >>> data = {"A": ["A := t", "LSQ", "1980Y1:2010Y1", 1.0, 0.79, 0.05, "12-06-1998"],
                ...         "B": ["B := grt A", "LSQ", "1980Y1:2010Y1", 1.0, 0.896, 0.15, "12-06-1998"],
                ...         "C": ["C := exp t", "ZELLNER", "1990Y1:", 1.0, 0.1258, 0.0225, ""],
                ...         "D": ["ln D := t", "MAX_LIKELIHOOD", "", 1.0, 0.358, 0.0698, "12-06-1998"],
                ...         "E": ["1/E := B", "LSQ", ":2010Y1", 1.0, 0.689, 2.23e-03, ""],
                ...         "F": ["d(F) := t + 1", "GLS", "", 1.0, 0.05, 4.568e-04, ""]}
                >>> df = pd.DataFrame.from_dict(data, orient=\'index\', columns=columns)
                >>> # display the pandas series
                >>> df                  # doctest: +NORMALIZE_WHITESPACE
                             lec          method         sample  ...   fstat        r2        date
                A         A := t             LSQ  1980Y1:2010Y1  ...  0.7900  0.050000  12-06-1998
                B     B := grt A             LSQ  1980Y1:2010Y1  ...  0.8960  0.150000  12-06-1998
                C     C := exp t         ZELLNER        1990Y1:  ...  0.1258  0.022500
                D      ln D := t  MAX_LIKELIHOOD                 ...  0.3580  0.069800  12-06-1998
                E       1/E := B             LSQ        :2010Y1  ...  0.6890  0.002230
                F  d(F) := t + 1             GLS                 ...  0.0500  0.000457
                <BLANKLINE>
                [6 rows x 7 columns]

                >>> # load into the IODE Equations database
                >>> equations.from_frame(df)
                >>> len(equations)
                6
        
                >>> equations.names
                [\'A\', \'B\', \'C\', \'D\', \'E\', \'F\']
                >>> df.loc["A"]         # doctest: +NORMALIZE_WHITESPACE  
                lec              A := t
                method              LSQ
                sample    1980Y1:2010Y1
                corr                1.0
                fstat              0.79
                r2                 0.05
                date         12-06-1998
                Name: A, dtype: object
                >>> equations["A"]      # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'A\',
                         lec = \'A := t\',
                         method = \'LSQ\',
                         from_period = \'1980Y1\',
                         to_period = \'2010Y1\',
                         tests = {corr = 1,
                                  dw = 0,
                                  fstat = 0.79,
                                  loglik = 0,
                                  meany = 0,
                                  r2 = 0.05,
                                  r2adj = 0,
                                  ssres = 0,
                                  stderr = 0,
                                  stderrp = 0,
                                  stdev = 0},
                         date = \'12-06-1998\')
                >>> df.loc["F"]         # doctest: +NORMALIZE_WHITESPACE
                lec       d(F) := t + 1
                method              GLS
                sample
                corr                1.0
                fstat              0.05
                r2             0.000457
                date
                Name: F, dtype: object
                >>> equations["F"]      # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'F\',
                         lec = \'d(F) := t + 1\',
                         method = \'GLS\',
                         from_period = \'1960Y1\',
                         to_period = \'2015Y1\',
                         tests = {corr = 1,
                                  dw = 0,
                                  fstat = 0.05,
                                  loglik = 0,
                                  meany = 0,
                                  r2 = 0.0004568,
                                  r2adj = 0,
                                  ssres = 0,
                                  stderr = 0,
                                  stderrp = 0,
                                  stdev = 0})
        '''
    @overload
    def from_frame(cls, df) -> Any:
        '''from_frame(self, df: DataFrame)

                Copy the pandas DataFrame `df` into the IODE Equations database.
                The equation names to copy are deduced from the index of the DataFrame.

                Parameters
                ----------
                df: DataFrame
                    pandas DataFrame containing the equations to copy into the Equations database.
                    The passed DataFrame must at least contains:
                        - an index containing the names of the equations
                        - a column labeled \'lec\' containing the LEC expressions of the equations.
                    Other possible columns are \'method\', \'sample\', \'comment\', \'instruments\' and \'block\'
                    (see :class:`Equation`).

                Notes
                -----
                The index of the passed DataFrame is sorted in alphabetical order before 
                copying to IODE Equations database.

                See Also
                --------
                Equations.from_series
                Equations.to_frame

                Examples
                --------
                >>> from iode import equations, variables
                >>> import pandas as pd
                >>> import numpy as np
                >>> variables.clear()
                >>> variables.sample = "1960Y1:2015Y1"
                >>> equations.clear()
                >>> len(equations)
                0

                >>> # create the pandas DataFrame
                >>> columns=["lec", "method", "sample", "corr", "fstat", "r2", "date"]
                >>> data = {"A": ["A := t", "LSQ", "1980Y1:2010Y1", 1.0, 0.79, 0.05, "12-06-1998"],
                ...         "B": ["B := grt A", "LSQ", "1980Y1:2010Y1", 1.0, 0.896, 0.15, "12-06-1998"],
                ...         "C": ["C := exp t", "ZELLNER", "1990Y1:", 1.0, 0.1258, 0.0225, ""],
                ...         "D": ["ln D := t", "MAX_LIKELIHOOD", "", 1.0, 0.358, 0.0698, "12-06-1998"],
                ...         "E": ["1/E := B", "LSQ", ":2010Y1", 1.0, 0.689, 2.23e-03, ""],
                ...         "F": ["d(F) := t + 1", "GLS", "", 1.0, 0.05, 4.568e-04, ""]}
                >>> df = pd.DataFrame.from_dict(data, orient=\'index\', columns=columns)
                >>> # display the pandas series
                >>> df                  # doctest: +NORMALIZE_WHITESPACE
                             lec          method         sample  ...   fstat        r2        date
                A         A := t             LSQ  1980Y1:2010Y1  ...  0.7900  0.050000  12-06-1998
                B     B := grt A             LSQ  1980Y1:2010Y1  ...  0.8960  0.150000  12-06-1998
                C     C := exp t         ZELLNER        1990Y1:  ...  0.1258  0.022500
                D      ln D := t  MAX_LIKELIHOOD                 ...  0.3580  0.069800  12-06-1998
                E       1/E := B             LSQ        :2010Y1  ...  0.6890  0.002230
                F  d(F) := t + 1             GLS                 ...  0.0500  0.000457
                <BLANKLINE>
                [6 rows x 7 columns]

                >>> # load into the IODE Equations database
                >>> equations.from_frame(df)
                >>> len(equations)
                6
        
                >>> equations.names
                [\'A\', \'B\', \'C\', \'D\', \'E\', \'F\']
                >>> df.loc["A"]         # doctest: +NORMALIZE_WHITESPACE  
                lec              A := t
                method              LSQ
                sample    1980Y1:2010Y1
                corr                1.0
                fstat              0.79
                r2                 0.05
                date         12-06-1998
                Name: A, dtype: object
                >>> equations["A"]      # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'A\',
                         lec = \'A := t\',
                         method = \'LSQ\',
                         from_period = \'1980Y1\',
                         to_period = \'2010Y1\',
                         tests = {corr = 1,
                                  dw = 0,
                                  fstat = 0.79,
                                  loglik = 0,
                                  meany = 0,
                                  r2 = 0.05,
                                  r2adj = 0,
                                  ssres = 0,
                                  stderr = 0,
                                  stderrp = 0,
                                  stdev = 0},
                         date = \'12-06-1998\')
                >>> df.loc["F"]         # doctest: +NORMALIZE_WHITESPACE
                lec       d(F) := t + 1
                method              GLS
                sample
                corr                1.0
                fstat              0.05
                r2             0.000457
                date
                Name: F, dtype: object
                >>> equations["F"]      # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'F\',
                         lec = \'d(F) := t + 1\',
                         method = \'GLS\',
                         from_period = \'1960Y1\',
                         to_period = \'2015Y1\',
                         tests = {corr = 1,
                                  dw = 0,
                                  fstat = 0.05,
                                  loglik = 0,
                                  meany = 0,
                                  r2 = 0.0004568,
                                  r2adj = 0,
                                  ssres = 0,
                                  stderr = 0,
                                  stderrp = 0,
                                  stdev = 0})
        '''
    @overload
    def from_series(self, s: Series) -> Any:
        '''from_series(self, s: Series)

                Copy the pandas Series `s` into the IODE Equations database.
                The equation names to copy are deduced from the index of the Series.

                Parameters
                ----------
                s: Series
                    pandas Series containing the equations as LEC expressions to copy into the Equations database.
                    All other Equation attributes (method, sample, comment, block, ...) will be set to their default 
                    values (see :class:`Equation`).

                Notes
                -----
                The index of the passed Series is sorted in alphabetical order before 
                copying to IODE Equations database.

                See Also
                --------
                Equations.from_frame
                Equations.to_frame

                Examples
                --------
                >>> from iode import equations, variables
                >>> import pandas as pd
                >>> variables.clear()
                >>> variables.sample = "1960Y1:2015Y1"
                >>> equations.clear()
                >>> len(equations)
                0

                >>> # create the pandas Series
                >>> data = {"A": "A := t",
                ...         "B": "B := grt A",
                ...         "C": "C := exp t",
                ...         "D": "ln D := t",
                ...         "E": "1/E := B",
                ...         "F": "d(F) := t + 1"}
                >>> s = pd.Series(data)
                >>> # display the pandas series
                >>> s                   # doctest: +NORMALIZE_WHITESPACE
                A           A := t
                B       B := grt A
                C       C := exp t
                D        ln D := t
                E         1/E := B
                F    d(F) := t + 1
                dtype: object

                >>> # load into the IODE Equations database
                >>> equations.from_series(s)
                >>> len(equations)
                6
        
                >>> equations.names
                [\'A\', \'B\', \'C\', \'D\', \'E\', \'F\']
                >>> equations["A"]      # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'A\',
                        lec = \'A := t\',
                        method = \'LSQ\',
                        from_period = \'1960Y1\',
                        to_period = \'2015Y1\')
                >>> equations["F"]      # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'F\',
                        lec = \'d(F) := t + 1\',
                        method = \'LSQ\',
                        from_period = \'1960Y1\',
                        to_period = \'2015Y1\')
        '''
    @overload
    def from_series(cls, s) -> Any:
        '''from_series(self, s: Series)

                Copy the pandas Series `s` into the IODE Equations database.
                The equation names to copy are deduced from the index of the Series.

                Parameters
                ----------
                s: Series
                    pandas Series containing the equations as LEC expressions to copy into the Equations database.
                    All other Equation attributes (method, sample, comment, block, ...) will be set to their default 
                    values (see :class:`Equation`).

                Notes
                -----
                The index of the passed Series is sorted in alphabetical order before 
                copying to IODE Equations database.

                See Also
                --------
                Equations.from_frame
                Equations.to_frame

                Examples
                --------
                >>> from iode import equations, variables
                >>> import pandas as pd
                >>> variables.clear()
                >>> variables.sample = "1960Y1:2015Y1"
                >>> equations.clear()
                >>> len(equations)
                0

                >>> # create the pandas Series
                >>> data = {"A": "A := t",
                ...         "B": "B := grt A",
                ...         "C": "C := exp t",
                ...         "D": "ln D := t",
                ...         "E": "1/E := B",
                ...         "F": "d(F) := t + 1"}
                >>> s = pd.Series(data)
                >>> # display the pandas series
                >>> s                   # doctest: +NORMALIZE_WHITESPACE
                A           A := t
                B       B := grt A
                C       C := exp t
                D        ln D := t
                E         1/E := B
                F    d(F) := t + 1
                dtype: object

                >>> # load into the IODE Equations database
                >>> equations.from_series(s)
                >>> len(equations)
                6
        
                >>> equations.names
                [\'A\', \'B\', \'C\', \'D\', \'E\', \'F\']
                >>> equations["A"]      # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'A\',
                        lec = \'A := t\',
                        method = \'LSQ\',
                        from_period = \'1960Y1\',
                        to_period = \'2015Y1\')
                >>> equations["F"]      # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'F\',
                        lec = \'d(F) := t + 1\',
                        method = \'LSQ\',
                        from_period = \'1960Y1\',
                        to_period = \'2015Y1\')
        '''
    def get_lec(self, key: str | int) -> str:
        '''get_lec(self, key: Union[str, int]) -> str

                Get the LEC of an equation in the database.

                Parameters
                ----------
                key : str or int
                    Name or index of the equation in the database.

                Returns
                -------
                str

                Examples
                --------
                >>> from iode import equations, SAMPLE_DATA_DIR
                >>> equations.load(f"{SAMPLE_DATA_DIR}/fun.eqs")
                >>> equations.get_lec("ACAF")   # doctest: +NORMALIZE_WHITESPACE
                \'(ACAF/VAF[-1]) :=acaf1+acaf2*GOSF[-1]+\\nacaf4*(TIME=1995)\'
                >>> equations.get_lec(0)       # doctest: +NORMALIZE_WHITESPACE
                \'(ACAF/VAF[-1]) :=acaf1+acaf2*GOSF[-1]+\\nacaf4*(TIME=1995)\'
        '''
    @overload
    def to_frame(self) -> DataFrame:
        '''to_frame(self) -> DataFrame

                Create a pandas DataFrame from the current Equations database.
                The index of the returned DataFrame is build from the Equations names.

                See Also
                --------
                Equations.from_series
                Equations.from_frame

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, equations
                >>> import pandas as pd
                >>> equations.load(f"{SAMPLE_DATA_DIR}/fun.eqs")
                >>> len(equations)
                274

                >>> # Export the IODE Equations database as a pandas DataFrame
                >>> df = equations.to_frame()
                >>> len(df)
                274

                >>> df.index.to_list()              # doctest: +ELLIPSIS
                [\'ACAF\', \'ACAG\', \'AOUC\', ..., \'YSSG\', \'ZF\', \'ZJ\', \'ZZF_\']
                >>> equations["ACAF"]               # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'ACAF\',
                         lec = \'(ACAF/VAF[-1]) :=acaf1+acaf2*GOSF[-1]+\\nacaf4*(TIME=1995)\',
                         method = \'LSQ\',
                         from_period = \'1980Y1\',
                         to_period = \'1996Y1\',
                         block = \'ACAF\',
                         tests = {corr = 1,
                                  dw = 2.32935,
                                  fstat = 32.2732,
                                  loglik = 83.8075,
                                  meany = 0.00818467,
                                  r2 = 0.821761,
                                  r2adj = 0.796299,
                                  ssres = 5.19945e-05,
                                  stderr = 0.00192715,
                                  stderrp = 23.5458,
                                  stdev = 0.0042699},
                         date = \'12-06-1998\')
                >>> df.loc["ACAF"]                  # doctest: +NORMALIZE_WHITESPACE
                lec            (ACAF/VAF[-1]) :=acaf1+acaf2*GOSF[-1]+\\nacaf4*...
                method                                                       LSQ
                sample                                             1980Y1:1996Y1
                comment
                instruments
                block                                                       ACAF
                corr                                                         1.0
                stdev                                                    0.00427
                meany                                                   0.008185
                ssres                                                   0.000052
                stderr                                                  0.001927
                stderrp                                                23.545813
                fstat                                                  32.273193
                r2                                                      0.821761
                r2adj                                                   0.796299
                dw                                                      2.329346
                loglik                                                 83.807526
                date                                                  12-06-1998
                Name: ACAF, dtype: object  
                >>> equations["YDH_"]               # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'YDH_\',
                         lec = \'grt YDH_ :=grt((WBU_+YN+GOSH_+IDH)-(SSF+SSH+DTH)+(SBH+OCUH))\',
                         method = \'LSQ\',
                         comment = \' \',
                         block = \'YDH_\')
                >>> df.loc["YDH_"]                  # doctest: +NORMALIZE_WHITESPACE
                lec            grt YDH_ :=grt((WBU_+YN+GOSH_+IDH)-(SSF+SSH+DT...
                method                                                       LSQ
                sample                                                      None
                comment
                instruments
                block                                                       YDH_
                corr                                                         0.0
                stdev                                                        0.0
                meany                                                        0.0
                ssres                                                        0.0
                stderr                                                       0.0
                stderrp                                                      0.0
                fstat                                                        0.0
                r2                                                           0.0
                r2adj                                                        0.0
                dw                                                           0.0
                loglik                                                       0.0
                date
                Name: YDH_, dtype: object

                >>> # Export a subset of the IODE Equations database as a pandas DataFrame
                >>> df = equations["A*;*_"].to_frame()
                >>> len(df)
                29

                >>> df.index.to_list()              # doctest: +ELLIPSIS
                [\'ACAF\', \'ACAG\', \'AOUC\', ..., \'WNF_\', \'YDH_\', \'ZZF_\']
                >>> equations["ACAF"]               # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'ACAF\',
                         lec = \'(ACAF/VAF[-1]) :=acaf1+acaf2*GOSF[-1]+\\nacaf4*(TIME=1995)\',
                         method = \'LSQ\',
                         from_period = \'1980Y1\',
                         to_period = \'1996Y1\',
                         block = \'ACAF\',
                         tests = {corr = 1,
                                  dw = 2.32935,
                                  fstat = 32.2732,
                                  loglik = 83.8075,
                                  meany = 0.00818467,
                                  r2 = 0.821761,
                                  r2adj = 0.796299,
                                  ssres = 5.19945e-05,
                                  stderr = 0.00192715,
                                  stderrp = 23.5458,
                                  stdev = 0.0042699},
                         date = \'12-06-1998\')
                >>> df.loc["ACAF"]                  # doctest: +NORMALIZE_WHITESPACE
                lec            (ACAF/VAF[-1]) :=acaf1+acaf2*GOSF[-1]+\\nacaf4*...
                method                                                       LSQ
                sample                                             1980Y1:1996Y1
                comment
                instruments
                block                                                       ACAF
                corr                                                         1.0
                stdev                                                    0.00427
                meany                                                   0.008185
                ssres                                                   0.000052
                stderr                                                  0.001927
                stderrp                                                23.545813
                fstat                                                  32.273193
                r2                                                      0.821761
                r2adj                                                   0.796299
                dw                                                      2.329346
                loglik                                                 83.807526
                date                                                  12-06-1998
                Name: ACAF, dtype: object
                >>> equations["YDH_"]               # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'YDH_\',
                         lec = \'grt YDH_ :=grt((WBU_+YN+GOSH_+IDH)-(SSF+SSH+DTH)+(SBH+OCUH))\',
                         method = \'LSQ\',
                         comment = \' \',
                         block = \'YDH_\')
                >>> df.loc["YDH_"]                  # doctest: +NORMALIZE_WHITESPACE
                lec            grt YDH_ :=grt((WBU_+YN+GOSH_+IDH)-(SSF+SSH+DT...
                method                                                       LSQ
                sample                                                      None
                comment
                instruments
                block                                                       YDH_
                corr                                                         0.0
                stdev                                                        0.0
                meany                                                        0.0
                ssres                                                        0.0
                stderr                                                       0.0
                stderrp                                                      0.0
                fstat                                                        0.0
                r2                                                           0.0
                r2adj                                                        0.0
                dw                                                           0.0
                loglik                                                       0.0
                date
                Name: YDH_, dtype: object
        '''
    @overload
    def to_frame(cls) -> Any:
        '''to_frame(self) -> DataFrame

                Create a pandas DataFrame from the current Equations database.
                The index of the returned DataFrame is build from the Equations names.

                See Also
                --------
                Equations.from_series
                Equations.from_frame

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, equations
                >>> import pandas as pd
                >>> equations.load(f"{SAMPLE_DATA_DIR}/fun.eqs")
                >>> len(equations)
                274

                >>> # Export the IODE Equations database as a pandas DataFrame
                >>> df = equations.to_frame()
                >>> len(df)
                274

                >>> df.index.to_list()              # doctest: +ELLIPSIS
                [\'ACAF\', \'ACAG\', \'AOUC\', ..., \'YSSG\', \'ZF\', \'ZJ\', \'ZZF_\']
                >>> equations["ACAF"]               # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'ACAF\',
                         lec = \'(ACAF/VAF[-1]) :=acaf1+acaf2*GOSF[-1]+\\nacaf4*(TIME=1995)\',
                         method = \'LSQ\',
                         from_period = \'1980Y1\',
                         to_period = \'1996Y1\',
                         block = \'ACAF\',
                         tests = {corr = 1,
                                  dw = 2.32935,
                                  fstat = 32.2732,
                                  loglik = 83.8075,
                                  meany = 0.00818467,
                                  r2 = 0.821761,
                                  r2adj = 0.796299,
                                  ssres = 5.19945e-05,
                                  stderr = 0.00192715,
                                  stderrp = 23.5458,
                                  stdev = 0.0042699},
                         date = \'12-06-1998\')
                >>> df.loc["ACAF"]                  # doctest: +NORMALIZE_WHITESPACE
                lec            (ACAF/VAF[-1]) :=acaf1+acaf2*GOSF[-1]+\\nacaf4*...
                method                                                       LSQ
                sample                                             1980Y1:1996Y1
                comment
                instruments
                block                                                       ACAF
                corr                                                         1.0
                stdev                                                    0.00427
                meany                                                   0.008185
                ssres                                                   0.000052
                stderr                                                  0.001927
                stderrp                                                23.545813
                fstat                                                  32.273193
                r2                                                      0.821761
                r2adj                                                   0.796299
                dw                                                      2.329346
                loglik                                                 83.807526
                date                                                  12-06-1998
                Name: ACAF, dtype: object  
                >>> equations["YDH_"]               # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'YDH_\',
                         lec = \'grt YDH_ :=grt((WBU_+YN+GOSH_+IDH)-(SSF+SSH+DTH)+(SBH+OCUH))\',
                         method = \'LSQ\',
                         comment = \' \',
                         block = \'YDH_\')
                >>> df.loc["YDH_"]                  # doctest: +NORMALIZE_WHITESPACE
                lec            grt YDH_ :=grt((WBU_+YN+GOSH_+IDH)-(SSF+SSH+DT...
                method                                                       LSQ
                sample                                                      None
                comment
                instruments
                block                                                       YDH_
                corr                                                         0.0
                stdev                                                        0.0
                meany                                                        0.0
                ssres                                                        0.0
                stderr                                                       0.0
                stderrp                                                      0.0
                fstat                                                        0.0
                r2                                                           0.0
                r2adj                                                        0.0
                dw                                                           0.0
                loglik                                                       0.0
                date
                Name: YDH_, dtype: object

                >>> # Export a subset of the IODE Equations database as a pandas DataFrame
                >>> df = equations["A*;*_"].to_frame()
                >>> len(df)
                29

                >>> df.index.to_list()              # doctest: +ELLIPSIS
                [\'ACAF\', \'ACAG\', \'AOUC\', ..., \'WNF_\', \'YDH_\', \'ZZF_\']
                >>> equations["ACAF"]               # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'ACAF\',
                         lec = \'(ACAF/VAF[-1]) :=acaf1+acaf2*GOSF[-1]+\\nacaf4*(TIME=1995)\',
                         method = \'LSQ\',
                         from_period = \'1980Y1\',
                         to_period = \'1996Y1\',
                         block = \'ACAF\',
                         tests = {corr = 1,
                                  dw = 2.32935,
                                  fstat = 32.2732,
                                  loglik = 83.8075,
                                  meany = 0.00818467,
                                  r2 = 0.821761,
                                  r2adj = 0.796299,
                                  ssres = 5.19945e-05,
                                  stderr = 0.00192715,
                                  stderrp = 23.5458,
                                  stdev = 0.0042699},
                         date = \'12-06-1998\')
                >>> df.loc["ACAF"]                  # doctest: +NORMALIZE_WHITESPACE
                lec            (ACAF/VAF[-1]) :=acaf1+acaf2*GOSF[-1]+\\nacaf4*...
                method                                                       LSQ
                sample                                             1980Y1:1996Y1
                comment
                instruments
                block                                                       ACAF
                corr                                                         1.0
                stdev                                                    0.00427
                meany                                                   0.008185
                ssres                                                   0.000052
                stderr                                                  0.001927
                stderrp                                                23.545813
                fstat                                                  32.273193
                r2                                                      0.821761
                r2adj                                                   0.796299
                dw                                                      2.329346
                loglik                                                 83.807526
                date                                                  12-06-1998
                Name: ACAF, dtype: object
                >>> equations["YDH_"]               # doctest: +NORMALIZE_WHITESPACE
                Equation(endogenous = \'YDH_\',
                         lec = \'grt YDH_ :=grt((WBU_+YN+GOSH_+IDH)-(SSF+SSH+DTH)+(SBH+OCUH))\',
                         method = \'LSQ\',
                         comment = \' \',
                         block = \'YDH_\')
                >>> df.loc["YDH_"]                  # doctest: +NORMALIZE_WHITESPACE
                lec            grt YDH_ :=grt((WBU_+YN+GOSH_+IDH)-(SSF+SSH+DT...
                method                                                       LSQ
                sample                                                      None
                comment
                instruments
                block                                                       YDH_
                corr                                                         0.0
                stdev                                                        0.0
                meany                                                        0.0
                ssres                                                        0.0
                stderr                                                       0.0
                stderrp                                                      0.0
                fstat                                                        0.0
                r2                                                           0.0
                r2adj                                                        0.0
                dw                                                           0.0
                loglik                                                       0.0
                date
                Name: YDH_, dtype: object
        '''
    def __delitem__(self, other) -> None:
        '''
        Remove the (subset of) equation(s) referenced by `key` from the Equations database.

        Parameters
        ----------
        key: str or list(str)
            (list of) name(s) of the equation(s) to be removed.
            The list of names can be given as a string pattern (e.g. "A*;*_").

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import equations
        >>> equations.load(f"{SAMPLE_DATA_DIR}/fun.eqs")

        >>> # a) delete one equation
        >>> equations.get_names("A*")
        [\'ACAF\', \'ACAG\', \'AOUC\']
        >>> del equations["ACAF"]
        >>> equations.get_names("A*")
        [\'ACAG\', \'AOUC\']

        >>> # b) delete several equations at once using a pattern
        >>> del equations["A*"]
        >>> equations.get_names("A*")
        []

        >>> # c) delete several equations at once using a list of names
        >>> equations.get_names("B*")
        [\'BENEF\', \'BQY\', \'BRUGP\', \'BVY\']
        >>> del equations[["BENEF", "BQY"]]
        >>> equations.get_names("B*")
        [\'BRUGP\', \'BVY\']

        >>> # delete one equation from a subset of the global database
        >>> equations_subset = equations["D*"]
        >>> equations_subset.names
        [\'DEBT\', \'DPU\', \'DPUF\', \'DPUG\', \'DPUGO\', \'DPUH\', \'DPUU\', \'DTF\', \'DTH\', \'DTH1\', \'DTH1C\']
        >>> del equations_subset["DPUGO"]
        >>> equations_subset.names
        [\'DEBT\', \'DPU\', \'DPUF\', \'DPUG\', \'DPUH\', \'DPUU\', \'DTF\', \'DTH\', \'DTH1\', \'DTH1C\']
        >>> # NOTE: the equation has also been deleted from the global database
        >>> "DPUGO" in equations
        False
        >>> equations.get_names("D*")
        [\'DEBT\', \'DPU\', \'DPUF\', \'DPUG\', \'DPUH\', \'DPUU\', \'DTF\', \'DTH\', \'DTH1\', \'DTH1C\']
        '''
    def __getitem__(self, index):
        '''
        Return the (subset of) equation(s) referenced by `key`.

        The `key` can represent a single object name (e.g. "ACAF") or a list of object names ("ACAF;ACAG;AOUC") 
        or a pattern (e.g. "A*") or a list of sub-patterns (e.g. "A*;*_").
        
        If the `key` represents a list of object names or of sub-patterns, each name or sub-pattern is separated 
        by a `separator` character which is either a whitespace ` `, or a comma `,`, or a semi-colon `;`, or a 
        tabulation `\\t`, or a newline `\\n`.

        A (sub-)`pattern` is a list of characters representing a group of object names. 
        It includes some special characters which have a special meaning:
        
            - `*` : any character sequence, even empty
            - `?` : any character (one and only one)
            - `@` : any alphanumerical char [A-Za-z0-9]
            - `&` : any non alphanumerical char
            - `|` : any alphanumeric character or none at the beginning and end of a string 
            - `!` : any non-alphanumeric character or none at the beginning and end of a string 
            - `\\` : escape the next character

        Note that the `key` can contain references to IODE lists which are prefixed with the symbol `$`.

        Parameters
        ----------
        key: str or list(str)
            (the list of) name(s) of the equation(s) to get.
            The list of objects to get can be specified by a pattern or by a list of sub-patterns (e.g. "A*;*_").

        Returns
        -------
        Single equation or a subset of the database.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import equations
        >>> equations.load(f"{SAMPLE_DATA_DIR}/fun.eqs")

        >>> # a) get one equation
        >>> equations["ACAF"]                  # doctest: +NORMALIZE_WHITESPACE
        Equation(endogenous = \'ACAF\',
                 lec = \'(ACAF/VAF[-1]) :=acaf1+acaf2*GOSF[-1]+\\nacaf4*(TIME=1995)\',
                 method = \'LSQ\',
                 from_period = \'1980Y1\',
                 to_period = \'1996Y1\',
                 block = \'ACAF\',
                 tests = {corr = 1,
                          dw = 2.32935,
                          fstat = 32.2732,
                          loglik = 83.8075,
                          meany = 0.00818467,
                          r2 = 0.821761,
                          r2adj = 0.796299,
                          ssres = 5.19945e-05,
                          stderr = 0.00192715,
                          stderrp = 23.5458,
                          stdev = 0.0042699},
                 date = \'12-06-1998\')

        >>> # b) get a subset of the Equations database using a pattern
        >>> equations_subset = equations["A*"]
        >>> equations_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\']

        >>> # c) get a subset of the Equations database using a list of names
        >>> equations_subset = equations[["ACAF", "AOUC", "BQY", "BVY"]]
        >>> equations_subset.names
        [\'ACAF\', \'AOUC\', \'BQY\', \'BVY\']
        '''
    def __hash__(self) -> int:
        '''
        Return a hash value for the current Equations database.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR, equations
        >>> equations.load(f"{SAMPLE_DATA_DIR}/fun.eqs")
        >>> len(equations)
        274
        >>> original_hash = hash(equations)

        >>> # modify one equation
        >>> eq = equations["ACAF"]
        >>> original_lec = eq.lec
        >>> eq.lec = "ACAF := 1"
        >>> equations["ACAF"] = eq
        >>> original_hash == hash(equations)
        False
        >>> eq.lec = original_lec  # revert the change
        >>> equations["ACAF"] = eq
        >>> original_hash == hash(equations)
        True

        >>> # delete a equation
        >>> original_eq = equations["ACAF"]
        >>> del equations["ACAF"]
        >>> original_hash == hash(equations)
        False
        >>> equations["ACAF"] = original_eq
        >>> original_hash == hash(equations)
        True

        >>> # add a equation
        >>> equations["NEW"] = "NEW := 0"
        >>> original_hash == hash(equations)
        False
        >>> del equations["NEW"]
        >>> original_hash == hash(equations)
        True
        '''
    def __reduce__(self):
        """__reduce_cython__(self)"""
    def __setitem__(self, index, object) -> None:
        '''
        Update/add a (subset of) equation(s) referenced by `key` from/to the Equations database.

        The `key` can represent a single object name (e.g. "ACAF") or a list of object names ("ACAF;ACAG;AOUC") 
        or a pattern (e.g. "A*") or a list of sub-patterns (e.g. "A*;*_").
        
        If the `key` represents a list of object names or of sub-patterns, each name or sub-pattern is separated 
        by a `separator` character which is either a whitespace ` `, or a comma `,`, or a semi-colon `;`, or a 
        tabulation `\\t`, or a newline `\\n`.

        A (sub-)`pattern` is a list of characters representing a group of object names. 
        It includes some special characters which have a special meaning:
        
            - `*` : any character sequence, even empty
            - `?` : any character (one and only one)
            - `@` : any alphanumerical char [A-Za-z0-9]
            - `&` : any non alphanumerical char
            - `|` : any alphanumeric character or none at the beginning and end of a string 
            - `!` : any non-alphanumeric character or none at the beginning and end of a string 
            - `\\` : escape the next character

        Note that the `key` can contain references to IODE lists which are prefixed with the symbol `$`.

        Parameters
        ----------
        key: str or list(str)
            (the list of) name(s) of the equation(s) to update/add.
            The list of objects to update/add can be specified by a pattern or by a list of sub-patterns 
            (e.g. "A*;*_").
        value: str or Equation or dict(str, ...) or list of any of those
            If str, then it is interpreted as the LEC of the equation.
            If dictionary, then it is interpreted as the values of the attributes of the equation to update or add. 
            The key \'endogenous\' is added automatically and does not need to be specified.
            If Equation, then it is used to update an existing equation or to create a new equation if it does not exist yet.
        
        See Also
        --------
        iode.Equation

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import equations, EqMethod
        >>> equations.load(f"{SAMPLE_DATA_DIR}/fun.eqs")
        
        >>> # a) add one equation
        >>> equations["BDY"] = "BDY := YN - YK"
        >>> equations["BDY"]                    # doctest: +NORMALIZE_WHITESPACE
        Equation(endogenous = \'BDY\',
                 lec = \'BDY := YN - YK\',
                 method = \'LSQ\',
                 from_period = \'1960Y1\',
                 to_period = \'2015Y1\')
        
        >>> # b) update one equation        
        >>> equations["ACAF"]                  # doctest: +NORMALIZE_WHITESPACE
        Equation(endogenous = \'ACAF\',
                 lec = \'(ACAF/VAF[-1]) :=acaf1+acaf2*GOSF[-1]+\\nacaf4*(TIME=1995)\',
                 method = \'LSQ\',
                 from_period = \'1980Y1\',
                 to_period = \'1996Y1\',
                 block = \'ACAF\',
                 tests = {corr = 1,
                          dw = 2.32935,
                          fstat = 32.2732,
                          loglik = 83.8075,
                          meany = 0.00818467,
                          r2 = 0.821761,
                          r2adj = 0.796299,
                          ssres = 5.19945e-05,
                          stderr = 0.00192715,
                          stderrp = 23.5458,
                          stdev = 0.0042699},
                 date = \'12-06-1998\')
        >>> # update only the LEC
        >>> equations["ACAF"] = "(ACAF/VAF[-1]) := acaf1 + acaf2 * GOSF[-1] + acaf4 * (TIME=1995)"
        >>> equations["ACAF"]                  # doctest: +NORMALIZE_WHITESPACE
        Equation(endogenous = \'ACAF\',
                 lec = \'(ACAF/VAF[-1]) := acaf1 + acaf2 * GOSF[-1] + acaf4 * (TIME=1995)\',
                 method = \'LSQ\',
                 from_period = \'1980Y1\',
                 to_period = \'1996Y1\',
                 block = \'ACAF\',
                 tests = {corr = 1,
                          dw = 2.32935,
                          fstat = 32.2732,
                          loglik = 83.8075,
                          meany = 0.00818467,
                          r2 = 0.821761,
                          r2adj = 0.796299,
                          ssres = 5.19945e-05,
                          stderr = 0.00192715,
                          stderrp = 23.5458,
                          stdev = 0.0042699},
                 date = \'12-06-1998\')
        >>> # update block and sample of a block of equations to estimation (dictionary)
        >>> estim_sample = "2000Y1:2010Y1"
        >>> block = "ACAF; ACAG; AOUC"
        >>> for eq_name in block.split(\';\'):
        ...     equations[eq_name] = {"sample": estim_sample, "block": block}
        >>> (equations["ACAF"].sample, equations["ACAG"].sample, equations["AOUC"].sample)
        (Sample("2000Y1:2010Y1"), Sample("2000Y1:2010Y1"), Sample("2000Y1:2010Y1"))
        >>> (equations["ACAF"].block, equations["ACAG"].block, equations["AOUC"].block)
        (\'ACAF; ACAG; AOUC\', \'ACAF; ACAG; AOUC\', \'ACAF; ACAG; AOUC\')
        >>> # update lec, method, sample and block
        >>> equations["ACAF"].lec = "(ACAF/VAF[-1]) := acaf2 * GOSF[-1] + acaf4 * (TIME=1995)"
        >>> equations["ACAF"].method = EqMethod.MAX_LIKELIHOOD
        >>> # new equation sample is from 1990Y1 to the last year of Variables
        >>> equations["ACAF"].sample = "1990Y1:"
        >>> equations["ACAF"].block = "ACAF"
        >>> equations["ACAF"]                  # doctest: +NORMALIZE_WHITESPACE
        Equation(endogenous = \'ACAF\',
                 lec = \'(ACAF/VAF[-1]) := acaf2 * GOSF[-1] + acaf4 * (TIME=1995)\',
                 method = \'MAX_LIKELIHOOD\',
                 from_period = \'1990Y1\',
                 to_period = \'2015Y1\',
                 block = \'ACAF\',
                 tests = {corr = 1,
                          dw = 2.32935,
                          fstat = 32.2732,
                          loglik = 83.8075,
                          meany = 0.00818467,
                          r2 = 0.821761,
                          r2adj = 0.796299,
                          ssres = 5.19945e-05,
                          stderr = 0.00192715,
                          stderrp = 23.5458,
                          stdev = 0.0042699},
                 date = \'12-06-1998\')

        >>> # c) working on a subset
        >>> # 1) get subset
        >>> equations_subset = equations["A*"]
        >>> equations_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\']
        >>> # 2) add a equation to the subset 
        >>> equations_subset["AOUC_"] = "AOUC_ := ((WCRH/QL)/(WCRH/QL)[1990Y1]) * (VAFF/(VM+VAFF))[-1] + PM * (VM/(VAFF+VM))[-1]"
        >>> equations_subset["AOUC_"]               # doctest: +NORMALIZE_WHITESPACE
        Equation(endogenous = \'AOUC_\',
                 lec = \'AOUC_ := ((WCRH/QL)/(WCRH/QL)[1990Y1]) * (VAFF/(VM+VAFF))[-1] + PM * (VM/(VAFF+VM))[-1]\',
                 method = \'LSQ\',
                 from_period = \'1960Y1\',
                 to_period = \'2015Y1\')
        >>> # --> new equation also appears in the global workspace
        >>> "AOUC_" in equations
        True
        >>> equations["AOUC_"]                      # doctest: +NORMALIZE_WHITESPACE
        Equation(endogenous = \'AOUC_\',
                 lec = \'AOUC_ := ((WCRH/QL)/(WCRH/QL)[1990Y1]) * (VAFF/(VM+VAFF))[-1] + PM * (VM/(VAFF+VM))[-1]\',
                 method = \'LSQ\',
                 from_period = \'1960Y1\',
                 to_period = \'2015Y1\')
        >>> # 3) update a equation in the subset
        >>> equations_subset["AOUC_"] = "AOUC_ := ((WCRH/QL)/(WCRH/QL)[1990Y1]) * (VAFF/(VM+VAFF))[-1]"
        >>> equations_subset["AOUC_"]           # doctest: +NORMALIZE_WHITESPACE
        Equation(endogenous = \'AOUC_\',
                 lec = \'AOUC_ := ((WCRH/QL)/(WCRH/QL)[1990Y1]) * (VAFF/(VM+VAFF))[-1]\',
                 method = \'LSQ\',
                 from_period = \'1960Y1\',
                 to_period = \'2015Y1\')
        >>> # --> equation is also updated in the global workspace
        >>> equations["AOUC_"]                  # doctest: +NORMALIZE_WHITESPACE
        Equation(endogenous = \'AOUC_\',
                 lec = \'AOUC_ := ((WCRH/QL)/(WCRH/QL)[1990Y1]) * (VAFF/(VM+VAFF))[-1]\',
                 method = \'LSQ\',
                 from_period = \'1960Y1\',
                 to_period = \'2015Y1\')  
        '''

class HighToLowType(enum.IntEnum):
    __new__: ClassVar[Callable] = ...
    LAST: ClassVar[HighToLowType] = ...
    MEAN: ClassVar[HighToLowType] = ...
    SUM: ClassVar[HighToLowType] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[int]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    def __format__(self, *args, **kwargs) -> str: ...

class Identities(_AbstractDatabase):
    __pyx_vtable__: ClassVar[PyCapsule] = ...
    i: PositionalIndexer
    series: Series
    def __init__(self, *args, **kwargs) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""
    def execute(self, identities: str | list[str] = ..., from_period: str | Period = ..., to_period: str | Period = ..., var_files: str | list[str] = ..., scalar_files: str | list[str] = ..., trace: bool = ...) -> Any:
        '''execute(self, identities: Union[str, List[str]] = None, from_period: Union[str, Period] = None, to_period: Union[str, Period] = None, var_files: Union[str, List[str]] = None, scalar_files: Union[str, List[str]] = None, trace: bool = False)

                Execute the specified identity(ies).

                Parameters
                ----------
                identities : str or List[str], optional
                    The identities to execute. If not provided, all identities are executed.
                from_period : str or Period, optional
                    The starting period of the execution range. 
                    Defaults to the starting period of the current Variables sample.
                to_period : str or Period, optional
                    The ending period of the execution range.
                    Defaults to the ending period of the current Variables sample.
                var_files : str or List[str], optional
                    The variable files from which the required Variables are temporarily loaded.
                    The Variables database is left unchanged after the execution of the identity(ies).
                    Defaults to None (current content of the Variables database is used). 
                scalar_files : str or List[str], optional
                    The scalar files from which the required Scalars are temporarily loaded.
                    The Variables database is left unchanged after the execution of the identity(ies).
                    Defaults to None (current content of the Scalars database is used).
                trace : bool, optional
                    If True, prints the source of the variables and scalars.

                Returns
                -------
                None

                Examples
                --------
                >>> from iode import identities, variables, SAMPLE_DATA_DIR
                >>> identities.load(f"{SAMPLE_DATA_DIR}/fun.idt")
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> sample = variables.sample
                >>> sample
                Sample("1960Y1:2015Y1")
                >>> identities["GAP_"]
                Identity(\'100*((QAF_/Q_F)-1)\')
                >>> identities["GAP2"]
                Identity(\'100*(QAFF_/(Q_F+Q_I))\')
                >>> # reset variables GAP_ and GAP2
                >>> variables["GAP_"] = 0.
                >>> variables["GAP_"]                   # doctest: +ELLIPSIS 
                [0.0, 0.0, 0.0, ..., 0.0, 0.0, 0.0]
                >>> variables["GAP2"] = 0.
                >>> variables["GAP2"]                   # doctest: +ELLIPSIS 
                [0.0, 0.0, 0.0, ..., 0.0, 0.0, 0.0]

                >>> # compute GAP_ and GAP2 (assuming Scalars and Variables are already loaded)
                >>> identities.execute("GAP_;GAP2")
                >>> variables["GAP_"]                   # doctest: +ELLIPSIS 
                [-3.20493949860704, -3.981808446333557, ..., 3.7800671441993616, 3.2396415884531793]
                >>> variables["GAP2"]                   # doctest: +ELLIPSIS 
                [96.92655844699298, 97.39603007168847, ..., 102.14581982070376, 101.58578527761608]

                >>> # compute GAP_ and GAP2 over a subset of the sample
                >>> variables["GAP_"] = 0.
                >>> variables["GAP2"] = 0.
                >>> identities.execute("GAP_;GAP2", "2000Y1", "2005Y1")
                >>> variables["GAP_", "2000Y1:2005Y1"]
                [4.510095736743436, 3.312304975734315, 2.6151793579969107, 3.464117181974924, 5.478645527985804, 5.578699398837528]
                >>> variables["GAP2", "2000Y1:2005Y1"]
                [104.60957761618035, 103.05782573291968, 102.17336700422976, 102.82322081548728, 104.4719275849864, 104.3586710898436]

                >>> # compute GAP_ and GAP2 assuming Variables are not already loaded
                >>> variables.clear()
                >>> variables.sample = "1960Y1:2015Y1"
                >>> variables.names
                []
                >>> # setting the var_files argument will fetch the required values of 
                >>> # \'QAF_\', \'QAFF_\', \'Q_F\' and \'Q_I\' from the passed Variables file
                >>> identities.execute("GAP_;GAP2", var_files=f"{SAMPLE_DATA_DIR}/fun.var")
                >>> variables["GAP_"]                   # doctest: +ELLIPSIS 
                [-3.20493949860704, -3.981808446333557, ..., 3.7800671441993616, 3.2396415884531793]
                >>> variables["GAP2"]                   # doctest: +ELLIPSIS 
                [96.92655844699298, 97.39603007168847, ..., 102.14581982070376, 101.58578527761608]
                >>> # note that the variables \'QAF_\', \'QAFF_\', \'Q_F\' and \'Q_I\' are not 
                >>> # present in the Variables database after running identities.execute
                >>> variables.names
                [\'GAP2\', \'GAP_\']
        '''
    @overload
    def from_series(self, s: Series) -> Any:
        '''from_series(self, s: Series)

                Copy the pandas Series `s` into the IODE Identities database.
                The identity names to copy are deduced from the index of the Series.

                Parameters
                ----------
                s: Series
                    pandas Series containing the identities to copy into the IODE Identities database.

                Notes
                -----
                The index of the passed Series is sorted in alphabetical order before 
                copying to IODE Identities database.

                See Also
                --------
                Identities.to_series

                Examples
                --------
                >>> from iode import identities
                >>> import pandas as pd
                >>> identities.clear()
                >>> len(identities)
                0

                >>> # create the pandas Series
                >>> names = ["CONST", "LOG_T", "EXP_T", "GRT_T", "MAVG_T", "DER_LOG_T"]
                >>> data = ["1", "ln t", "exp t", "grt t", "mavg t", "d(ln t)"]
                >>> s = pd.Series(data=data, index=names, dtype=str, name="Identities")
                >>> # display the pandas series
                >>> s          # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                CONST              1
                LOG_T           ln t
                EXP_T          exp t
                GRT_T          grt t
                MAVG_T        mavg t
                DER_LOG_T    d(ln t)
                Name: Identities, dtype: object

                >>> # load into the IODE Identities database
                >>> identities.from_series(s)
                >>> len(identities)
                6
        
                >>> identities.names             # doctest: +ELLIPSIS
                [\'CONST\', \'DER_LOG_T\', \'EXP_T\', \'GRT_T\', \'LOG_T\', \'MAVG_T\']
                >>> identities["LOG_T"]
                Identity(\'ln t\')
                >>> identities["DER_LOG_T"]
                Identity(\'d(ln t)\')
        '''
    @overload
    def from_series(cls, s) -> Any:
        '''from_series(self, s: Series)

                Copy the pandas Series `s` into the IODE Identities database.
                The identity names to copy are deduced from the index of the Series.

                Parameters
                ----------
                s: Series
                    pandas Series containing the identities to copy into the IODE Identities database.

                Notes
                -----
                The index of the passed Series is sorted in alphabetical order before 
                copying to IODE Identities database.

                See Also
                --------
                Identities.to_series

                Examples
                --------
                >>> from iode import identities
                >>> import pandas as pd
                >>> identities.clear()
                >>> len(identities)
                0

                >>> # create the pandas Series
                >>> names = ["CONST", "LOG_T", "EXP_T", "GRT_T", "MAVG_T", "DER_LOG_T"]
                >>> data = ["1", "ln t", "exp t", "grt t", "mavg t", "d(ln t)"]
                >>> s = pd.Series(data=data, index=names, dtype=str, name="Identities")
                >>> # display the pandas series
                >>> s          # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                CONST              1
                LOG_T           ln t
                EXP_T          exp t
                GRT_T          grt t
                MAVG_T        mavg t
                DER_LOG_T    d(ln t)
                Name: Identities, dtype: object

                >>> # load into the IODE Identities database
                >>> identities.from_series(s)
                >>> len(identities)
                6
        
                >>> identities.names             # doctest: +ELLIPSIS
                [\'CONST\', \'DER_LOG_T\', \'EXP_T\', \'GRT_T\', \'LOG_T\', \'MAVG_T\']
                >>> identities["LOG_T"]
                Identity(\'ln t\')
                >>> identities["DER_LOG_T"]
                Identity(\'d(ln t)\')
        '''
    @overload
    def to_series(self) -> Series:
        '''to_series(self) -> Series

                Create a pandas Series from the current Identities database.
                The index of the returned Series is build from the Identities names.

                See Also
                --------
                Identities.from_series

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, identities
                >>> import pandas as pd
                >>> identities.load(f"{SAMPLE_DATA_DIR}/fun.idt")
                >>> len(identities)
                48

                >>> # Export the IODE Identities database as a pandas Series
                >>> s = identities.to_series()
                >>> len(s)
                48

                >>> s.index.to_list()               # doctest: +ELLIPSIS
                [\'AOUC\', \'AOUC_\', \'FLGR\', ..., \'XW\', \'Y\', \'YSEFPR\', \'YSFICR\']
                >>> identities["GAP_"]              # doctest: +NORMALIZE_WHITESPACE
                Identity(\'100*((QAF_/Q_F)-1)\')
                >>> s["GAP_"]                       # doctest: +NORMALIZE_WHITESPACE
                \'100*((QAF_/Q_F)-1)\'
                >>> identities["XTFP"]              # doctest: +NORMALIZE_WHITESPACE
                Identity(\'grt TFPFHP_\')
                >>> s["XTFP"]                       # doctest: +NORMALIZE_WHITESPACE
                \'grt TFPFHP_\'

                >>> # Export a subset of the IODE Identities database as a pandas Series
                >>> s = identities["X*;*_"].to_series()
                >>> len(s)
                15

                >>> s.index.to_list()               # doctest: +ELLIPSIS
                [\'AOUC_\', \'GAP_\', \'XEX\', \'XNATY\', ..., \'XTFP\', \'XW\']
                >>> identities["GAP_"]              # doctest: +NORMALIZE_WHITESPACE
                Identity(\'100*((QAF_/Q_F)-1)\')
                >>> s["GAP_"]                       # doctest: +NORMALIZE_WHITESPACE
                \'100*((QAF_/Q_F)-1)\'
                >>> identities["XTFP"]              # doctest: +NORMALIZE_WHITESPACE
                Identity(\'grt TFPFHP_\')
                >>> s["XTFP"]                       # doctest: +NORMALIZE_WHITESPACE
                \'grt TFPFHP_\'
        '''
    @overload
    def to_series(cls) -> Any:
        '''to_series(self) -> Series

                Create a pandas Series from the current Identities database.
                The index of the returned Series is build from the Identities names.

                See Also
                --------
                Identities.from_series

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, identities
                >>> import pandas as pd
                >>> identities.load(f"{SAMPLE_DATA_DIR}/fun.idt")
                >>> len(identities)
                48

                >>> # Export the IODE Identities database as a pandas Series
                >>> s = identities.to_series()
                >>> len(s)
                48

                >>> s.index.to_list()               # doctest: +ELLIPSIS
                [\'AOUC\', \'AOUC_\', \'FLGR\', ..., \'XW\', \'Y\', \'YSEFPR\', \'YSFICR\']
                >>> identities["GAP_"]              # doctest: +NORMALIZE_WHITESPACE
                Identity(\'100*((QAF_/Q_F)-1)\')
                >>> s["GAP_"]                       # doctest: +NORMALIZE_WHITESPACE
                \'100*((QAF_/Q_F)-1)\'
                >>> identities["XTFP"]              # doctest: +NORMALIZE_WHITESPACE
                Identity(\'grt TFPFHP_\')
                >>> s["XTFP"]                       # doctest: +NORMALIZE_WHITESPACE
                \'grt TFPFHP_\'

                >>> # Export a subset of the IODE Identities database as a pandas Series
                >>> s = identities["X*;*_"].to_series()
                >>> len(s)
                15

                >>> s.index.to_list()               # doctest: +ELLIPSIS
                [\'AOUC_\', \'GAP_\', \'XEX\', \'XNATY\', ..., \'XTFP\', \'XW\']
                >>> identities["GAP_"]              # doctest: +NORMALIZE_WHITESPACE
                Identity(\'100*((QAF_/Q_F)-1)\')
                >>> s["GAP_"]                       # doctest: +NORMALIZE_WHITESPACE
                \'100*((QAF_/Q_F)-1)\'
                >>> identities["XTFP"]              # doctest: +NORMALIZE_WHITESPACE
                Identity(\'grt TFPFHP_\')
                >>> s["XTFP"]                       # doctest: +NORMALIZE_WHITESPACE
                \'grt TFPFHP_\'
        '''
    @overload
    def to_series(cls) -> Any:
        '''to_series(self) -> Series

                Create a pandas Series from the current Identities database.
                The index of the returned Series is build from the Identities names.

                See Also
                --------
                Identities.from_series

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, identities
                >>> import pandas as pd
                >>> identities.load(f"{SAMPLE_DATA_DIR}/fun.idt")
                >>> len(identities)
                48

                >>> # Export the IODE Identities database as a pandas Series
                >>> s = identities.to_series()
                >>> len(s)
                48

                >>> s.index.to_list()               # doctest: +ELLIPSIS
                [\'AOUC\', \'AOUC_\', \'FLGR\', ..., \'XW\', \'Y\', \'YSEFPR\', \'YSFICR\']
                >>> identities["GAP_"]              # doctest: +NORMALIZE_WHITESPACE
                Identity(\'100*((QAF_/Q_F)-1)\')
                >>> s["GAP_"]                       # doctest: +NORMALIZE_WHITESPACE
                \'100*((QAF_/Q_F)-1)\'
                >>> identities["XTFP"]              # doctest: +NORMALIZE_WHITESPACE
                Identity(\'grt TFPFHP_\')
                >>> s["XTFP"]                       # doctest: +NORMALIZE_WHITESPACE
                \'grt TFPFHP_\'

                >>> # Export a subset of the IODE Identities database as a pandas Series
                >>> s = identities["X*;*_"].to_series()
                >>> len(s)
                15

                >>> s.index.to_list()               # doctest: +ELLIPSIS
                [\'AOUC_\', \'GAP_\', \'XEX\', \'XNATY\', ..., \'XTFP\', \'XW\']
                >>> identities["GAP_"]              # doctest: +NORMALIZE_WHITESPACE
                Identity(\'100*((QAF_/Q_F)-1)\')
                >>> s["GAP_"]                       # doctest: +NORMALIZE_WHITESPACE
                \'100*((QAF_/Q_F)-1)\'
                >>> identities["XTFP"]              # doctest: +NORMALIZE_WHITESPACE
                Identity(\'grt TFPFHP_\')
                >>> s["XTFP"]                       # doctest: +NORMALIZE_WHITESPACE
                \'grt TFPFHP_\'
        '''
    def __delitem__(self, other) -> None:
        '''
        Remove the (subset of) identity(ies) referenced by `key` from the Identities database.

        Parameters
        ----------
        key: str or list(str)
            (list of) name(s) of the identity(ies) to be removed.
            The list of names can be given as a string pattern (e.g. "A*;*_").

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import identities
        >>> identities.load(f"{SAMPLE_DATA_DIR}/fun.idt")

        >>> # a) delete one identity
        >>> identities.get_names("W*")
        [\'W\', \'WBGR\', \'WCRH\', \'WMINR\', \'WO\']
        >>> del identities["W"]
        >>> identities.get_names("W*")
        [\'WBGR\', \'WCRH\', \'WMINR\', \'WO\']

        >>> # b) delete several identities at once using a pattern
        >>> del identities["W*"]
        >>> identities.get_names("W*")
        []

        >>> # c) delete several identities at once using a list of names
        >>> identities.get_names("XP*")
        [\'XPOIL\', \'XPWMAB\', \'XPWMS\', \'XPWXAB\', \'XPWXS\']
        >>> del identities[["XPOIL", "XPWXS"]]
        >>> identities.get_names("XP*")
        [\'XPWMAB\', \'XPWMS\', \'XPWXAB\']

        >>> # delete one identity from a subset of the global database
        >>> identities_subset = identities["Y*"]
        >>> identities_subset.names
        [\'Y\', \'YSEFPR\', \'YSFICR\']
        >>> del identities_subset["Y"]
        >>> identities_subset.names
        [\'YSEFPR\', \'YSFICR\']
        >>> # NOTE: the identity has also been deleted from the global database
        >>> "Y" in identities
        False
        >>> identities.get_names("Y*")
        [\'YSEFPR\', \'YSFICR\']
        '''
    def __getitem__(self, index):
        '''
        Return the (subset of) identity(ies) referenced by `key`.

        The `key` can represent a single object name (e.g. "ACAF") or a list of object names ("ACAF;ACAG;AOUC") 
        or a pattern (e.g. "A*") or a list of sub-patterns (e.g. "A*;*_").
        
        If the `key` represents a list of object names or of sub-patterns, each name or sub-pattern is separated 
        by a `separator` character which is either a whitespace ` `, or a comma `,`, or a semi-colon `;`, or a 
        tabulation `\\t`, or a newline `\\n`.

        A (sub-)`pattern` is a list of characters representing a group of object names. 
        It includes some special characters which have a special meaning:
        
            - `*` : any character sequence, even empty
            - `?` : any character (one and only one)
            - `@` : any alphanumerical char [A-Za-z0-9]
            - `&` : any non alphanumerical char
            - `|` : any alphanumeric character or none at the beginning and end of a string 
            - `!` : any non-alphanumeric character or none at the beginning and end of a string 
            - `\\` : escape the next character

        Note that the `key` can contain references to IODE lists which are prefixed with the symbol `$`.

        Parameters
        ----------
        key: str or list(str)
            (the list of) name(s) of the identity(ies) to get.
            The list of objects to get can be specified by a pattern or by a list of sub-patterns (e.g. "A*;*_").

        Returns
        -------
        Single identity or a subset of the database.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import identities
        >>> identities.load(f"{SAMPLE_DATA_DIR}/fun.idt")

        >>> # a) get one Identity
        >>> identities["AOUC"]
        Identity(\'((WCRH/QL)/(WCRH/QL)[1990Y1])*(VAFF/(VM+VAFF))[-1]+PM*(VM/(VM+VAFF))[-1]\')

        >>> # b) get a subset of the Identities database using a pattern
        >>> identities_subset = identities["X*"]
        >>> identities_subset.names
        [\'XEX\', \'XNATY\', \'XPOIL\', \'XPWMAB\', \'XPWMS\', \'XPWXAB\', \'XPWXS\', \'XQWXAB\', \'XQWXS\', \'XQWXSS\', \'XRLBER\', \'XTFP\', \'XW\']

        >>> # c) get a subset of the Identities database using a list of names
        >>> identities_subset = identities[["XEX", "XPWMAB", "XPWMS", "XQWXS", "XTFP"]]
        >>> identities_subset.names
        [\'XEX\', \'XPWMAB\', \'XPWMS\', \'XQWXS\', \'XTFP\']
        '''
    def __hash__(self) -> int:
        '''
        Return a hash value for the current Identities database.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR, identities
        >>> identities.load(f"{SAMPLE_DATA_DIR}/fun.idt")
        >>> len(identities)
        48
        >>> original_hash = hash(identities)
        
        >>> # rename 1 identity
        >>> identities.rename("FLGR", "FLGR_")
        >>> original_hash == hash(identities)
        False
        >>> identities.rename("FLGR_", "FLGR")  # revert the change
        >>> original_hash == hash(identities)
        True

        >>> # modify one identity
        >>> original_lec = identities["FLGR"]
        >>> identities["FLGR"] = "1"
        >>> original_hash == hash(identities)
        False
        >>> identities["FLGR"] = original_lec  # revert the change
        >>> original_hash == hash(identities)
        True

        >>> # delete a identity
        >>> original_identity = identities["FLGR"]
        >>> del identities["FLGR"]
        >>> original_hash == hash(identities)
        False
        >>> identities["FLGR"] = original_identity
        >>> original_hash == hash(identities)
        True

        >>> # add a identity
        >>> identities["NEW"] = "1"
        >>> original_hash == hash(identities)
        False
        >>> del identities["NEW"]
        >>> original_hash == hash(identities)
        True
        '''
    def __reduce__(self):
        """__reduce_cython__(self)"""
    def __setitem__(self, index, object) -> None:
        '''
        Update/add a (subset of) identity(ies) referenced by `key` from/to the Identities database.

        The `key` can represent a single object name (e.g. "ACAF") or a list of object names ("ACAF;ACAG;AOUC") 
        or a pattern (e.g. "A*") or a list of sub-patterns (e.g. "A*;*_").
        
        If the `key` represents a list of object names or of sub-patterns, each name or sub-pattern is separated 
        by a `separator` character which is either a whitespace ` `, or a comma `,`, or a semi-colon `;`, or a 
        tabulation `\\t`, or a newline `\\n`.

        A (sub-)`pattern` is a list of characters representing a group of object names. 
        It includes some special characters which have a special meaning:
        
            - `*` : any character sequence, even empty
            - `?` : any character (one and only one)
            - `@` : any alphanumerical char [A-Za-z0-9]
            - `&` : any non alphanumerical char
            - `|` : any alphanumeric character or none at the beginning and end of a string 
            - `!` : any non-alphanumeric character or none at the beginning and end of a string 
            - `\\` : escape the next character

        Note that the `key` can contain references to IODE lists which are prefixed with the symbol `$`.

        Parameters
        ----------
        key: str or list(str)
            (the list of) name(s) of the identity(ies) to update/add.
            The list of objects to update/add can be specified by a pattern or by a list of sub-patterns 
            (e.g. "A*;*_").
        value: str, Identity, list(str) or list(Identity)
            (new) identity value(s).

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import identities
        >>> identities.load(f"{SAMPLE_DATA_DIR}/fun.idt")
        
        >>> # a) add one identity
        >>> identities["BDY"] = "YN - YK"
        >>> identities["BDY"]
        Identity(\'YN - YK\')

        >>> # b) update one identity
        >>> identities["AOUC"]
        Identity(\'((WCRH/QL)/(WCRH/QL)[1990Y1])*(VAFF/(VM+VAFF))[-1]+PM*(VM/(VM+VAFF))[-1]\')
        >>> identities["AOUC"] = \'(WCRH / WCRH[1990Y1]) * (VAFF / (VM+VAFF))[-1] + PM * (VM / (VM+VAFF))[-1]\'
        >>> identities["AOUC"]
        Identity(\'(WCRH / WCRH[1990Y1]) * (VAFF / (VM+VAFF))[-1] + PM * (VM / (VM+VAFF))[-1]\')

        >>> # c) working on a subset
        >>> # 1) get subset
        >>> identities_subset = identities["X*"]
        >>> identities_subset.names
        [\'XEX\', \'XNATY\', \'XPOIL\', \'XPWMAB\', \'XPWMS\', \'XPWXAB\', \'XPWXS\', \'XQWXAB\', \'XQWXS\', \'XQWXSS\', \'XRLBER\', \'XTFP\', \'XW\']
        >>> # 2) add an identity to the subset 
        >>> identities_subset["XDPU"] = "grt DPU"
        >>> identities_subset["XDPU"]
        Identity(\'grt DPU\')
        >>> # --> new identity also appears in the global workspace
        >>> "XDPU" in identities
        True
        >>> identities["XDPU"]
        Identity(\'grt DPU\')
        >>> # 3) update an identity in the subset
        >>> identities_subset["XDPU"] = "0"
        >>> identities_subset["XDPU"]
        Identity(\'0\')
        >>> # --> identity is also updated in the global workspace
        >>> identities["XDPU"]
        Identity(\'0\')
        '''

class Identity:
    __pyx_vtable__: ClassVar[PyCapsule] = ...
    coefficients: list[str]
    variables: list[str]
    def __init__(self, *args, **kwargs) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""
    def __copy__(self) -> Identity:
        '''__copy__(self) -> Identity

                Return a copy of the current Identity.

                Examples
                --------
                >>> import copy
                >>> from iode import Identity
                >>> idt = Identity("FLG/VBBP")
                >>> idt
                Identity(\'FLG/VBBP\')
                >>> copied_idt = copy.copy(idt)
                >>> copied_idt
                Identity(\'FLG/VBBP\')
        '''
    def __eq__(self, other: object) -> bool:
        """Return self==value."""
    def __ge__(self, other: object) -> bool:
        """Return self>=value."""
    def __gt__(self, other: object) -> bool:
        """Return self>value."""
    def __hash__(self) -> int:
        """Return hash(self)."""
    def __le__(self, other: object) -> bool:
        """Return self<=value."""
    def __lt__(self, other: object) -> bool:
        """Return self<value."""
    def __ne__(self, other: object) -> bool:
        """Return self!=value."""
    def __reduce__(self):
        """__reduce_cython__(self)"""

class IodeFileType(enum.IntEnum):
    __new__: ClassVar[Callable] = ...
    DIRECTORY: ClassVar[IodeFileType] = ...
    FILE_A2M: ClassVar[IodeFileType] = ...
    FILE_AAS: ClassVar[IodeFileType] = ...
    FILE_AGL: ClassVar[IodeFileType] = ...
    FILE_ANY: ClassVar[IodeFileType] = ...
    FILE_COMMENTS: ClassVar[IodeFileType] = ...
    FILE_CSV: ClassVar[IodeFileType] = ...
    FILE_DIF: ClassVar[IodeFileType] = ...
    FILE_EQUATIONS: ClassVar[IodeFileType] = ...
    FILE_HTML: ClassVar[IodeFileType] = ...
    FILE_IDENTITIES: ClassVar[IodeFileType] = ...
    FILE_LISTS: ClassVar[IodeFileType] = ...
    FILE_LOG: ClassVar[IodeFileType] = ...
    FILE_MIF: ClassVar[IodeFileType] = ...
    FILE_PRF: ClassVar[IodeFileType] = ...
    FILE_PS: ClassVar[IodeFileType] = ...
    FILE_REF: ClassVar[IodeFileType] = ...
    FILE_REP: ClassVar[IodeFileType] = ...
    FILE_RTF: ClassVar[IodeFileType] = ...
    FILE_SCALARS: ClassVar[IodeFileType] = ...
    FILE_SETTINGS: ClassVar[IodeFileType] = ...
    FILE_TABLES: ClassVar[IodeFileType] = ...
    FILE_TXT: ClassVar[IodeFileType] = ...
    FILE_VARIABLES: ClassVar[IodeFileType] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[int]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    def __format__(self, *args, **kwargs) -> str: ...

class IodeType(enum.IntEnum):
    __new__: ClassVar[Callable] = ...
    COMMENTS: ClassVar[IodeType] = ...
    EQUATIONS: ClassVar[IodeType] = ...
    IDENTITIES: ClassVar[IodeType] = ...
    LISTS: ClassVar[IodeType] = ...
    SCALARS: ClassVar[IodeType] = ...
    TABLES: ClassVar[IodeType] = ...
    VARIABLES: ClassVar[IodeType] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[int]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    def __format__(self, *args, **kwargs) -> str: ...

class Lists(_AbstractDatabase):
    __pyx_vtable__: ClassVar[PyCapsule] = ...
    i: PositionalIndexer
    series: Series
    def __init__(self, *args, **kwargs) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""
    @overload
    def from_series(self, s: Series) -> Any:
        '''from_series(self, s: Series)

                Copy the pandas Series `s` into the IODE Lists database.
                The IODE list names to copy are deduced from the index of the Series.

                Parameters
                ----------
                s: Series
                    pandas Series containing the IODE lists to copy into the IODE Lists database.

                Notes
                -----
                The index of the passed Series is sorted in alphabetical order before 
                copying to IODE Lists database.

                See Also
                --------
                Lists.to_series

                Examples
                --------
                >>> from iode import lists
                >>> import pandas as pd
                >>> lists.clear()
                >>> len(lists)
                0

                >>> # create the pandas Series
                >>> names = ["SEMICOLON_LST", "COMMA_LST", "WHITESPACE_LST", "TAB_LST", "MIX_LST"]
                >>> semicolon_lst = "A;B;C;D;E;F"
                >>> comma_lst = "A,B,C,D,E,F"
                >>> whitespace_lst = "A B C D E F" 
                >>> tab_lst = "A    B    C    D    E    F"
                >>> mix_lst = "  A;B,C D    E;F  "
                >>> data = [semicolon_lst, comma_lst, whitespace_lst, tab_lst, mix_lst]
                >>> s = pd.Series(data=data, index=names, dtype=str, name="Lists")
                >>> # display the pandas series
                >>> s          # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                SEMICOLON_LST                        A;B;C;D;E;F
                COMMA_LST                            A,B,C,D,E,F
                WHITESPACE_LST                       A B C D E F
                TAB_LST               A    B    C    D    E    F
                MIX_LST                         A;B,C D    E;F
                Name: Lists, dtype: object

                >>> # load into the IODE Lists database
                >>> lists.from_series(s)
                >>> len(lists)
                5
        
                >>> lists.names
                [\'COMMA_LST\', \'MIX_LST\', \'SEMICOLON_LST\', \'TAB_LST\', \'WHITESPACE_LST\']
                >>> lists["SEMICOLON_LST"]
                [\'A\', \'B\', \'C\', \'D\', \'E\', \'F\']
                >>> # note: when added or updated, an IODE list is normalized.
                >>> #       - leading and trailing whitespaces are stripped
                >>> #       - items of the IODE list are separated by a semicolon
                >>> lists["COMMA_LST"]
                [\'A\', \'B\', \'C\', \'D\', \'E\', \'F\']
                >>> lists["MIX_LST"]
                [\'A\', \'B\', \'C\', \'D\', \'E\', \'F\']
        '''
    @overload
    def from_series(cls, s) -> Any:
        '''from_series(self, s: Series)

                Copy the pandas Series `s` into the IODE Lists database.
                The IODE list names to copy are deduced from the index of the Series.

                Parameters
                ----------
                s: Series
                    pandas Series containing the IODE lists to copy into the IODE Lists database.

                Notes
                -----
                The index of the passed Series is sorted in alphabetical order before 
                copying to IODE Lists database.

                See Also
                --------
                Lists.to_series

                Examples
                --------
                >>> from iode import lists
                >>> import pandas as pd
                >>> lists.clear()
                >>> len(lists)
                0

                >>> # create the pandas Series
                >>> names = ["SEMICOLON_LST", "COMMA_LST", "WHITESPACE_LST", "TAB_LST", "MIX_LST"]
                >>> semicolon_lst = "A;B;C;D;E;F"
                >>> comma_lst = "A,B,C,D,E,F"
                >>> whitespace_lst = "A B C D E F" 
                >>> tab_lst = "A    B    C    D    E    F"
                >>> mix_lst = "  A;B,C D    E;F  "
                >>> data = [semicolon_lst, comma_lst, whitespace_lst, tab_lst, mix_lst]
                >>> s = pd.Series(data=data, index=names, dtype=str, name="Lists")
                >>> # display the pandas series
                >>> s          # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                SEMICOLON_LST                        A;B;C;D;E;F
                COMMA_LST                            A,B,C,D,E,F
                WHITESPACE_LST                       A B C D E F
                TAB_LST               A    B    C    D    E    F
                MIX_LST                         A;B,C D    E;F
                Name: Lists, dtype: object

                >>> # load into the IODE Lists database
                >>> lists.from_series(s)
                >>> len(lists)
                5
        
                >>> lists.names
                [\'COMMA_LST\', \'MIX_LST\', \'SEMICOLON_LST\', \'TAB_LST\', \'WHITESPACE_LST\']
                >>> lists["SEMICOLON_LST"]
                [\'A\', \'B\', \'C\', \'D\', \'E\', \'F\']
                >>> # note: when added or updated, an IODE list is normalized.
                >>> #       - leading and trailing whitespaces are stripped
                >>> #       - items of the IODE list are separated by a semicolon
                >>> lists["COMMA_LST"]
                [\'A\', \'B\', \'C\', \'D\', \'E\', \'F\']
                >>> lists["MIX_LST"]
                [\'A\', \'B\', \'C\', \'D\', \'E\', \'F\']
        '''
    @overload
    def to_series(self) -> Series:
        '''to_series(self) -> Series

                Create a pandas Series from the current Lists database.
                The index of the returned Series is build from the Lists names.

                See Also
                --------
                Lists.from_series

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, lists
                >>> import pandas as pd
                >>> lists.load(f"{SAMPLE_DATA_DIR}/fun.lst")
                >>> len(lists)
                17

                >>> # Export the IODE Lists database as a pandas Series
                >>> s = lists.to_series()
                >>> len(s)
                17

                >>> s.index.to_list()               # doctest: +ELLIPSIS
                [\'COPY\', \'COPY0\', \'COPY1\', ..., \'XSCENARIO\', \'_SCAL\', \'_SEARCH\']
                >>> lists["ENVI"]                   # doctest: +NORMALIZE_WHITESPACE
                [\'EX\', \'PWMAB\', \'PWMS\', \'PWXAB\', \'PWXS\', \'QWXAB\', \'QWXS\', \'POIL\', \'NATY\', \'TFPFHP_\'] 
                >>> s["ENVI"]                       # doctest: +NORMALIZE_WHITESPACE
                "[\'EX\', \'PWMAB\', \'PWMS\', \'PWXAB\', \'PWXS\', \'QWXAB\', \'QWXS\', \'POIL\', \'NATY\', \'TFPFHP_\']"
                >>> lists["MAINEQ"]                 # doctest: +NORMALIZE_WHITESPACE
                [\'W\', \'NFYH\', \'KNFF\', \'PC\', \'PXAB\', \'PMAB\', \'QXAB\', \'QMAB\', \'QC_\']
                >>> s["MAINEQ"]                     # doctest: +NORMALIZE_WHITESPACE
                "[\'W\', \'NFYH\', \'KNFF\', \'PC\', \'PXAB\', \'PMAB\', \'QXAB\', \'QMAB\', \'QC_\']"

                >>> # Export a subset of the IODE Lists database as a pandas Series
                >>> s = lists["E*;MA*"].to_series()
                >>> len(s)
                5

                >>> s.index.to_list()
                [\'ENDO\', \'ENDO0\', \'ENDO1\', \'ENVI\', \'MAINEQ\']
                >>> lists["ENVI"]                   # doctest: +NORMALIZE_WHITESPACE
                [\'EX\', \'PWMAB\', \'PWMS\', \'PWXAB\', \'PWXS\', \'QWXAB\', \'QWXS\', \'POIL\', \'NATY\', \'TFPFHP_\'] 
                >>> s["ENVI"]                       # doctest: +NORMALIZE_WHITESPACE
                "[\'EX\', \'PWMAB\', \'PWMS\', \'PWXAB\', \'PWXS\', \'QWXAB\', \'QWXS\', \'POIL\', \'NATY\', \'TFPFHP_\']"
                >>> lists["MAINEQ"]                 # doctest: +NORMALIZE_WHITESPACE
                [\'W\', \'NFYH\', \'KNFF\', \'PC\', \'PXAB\', \'PMAB\', \'QXAB\', \'QMAB\', \'QC_\']
                >>> s["MAINEQ"]                     # doctest: +NORMALIZE_WHITESPACE
                "[\'W\', \'NFYH\', \'KNFF\', \'PC\', \'PXAB\', \'PMAB\', \'QXAB\', \'QMAB\', \'QC_\']"
        '''
    @overload
    def to_series(cls) -> Any:
        '''to_series(self) -> Series

                Create a pandas Series from the current Lists database.
                The index of the returned Series is build from the Lists names.

                See Also
                --------
                Lists.from_series

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, lists
                >>> import pandas as pd
                >>> lists.load(f"{SAMPLE_DATA_DIR}/fun.lst")
                >>> len(lists)
                17

                >>> # Export the IODE Lists database as a pandas Series
                >>> s = lists.to_series()
                >>> len(s)
                17

                >>> s.index.to_list()               # doctest: +ELLIPSIS
                [\'COPY\', \'COPY0\', \'COPY1\', ..., \'XSCENARIO\', \'_SCAL\', \'_SEARCH\']
                >>> lists["ENVI"]                   # doctest: +NORMALIZE_WHITESPACE
                [\'EX\', \'PWMAB\', \'PWMS\', \'PWXAB\', \'PWXS\', \'QWXAB\', \'QWXS\', \'POIL\', \'NATY\', \'TFPFHP_\'] 
                >>> s["ENVI"]                       # doctest: +NORMALIZE_WHITESPACE
                "[\'EX\', \'PWMAB\', \'PWMS\', \'PWXAB\', \'PWXS\', \'QWXAB\', \'QWXS\', \'POIL\', \'NATY\', \'TFPFHP_\']"
                >>> lists["MAINEQ"]                 # doctest: +NORMALIZE_WHITESPACE
                [\'W\', \'NFYH\', \'KNFF\', \'PC\', \'PXAB\', \'PMAB\', \'QXAB\', \'QMAB\', \'QC_\']
                >>> s["MAINEQ"]                     # doctest: +NORMALIZE_WHITESPACE
                "[\'W\', \'NFYH\', \'KNFF\', \'PC\', \'PXAB\', \'PMAB\', \'QXAB\', \'QMAB\', \'QC_\']"

                >>> # Export a subset of the IODE Lists database as a pandas Series
                >>> s = lists["E*;MA*"].to_series()
                >>> len(s)
                5

                >>> s.index.to_list()
                [\'ENDO\', \'ENDO0\', \'ENDO1\', \'ENVI\', \'MAINEQ\']
                >>> lists["ENVI"]                   # doctest: +NORMALIZE_WHITESPACE
                [\'EX\', \'PWMAB\', \'PWMS\', \'PWXAB\', \'PWXS\', \'QWXAB\', \'QWXS\', \'POIL\', \'NATY\', \'TFPFHP_\'] 
                >>> s["ENVI"]                       # doctest: +NORMALIZE_WHITESPACE
                "[\'EX\', \'PWMAB\', \'PWMS\', \'PWXAB\', \'PWXS\', \'QWXAB\', \'QWXS\', \'POIL\', \'NATY\', \'TFPFHP_\']"
                >>> lists["MAINEQ"]                 # doctest: +NORMALIZE_WHITESPACE
                [\'W\', \'NFYH\', \'KNFF\', \'PC\', \'PXAB\', \'PMAB\', \'QXAB\', \'QMAB\', \'QC_\']
                >>> s["MAINEQ"]                     # doctest: +NORMALIZE_WHITESPACE
                "[\'W\', \'NFYH\', \'KNFF\', \'PC\', \'PXAB\', \'PMAB\', \'QXAB\', \'QMAB\', \'QC_\']"
        '''
    @overload
    def to_series(cls) -> Any:
        '''to_series(self) -> Series

                Create a pandas Series from the current Lists database.
                The index of the returned Series is build from the Lists names.

                See Also
                --------
                Lists.from_series

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, lists
                >>> import pandas as pd
                >>> lists.load(f"{SAMPLE_DATA_DIR}/fun.lst")
                >>> len(lists)
                17

                >>> # Export the IODE Lists database as a pandas Series
                >>> s = lists.to_series()
                >>> len(s)
                17

                >>> s.index.to_list()               # doctest: +ELLIPSIS
                [\'COPY\', \'COPY0\', \'COPY1\', ..., \'XSCENARIO\', \'_SCAL\', \'_SEARCH\']
                >>> lists["ENVI"]                   # doctest: +NORMALIZE_WHITESPACE
                [\'EX\', \'PWMAB\', \'PWMS\', \'PWXAB\', \'PWXS\', \'QWXAB\', \'QWXS\', \'POIL\', \'NATY\', \'TFPFHP_\'] 
                >>> s["ENVI"]                       # doctest: +NORMALIZE_WHITESPACE
                "[\'EX\', \'PWMAB\', \'PWMS\', \'PWXAB\', \'PWXS\', \'QWXAB\', \'QWXS\', \'POIL\', \'NATY\', \'TFPFHP_\']"
                >>> lists["MAINEQ"]                 # doctest: +NORMALIZE_WHITESPACE
                [\'W\', \'NFYH\', \'KNFF\', \'PC\', \'PXAB\', \'PMAB\', \'QXAB\', \'QMAB\', \'QC_\']
                >>> s["MAINEQ"]                     # doctest: +NORMALIZE_WHITESPACE
                "[\'W\', \'NFYH\', \'KNFF\', \'PC\', \'PXAB\', \'PMAB\', \'QXAB\', \'QMAB\', \'QC_\']"

                >>> # Export a subset of the IODE Lists database as a pandas Series
                >>> s = lists["E*;MA*"].to_series()
                >>> len(s)
                5

                >>> s.index.to_list()
                [\'ENDO\', \'ENDO0\', \'ENDO1\', \'ENVI\', \'MAINEQ\']
                >>> lists["ENVI"]                   # doctest: +NORMALIZE_WHITESPACE
                [\'EX\', \'PWMAB\', \'PWMS\', \'PWXAB\', \'PWXS\', \'QWXAB\', \'QWXS\', \'POIL\', \'NATY\', \'TFPFHP_\'] 
                >>> s["ENVI"]                       # doctest: +NORMALIZE_WHITESPACE
                "[\'EX\', \'PWMAB\', \'PWMS\', \'PWXAB\', \'PWXS\', \'QWXAB\', \'QWXS\', \'POIL\', \'NATY\', \'TFPFHP_\']"
                >>> lists["MAINEQ"]                 # doctest: +NORMALIZE_WHITESPACE
                [\'W\', \'NFYH\', \'KNFF\', \'PC\', \'PXAB\', \'PMAB\', \'QXAB\', \'QMAB\', \'QC_\']
                >>> s["MAINEQ"]                     # doctest: +NORMALIZE_WHITESPACE
                "[\'W\', \'NFYH\', \'KNFF\', \'PC\', \'PXAB\', \'PMAB\', \'QXAB\', \'QMAB\', \'QC_\']"
        '''
    def __delitem__(self, other) -> None:
        '''
        Remove the (subset of) IODE list(s) referenced by `key` from the Lists database.

        Parameters
        ----------
        key: str or list(str)
            (list of) name(s) of the IODE list(s) to be removed.
            The list of names can be given as a string pattern (e.g. "A*;*_").

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import lists
        >>> lists.load(f"{SAMPLE_DATA_DIR}/fun.lst")

        >>> # a) delete one IODE list
        >>> lists.get_names("C*")
        [\'COPY\', \'COPY0\', \'COPY1\']
        >>> del lists["COPY"]
        >>> lists.get_names("C*")
        [\'COPY0\', \'COPY1\']

        >>> # b) delete several lists at once using a pattern
        >>> del lists["C*"]
        >>> lists.get_names("C*")
        []

        >>> # c) delete several lists at once using a list of names
        >>> lists.get_names("T*")
        [\'TOTAL\', \'TOTAL0\', \'TOTAL1\']
        >>> del lists[["TOTAL", "TOTAL0"]]
        >>> lists.get_names("T*")
        [\'TOTAL1\']

        >>> # delete one IODE list from a subset of the global database
        >>> lists_subset = lists["E*"]
        >>> lists_subset.names
        [\'ENDO\', \'ENDO0\', \'ENDO1\', \'ENVI\']
        >>> del lists_subset["ENVI"]
        >>> lists_subset.names
        [\'ENDO\', \'ENDO0\', \'ENDO1\']
        >>> # NOTE: the IODE list has also been deleted from the global database
        >>> "ENVI" in lists
        False
        >>> lists.get_names("E*")
        [\'ENDO\', \'ENDO0\', \'ENDO1\']
        '''
    def __getitem__(self, index):
        '''
        Return the (subset of) IODE list(s) referenced by `key`.

        The `key` can represent a single object name (e.g. "ACAF") or a list of object names ("ACAF;ACAG;AOUC") 
        or a pattern (e.g. "A*") or a list of sub-patterns (e.g. "A*;*_").
        
        If the `key` represents a list of object names or of sub-patterns, each name or sub-pattern is separated 
        by a `separator` character which is either a whitespace ` `, or a comma `,`, or a semi-colon `;`, or a 
        tabulation `\\t`, or a newline `\\n`.

        A (sub-)`pattern` is a list of characters representing a group of object names. 
        It includes some special characters which have a special meaning:
        
            - `*` : any character sequence, even empty
            - `?` : any character (one and only one)
            - `@` : any alphanumerical char [A-Za-z0-9]
            - `&` : any non alphanumerical char
            - `|` : any alphanumeric character or none at the beginning and end of a string 
            - `!` : any non-alphanumeric character or none at the beginning and end of a string 
            - `\\` : escape the next character

        Note that the `key` can contain references to IODE lists which are prefixed with the symbol `$`.

        Parameters
        ----------
        key: str or list(str)
            (the list of) name(s) of the IODE list(s) to get.
            The list of objects to get can be specified by a pattern or by a list of sub-patterns (e.g. "A*;*_").

        Returns
        -------
        Single IODE list or a subset of the database.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import lists
        >>> lists.load(f"{SAMPLE_DATA_DIR}/fun.lst")

        >>> # a) get one list
        >>> lists["ENVI"]
        [\'EX\', \'PWMAB\', \'PWMS\', \'PWXAB\', \'PWXS\', \'QWXAB\', \'QWXS\', \'POIL\', \'NATY\', \'TFPFHP_\']

        >>> # b) get a subset of the Lists database using a pattern
        >>> lists_subset = lists["E*"]
        >>> lists_subset.names
        [\'ENDO\', \'ENDO0\', \'ENDO1\', \'ENVI\']

        >>> # c) get a subset of the Lists database using a list of names
        >>> lists_subset = lists[["COPY", "ENDO", "ENVI", "TOTAL"]]
        >>> lists_subset.names
        [\'COPY\', \'ENDO\', \'ENVI\', \'TOTAL\']
        '''
    def __hash__(self) -> int:
        '''
        Return a hash value for the current Lists database.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR, lists
        >>> lists.load(f"{SAMPLE_DATA_DIR}/fun.lst")
        >>> len(lists)
        17
        >>> original_hash = hash(lists)
        
        >>> # rename 1 IODE list
        >>> lists.rename("COPY", "COPY_")
        >>> original_hash == hash(lists)
        False
        >>> lists.rename("COPY_", "COPY")  # revert the change
        >>> original_hash == hash(lists)
        True

        >>> # modify one IODE list
        >>> original_list = lists["COPY"]
        >>> lists["COPY"] = "$COPY0 $COPY1"
        >>> original_hash == hash(lists)
        False
        >>> lists["COPY"] = original_list  # revert the change
        >>> original_hash == hash(lists)
        True

        >>> # delete a IODE list
        >>> original_list = lists["COPY"]
        >>> del lists["COPY"]
        >>> original_hash == hash(lists)
        False
        >>> lists["COPY"] = original_list
        >>> original_hash == hash(lists)
        True

        >>> # add a IODE list
        >>> lists["NEW"] = "$COPY0"
        >>> original_hash == hash(lists)
        False
        >>> del lists["NEW"]
        >>> original_hash == hash(lists)
        True
        '''
    def __reduce__(self):
        """__reduce_cython__(self)"""
    def __setitem__(self, index, object) -> None:
        '''
        Update/add a (subset of) IODE list(s) referenced by `key` from/to the Lists database.

        The `key` can represent a single object name (e.g. "ACAF") or a list of object names ("ACAF;ACAG;AOUC") 
        or a pattern (e.g. "A*") or a list of sub-patterns (e.g. "A*;*_").
        
        If the `key` represents a list of object names or of sub-patterns, each name or sub-pattern is separated 
        by a `separator` character which is either a whitespace ` `, or a comma `,`, or a semi-colon `;`, or a 
        tabulation `\\t`, or a newline `\\n`.

        A (sub-)`pattern` is a list of characters representing a group of object names. 
        It includes some special characters which have a special meaning:
        
            - `*` : any character sequence, even empty
            - `?` : any character (one and only one)
            - `@` : any alphanumerical char [A-Za-z0-9]
            - `&` : any non alphanumerical char
            - `|` : any alphanumeric character or none at the beginning and end of a string 
            - `!` : any non-alphanumeric character or none at the beginning and end of a string 
            - `\\` : escape the next character

        Note that the `key` can contain references to IODE lists which are prefixed with the symbol `$`.

        Parameters
        ----------
        key: str or list(str)
            (the list of) name(s) of the IODE list(s) to update/add.
            The list of objects to update/add can be specified by a pattern or by a list of sub-patterns 
            (e.g. "A*;*_").
        value: str or list(str)
            (new) IODE list value(s).

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import lists, variables
        >>> lists.load(f"{SAMPLE_DATA_DIR}/fun.lst")
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
        
        >>> # a) add one list
        >>> # --- by passing a string 
        >>> lists["A_VAR"] = "ACAF;ACAG;AOUC;AOUC_;AQC"
        >>> lists["A_VAR"]
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> # --- by passing a Python list
        >>> b_vars = variables.get_names("B*")
        >>> b_vars
        [\'BENEF\', \'BQY\', \'BRUGP\', \'BVY\']
        >>> lists["B_VAR"] = b_vars
        >>> lists["B_VAR"]
        [\'BENEF\', \'BQY\', \'BRUGP\', \'BVY\']

        >>> # b) update one list
        >>> # --- by passing a string
        >>> lists["A_VAR"] = "ACAF;ACAG;AOUC;AQC"
        >>> lists["A_VAR"]
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AQC\']
        >>> # --- by passing a Python list
        >>> b_y_vars = variables.get_names("B*Y")
        >>> b_y_vars
        [\'BQY\', \'BVY\']
        >>> lists["B_VAR"] = b_y_vars
        >>> lists["B_VAR"]
        [\'BQY\', \'BVY\']

        >>> # c) working on a subset
        >>> # 1) get subset
        >>> lists_subset = lists["E*"]
        >>> lists_subset.names
        [\'ENDO\', \'ENDO0\', \'ENDO1\', \'ENVI\']
        >>> # 2) add a list to the subset 
        >>> lists_subset["E_VAR"] = variables.get_names("E*")
        >>> lists_subset["E_VAR"]
        [\'EFMY\', \'EFXY\', \'EX\', \'EXC\', \'EXCC\', \'EXCCR\']
        >>> # --> new list also appears in the global workspace
        >>> "E_VAR" in lists
        True
        >>> lists["E_VAR"]
        [\'EFMY\', \'EFXY\', \'EX\', \'EXC\', \'EXCC\', \'EXCCR\']
        >>> # 3) update a list in the subset
        >>> lists_subset["E_VAR"] = "EX;EXC;EXCC;EXCCR"
        >>> lists_subset["E_VAR"]
        [\'EX\', \'EXC\', \'EXCC\', \'EXCCR\']
        >>> # --> list is also updated in the global workspace
        >>> lists["E_VAR"]
        [\'EX\', \'EXC\', \'EXCC\', \'EXCCR\']
        '''

class LowToHighMethod(enum.StrEnum):
    __new__: ClassVar[Callable] = ...
    CUBIC_SPLINES: ClassVar[LowToHighMethod] = ...
    LINEAR: ClassVar[LowToHighMethod] = ...
    STEP: ClassVar[LowToHighMethod] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[str]] = ...
    _new_member_: ClassVar[Callable] = ...
    _unhashable_values_: ClassVar[list] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    def __format__(self, *args, **kwargs) -> str:
        """Return a formatted version of the string as described by format_spec."""

class LowToHighType(enum.IntEnum):
    __new__: ClassVar[Callable] = ...
    FLOW: ClassVar[LowToHighType] = ...
    STOCK: ClassVar[LowToHighType] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[int]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    def __format__(self, *args, **kwargs) -> str: ...

class Period:
    nb_periods_per_year: int
    periodicity: str
    step: int
    year: int
    def __init__(self, period_or_year: str | int | Period, periodicity: str = ..., step: int = ...) -> Period:
        '''__init__(self, period_or_year: Union[str, int, Period], periodicity: str = u\'Y\', step: int = 1) -> Period

                An IODE period is defined by a year, a periodicity and by a position in the year.

                Possible values for the periodicity are:

                    - Y: yearly
                    - S: semestrial
                    - Q: quarterly
                    - M: monthly
                    - W: weekly
                    - D: daily

                Parameters
                ----------
                period_or_year: str or int
                    If str, it is consided as a string representing the year, the periodicity 
                    and the position of the period in the year (step).
                    If int, it is considered as the year of the period.
                periodicity: str, optional
                        Periodicity of the period. Possible values are \'Y\', \'S\', \'Q\', \'M\', \'W\' and \'D\'.
                        Default to \'Y\' (yearly).
                step: int, optional
                        Position of the period in the year. 
                        Default to 1.

                Attributes
                ----------
                year: int
                periodicity: str
                step: int
                    position in the year

                Examples
                --------
                >>> from iode import Period
                >>> # passing a string
                >>> period = Period("2000Y1")
                >>> period
                Period("2000Y1")
                >>> # passing year, periodicity and step
                >>> period = Period(2010, \'Q\', 1)
                >>> period
                Period("2010Q1")
                >>> # passing only year
                >>> period = Period(2010)
                >>> period
                Period("2010Y1")
                >>> # passing year and periodicity
                >>> period = Period(2010, \'Q\')
                >>> period
                Period("2010Q1")
                >>> # copy a period
                >>> period_2 = Period(period)
                >>> period_2
                Period("2010Q1")
        '''
    @overload
    def difference(self, other: Period) -> int:
        """difference(self, other: Period) -> int

                Number of sub periods between two periods. 
                The two periods must have the same periodicity.

                Parameters
                ----------
                other: Period

                Returns
                -------
                int

                Examples
                --------
                >>> from iode import Period
                >>> period = Period(2000, 'Q', 1)
                >>> period_2 = Period(2001, 'Q', 3)
                >>> period.difference(period_2)
                -6
                >>> period_2.difference(period)
                6
        """
    @overload
    def difference(cls, period_2) -> Any:
        """difference(self, other: Period) -> int

                Number of sub periods between two periods. 
                The two periods must have the same periodicity.

                Parameters
                ----------
                other: Period

                Returns
                -------
                int

                Examples
                --------
                >>> from iode import Period
                >>> period = Period(2000, 'Q', 1)
                >>> period_2 = Period(2001, 'Q', 3)
                >>> period.difference(period_2)
                -6
                >>> period_2.difference(period)
                6
        """
    @overload
    def difference(cls, period) -> Any:
        """difference(self, other: Period) -> int

                Number of sub periods between two periods. 
                The two periods must have the same periodicity.

                Parameters
                ----------
                other: Period

                Returns
                -------
                int

                Examples
                --------
                >>> from iode import Period
                >>> period = Period(2000, 'Q', 1)
                >>> period_2 = Period(2001, 'Q', 3)
                >>> period.difference(period_2)
                -6
                >>> period_2.difference(period)
                6
        """
    def shift(self, nb_periods: int) -> Period:
        '''shift(self, nb_periods: int) -> Period

                Shift the current period by a number of sub periods.
                If the number of sub periods is positive, the shift is time forward.  
                Conversely, if the number of sub periods is negative, the shift is time backward.  

                Parameters
                ----------
                nb_periods: int
                    Number of sub periods. 
                    The shift is time forward if positive and time backward if negative. 

                Returns
                -------
                shifted_period: Period

                Examples
                --------
                >>> from iode import Period
                >>> period = Period(2000, \'Q\', 1)
                >>> period
                Period("2000Q1")
                >>> # shift forward
                >>> shifted_period = period.shift(7)
                >>> shifted_period
                Period("2001Q4")
                >>> # shift backward
                >>> shifted_period = period.shift(-7)
                >>> shifted_period
                Period("1998Q2")
        '''
    @overload
    def to_float(self) -> float:
        """to_float(self) -> float

                Returns a numerical representation of the period.

                Returns
                -------
                float

                Examples
                --------
                >>> from iode import Period
                >>> period = Period(2000, 'Q', 1)
                >>> period.to_float()
                2000.0
                >>> period = Period(2000, 'Q', 3)
                >>> period.to_float()
                2000.5
        """
    @overload
    def to_float(cls) -> Any:
        """to_float(self) -> float

                Returns a numerical representation of the period.

                Returns
                -------
                float

                Examples
                --------
                >>> from iode import Period
                >>> period = Period(2000, 'Q', 1)
                >>> period.to_float()
                2000.0
                >>> period = Period(2000, 'Q', 3)
                >>> period.to_float()
                2000.5
        """
    @overload
    def to_float(cls) -> Any:
        """to_float(self) -> float

                Returns a numerical representation of the period.

                Returns
                -------
                float

                Examples
                --------
                >>> from iode import Period
                >>> period = Period(2000, 'Q', 1)
                >>> period.to_float()
                2000.0
                >>> period = Period(2000, 'Q', 3)
                >>> period.to_float()
                2000.5
        """
    def __eq__(self, other: object) -> bool:
        """Return self==value."""
    def __ge__(self, other: object) -> bool:
        """Return self>=value."""
    def __gt__(self, other: object) -> bool:
        """Return self>value."""
    def __le__(self, other: object) -> bool:
        """Return self<=value."""
    def __lt__(self, other: object) -> bool:
        """Return self<value."""
    def __ne__(self, other: object) -> bool:
        """Return self!=value."""
    def __reduce__(self):
        """__reduce_cython__(self)"""

class PositionalIndexer:
    def __init__(self, database) -> Any:
        """__init__(self, database)"""
    def __getitem__(self, index: int) -> Any:
        """__getitem__(self, index: int)"""
    def __iter__(self) -> typing.Iterator[Any]:
        """def __iter__(self) -> typing.Iterator[Any]"""
    def __setitem__(self, index: int, value) -> Any:
        """__setitem__(self, index: int, value)"""

class Sample:
    __pyx_vtable__: ClassVar[PyCapsule] = ...
    end: Incomplete
    nb_periods: Incomplete
    start: Incomplete
    def __init__(self, *args, **kwargs) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""
    @overload
    def get_period_list(self, astype=...) -> list[Any]:
        '''get_period_list(self, astype: Union[type(Any), str] = str) -> List[Any]

                List of all periods of the sample.
                Periods are exported as string (default) or as float.

                Parameters
                ----------
                astype: type or str
                    Allowed returned type for periods are str and float.
                    Default to str.

                Returns
                -------
                list(str) or list(float)

                Examples
                --------
                >>> from iode import variables, SAMPLE_DATA_DIR
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> variables.sample.get_period_list()    #doctest: +ELLIPSIS
                [\'1960Y1\', \'1961Y1\', ..., \'2014Y1\', \'2015Y1\']
                >>> variables.sample.get_period_list(astype=float)    #doctest: +ELLIPSIS
                [1960.0, 1961.0, ..., 2014.0, 2015.0]
        '''
    @overload
    def get_period_list(cls) -> Any:
        '''get_period_list(self, astype: Union[type(Any), str] = str) -> List[Any]

                List of all periods of the sample.
                Periods are exported as string (default) or as float.

                Parameters
                ----------
                astype: type or str
                    Allowed returned type for periods are str and float.
                    Default to str.

                Returns
                -------
                list(str) or list(float)

                Examples
                --------
                >>> from iode import variables, SAMPLE_DATA_DIR
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> variables.sample.get_period_list()    #doctest: +ELLIPSIS
                [\'1960Y1\', \'1961Y1\', ..., \'2014Y1\', \'2015Y1\']
                >>> variables.sample.get_period_list(astype=float)    #doctest: +ELLIPSIS
                [1960.0, 1961.0, ..., 2014.0, 2015.0]
        '''
    @overload
    def get_period_list(cls, astype=...) -> Any:
        '''get_period_list(self, astype: Union[type(Any), str] = str) -> List[Any]

                List of all periods of the sample.
                Periods are exported as string (default) or as float.

                Parameters
                ----------
                astype: type or str
                    Allowed returned type for periods are str and float.
                    Default to str.

                Returns
                -------
                list(str) or list(float)

                Examples
                --------
                >>> from iode import variables, SAMPLE_DATA_DIR
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> variables.sample.get_period_list()    #doctest: +ELLIPSIS
                [\'1960Y1\', \'1961Y1\', ..., \'2014Y1\', \'2015Y1\']
                >>> variables.sample.get_period_list(astype=float)    #doctest: +ELLIPSIS
                [1960.0, 1961.0, ..., 2014.0, 2015.0]
        '''
    def index(self, period: str | Period) -> int:
        '''index(self, period: Union[str, Period]) -> int

                Position of the \'period\' in the sample.

                Returns
                -------
                int

                Raises
                ------
                IndexError
                    If the \'period\' has not been found in the sample.

                Examples
                --------
                >>> from iode import variables, SAMPLE_DATA_DIR
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> variables.sample.index("1982Y1")
                22
                >>> variables.sample.index("2020Y1")
                Traceback (most recent call last):
                ... 
                IndexError: The period \'2020Y1\' is not in the sample \'1960Y1:2015Y1\'
        '''
    def intersection(self, other_sample: Sample) -> Sample:
        '''intersection(self, other_sample: Sample) -> Sample

                Compute the intersection between two samples.

                Returns
                -------
                Sample

                Examples
                --------
                >>> from iode import variables, SAMPLE_DATA_DIR
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> variables.sample
                Sample("1960Y1:2015Y1")
                >>> variables_2 = variables.copy()
                >>> variables_2.sample = "2000Y1:2040Y1"
                >>> variables_2.sample
                Sample("2000Y1:2040Y1")
                >>> sample_intersec = variables.sample.intersection(variables_2.sample)
                >>> sample_intersec
                Sample("2000Y1:2015Y1")
        '''
    def __eq__(self, other: object) -> bool:
        """Return self==value."""
    def __ge__(self, other: object) -> bool:
        """Return self>=value."""
    def __gt__(self, other: object) -> bool:
        """Return self>value."""
    def __le__(self, other: object) -> bool:
        """Return self<=value."""
    def __len__(self) -> int:
        """Return len(self)."""
    def __lt__(self, other: object) -> bool:
        """Return self<value."""
    def __ne__(self, other: object) -> bool:
        """Return self!=value."""
    def __reduce__(self):
        """__reduce_cython__(self)"""

class Scalar:
    __pyx_vtable__: ClassVar[PyCapsule] = ...
    relax: Incomplete
    std: Incomplete
    value: Incomplete
    def __init__(self, *args, **kwargs) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""
    def __copy__(self) -> Scalar:
        '''__copy__(self) -> Scalar

                Return a copy of the current Scalar.

                Examples
                --------
                >>> import copy
                >>> from iode import Scalar, scalars, SAMPLE_DATA_DIR
                >>> scalars.load(f"{SAMPLE_DATA_DIR}/fun.scl")
                >>> scalars["acaf1"]
                Scalar(0.0157684, 1, 0.00136871)
                >>> copied_scl = copy.copy(scalars["acaf1"])
                >>> copied_scl
                Scalar(0.0157684, 1, 0.00136871)
        '''
    def __eq__(self, other: object) -> bool:
        """Return self==value."""
    def __ge__(self, other: object) -> bool:
        """Return self>=value."""
    def __gt__(self, other: object) -> bool:
        """Return self>value."""
    def __hash__(self) -> int:
        """Return hash(self)."""
    def __le__(self, other: object) -> bool:
        """Return self<=value."""
    def __lt__(self, other: object) -> bool:
        """Return self<value."""
    def __ne__(self, other: object) -> bool:
        """Return self!=value."""
    def __reduce__(self):
        """__reduce_cython__(self)"""

class Scalars(_AbstractDatabase):
    __pyx_vtable__: ClassVar[PyCapsule] = ...
    df: DataFrame
    i: PositionalIndexer
    def __init__(self, *args, **kwargs) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""
    @overload
    def from_frame(self, df: DataFrame) -> Any:
        '''from_frame(self, df: DataFrame)

                Copy the pandas DataFrame `df` into the IODE Scalars database.
                The scalar names to copy are deduced from the index of the DataFrame.

                Parameters
                ----------
                df: DataFrame
                    pandas DataFrame containing the scalars to copy into the Scalars database.
                    The passed DataFrame should have the following structure:
                        - index: names of the scalars
                        - a column labeled \'value\': values of the scalars.
                        - a column labeled \'relax\' (optional): relaxation parameter values of the scalars. 
                        - a column labeled \'std\' (optional): standard deviation values resulting from estimation.
                    If the DataFrame does not contain a \'relax\' column, relax values will be all set to 1.
                    If the DataFrame does not contain a \'std\' column, std values will be all set to :math:`NA`.

                Notes
                -----
                The index of the passed DataFrame is sorted in alphabetical order before 
                copying to IODE Scalars database.

                See Also
                --------
                Scalars.from_series
                Scalars.to_frame

                Examples
                --------
                >>> from iode import scalars
                >>> import pandas as pd
                >>> import numpy as np
                >>> scalars.clear()
                >>> len(scalars)
                0

                >>> # create the pandas DataFrame
                >>> data = {"alpha_0": [0.9, 1., np.nan],
                ...         "alpha_1": [0.1, 0.8, np.nan],
                ...         "alpha_": [5.5, 0., np.nan],
                ...         "beta_0": [0.9, 1., 0.05246],
                ...         "beta_1": [0.01, 0.8, 0.25687],
                ...         "beta_": [3.6, 0., 0.]}
                >>> df = pd.DataFrame.from_dict(data, orient=\'index\', dtype="float64", 
                ...                             columns=["value", "relax", "std"])
                >>> # display the pandas series
                >>> df          # doctest: +NORMALIZE_WHITESPACE
                        value  relax      std
                alpha_0   0.90    1.0      NaN
                alpha_1   0.10    0.8      NaN
                alpha_    5.50    0.0      NaN
                beta_0    0.90    1.0  0.05246
                beta_1    0.01    0.8  0.25687
                beta_     3.60    0.0  0.00000

                >>> # load into the IODE Scalars database
                >>> scalars.from_frame(df)
                >>> len(scalars)
                6
        
                >>> scalars.names
                [\'alpha_\', \'alpha_0\', \'alpha_1\', \'beta_\', \'beta_0\', \'beta_1\']
                >>> df.loc["alpha_1"]
                value    0.1
                relax    0.8
                std      NaN
                Name: alpha_1, dtype: float64
                >>> scalars["alpha_1"]
                Scalar(0.1, 0.8, na)
                >>> df.loc["beta_1"]
                value    0.01000
                relax    0.80000
                std      0.25687
                Name: beta_1, dtype: float64
                >>> scalars["beta_1"]
                Scalar(0.01, 0.8, 0.25687)
        '''
    @overload
    def from_frame(cls, df) -> Any:
        '''from_frame(self, df: DataFrame)

                Copy the pandas DataFrame `df` into the IODE Scalars database.
                The scalar names to copy are deduced from the index of the DataFrame.

                Parameters
                ----------
                df: DataFrame
                    pandas DataFrame containing the scalars to copy into the Scalars database.
                    The passed DataFrame should have the following structure:
                        - index: names of the scalars
                        - a column labeled \'value\': values of the scalars.
                        - a column labeled \'relax\' (optional): relaxation parameter values of the scalars. 
                        - a column labeled \'std\' (optional): standard deviation values resulting from estimation.
                    If the DataFrame does not contain a \'relax\' column, relax values will be all set to 1.
                    If the DataFrame does not contain a \'std\' column, std values will be all set to :math:`NA`.

                Notes
                -----
                The index of the passed DataFrame is sorted in alphabetical order before 
                copying to IODE Scalars database.

                See Also
                --------
                Scalars.from_series
                Scalars.to_frame

                Examples
                --------
                >>> from iode import scalars
                >>> import pandas as pd
                >>> import numpy as np
                >>> scalars.clear()
                >>> len(scalars)
                0

                >>> # create the pandas DataFrame
                >>> data = {"alpha_0": [0.9, 1., np.nan],
                ...         "alpha_1": [0.1, 0.8, np.nan],
                ...         "alpha_": [5.5, 0., np.nan],
                ...         "beta_0": [0.9, 1., 0.05246],
                ...         "beta_1": [0.01, 0.8, 0.25687],
                ...         "beta_": [3.6, 0., 0.]}
                >>> df = pd.DataFrame.from_dict(data, orient=\'index\', dtype="float64", 
                ...                             columns=["value", "relax", "std"])
                >>> # display the pandas series
                >>> df          # doctest: +NORMALIZE_WHITESPACE
                        value  relax      std
                alpha_0   0.90    1.0      NaN
                alpha_1   0.10    0.8      NaN
                alpha_    5.50    0.0      NaN
                beta_0    0.90    1.0  0.05246
                beta_1    0.01    0.8  0.25687
                beta_     3.60    0.0  0.00000

                >>> # load into the IODE Scalars database
                >>> scalars.from_frame(df)
                >>> len(scalars)
                6
        
                >>> scalars.names
                [\'alpha_\', \'alpha_0\', \'alpha_1\', \'beta_\', \'beta_0\', \'beta_1\']
                >>> df.loc["alpha_1"]
                value    0.1
                relax    0.8
                std      NaN
                Name: alpha_1, dtype: float64
                >>> scalars["alpha_1"]
                Scalar(0.1, 0.8, na)
                >>> df.loc["beta_1"]
                value    0.01000
                relax    0.80000
                std      0.25687
                Name: beta_1, dtype: float64
                >>> scalars["beta_1"]
                Scalar(0.01, 0.8, 0.25687)
        '''
    @overload
    def from_series(self, s: Series) -> Any:
        '''from_series(self, s: Series)

                Copy the pandas Series `s` into the IODE Scalars database.
                The scalar names to copy are deduced from the index of the Series.

                Parameters
                ----------
                s: Series
                    pandas Series containing the scalars values to copy into the Scalars database.
                    The \'relax\' value of all scalars will be set to 1.

                Notes
                -----
                The index of the passed Series is sorted in alphabetical order before 
                copying to IODE Scalars database.

                See Also
                --------
                Scalars.from_frame
                Scalars.to_frame

                Examples
                --------
                >>> from iode import scalars
                >>> import pandas as pd
                >>> scalars.clear()
                >>> len(scalars)
                0

                >>> # create the pandas Series
                >>> data = {"alpha_0": 0.9, "alpha_1": 0.1, "alpha_": 5.5,
                ...         "beta_0": 0.9, "beta_1": 0.01, "beta_": 3.6}
                >>> s = pd.Series(data, dtype="float64")
                >>> # display the pandas series
                >>> s          # doctest: +NORMALIZE_WHITESPACE
                alpha_0    0.90
                alpha_1    0.10
                alpha_     5.50
                beta_0     0.90
                beta_1     0.01
                beta_      3.60
                dtype: float64

                >>> # load into the IODE Scalars database
                >>> scalars.from_series(s)
                >>> len(scalars)
                6
        
                >>> scalars.names
                [\'alpha_\', \'alpha_0\', \'alpha_1\', \'beta_\', \'beta_0\', \'beta_1\']
                >>> scalars["alpha_1"]
                Scalar(0.1, 1, na)
                >>> scalars["beta_"]
                Scalar(3.6, 1, na)
        '''
    @overload
    def from_series(cls, s) -> Any:
        '''from_series(self, s: Series)

                Copy the pandas Series `s` into the IODE Scalars database.
                The scalar names to copy are deduced from the index of the Series.

                Parameters
                ----------
                s: Series
                    pandas Series containing the scalars values to copy into the Scalars database.
                    The \'relax\' value of all scalars will be set to 1.

                Notes
                -----
                The index of the passed Series is sorted in alphabetical order before 
                copying to IODE Scalars database.

                See Also
                --------
                Scalars.from_frame
                Scalars.to_frame

                Examples
                --------
                >>> from iode import scalars
                >>> import pandas as pd
                >>> scalars.clear()
                >>> len(scalars)
                0

                >>> # create the pandas Series
                >>> data = {"alpha_0": 0.9, "alpha_1": 0.1, "alpha_": 5.5,
                ...         "beta_0": 0.9, "beta_1": 0.01, "beta_": 3.6}
                >>> s = pd.Series(data, dtype="float64")
                >>> # display the pandas series
                >>> s          # doctest: +NORMALIZE_WHITESPACE
                alpha_0    0.90
                alpha_1    0.10
                alpha_     5.50
                beta_0     0.90
                beta_1     0.01
                beta_      3.60
                dtype: float64

                >>> # load into the IODE Scalars database
                >>> scalars.from_series(s)
                >>> len(scalars)
                6
        
                >>> scalars.names
                [\'alpha_\', \'alpha_0\', \'alpha_1\', \'beta_\', \'beta_0\', \'beta_1\']
                >>> scalars["alpha_1"]
                Scalar(0.1, 1, na)
                >>> scalars["beta_"]
                Scalar(3.6, 1, na)
        '''
    @overload
    def to_frame(self) -> DataFrame:
        '''to_frame(self) -> DataFrame

                Create a pandas DataFrame from the current Scalars database.
                The index of the returned DataFrame is build from the Scalars names.

                See Also
                --------
                Scalars.from_series
                Scalars.from_frame

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, scalars
                >>> import pandas as pd
                >>> scalars.load(f"{SAMPLE_DATA_DIR}/fun.scl")
                >>> len(scalars)
                161

                >>> # Export the IODE Scalars database as a pandas DataFrame
                >>> df = scalars.to_frame()
                >>> len(df)
                161

                >>> df.index.to_list()              # doctest: +ELLIPSIS
                [\'acaf1\', \'acaf2\', \'acaf3\', ..., \'zkf1\', \'zkf2\', \'zkf3\']
                >>> scalars["acaf1"]                # doctest: +NORMALIZE_WHITESPACE
                Scalar(0.0157684, 1, 0.00136871)
                >>> df.loc["acaf1"]                 # doctest: +NORMALIZE_WHITESPACE
                value    0.015768
                relax    1.000000
                std      0.001369
                Name: acaf1, dtype: float64
                >>> scalars["qc0_"]                 # doctest: +NORMALIZE_WHITESPACE
                Scalar(0.178165, 1, 0.102838)
                >>> df.loc["qc0_"]                  # doctest: +NORMALIZE_WHITESPACE
                value    0.178165
                relax    1.000000
                std      0.102838
                Name: qc0_, dtype: float64

                >>> # Export a subset of the IODE Scalars database as a pandas DataFrame
                >>> df = scalars["a*;*_"].to_frame()
                >>> len(df)
                19

                >>> df.index.to_list()              # doctest: +ELLIPSIS
                [\'acaf1\', \'acaf2\', \'acaf3\', ..., \'vs1_\', \'vs2_\', \'vs3_\']
                >>> scalars["acaf1"]                # doctest: +NORMALIZE_WHITESPACE
                Scalar(0.0157684, 1, 0.00136871)
                >>> df.loc["acaf1"]                 # doctest: +NORMALIZE_WHITESPACE
                value    0.015768
                relax    1.000000
                std      0.001369
                Name: acaf1, dtype: float64
                >>> scalars["qc0_"]                 # doctest: +NORMALIZE_WHITESPACE
                Scalar(0.178165, 1, 0.102838)
                >>> df.loc["qc0_"]                  # doctest: +NORMALIZE_WHITESPACE
                value    0.178165
                relax    1.000000
                std      0.102838
                Name: qc0_, dtype: float64
        '''
    @overload
    def to_frame(cls) -> Any:
        '''to_frame(self) -> DataFrame

                Create a pandas DataFrame from the current Scalars database.
                The index of the returned DataFrame is build from the Scalars names.

                See Also
                --------
                Scalars.from_series
                Scalars.from_frame

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, scalars
                >>> import pandas as pd
                >>> scalars.load(f"{SAMPLE_DATA_DIR}/fun.scl")
                >>> len(scalars)
                161

                >>> # Export the IODE Scalars database as a pandas DataFrame
                >>> df = scalars.to_frame()
                >>> len(df)
                161

                >>> df.index.to_list()              # doctest: +ELLIPSIS
                [\'acaf1\', \'acaf2\', \'acaf3\', ..., \'zkf1\', \'zkf2\', \'zkf3\']
                >>> scalars["acaf1"]                # doctest: +NORMALIZE_WHITESPACE
                Scalar(0.0157684, 1, 0.00136871)
                >>> df.loc["acaf1"]                 # doctest: +NORMALIZE_WHITESPACE
                value    0.015768
                relax    1.000000
                std      0.001369
                Name: acaf1, dtype: float64
                >>> scalars["qc0_"]                 # doctest: +NORMALIZE_WHITESPACE
                Scalar(0.178165, 1, 0.102838)
                >>> df.loc["qc0_"]                  # doctest: +NORMALIZE_WHITESPACE
                value    0.178165
                relax    1.000000
                std      0.102838
                Name: qc0_, dtype: float64

                >>> # Export a subset of the IODE Scalars database as a pandas DataFrame
                >>> df = scalars["a*;*_"].to_frame()
                >>> len(df)
                19

                >>> df.index.to_list()              # doctest: +ELLIPSIS
                [\'acaf1\', \'acaf2\', \'acaf3\', ..., \'vs1_\', \'vs2_\', \'vs3_\']
                >>> scalars["acaf1"]                # doctest: +NORMALIZE_WHITESPACE
                Scalar(0.0157684, 1, 0.00136871)
                >>> df.loc["acaf1"]                 # doctest: +NORMALIZE_WHITESPACE
                value    0.015768
                relax    1.000000
                std      0.001369
                Name: acaf1, dtype: float64
                >>> scalars["qc0_"]                 # doctest: +NORMALIZE_WHITESPACE
                Scalar(0.178165, 1, 0.102838)
                >>> df.loc["qc0_"]                  # doctest: +NORMALIZE_WHITESPACE
                value    0.178165
                relax    1.000000
                std      0.102838
                Name: qc0_, dtype: float64
        '''
    @overload
    def to_frame(cls) -> Any:
        '''to_frame(self) -> DataFrame

                Create a pandas DataFrame from the current Scalars database.
                The index of the returned DataFrame is build from the Scalars names.

                See Also
                --------
                Scalars.from_series
                Scalars.from_frame

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, scalars
                >>> import pandas as pd
                >>> scalars.load(f"{SAMPLE_DATA_DIR}/fun.scl")
                >>> len(scalars)
                161

                >>> # Export the IODE Scalars database as a pandas DataFrame
                >>> df = scalars.to_frame()
                >>> len(df)
                161

                >>> df.index.to_list()              # doctest: +ELLIPSIS
                [\'acaf1\', \'acaf2\', \'acaf3\', ..., \'zkf1\', \'zkf2\', \'zkf3\']
                >>> scalars["acaf1"]                # doctest: +NORMALIZE_WHITESPACE
                Scalar(0.0157684, 1, 0.00136871)
                >>> df.loc["acaf1"]                 # doctest: +NORMALIZE_WHITESPACE
                value    0.015768
                relax    1.000000
                std      0.001369
                Name: acaf1, dtype: float64
                >>> scalars["qc0_"]                 # doctest: +NORMALIZE_WHITESPACE
                Scalar(0.178165, 1, 0.102838)
                >>> df.loc["qc0_"]                  # doctest: +NORMALIZE_WHITESPACE
                value    0.178165
                relax    1.000000
                std      0.102838
                Name: qc0_, dtype: float64

                >>> # Export a subset of the IODE Scalars database as a pandas DataFrame
                >>> df = scalars["a*;*_"].to_frame()
                >>> len(df)
                19

                >>> df.index.to_list()              # doctest: +ELLIPSIS
                [\'acaf1\', \'acaf2\', \'acaf3\', ..., \'vs1_\', \'vs2_\', \'vs3_\']
                >>> scalars["acaf1"]                # doctest: +NORMALIZE_WHITESPACE
                Scalar(0.0157684, 1, 0.00136871)
                >>> df.loc["acaf1"]                 # doctest: +NORMALIZE_WHITESPACE
                value    0.015768
                relax    1.000000
                std      0.001369
                Name: acaf1, dtype: float64
                >>> scalars["qc0_"]                 # doctest: +NORMALIZE_WHITESPACE
                Scalar(0.178165, 1, 0.102838)
                >>> df.loc["qc0_"]                  # doctest: +NORMALIZE_WHITESPACE
                value    0.178165
                relax    1.000000
                std      0.102838
                Name: qc0_, dtype: float64
        '''
    def __delitem__(self, other) -> None:
        '''
        Remove the (subset of) scalar(s) referenced by `key` from the Scalars database.

        Parameters
        ----------
        key: str or list(str)
            (list of) name(s) of the scalar(s) to be removed.
            The list of names can be given as a string pattern (e.g. "A*;*_").

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import scalars
        >>> scalars.load(f"{SAMPLE_DATA_DIR}/fun.scl")

        >>> # a) delete one scalar
        >>> scalars.get_names("a*")
        [\'acaf1\', \'acaf2\', \'acaf3\', \'acaf4\']
        >>> del scalars["acaf4"]
        >>> scalars.get_names("a*")
        [\'acaf1\', \'acaf2\', \'acaf3\']

        >>> # b) delete several scalars at once using a pattern
        >>> del scalars["a*"]
        >>> scalars.get_names("a*")
        []

        >>> # c) delete several scalars at once using a list of names
        >>> scalars.get_names("g*")
        [\'gamma0\', \'gamma1\', \'gamma2\', \'gamma3\', \'gamma4\', \'gamma5\', \'gamma_\']
        >>> del scalars[["gamma3", "gamma_"]]
        >>> scalars.get_names("g*")
        [\'gamma0\', \'gamma1\', \'gamma2\', \'gamma4\', \'gamma5\']

        >>> # delete one scalar from a subset of the global database
        >>> scalars_subset = scalars["z*"]
        >>> scalars_subset.names
        [\'zkf1\', \'zkf2\', \'zkf3\']
        >>> del scalars_subset["zkf2"]
        >>> scalars_subset.names
        [\'zkf1\', \'zkf3\']
        >>> # NOTE: the scalar has also been deleted from the global database
        >>> "zkf2" in scalars
        False
        >>> scalars.get_names("z*")
        [\'zkf1\', \'zkf3\']
        '''
    def __getitem__(self, index):
        '''
        Return the (subset of) scalar(s) referenced by `key`.

        The `key` can represent a single object name (e.g. "ACAF") or a list of object names ("ACAF;ACAG;AOUC") 
        or a pattern (e.g. "A*") or a list of sub-patterns (e.g. "A*;*_").
        
        If the `key` represents a list of object names or of sub-patterns, each name or sub-pattern is separated 
        by a `separator` character which is either a whitespace ` `, or a comma `,`, or a semi-colon `;`, or a 
        tabulation `\\t`, or a newline `\\n`.

        A (sub-)`pattern` is a list of characters representing a group of object names. 
        It includes some special characters which have a special meaning:
        
            - `*` : any character sequence, even empty
            - `?` : any character (one and only one)
            - `@` : any alphanumerical char [A-Za-z0-9]
            - `&` : any non alphanumerical char
            - `|` : any alphanumeric character or none at the beginning and end of a string 
            - `!` : any non-alphanumeric character or none at the beginning and end of a string 
            - `\\` : escape the next character

        Note that the `key` can contain references to IODE lists which are prefixed with the symbol `$`.

        Parameters
        ----------
        key: str or list(str)
            (the list of) name(s) of the scalar(s) to get.
            The list of objects to get can be specified by a pattern or by a list of sub-patterns (e.g. "A*;*_").

        Returns
        -------
        Single scalar or a subset of the database.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import scalars
        >>> scalars.load(f"{SAMPLE_DATA_DIR}/fun.scl")

        >>> # a) get one scalar
        >>> acaf1 = scalars["acaf1"]
        >>> acaf1
        Scalar(0.0157684, 1, 0.00136871)
        >>> acaf1.value
        0.01576840691268444
        >>> acaf1.relax
        1.0
        >>> acaf1.std
        0.0013687137980014086

        >>> # b) get a subset of the Scalars database using a pattern
        >>> scalars_subset = scalars["a*"]
        >>> scalars_subset.names
        [\'acaf1\', \'acaf2\', \'acaf3\', \'acaf4\']

        >>> # c) get a subset of the Scalars database using a list of names
        >>> scalars_subset = scalars[["acaf1", "acaf4", "dpuh_1", "dpuh_2"]]
        >>> scalars_subset.names
        [\'acaf1\', \'acaf4\', \'dpuh_1\', \'dpuh_2\']
        '''
    def __hash__(self) -> int:
        '''
        Return a hash value for the current Scalars database.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR, scalars
        >>> scalars.load(f"{SAMPLE_DATA_DIR}/fun.scl")
        >>> len(scalars)
        161
        >>> original_hash = hash(scalars)
        
        >>> # rename 1 scalar
        >>> scalars.rename("acaf1", "acaf1_")
        >>> original_hash == hash(scalars)
        False
        >>> scalars.rename("acaf1_", "acaf1")  # revert the change
        >>> original_hash == hash(scalars)
        True

        >>> # modify one scalar
        >>> original_value = scalars["acaf1"].value
        >>> scalars["acaf1"] = 0.02
        >>> original_hash == hash(scalars)
        False
        >>> scalars["acaf1"].value = original_value  # revert the change
        >>> original_hash == hash(scalars)
        True

        >>> # delete a scalar
        >>> original_scalar = scalars["acaf1"]
        >>> del scalars["acaf1"]
        >>> original_hash == hash(scalars)
        False
        >>> scalars["acaf1"] = original_scalar
        >>> original_hash == hash(scalars)
        True

        >>> # add a scalar
        >>> scalars["new_scalar"] = (0.9, 1.0)
        >>> original_hash == hash(scalars)
        False
        >>> del scalars["new_scalar"]
        >>> original_hash == hash(scalars)
        True
        '''
    def __reduce__(self):
        """__reduce_cython__(self)"""
    def __setitem__(self, index, object) -> None:
        '''
        Update/add a (subset of) scalar(s) referenced by `key` from/to the Scalars database.

        The `key` can represent a single object name (e.g. "ACAF") or a list of object names ("ACAF;ACAG;AOUC") 
        or a pattern (e.g. "A*") or a list of sub-patterns (e.g. "A*;*_").
        
        If the `key` represents a list of object names or of sub-patterns, each name or sub-pattern is separated 
        by a `separator` character which is either a whitespace ` `, or a comma `,`, or a semi-colon `;`, or a 
        tabulation `\\t`, or a newline `\\n`.

        A (sub-)`pattern` is a list of characters representing a group of object names. 
        It includes some special characters which have a special meaning:
        
            - `*` : any character sequence, even empty
            - `?` : any character (one and only one)
            - `@` : any alphanumerical char [A-Za-z0-9]
            - `&` : any non alphanumerical char
            - `|` : any alphanumeric character or none at the beginning and end of a string 
            - `!` : any non-alphanumeric character or none at the beginning and end of a string 
            - `\\` : escape the next character

        Note that the `key` can contain references to IODE lists which are prefixed with the symbol `$`.

        Parameters
        ----------
        key: str or list(str)
            (the list of) name(s) of the scalar(s) to update/add.
            The list of objects to update/add can be specified by a pattern or by a list of sub-patterns 
            (e.g. "A*;*_").
        value: float or tuple(float, float) Scalar or dict(str, ...) or list of any of those
            If float, then it is interpreted as the value of the scalar.
            If tuple of two float, then it is interpreted as the value and relax of the scalar.
            If dictionary, then it is interpreted as the values of the attributes value and relax of the scalar to update or add.
            The standard deviation is computed during the estimation process and cannot be modified manually.
            If Scalar, then it is used to update an existing scalar or to create a new scalar if it does not exist yet.

        See Also
        --------
        iode.Scalar

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import scalars
        >>> scalars.load(f"{SAMPLE_DATA_DIR}/fun.scl")
        
        >>> # a) -------- add one scalar -------- 
        >>> # 1. default relax to 1.0
        >>> scalars["a0"] = 0.1
        >>> scalars["a0"]
        Scalar(0.1, 1, na)
        >>> # 2. value + relax
        >>> scalars["a1"] = 0.1, 0.9
        >>> scalars["a1"]
        Scalar(0.1, 0.9, na)

        >>> # b) -------- update one scalar --------
        >>> scalars["acaf1"]
        Scalar(0.0157684, 1, 0.00136871)
        >>> # only update the value
        >>> scalars["acaf1"] = 0.8
        >>> scalars["acaf1"]
        Scalar(0.8, 1, 0.00136871)
        >>> # update value and relax (tuple)
        >>> scalars["acaf2"] = 0.8, 0.9
        >>> scalars["acaf2"]
        Scalar(0.8, 0.9, na)
        >>> # update value and relax (list)
        >>> scalars["acaf2"] = (0.7, 0.8)
        >>> scalars["acaf2"]
        Scalar(0.7, 0.8, na)
        >>> # update value and relax (dictionary)
        >>> scalars["acaf3"] = {"relax": 0.9, "value": 0.8}
        >>> scalars["acaf3"]
        Scalar(0.8, 0.9, 0.87301)
        >>> # update value and/or relax (Scalar object)
        >>> # NOTE: the standard deviation (std) cannot be changed manually
        >>> scalars["acaf4"]
        Scalar(-0.00850518, 1, 0.0020833)
        >>> scalars["acaf4"].value = 0.8
        >>> scalars["acaf4"].relax = 0.9
        >>> scalars["acaf4"]
        Scalar(0.8, 0.9, 0.0020833)

        >>> # c) working on a subset
        >>> # 1) get subset
        >>> scalars_subset = scalars["a*"]
        >>> scalars_subset.names
        [\'a0\', \'a1\', \'acaf1\', \'acaf2\', \'acaf3\', \'acaf4\']
        >>> # 2) add a scalar to the subset 
        >>> scalars_subset["acaf0"] = 1.0, 1.0
        >>> scalars_subset["acaf0"]
        Scalar(1, 1, na)
        >>> # --> new scalar also appears in the global workspace
        >>> "acaf0" in scalars
        True
        >>> scalars["acaf0"]
        Scalar(1, 1, na)
        >>> # 3) update a scalar in the subset
        >>> scalars_subset["acaf0"] = 0.1
        >>> scalars_subset["acaf0"]
        Scalar(0.1, 1, na)
        >>> # --> scalar is also updated in the global workspace
        >>> scalars["acaf0"]
        Scalar(0.1, 1, na)
        '''

class Simulation:
    convergence_threshold: float
    debug: bool
    debug_newton: bool
    initialization_method: str
    max_nb_iterations: int
    max_nb_iterations_newton: int
    nb_passes: int
    relax: float
    sort_algorithm: str
    def __init__(self) -> Any:
        """Create and return a new object.  See help(type) for accurate signature."""
    def model_calculate_SCC(self, nb_iterations: int, pre_name: str = ..., inter_name: str = ..., post_name: str = ..., list_eqs: str | list[str] = ...) -> Any:
        '''model_calculate_SCC(self, nb_iterations: int, pre_name: str = u\'_PRE\', inter_name: str = u\'_INTER\', post_name: str = u\'_POST\', list_eqs: Union[str, List[str]] = None)

                Decompose the model into Strongly Connex Components (SCC) and reorder it. 
        
                Three lists are created: pre-recursive, interdependent and post-recursive equations.
                When reordering the model, the number of triangulation iterations for the interdependent 
                block must be specified. This value only affects the list of interdependent equations.
                These 3 lists only need to be constructed once for a given model.

                Parameters
                ----------
                nb_iterations : int
                    Number of triangulation iterations to be performed. 
                pre_name : str, optional
                    Name of the list representing the pre-recursive equations.
                    Default to "_PRE".
                inter_name : str, optional
                    Name of the list representing the interdependent equations. 
                    Default to "_INTER".
                post_name : str, optional
                    Name of the list representing the post-recursive equations.
                    Default to "_POST".
                list_eqs : str or list(str), optional
                    List of equations representing the model.
                    Default to empty (all equations).

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, equations, lists, scalars, variables 
                >>> from iode import Simulation
                >>> equations.load(f"{SAMPLE_DATA_DIR}/fun.eqs")
                >>> lists.load(f"{SAMPLE_DATA_DIR}/fun.lst")
                >>> scalars.load(f"{SAMPLE_DATA_DIR}/fun.scl")
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")

                >>> simu = Simulation()
                >>> simu.model_calculate_SCC(10);

                >>> lists["_PRE"]           # doctest: +ELLIPSIS
                [\'BRUGP\', \'DTH1C\', \'EX\', \'ITCEE\', ..., \'DTH1\', \'PME\', \'PMS\', \'PMT\']
                >>> len(lists["_PRE"])
                31

                >>> lists["_INTER"]         # doctest: +ELLIPSIS
                [\'PMAB\', \'PXAB\', \'ULCP\', \'SSH3P\', ..., \'WCF_\', \'ITEP\', \'EXC\', \'ITT\']
                >>> len(lists["_INTER"])
                204

                >>> lists["_POST"]          # doctest: +ELLIPSIS
                [\'IFU\', \'SSHFF\', \'PBBP\', \'OCUF\', ..., \'FLGR\', \'FLF\', \'DPUU\', \'BENEF\']
                >>> len(lists["_POST"])
                39
        '''
    def model_compile(self, list_eqs: str | list[str] = ...) -> Any:
        """model_compile(self, list_eqs: Union[str, List[str]] = None)

                Recompiles a list of equations, or all equations if no list is specified. 
                It is only useful if (the list of) equations contain macros in their LEC form (like $list).

                The set of equations stored in memory are made up of several elements: the LEC form, 
                the estimation inputs and the **compiled form** of the equation. If the LEC form 
                contains macros, the **compiled form** may no longer correspond to the LEC form.
                This can happen when a new list file is loaded into memory, when a list is manually 
                modified or copied from disk, or when an equation file is loaded into memory.

                Parameters
                ----------
                list_eqs : str or list(str), optional
                    List of equations to recompile. 
                    Default to empty (recompile all equations).
        """
    def model_exchange(self, list_exo: str | list[str] = ...) -> Any:
        '''model_exchange(self, list_exo: Union[str, List[str]] = None)

                Set a list of endogenous-exogenous pairs for ``goal seeking``. 
                For each pair, the status of the variables is exchanged: endogenous becomes exogenous and vice versa. 
                This enables the model to be run through known endogenous values and to deduce the necessary values 
                for the associated exogenous variables. The exogenous variables take on the value calculated over the entire 
                workspace period.

                Parameters
                ----------
                list_exo : str or list(str)
                    List of endogenous-exogenous pairs. 
                    Each pair must be written as `endogenous_name-exogenous_name`.
                    If the passed list is empty, the internal list of endogenous-exogenous pairs is reset.
                    Default to None (empty).

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, equations, identities, scalars, variables 
                >>> from iode import Simulation
                >>> equations.load(f"{SAMPLE_DATA_DIR}/fun.eqs")
                >>> identities.load(f"{SAMPLE_DATA_DIR}/fun.idt")
                >>> scalars.load(f"{SAMPLE_DATA_DIR}/fun.scl")
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")

                >>> equations["UY"].lec
                \'UY := NATY-NDOMY-NIY-NGY-(EFXY-EFMY)-NFY\'
                >>> identities["XNATY"]
                Identity(\'grt NATY\')
                >>> # endogenous variable
                >>> variables["XNATY", "2000Y1:2015Y1"]     # doctest: +ELLIPSIS
                [0.22, 0.699999988079071, 0.4, 0.7, ..., -0.20000000298023224, -0.20000000298023224]
                >>> # reset exogeneous variable
                >>> variables["UY", "2000Y1:2015Y1"] = 0.0
                >>> variables["UY", "2000Y1:2015Y1"]    # doctest: +ELLIPSIS
                [0.0, 0.0, ..., 0.0, 0.0]

                >>> simu = Simulation()
                >>> simu.model_simulate("2000Y1", "2015Y1")
                >>> # endogenous variable (unchanged)
                >>> variables["XNATY", "2000Y1:2015Y1"]     # doctest: +ELLIPSIS
                [0.22, 0.699999988079071, 0.4, 0.7, ..., -0.20000000298023224, -0.20000000298023224]
                >>> # exogeneous variable
                >>> variables["UY", "2000Y1:2015Y1"]         # doctest: +ELLIPSIS
                [624.1781881798956, 645.0542977503601, 661.6074127102179, ..., 525.1258272992568]

                >>> # exchange UY and XNATY
                >>> simu.model_exchange("UY-XNATY")
                >>> # update endogenous variable (now UY)
                >>> variables["UY", "2000Y1:2002Y1"] = [630.0, 650.0, 670.0]
                >>> # rerun simulation
                >>> simu.model_simulate("2000Y1", "2015Y1")
                >>> # exogeneous variable (now XNATY)
                >>> variables["XNATY", "2000Y1:2015Y1"]     # doctest: +ELLIPSIS
                [0.3508274950028535, 0.6702440874066322, 0.5092369210188463, ..., -0.19926923081414763]
        '''
    def model_simulate(self, from_period: str | Period, to_period: str | Period, list_eqs: str | list[str] = ...) -> Any:
        '''model_simulate(self, from_period: Union[str, Period], to_period: Union[str, Period], list_eqs: Union[str, List[str]] = None)

                Run the simulation of a model for a given sample.

                To simulate a model, all equations of the model must have been loaded or created. In addition, 
                all the variables and scalars used in the model equations must have been loaded or created. 
                Values of exogenous variables and of scalar cannot be :math:`NA` (*Not Available*) over the 
                simulation sample.

                Parameters
                ----------
                from_period : str or Period
                    The starting period for the simulation.
                to_period : str or Period
                    The ending period for the simulation.
                list_eqs : str or list(str), optional
                    List of equations representing the model.
                    Default to empty (all equations).

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, equations, identities, lists, scalars, variables 
                >>> from iode import Simulation
                >>> equations.load(f"{SAMPLE_DATA_DIR}/fun.eqs")
                >>> identities.load(f"{SAMPLE_DATA_DIR}/fun.idt")
                >>> lists.load(f"{SAMPLE_DATA_DIR}/fun.lst")
                >>> scalars.load(f"{SAMPLE_DATA_DIR}/fun.scl")
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")

                >>> # exogeneous variable 
                >>> equations["UY"].lec
                \'UY := NATY-NDOMY-NIY-NGY-(EFXY-EFMY)-NFY\'
                >>> # reset values
                >>> variables["UY", "2000Y1:2015Y1"] = 0.0
                >>> variables["UY", "2000Y1:2015Y1"]    # doctest: +ELLIPSIS
                [0.0, 0.0, ..., 0.0, 0.0]

                >>> # endogenous variable
                >>> identities["XNATY"]
                Identity(\'grt NATY\')
                >>> variables["XNATY", "2000Y1:2015Y1"]     # doctest: +ELLIPSIS
                [0.22, 0.699999988079071, 0.4, 0.7, ..., -0.20000000298023224, -0.20000000298023224]

                >>> # simulate the model (no reordering)
                >>> simu = Simulation()
                >>> simu.model_simulate("2000Y1", "2015Y1")
                >>> variables["UY", "2000Y1:2015Y1"]    # doctest: +ELLIPSIS
                [624.1781881798956, 645.0542977503601, 661.6074127102179, ..., 525.1258272992568]
        '''
    def model_simulate_SCC(self, from_period: str | Period, to_period: str | Period, pre_name: str = ..., inter_name: str = ..., post_name: str = ...) -> Any:
        '''model_simulate_SCC(self, from_period: Union[str, Period], to_period: Union[str, Period], pre_name: str = u\'_PRE\', inter_name: str = u\'_INTER\', post_name: str = u\'_POST\')

                Simulates a model previously decomposed into Strongly Connex Components (SCC) and reordered.
                It is intended to be called after the method :meth:`Simulation.model_calculate_SCC`.

                To improve performance at simulation start-up, the simulation process has been divided into 2 stages. 
                The first only deals with model reordering (see method :meth:`Simulation.model_calculate_SCC`), 
                the second with simulation (see method :meth:`Simulation.model_simulate`).

                Parameters
                ----------
                from_period : str or Period
                    The starting period for the simulation.
                to_period : str or Period
                    The ending period for the simulation.
                pre_name : str, optional
                    Name of the list representing the pre-recursive equations.
                    Default to "_PRE".
                inter_name : str, optional
                    Name of the list representing the interdependent equations. 
                    Default to "_INTER".
                post_name : str, optional
                    Name of the list representing the post-recursive equations.
                    Default to "_POST".

                See Also
                --------
                Simulation.model_calculate_SCC
                Simulation.model_simulate

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, equations, identities, lists, scalars, variables 
                >>> from iode import Simulation
                >>> equations.load(f"{SAMPLE_DATA_DIR}/fun.eqs")
                >>> identities.load(f"{SAMPLE_DATA_DIR}/fun.idt")
                >>> lists.load(f"{SAMPLE_DATA_DIR}/fun.lst")
                >>> scalars.load(f"{SAMPLE_DATA_DIR}/fun.scl")
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> simu = Simulation()

                Step 1 - Compute the Strongly Connex Components (SCC) decomposition

                >>> simu.model_calculate_SCC(10);
                >>> lists["_PRE"]           # doctest: +ELLIPSIS
                [\'BRUGP\', \'DTH1C\', \'EX\', \'ITCEE\', ..., \'ZZF_\', \'DTH1\', \'PME\', \'PMS\', \'PMT\']
                >>> len(lists["_PRE"])
                31
                >>> lists["_INTER"]         # doctest: +ELLIPSIS
                [\'PMAB\', \'PXAB\', \'ULCP\', \'SSH3P\', ..., \'YSSG\', \'WCF_\', \'ITEP\', \'EXC\', \'ITT\']
                >>> len(lists["_INTER"])
                204
                >>> lists["_POST"]          # doctest: +ELLIPSIS
                [\'IFU\', \'SSHFF\', \'PBBP\', \'OCUF\', ..., \'GAP\', \'FLGR\', \'FLF\', \'DPUU\', \'BENEF\']
                >>> len(lists["_POST"])
                39

                Step 2 - Simulate the model based on the 3 lists _PRE, _INTER and _POST

                >>> # exogeneous variable 
                >>> equations["UY"].lec
                \'UY := NATY-NDOMY-NIY-NGY-(EFXY-EFMY)-NFY\'
                >>> # reset values
                >>> variables["UY", "2000Y1:2015Y1"] = 0.0
                >>> variables["UY", "2000Y1:2015Y1"]    # doctest: +ELLIPSIS
                [0.0, 0.0, ..., 0.0, 0.0]
                >>> # endogenous variable
                >>> identities["XNATY"]
                Identity(\'grt NATY\')
                >>> variables["XNATY", "2000Y1:2015Y1"]     # doctest: +ELLIPSIS
                [0.22, 0.699999988079071, 0.4, 0.7, ..., -0.20000000298023224, -0.20000000298023224]
        
                >>> simu.model_simulate_SCC("2000Y1", "2015Y1")

                >>> variables["UY", "2000Y1:2015Y1"]    # doctest: +ELLIPSIS
                [624.1822680369564, 645.0533712662168, 661.5778936891761, ..., 525.2940777398553]
        '''
    def __reduce__(self):
        """__reduce_cython__(self)"""

class SimulationInitialization(enum.IntEnum):
    __new__: ClassVar[Callable] = ...
    ASIS: ClassVar[SimulationInitialization] = ...
    EXTRA: ClassVar[SimulationInitialization] = ...
    EXTRA_A: ClassVar[SimulationInitialization] = ...
    EXTRA_NA: ClassVar[SimulationInitialization] = ...
    TM1: ClassVar[SimulationInitialization] = ...
    TM1_A: ClassVar[SimulationInitialization] = ...
    TM1_NA: ClassVar[SimulationInitialization] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[int]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    def __format__(self, *args, **kwargs) -> str: ...

class SimulationSort(enum.IntEnum):
    __new__: ClassVar[Callable] = ...
    BOTH: ClassVar[SimulationSort] = ...
    CONNEX: ClassVar[SimulationSort] = ...
    NONE: ClassVar[SimulationSort] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[int]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    def __format__(self, *args, **kwargs) -> str: ...

class Table:
    __pyx_vtable__: ClassVar[PyCapsule] = ...
    box: bool
    graph_alignment: str
    graph_axis: str
    gridx: str
    gridy: str
    language: str
    nb_columns: Incomplete
    nb_lines: Incomplete
    shadow: bool
    ymax: float
    ymin: float
    def __init__(self) -> Any:
        """Initialize self.  See help(type(self)) for accurate signature."""
    def compute(self, generalized_sample: str, extra_files: str | Path | list[str] | list[Path] = ..., nb_decimals: int = ...) -> ComputedTable:
        '''compute(self, generalized_sample: str, extra_files: Union[str, Path, List[str], List[Path]] = None, nb_decimals: int = 2) -> ComputedTable

                Compute the values corresponding to LEC expressions in cells.
        
                The values are calculated for given a *generalized sample*. 
                This sample contains the following information:

                  - the sampling of the periods to take into account 
                  - the operations to be performed on the periods
                  - the list of files involved in the computation of the table
                  - the operations to be performed between files
                  - the repetition factor

                The syntax of the *generalized sample* follows the rules described below.

                The syntax of a period: 
        
                  - a period is indicated as in LEC: \'yyPpp\' or \'yyyyPpp\' where yyyy indicates the year, 
                    P the periodicity and pp the sub-period (e.g. 1990Y1) 
                  - a period can be shifted n periods to the left or right using the operators <n and >n 
                  - when used with a zero argument, the shift oprerators have a special meaning: 
                      - <0 means "first period of the year" 
                      - >0 means "last period of the year" 
                  - the special periods \'BOS\', \'EOS\' and \'NOW\' can be used to represent the beginning 
                    or end of the current sample or the current period (PC clock)
                  - the special periods \'BOS1\', \'EOS1\' and \'NOW1\' are equivalent to the previous ones, 
                    except that they are moved to the first sub-period of the year of \'BOS\', \'EOS\' and 
                    \'NOW\' respectively (if NOW = 2012M5, NOW1 = 2012M1)
                  - each period is separated from the next by a semicolon 
                  - a period or group of periods can be repeated: simply place the colon character (:) after 
                    the definition of the column or group of columns, followed by the desired number of repetitions. 
                    Repetitions are made with an increment of one period, unless followed by an asterisk and a value. 
                    This value is then the repeat increment. It can be negative, in which case the periods are 
                    presented in decreasing order
                  - the repeat, increment and shift can be the words PER (or P) or SUB (or S), which respectively indicate 
                    the number of periods in a year of the current sample and the current sub-period
                  - the file definition is optional and is enclosed in square brackets. 
                    It applies to all preceding period definitions.

                The following file operations are possible:

                  - absolute value: [1] 
                  - difference: [1-2] 
                  - difference in percent: [1/2] 
                  - sum: [1+2] 
                  - average: [1~2] or [1^2]. 

                The file [1] always refers to the current workspace. 
                Extra files (if passed as argument) are numeroted from 2 to 5.

                The following period operations are possible:

                  - value: (75) 
                  - growth rate over one or more periods: (75/74, 75/70) 
                  - average growth rate: (75//70) 
                  - difference: (75-74, 75-70) 
                  - average difference: (75--70) 
                  - average: (75~74) or (75^74) 
                  - sum of consecutive periods: (70Q1+70Q4) 
                  - index or base value: (76=70) 

                Repetition can be performed with an increment greater than 1 or less than 0: 
                simply place a star followed by the step after the number of repetitions (70:3*5 = 70, 75, 80).

                Examples::

                    70; 75; 80:6 = 70:3*5; 81:5 = 70; 75; 80; 81; 82; 83; 84; 85
                    70/69:2 = 70/69; 71/70
                    (70; 70-69):2 = 70; 70-69; 71; 71-70;
                    70[1,2]:2*5 = 70[1]; 70[2]; 75[1]; 75[2]
                    (70;75)[1,2-1] = 70[1]; 75[1]; 70[2-1]; 75[2-1]
                    (70;75;(80; 80/79):2)[1,2] = 70[1]; 70[2]; 75[1]; 75[2]; 80[1]; 80[2]; 80/79[1]; 80/79[2] 81[1]; 81[2]; 81/80[1]; 81/80[2]
                    2000Y1>5 = 2005Y1
                    1999M1>12 = 2000M1
                    EOS<1 = 2019Y1                         (if EOS == 2020Y1)
                    BOS<1 = 1959Y1                         (if BOS == 1960Y1)
                    EOS<4:5*-1 =2016;2017;2018;2019;2020   (if EOS = 2020Y1)

                Parameters
                ----------
                generalized_sample: str
                    Generalized sample (see above).
                extra_files: str or Path or list(str) or list(Path), optional
                    (List of) extra file(s) referenced in the generalized sample.
                    Maximum 4 files can be passed as argument.
                    The file [1] always refers to the current workspace. 
                    Extra files (if passed as argument) are numeroted from 2 to 5.
                    Default to None (no extra files).
                nb_decimals: int, optional
                    The number of decimals to display.
                    Default to 2.

                Returns
                -------
                ComputedTable

                Examples
                --------
                >>> from pathlib import Path
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import Table, tables, variables
                >>> tables.load(f"{SAMPLE_DATA_DIR}/fun.tbl")
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> tables["C8_1"]          # doctest: +NORMALIZE_WHITESPACE
                DIVIS | 1                                  |
                TITLE |      "Déterminants de l\'output potentiel"
                ----- | ---------------------------------------------
                CELL  |                                    |   "#s"
                ----- | ---------------------------------------------
                CELL  | "Output potentiel"                 |  Q_F+Q_I
                CELL  | "Stock de capital"                 | KNFF[-1]
                CELL  | "Intensité de capital"             |    KLFHP
                CELL  | "Productivité totale des facteurs" |  TFPFHP_
                <BLANKLINE>
                nb lines: 8
                nb columns: 2
                language: \'ENGLISH\'
                gridx: \'MAJOR\'
                gridy: \'MAJOR\'
                graph_axis: \'VALUES\'
                graph_alignment: \'LEFT\'
                <BLANKLINE>

                >>> # simple time series (current workspace) - 6 observations - 4 decimals
                >>> computed_table = tables["C8_1"].compute("2000:6", nb_decimals=4)
                >>> computed_table              # doctest: +NORMALIZE_WHITESPACE
                   line title \\ period[file]     |     00    |     01    |     02    |     03    |     04    |     05
                ---------------------------------------------------------------------------------------------------------
                Output potentiel                 | 5495.2128 | 5627.8589 | 5748.7804 | 5857.9529 | 5966.1999 | 6103.6318
                Stock de capital                 | 8083.5517 | 8359.8908 | 8647.9354 | 8910.3393 | 9175.8106 | 9468.8865
                Intensité de capital             |    0.5032 |    0.4896 |    0.4758 |    0.4623 |    0.4481 |    0.4349
                Productivité totale des facteurs |    0.9938 |    1.0037 |    1.0137 |    1.0239 |    1.0341 |    1.0445
                <BLANKLINE>

                >>> # two time series (current workspace) - 5 observations - 2 decimals (default)
                >>> computed_table = tables["C8_1"].compute("(2010;2010/2009):5")
                >>> computed_table              # doctest: +NORMALIZE_WHITESPACE
                   line title \\ period[file]     |    10    | 10/09 |    11    | 11/10 |    12    | 12/11 |    13    | 13/12 |    14    | 14/13
                --------------------------------------------------------------------------------------------------------------------------------
                Output potentiel                 |  6936.11 |  1.74 |  7045.34 |  1.57 |  7161.54 |  1.65 |  7302.29 |  1.97 |  7460.12 |  2.16
                Stock de capital                 | 11293.85 |  2.82 | 11525.01 |  2.05 | 11736.78 |  1.84 | 11975.49 |  2.03 | 12263.95 |  2.41
                Intensité de capital             |     0.39 | -2.17 |     0.38 | -2.05 |     0.37 | -1.91 |     0.36 | -1.86 |     0.36 | -1.90
                Productivité totale des facteurs |     1.10 |  1.00 |     1.11 |  1.00 |     1.12 |  1.00 |     1.13 |  1.00 |     1.14 |  1.00
                <BLANKLINE>

                >>> # simple time series (current workspace + one extra file) - 5 observations - 2 decimals (default)
                >>> sample_data_dir = Path(SAMPLE_DATA_DIR)
                >>> computed_table = tables["C8_1"].compute("2010[1;2]:5", extra_files=sample_data_dir/"ref.av")
                >>> computed_table              # doctest: +NORMALIZE_WHITESPACE
                   line title \\ period[file]     |  10[1]   |  10[2]   |  11[1]   |  11[2]   |  12[1]   |  12[2]   |  13[1]   |  13[2]   |  14[1]   |  14[2]    
                ----------------------------------------------------------------------------------------------------------------------------------------------- 
                Output potentiel                 |  6936.11 |  6797.39 |  7045.34 |  6904.44 |  7161.54 |  7018.31 |  7302.29 |  7156.24 |  7460.12 |  7310.91  
                Stock de capital                 | 11293.85 | 11067.97 | 11525.01 | 11294.51 | 11736.78 | 11502.05 | 11975.49 | 11735.98 | 12263.95 | 12018.67  
                Intensité de capital             |     0.39 |     0.38 |     0.38 |     0.37 |     0.37 |     0.36 |     0.36 |     0.36 |     0.36 |     0.35  
                Productivité totale des facteurs |     1.10 |     1.08 |     1.11 |     1.09 |     1.12 |     1.10 |     1.13 |     1.11 |     1.14 |     1.12  
                <BLANKLINE>
                >>> computed_table.files            # doctest: +ELLIPSIS
                [\'...ref.av\']

                >>> # simple time series (current workspace + 4 extra files) - 5 observations - 2 decimals (default)
                >>> extra_files = [sample_data_dir / "ref.av", sample_data_dir / "fun.av", 
                ...                sample_data_dir / "fun2.av", sample_data_dir / "a.var"]
                >>> computed_table = tables["C8_1"].compute("2010[1;2;3;4;5]:1", extra_files=extra_files)
                >>> computed_table              # doctest: +NORMALIZE_WHITESPACE
                   line title \\ period[file]     |  10[1]   |  10[2]   |  10[3]   |  10[4]   | 10[5]         
                -------------------------------------------------------------------------------------
                Output potentiel                 |  6936.11 |  6797.39 |  6936.11 |  6936.11 |   --
                Stock de capital                 | 11293.85 | 11067.97 | 11293.85 | 11293.85 |   --
                Intensité de capital             |     0.39 |     0.38 |     0.39 |     0.39 |   --
                Productivité totale des facteurs |     1.10 |     1.08 |     1.10 |     1.10 |   --
                <BLANKLINE>
                >>> [Path(filepath).name for filepath in computed_table.files]
                [\'fun.var\', \'ref.av\', \'fun.av\', \'fun2.av\', \'a.var\']
        '''
    def index(self, key: str) -> int:
        '''index(self, key: str) -> int

                Return the index of the line containing the string `key`.

                Parameters
                ----------
                key: str
                    Text to be retrieved either as a title or as a cell content.

                Returns
                -------
                int

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import Table, comments, lists, variables
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> lists.load(f"{SAMPLE_DATA_DIR}/fun.lst")
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> table_title = "Table example"
                >>> lines_titles = ["GOSG:", "YSSG+COTRES:", "OCUG:"]
                >>> lines_lecs = ["GOSG", "YSSG+COTRES", "OCUG"]
                >>> table = Table(2, table_title, lines_lecs, lines_titles, True, True, True)
                >>> table           # doctest: +NORMALIZE_WHITESPACE
                DIVIS | 1              |
                TITLE |       "Table example"
                ----- | ----------------------------
                CELL  |                |     "#S"
                ----- | ----------------------------
                CELL  | "GOSG:"        |        GOSG
                CELL  | "YSSG+COTRES:" | YSSG+COTRES
                CELL  | "OCUG:"        |        OCUG
                ----- | ----------------------------
                MODE  |
                FILES |
                DATE  |
                <BLANKLINE>
                nb lines: 11
                nb columns: 2
                language: \'ENGLISH\'
                gridx: \'MAJOR\'
                gridy: \'MAJOR\'
                graph_axis: \'VALUES\'
                graph_alignment: \'LEFT\'
                <BLANKLINE>        

                >>> table.index("Table example")
                0
                >>> table.index("GOSG:")
                4
                >>> table.index("OCUG")
                6
        '''
    def insert(self, index: int, value: str | list[str] | tuple[str] | TableLine | TableLineType) -> Any:
        '''insert(self, index: int, value: Union[str, List[str], Tuple[str], TableLine, TableLineType])

                Insert a new line in the table.

                Parameters
                ----------
                index: int
                    index where to insert a line.
                value: str or list(str) or tuple(str) or TableLine or TableLineType
                    value of the new line.
                    If TableLineType, \'value\' represents the type of the new line: FILES, MODE, DATE or SEP.
                    If str, \'value\' represents either a separator line if it only contains characters \'-\' 
                    or a title line.
                    If an iterable of str, \'value\' represents the content of the cells of the new line.

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, TableLineType
                >>> from iode import Table, comments, lists, variables
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> lists.load(f"{SAMPLE_DATA_DIR}/fun.lst")
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> table_title = "Table example"
                >>> lines_titles = ["GOSG:", "YSSG+COTRES:", "OCUG:"]
                >>> lines_lecs = ["GOSG", "YSSG+COTRES", "OCUG"]
                >>> table = Table(2, table_title, lines_lecs, lines_titles, True, True, True)
                >>> table           # doctest: +NORMALIZE_WHITESPACE
                DIVIS | 1              |
                TITLE |       "Table example"
                ----- | ----------------------------
                CELL  |                |     "#S"
                ----- | ----------------------------
                CELL  | "GOSG:"        |        GOSG
                CELL  | "YSSG+COTRES:" | YSSG+COTRES
                CELL  | "OCUG:"        |        OCUG
                ----- | ----------------------------
                MODE  |
                FILES |
                DATE  |
                <BLANKLINE>
                nb lines: 11
                nb columns: 2
                language: \'ENGLISH\'
                gridx: \'MAJOR\'
                gridy: \'MAJOR\'
                graph_axis: \'VALUES\'
                graph_alignment: \'LEFT\'
                <BLANKLINE>

                >>> index = table.index("YSSG+COTRES:")
                >>> index
                5

                >>> # insert new separator line
                >>> index += 1
                >>> table.insert(index, \'-\')

                >>> # insert new title
                >>> index += 1
                >>> table.insert(index, "New Title")

                >>> # insert new separator line 
                >>> index += 1
                >>> table.insert(index, TableLineType.SEP)

                >>> # insert new line with cells
                >>> # "    -> STRING cell
                >>> # no " -> LEC cell
                >>> index += 1
                >>> table.insert(index, [\'"RIDG:\', \'RIDG\'])

                >>> table           # doctest: +NORMALIZE_WHITESPACE
                DIVIS | 1              |
                TITLE |       "Table example"
                ----- | ----------------------------
                CELL  |                |     "#S"
                ----- | ----------------------------
                CELL  | "GOSG:"        |        GOSG
                CELL  | "YSSG+COTRES:" | YSSG+COTRES
                ----- | ----------------------------
                TITLE |         "New Title"
                ----- | ----------------------------
                CELL  | "RIDG:"        |        RIDG
                CELL  | "OCUG:"        |        OCUG
                ----- | ----------------------------
                MODE  |
                FILES |
                DATE  |
                <BLANKLINE>
                nb lines: 15
                nb columns: 2
                language: \'ENGLISH\'
                gridx: \'MAJOR\'
                gridy: \'MAJOR\'
                graph_axis: \'VALUES\'
                graph_alignment: \'LEFT\'
                <BLANKLINE>
        '''
    def update_global_database(self) -> Any:
        """update_global_database(self)"""
    def __copy__(self) -> Table:
        '''__copy__(self) -> Table

                Return a copy of the current Table.

                Examples
                --------
                >>> import copy
                >>> from iode import SAMPLE_DATA_DIR, tables
                >>> tables.load(f"{SAMPLE_DATA_DIR}/fun.tbl")
                >>> tables["C8_1"]          # doctest: +NORMALIZE_WHITESPACE
                DIVIS | 1                                  |
                TITLE |      "Déterminants de l\'output potentiel"
                ----- | ---------------------------------------------
                CELL  |                                    |   "#s"
                ----- | ---------------------------------------------
                CELL  | "Output potentiel"                 |  Q_F+Q_I
                CELL  | "Stock de capital"                 | KNFF[-1]
                CELL  | "Intensité de capital"             |    KLFHP
                CELL  | "Productivité totale des facteurs" |  TFPFHP_
                <BLANKLINE>
                nb lines: 8
                nb columns: 2
                language: \'ENGLISH\'
                gridx: \'MAJOR\'
                gridy: \'MAJOR\'
                graph_axis: \'VALUES\'
                graph_alignment: \'LEFT\'
                <BLANKLINE>
                >>> copied_tbl = copy.copy(tables["C8_1"])
                >>> copied_tbl              # doctest: +NORMALIZE_WHITESPACE
                DIVIS | 1                                  |
                TITLE |      "Déterminants de l\'output potentiel"
                ----- | ---------------------------------------------
                CELL  | ""                                 |   "#s"
                ----- | ---------------------------------------------
                CELL  | "Output potentiel"                 |  Q_F+Q_I
                CELL  | "Stock de capital"                 | KNFF[-1]
                CELL  | "Intensité de capital"             |    KLFHP
                CELL  | "Productivité totale des facteurs" |  TFPFHP_
                <BLANKLINE>
                nb lines: 8
                nb columns: 2
                language: \'ENGLISH\'
                gridx: \'MAJOR\'
                gridy: \'MAJOR\'
                graph_axis: \'VALUES\'
                graph_alignment: \'LEFT\'
                <BLANKLINE>
        '''
    def __delitem__(self, other) -> None:
        '''
        Delete the line at the given index

        Parameters
        ----------
        index: int
            index of the line to be removed.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import Table, comments, lists, variables
        >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
        >>> lists.load(f"{SAMPLE_DATA_DIR}/fun.lst")
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
        >>> table_title = "Table example"
        >>> lines_titles = ["GOSG:", "YSSG+COTRES:", "OCUG:"]
        >>> lines_lecs = ["GOSG", "YSSG+COTRES", "OCUG"]
        >>> table = Table(2, table_title, lines_lecs, lines_titles, True, True, True)
        >>> table           # doctest: +NORMALIZE_WHITESPACE
        DIVIS | 1              |
        TITLE |       "Table example"
        ----- | ----------------------------
        CELL  |                |     "#S"
        ----- | ----------------------------
        CELL  | "GOSG:"        |        GOSG
        CELL  | "YSSG+COTRES:" | YSSG+COTRES
        CELL  | "OCUG:"        |        OCUG
        ----- | ----------------------------
        MODE  |
        FILES |
        DATE  |
        <BLANKLINE>
        nb lines: 11
        nb columns: 2
        language: \'ENGLISH\'
        gridx: \'MAJOR\'
        gridy: \'MAJOR\'
        graph_axis: \'VALUES\'
        graph_alignment: \'LEFT\'
        <BLANKLINE> 

        >>> # remove the MODE line
        >>> del table[-3]
        >>> table           # doctest: +NORMALIZE_WHITESPACE
        DIVIS | 1              |
        TITLE |       "Table example"
        ----- | ----------------------------
        CELL  |                |     "#S"
        ----- | ----------------------------
        CELL  | "GOSG:"        |        GOSG
        CELL  | "YSSG+COTRES:" | YSSG+COTRES
        CELL  | "OCUG:"        |        OCUG
        ----- | ----------------------------
        FILES |
        DATE  |
        <BLANKLINE>
        nb lines: 10
        nb columns: 2
        language: \'ENGLISH\'
        gridx: \'MAJOR\'
        gridy: \'MAJOR\'
        graph_axis: \'VALUES\'
        graph_alignment: \'LEFT\'
        <BLANKLINE>        
        '''
    def __getitem__(self, index):
        '''
        Get a line of the table.

        Parameters
        ----------
        index: int
            index of the line.

        Returns
        -------
        TableLine

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import Table, comments, lists, variables
        >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
        >>> lists.load(f"{SAMPLE_DATA_DIR}/fun.lst")
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
        >>> table_title = "Table example"
        >>> lines_titles = ["GOSG:", "YSSG+COTRES:", "OCUG:"]
        >>> lines_lecs = ["GOSG", "YSSG+COTRES", "OCUG"]
        >>> table = Table(2, table_title, lines_lecs, lines_titles, True, True, True)
        >>> table           # doctest: +NORMALIZE_WHITESPACE
        DIVIS | 1              |
        TITLE |       "Table example"
        ----- | ----------------------------
        CELL  |                |     "#S"
        ----- | ----------------------------
        CELL  | "GOSG:"        |        GOSG
        CELL  | "YSSG+COTRES:" | YSSG+COTRES
        CELL  | "OCUG:"        |        OCUG
        ----- | ----------------------------
        MODE  |
        FILES |
        DATE  |
        <BLANKLINE>
        nb lines: 11
        nb columns: 2
        language: \'ENGLISH\'
        gridx: \'MAJOR\'
        gridy: \'MAJOR\'
        graph_axis: \'VALUES\'
        graph_alignment: \'LEFT\'
        <BLANKLINE>

        >>> table[0]
        Table example
        >>> table[1]
        ---
        >>> table[4]
        (\'"GOSG:"\', \'GOSG\')
        >>> table[-3]
        <MODE>
        >>> table[-2]
        <FILES>
        >>> table[-1]
        <DATE>
        '''
    def __iadd__(self, other):
        '''
        Append a new line to the table.

        Parameters
        ----------
        value: str or list(str) or tuple(str) or TableLine or TableLineType
            value of the new line.
            If TableLineType, \'value\' represents the type of the new line: FILES, MODE, DATE or SEP.
            If str, \'value\' represents either a separator line if it only contains characters \'-\' 
            or a title line.
            If an iterable of str, \'value\' represents the content of the cells of the new line.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR, TableLineType
        >>> from iode import Table, comments, lists, variables
        >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
        >>> lists.load(f"{SAMPLE_DATA_DIR}/fun.lst")
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
        >>> table_title = "Table example"
        >>> lines_titles = ["GOSG:", "YSSG+COTRES:", "OCUG:"]
        >>> lines_lecs = ["GOSG", "YSSG+COTRES", "OCUG"]
        >>> table = Table(2, table_title, lines_lecs, lines_titles)
        >>> table           # doctest: +NORMALIZE_WHITESPACE
        DIVIS | 1              |
        TITLE |       "Table example"
        ----- | ----------------------------
        CELL  |                |     "#S"
        ----- | ----------------------------
        CELL  | "GOSG:"        |        GOSG
        CELL  | "YSSG+COTRES:" | YSSG+COTRES
        CELL  | "OCUG:"        |        OCUG
        <BLANKLINE>
        nb lines: 7
        nb columns: 2
        language: \'ENGLISH\'
        gridx: \'MAJOR\'
        gridy: \'MAJOR\'
        graph_axis: \'VALUES\'
        graph_alignment: \'LEFT\'
        <BLANKLINE>

        >>> # append a separator line
        >>> table += \'-\'

        >>> # append a title
        >>> table += "New Title"

        >>> # append a separator line 
        >>> table += TableLineType.SEP

        >>> # append a line with cells
        >>> # "    -> STRING cell
        >>> # no " -> LEC cell
        >>> table += [\'"RIDG:"\', \'RIDG\']

        >>> table           # doctest: +NORMALIZE_WHITESPACE
        DIVIS | 1              |
        TITLE |       "Table example"
        ----- | ----------------------------
        CELL  |                |     "#S"
        ----- | ----------------------------
        CELL  | "GOSG:"        |        GOSG
        CELL  | "YSSG+COTRES:" | YSSG+COTRES
        CELL  | "OCUG:"        |        OCUG
        ----- | ----------------------------
        TITLE |         "New Title"
        ----- | ----------------------------
        CELL  | "RIDG:"        |        RIDG
        <BLANKLINE>
        nb lines: 11
        nb columns: 2
        language: \'ENGLISH\'
        gridx: \'MAJOR\'
        gridy: \'MAJOR\'
        graph_axis: \'VALUES\'
        graph_alignment: \'LEFT\'
        <BLANKLINE>
        '''
    def __len__(self) -> int:
        '''
        Number of lines of the table.

        Equivalent to :py:meth:`Table.nb_lines`.

        Notes
        -----
        The special divider line is not taken into account in the returned number of lines.

        Examples
        --------
        >>> from iode import Table
        >>> table = Table()
        >>> table           # doctest: +NORMALIZE_WHITESPACE
        DIVIS | 1 |
        TITLE |
        ----- | --------
        CELL  |   | "#S"
        ----- | --------
        <BLANKLINE>
        nb lines: 4
        nb columns: 2
        language: \'ENGLISH\'
        gridx: \'MAJOR\'
        gridy: \'MAJOR\'
        graph_axis: \'VALUES\'
        graph_alignment: \'LEFT\'
        <BLANKLINE>

        >>> len(table)
        4
        >>> # equivalent to nb_lines
        >>> table.nb_lines
        4
        '''
    def __reduce__(self):
        """__reduce_cython__(self)"""
    def __setitem__(self, index, object) -> None:
        '''
        Update a line of the table.

        Parameters
        ----------
        index: int
            index of the line to be updated.
        value: str or list(str) or tuple(str) or TableLine
            New content for the line. 
        
        Warnings
        --------
        - Line of type \'MODE\', \'FILES\', \'DATE\' and \'SEP\' cannot be updated.
        - When updating a cell content, the cell is converted to a cell of type \'STRING\' 
          if the new text contains double quotes, otherwise the cell becomes of type \'LEC\'.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import Table, comments, lists, variables
        >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
        >>> lists.load(f"{SAMPLE_DATA_DIR}/fun.lst")
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
        >>> table_title = "Table example"
        >>> lines_titles = ["GOSG:", "YSSG+COTRES:", "OCUG:"]
        >>> lines_lecs = ["GOSG", "YSSG+COTRES", "OCUG"]
        >>> table = Table(2, table_title, lines_lecs, lines_titles, True, True, True)
        >>> table           # doctest: +NORMALIZE_WHITESPACE
        DIVIS | 1              |
        TITLE |       "Table example"
        ----- | ----------------------------
        CELL  |                |     "#S"
        ----- | ----------------------------
        CELL  | "GOSG:"        |        GOSG
        CELL  | "YSSG+COTRES:" | YSSG+COTRES
        CELL  | "OCUG:"        |        OCUG
        ----- | ----------------------------
        MODE  |
        FILES |
        DATE  |
        <BLANKLINE>
        nb lines: 11
        nb columns: 2
        language: \'ENGLISH\'
        gridx: \'MAJOR\'
        gridy: \'MAJOR\'
        graph_axis: \'VALUES\'
        graph_alignment: \'LEFT\'
        <BLANKLINE>

        >>> # warning: line of type \'MODE\', \'FILES\', \'DATE\' and \'SEP\' cannot be updated.

        >>> # update a title line
        >>> table[0] = "New title"
        >>> table[0]
        New title

        >>> # update cell contents
        >>> # warning: the cell is converted to a cell of type \'STRING\' if the new text contains 
        >>> # double quotes, otherwise the cell becomes of type \'LEC\'.
        >>> table[5] = (\'"YSSG:"\', \'YSSG\')
        >>> table[5]
        (\'"YSSG:"\', \'YSSG\')

        >>> table           # doctest: +NORMALIZE_WHITESPACE
        DIVIS | 1       |
        TITLE |  "New title"
        ----- | --------------
        CELL  |         | "#S"
        ----- | --------------
        CELL  | "GOSG:" | GOSG
        CELL  | "YSSG:" | YSSG
        CELL  | "OCUG:" | OCUG
        ----- | --------------
        MODE  |
        FILES |
        DATE  |
        <BLANKLINE>
        nb lines: 11
        nb columns: 2
        language: \'ENGLISH\'
        gridx: \'MAJOR\'
        gridy: \'MAJOR\'
        graph_axis: \'VALUES\'
        graph_alignment: \'LEFT\'
        <BLANKLINE>
        '''

class TableCell:
    __pyx_vtable__: ClassVar[PyCapsule] = ...
    align: str
    bold: bool
    cell_type: str
    italic: bool
    underline: bool
    variables: list[str]
    def __init__(self, *args, **kwargs) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""
    def __reduce__(self):
        """__reduce_cython__(self)"""

class TableCellAlign(enum.IntEnum):
    __new__: ClassVar[Callable] = ...
    CENTER: ClassVar[TableCellAlign] = ...
    DECIMAL: ClassVar[TableCellAlign] = ...
    LEFT: ClassVar[TableCellAlign] = ...
    RIGHT: ClassVar[TableCellAlign] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[int]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    def __format__(self, *args, **kwargs) -> str: ...

class TableCellFont(enum.IntEnum):
    __new__: ClassVar[Callable] = ...
    BOLD: ClassVar[TableCellFont] = ...
    ITALIC: ClassVar[TableCellFont] = ...
    NORMAL: ClassVar[TableCellFont] = ...
    UNDERLINE: ClassVar[TableCellFont] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[int]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    def __format__(self, *args, **kwargs) -> str: ...

class TableCellType(enum.IntEnum):
    __new__: ClassVar[Callable] = ...
    LEC: ClassVar[TableCellType] = ...
    STRING: ClassVar[TableCellType] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[int]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    def __format__(self, *args, **kwargs) -> str: ...

class TableGraphAlign(enum.IntEnum):
    __new__: ClassVar[Callable] = ...
    CENTER: ClassVar[TableGraphAlign] = ...
    LEFT: ClassVar[TableGraphAlign] = ...
    RIGHT: ClassVar[TableGraphAlign] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[int]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    def __format__(self, *args, **kwargs) -> str: ...

class TableGraphAxis(enum.IntEnum):
    __new__: ClassVar[Callable] = ...
    LOG: ClassVar[TableGraphAxis] = ...
    PERCENT: ClassVar[TableGraphAxis] = ...
    SEMILOG: ClassVar[TableGraphAxis] = ...
    VALUES: ClassVar[TableGraphAxis] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[int]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    def __format__(self, *args, **kwargs) -> str: ...

class TableGraphGrid(enum.IntEnum):
    __new__: ClassVar[Callable] = ...
    MAJOR: ClassVar[TableGraphGrid] = ...
    MINOR: ClassVar[TableGraphGrid] = ...
    NONE: ClassVar[TableGraphGrid] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[int]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    def __format__(self, *args, **kwargs) -> str: ...

class TableGraphType(enum.IntEnum):
    __new__: ClassVar[Callable] = ...
    BAR: ClassVar[TableGraphType] = ...
    LINE: ClassVar[TableGraphType] = ...
    SCATTER: ClassVar[TableGraphType] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[int]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    def __format__(self, *args, **kwargs) -> str: ...

class TableLang(enum.IntEnum):
    __new__: ClassVar[Callable] = ...
    DUTCH: ClassVar[TableLang] = ...
    ENGLISH: ClassVar[TableLang] = ...
    FRENCH: ClassVar[TableLang] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[int]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    def __format__(self, *args, **kwargs) -> str: ...

class TableLine:
    __pyx_vtable__: ClassVar[PyCapsule] = ...
    axis_left: bool
    graph_type: str
    line_type: str
    def __init__(self, *args, **kwargs) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""
    def __delitem__(self, other) -> None:
        """Delete self[key]."""
    def __getitem__(self, index):
        '''
        Get the cell of a line.

        Parameters
        ----------
        index: int
            index of the cell.

        Returns
        -------
        TableCell

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import tables, Table
        >>> tables.load(f"{SAMPLE_DATA_DIR}/fun.tbl")
        >>> table = tables["ANAPRIX"]
        >>> table           # doctest: +NORMALIZE_WHITESPACE
        DIVIS | 1                            |
        TITLE |                        "Analyse des prix"
        ----- | ---------------------------------------------------------------
        CELL  |                              |               "#s"
        ----- | ---------------------------------------------------------------
        CELL  | "GAP_"                       |                             GAP_
        CELL  | "dln (PC/(1+ITCR))-dln AOUC" | 100*(dln (PC/(1+ITCR))-dln AOUC)
        <BLANKLINE>
        nb lines: 6
        nb columns: 2
        language: \'ENGLISH\'
        gridx: \'MAJOR\'
        gridy: \'MAJOR\'
        graph_axis: \'VALUES\'
        graph_alignment: \'LEFT\'
        <BLANKLINE>

        >>> table[4][0]
        "GAP_"
        >>> table[4][1]
        GAP_
        '''
    def __len__(self) -> int:
        """Return len(self)."""
    def __reduce__(self):
        """__reduce_cython__(self)"""
    def __setitem__(self, index, object) -> None:
        '''
        Update the content of a cell of the line.

        Parameters
        ----------
        index: int
            index of the cell.
        value: str or TableCell
            new content of the cell.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import tables, Table
        >>> tables.load(f"{SAMPLE_DATA_DIR}/fun.tbl")
        >>> table = tables["ANAPRIX"]
        >>> table           # doctest: +NORMALIZE_WHITESPACE
        DIVIS | 1                            |
        TITLE |                        "Analyse des prix"
        ----- | ---------------------------------------------------------------
        CELL  |                              |               "#s"
        ----- | ---------------------------------------------------------------
        CELL  | "GAP_"                       |                             GAP_
        CELL  | "dln (PC/(1+ITCR))-dln AOUC" | 100*(dln (PC/(1+ITCR))-dln AOUC)
        <BLANKLINE>
        nb lines: 6
        nb columns: 2
        language: \'ENGLISH\'
        gridx: \'MAJOR\'
        gridy: \'MAJOR\'
        graph_axis: \'VALUES\'
        graph_alignment: \'LEFT\'
        <BLANKLINE>

        >>> # warning: content of cells of type \'STRING\' must be surrounded by double quotes
        >>> table[5][0] = \'"dln(PC/ITCR) - dln AOUC"\'
        >>> table[5][0]
        "dln(PC/ITCR) - dln AOUC"
        >>> # warning: if the new content is not surrounded by double quotes, 
        >>> #          it is considerer as \'LEC\' cell
        >>> table[5][1] = \'100*(dln(PC / ITCR) - dln AOUC)\'
        >>> table[5][1]
        100*(dln(PC / ITCR) - dln AOUC)
        '''

class TableLineType(enum.IntEnum):
    __new__: ClassVar[Callable] = ...
    CELL: ClassVar[TableLineType] = ...
    DATE: ClassVar[TableLineType] = ...
    FILES: ClassVar[TableLineType] = ...
    MODE: ClassVar[TableLineType] = ...
    SEP: ClassVar[TableLineType] = ...
    TITLE: ClassVar[TableLineType] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[int]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    def __format__(self, *args, **kwargs) -> str: ...

class Tables(_AbstractDatabase):
    __pyx_vtable__: ClassVar[PyCapsule] = ...
    i: PositionalIndexer
    def __init__(self, *args, **kwargs) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""
    def get_title(self, key: str | int) -> str:
        '''get_title(self, key: Union[str, int]) -> str

                Get the title of a table from its name or index.

                Parameters
                ----------
                key: str or int
                   Name or index of the table to get the title from.
        
                Returns
                -------
                str

                Examples
                --------
                >>> from iode import tables, SAMPLE_DATA_DIR
                >>> tables.load(f"{SAMPLE_DATA_DIR}/fun.tbl")
                >>> tables.get_title("ANAKNFF")
                \'Déterminants de la croissance de K\'
                >>> tables.get_title(0)
                \'Déterminants de la croissance de K\'
        '''
    def print_to_file(self, destination_file: str | Path, generalized_sample: str, names: str | Path | list[str | Path], nb_decimals: int, format: str = ...) -> Any:
        '''print_to_file(self, destination_file: Union[str, Path], generalized_sample: str, names: Union[str, Path, List[Union[str, Path]]], nb_decimals: int, format: str = None)

                Compute and print a list of IODE tables to a file.

                Argument `format` must be in the list:
                - \'H\' (HTML file)
                - \'M\' (MIF file)
                - \'R\' (RTF file)
                - \'C\' (CSV file)
                - \'D\' (DUMMY file)

                If argument `format` is not provided, the A2M format will be used to print the output.

                If the filename does not contain an extension, it is automatically added based on the format.

                Parameters
                ----------
                destination_file: str or Path 
                    The destination file path.
                generalized_sample: str
                    The generalized sample (see :py:meth:`` for details).
                names: str or list(str) 
                    The names of the IODE tables to be printed.
                nb_decimals: int 
                    The number of decimals to use in the output.
                format: str, optional 
                    The format of the output file. Deduced from the extension if not provided.
                    If destination_file has no extension and format is None, the A2M format is used.

                Examples
                --------
                >>> from pathlib import Path
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import Table, load_extra_files, tables, variables
                >>> tables.load(f"{SAMPLE_DATA_DIR}/fun.tbl")
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")

                >>> tables["C8_1"]              # doctest: +NORMALIZE_WHITESPACE
                DIVIS | 1                                  |
                TITLE |      "Déterminants de l\'output potentiel"
                ----- | ---------------------------------------------
                CELL  |                                    |   "#s"
                ----- | ---------------------------------------------
                CELL  | "Output potentiel"                 |  Q_F+Q_I
                CELL  | "Stock de capital"                 | KNFF[-1]
                CELL  | "Intensité de capital"             |    KLFHP
                CELL  | "Productivité totale des facteurs" |  TFPFHP_
                <BLANKLINE>
                nb lines: 8
                nb columns: 2
                language: \'ENGLISH\'
                gridx: \'MAJOR\'
                gridy: \'MAJOR\'
                graph_axis: \'VALUES\'
                graph_alignment: \'LEFT\'
                <BLANKLINE>

                >>> tables["C8_2"]              # doctest: +NORMALIZE_WHITESPACE
                DIVIS | 1                      |
                TITLE |     "Déterminants de la productivité"
                ----- | ------------------------------------------
                CELL  |                        |        "#s"
                ----- | ------------------------------------------
                CELL  | "Productivité"         | (QAFF_/(NFY+NIY))
                CELL  | "Intensité de capital" |    KLFHP**(-0.31)
                CELL  | "TFPFHP_"              |           TFPFHP_
                <BLANKLINE>
                nb lines: 7
                nb columns: 2
                language: \'ENGLISH\'
                gridx: \'MAJOR\'
                gridy: \'MAJOR\'
                graph_axis: \'VALUES\'
                graph_alignment: \'LEFT\'
                <BLANKLINE>

                >>> tables["C8_3"]              # doctest: +NORMALIZE_WHITESPACE
                DIVIS | 1                                |
                TITLE |               "Output gap"
                ----- | ---------------------------------------
                CELL  |                                  | "#s"
                ----- | ---------------------------------------
                CELL  | "Output gap (methodologie FPB)." | GAP_
                <BLANKLINE>
                nb lines: 5
                nb columns: 2
                language: \'ENGLISH\'
                gridx: \'MAJOR\'
                gridy: \'MAJOR\'
                graph_axis: \'VALUES\'
                graph_alignment: \'LEFT\'
                <BLANKLINE>

                >>> tables["C8_4"]              # doctest: +NORMALIZE_WHITESPACE
                DIVIS | 1                              |
                TITLE |  "Déterminants de la croissance du stock de capital"
                ----- | -----------------------------------------------------
                CELL  |                                |         "#s"
                ----- | -----------------------------------------------------
                CELL  | "Croissnce du stoc de capital" |             dln KNFF
                CELL  | "Rentabilité"                  | 0.04*ln mavg(3,RENT)
                <BLANKLINE>
                nb lines: 6
                nb columns: 2
                language: \'ENGLISH\'
                gridx: \'MAJOR\'
                gridy: \'MAJOR\'
                graph_axis: \'VALUES\'
                graph_alignment: \'LEFT\'
                <BLANKLINE>

                >>> names = ["C8_1", "C8_2", "C8_3", "C8_4"]
                >>> tables.print_to_file(tmp_dir / "tables_2_periods.csv", "(2010;2010/2009):5", names, 2)
                >>> with open(tmp_dir / "tables_2_periods.csv", "r") as f:    # doctest: +NORMALIZE_WHITESPACE
                ...     print(f.read())
                "Déterminants de l\'output potentiel"
                <BLANKLINE>
                " ","10","10/09","11","11/10","12","12/11","13","13/12","14","14/13",
                <BLANKLINE>
                "Output potentiel","6936.11","1.74","7045.34","1.57","7161.54","1.65","7302.29","1.97","7460.12","2.16",       
                "Stock de capital","11293.85","2.82","11525.01","2.05","11736.78","1.84","11975.49","2.03","12263.95","2.41",  
                "Intensité de capital","0.39","-2.17","0.38","-2.05","0.37","-1.91","0.36","-1.86","0.36","-1.90",
                "Productivité totale des facteurs","1.10","1.00","1.11","1.00","1.12","1.00","1.13","1.00","1.14","1.00",      
                "Déterminants de la productivité"
                <BLANKLINE>
                " ","10","10/09","11","11/10","12","12/11","13","13/12","14","14/13",
                <BLANKLINE>
                "Productivité","2.13","1.42","2.17","2.07","2.22","2.00","2.26","1.86","2.30","1.61",
                "Intensité de capital","1.34","0.68","1.35","0.64","1.36","0.60","1.37","0.59","1.38","0.60",
                "TFPFHP_","1.10","1.00","1.11","1.00","1.12","1.00","1.13","1.00","1.14","1.00",
                "Output gap"
                <BLANKLINE>
                " ","10","10/09","11","11/10","12","12/11","13","13/12","14","14/13",
                <BLANKLINE>
                "Output gap (methodologie FPB).","3.38","-11.68","3.57","5.62","3.74","4.84","3.66","-2.21","3.31","-9.37",    
                "Déterminants de la croissance du stock de capital"
                <BLANKLINE>
                " ","10","10/09","11","11/10","12","12/11","13","13/12","14","14/13",
                <BLANKLINE>
                "Croissnce du stoc de capital","0.02","-27.10","0.02","-10.13","0.02","10.58","0.02","18.22","0.02","2.48",    
                "Rentabilité","-0.01","34.56","-0.01","7.51","-0.01","0.31","-0.01","0.17","-0.01","2.46",
                <BLANKLINE>

                >>> extra_files = load_extra_files(Path(SAMPLE_DATA_DIR) / "ref.av")
                >>> extra_files[0].name
                \'ref.av\'
                >>> tables.print_to_file(tmp_dir / "tables_2_files.csv", "2010[1-2]:5", names, 2)    
                >>> with open(tmp_dir / "tables_2_files.csv", "r") as f:
                ...     print(f.read())
                "Déterminants de l\'output potentiel"
                <BLANKLINE>
                " ","10[1-2]","11[1-2]","12[1-2]","13[1-2]","14[1-2]",
                <BLANKLINE>
                "Output potentiel","138.72","140.91","143.23","146.05","149.20",
                "Stock de capital","225.88","230.50","234.74","239.51","245.28",
                "Intensité de capital","0.01","0.01","0.01","0.01","0.01",
                "Productivité totale des facteurs","0.02","0.02","0.02","0.02","0.02",
                "Déterminants de la productivité"
                <BLANKLINE>
                " ","10[1-2]","11[1-2]","12[1-2]","13[1-2]","14[1-2]",
                <BLANKLINE>
                "Productivité","0.00","-0.00","-0.00","0.00","0.00",
                "Intensité de capital","-0.01","-0.01","-0.01","-0.01","-0.01",
                "TFPFHP_","0.02","0.02","0.02","0.02","0.02",
                "Output gap"
                <BLANKLINE>
                " ","10[1-2]","11[1-2]","12[1-2]","13[1-2]","14[1-2]",
                <BLANKLINE>
                "Output gap (methodologie FPB).","-0.00","0.00","0.00","0.00","0.00",
                "Déterminants de la croissance du stock de capital"
                <BLANKLINE>
                " ","10[1-2]","11[1-2]","12[1-2]","13[1-2]","14[1-2]",
                <BLANKLINE>
                "Croissnce du stoc de capital","0.00","-0.00","-0.00","0.00","-0.00",
                "Rentabilité","0.00","-0.00","-0.00","-0.00","0.00",
                <BLANKLINE>
        '''
    def __delitem__(self, other) -> None:
        '''
        Remove the (subset of) table(s) referenced by `key` from the Tables database.

        Parameters
        ----------
        key: str or list(str)
            (list of) name(s) of the table(s) to be removed.
            The list of names can be given as a string pattern (e.g. "A*;*_").

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import tables
        >>> tables.load(f"{SAMPLE_DATA_DIR}/fun.tbl")

        >>> # a) delete one table
        >>> tables.get_names("G*")
        [\'GAP\', \'GDP\', \'GFR\', \'GFRLEVEL\', \'GFRPC\', \'GROWTH\']
        >>> del tables["GFRLEVEL"]
        >>> tables.get_names("G*")
        [\'GAP\', \'GDP\', \'GFR\', \'GFRPC\', \'GROWTH\']

        >>> # b) delete several tables at once using a pattern
        >>> del tables["G*"]
        >>> tables.get_names("G*")
        []

        >>> # c) delete several tables at once using a list of names
        >>> tables.get_names("C8_?")
        [\'C8_1\', \'C8_2\', \'C8_3\', \'C8_4\', \'C8_5\', \'C8_6\', \'C8_7\', \'C8_8\', \'C8_9\']
        >>> del tables[["C8_1", "C8_3", "C8_5", "C8_7", "C8_9"]]
        >>> tables.get_names("C8_?")
        [\'C8_2\', \'C8_4\', \'C8_6\', \'C8_8\']

        >>> # delete one table from a subset of the global database
        >>> tables_subset = tables["M*"]
        >>> tables_subset.names
        [\'MULT1FR\', \'MULT1RESU\', \'MULT2FR\', \'MULT2RESU\']
        >>> del tables_subset["MULT2RESU"]
        >>> tables_subset.names
        [\'MULT1FR\', \'MULT1RESU\', \'MULT2FR\']
        >>> # NOTE: the table has also been deleted from the global database
        >>> "MULT2RESU" in tables
        False
        >>> tables.get_names("M*")
        [\'MULT1FR\', \'MULT1RESU\', \'MULT2FR\']
        '''
    def __getitem__(self, index):
        '''
        Return the (subset of) table(s) referenced by `key`.

        The `key` can represent a single object name (e.g. "ACAF") or a list of object names ("ACAF;ACAG;AOUC") 
        or a pattern (e.g. "A*") or a list of sub-patterns (e.g. "A*;*_").
        
        If the `key` represents a list of object names or of sub-patterns, each name or sub-pattern is separated 
        by a `separator` character which is either a whitespace ` `, or a comma `,`, or a semi-colon `;`, or a 
        tabulation `\\t`, or a newline `\\n`.

        A (sub-)`pattern` is a list of characters representing a group of object names. 
        It includes some special characters which have a special meaning:
        
            - `*` : any character sequence, even empty
            - `?` : any character (one and only one)
            - `@` : any alphanumerical char [A-Za-z0-9]
            - `&` : any non alphanumerical char
            - `|` : any alphanumeric character or none at the beginning and end of a string 
            - `!` : any non-alphanumeric character or none at the beginning and end of a string 
            - `\\` : escape the next character

        Note that the `key` can contain references to IODE lists which are prefixed with the symbol `$`.

        Parameters
        ----------
        key: str or list(str)
            (the list of) name(s) of the table(s) to get.
            The list of objects to get can be specified by a pattern or by a list of sub-patterns (e.g. "A*;*_").

        Returns
        -------
        Single table or a subset of the database.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import tables
        >>> tables.load(f"{SAMPLE_DATA_DIR}/fun.tbl")

        >>> # a) get one table
        >>> tables["YDH"]           # doctest: +NORMALIZE_WHITESPACE
        DIVIS |                                                          1 |                                    PC_*40.34
        TITLE |                        "Tableau B-3. Revenu disponible des ménages à prix constant"
        ----- | ---------------------------------------------------------------------------------------------------------
        CELL  |                                                            |                     "#S"
        CELL  | "Revenus primaires"                                        |                            WBU_+YN+GOSH_+IDH
        CELL  | "   Masse salariale totale"                                |                                         WBU_
        CELL  | "   Revenu net du travail en provenance du Reste du monde" |                                           YN
        CELL  | "   Surplus brut d\'exploitation"                           |                                        GOSH_
        CELL  | "   Revenu net de la propriété"                            |                                          IDH
        CELL  | "Cotisations sociales et impôts"                           |                                  SSF+SSH+DTH
        CELL  | "   Cotisations patronales"                                |                                          SSF
        CELL  | "   Cotisations personnelles"                              |                                          SSH
        CELL  | "IPP"                                                      |                                          DTH
        CELL  | "Prestations sociales "                                    |                                     SBH+OCUH
        CELL  | "   Sécurité sociale"                                      |                                          SBH
        CELL  | "   Diverses prestations"                                  |                                         OCUH
        CELL  | "Total"                                                    | (WBU_+YN+GOSH_+IDH)-(SSF+SSH+DTH)+(SBH+OCUH)
        ----- | ---------------------------------------------------------------------------------------------------------
        FILES |
        DATE  |
        <BLANKLINE>
        nb lines: 19
        nb columns: 2
        language: \'ENGLISH\'
        gridx: \'MAJOR\'
        gridy: \'MAJOR\'
        graph_axis: \'VALUES\'
        graph_alignment: \'LEFT\'
        <BLANKLINE>

        >>> # b) get a subset of the Tables database using a pattern
        >>> tables_subset = tables["C8_*"]
        >>> tables_subset.names
        [\'C8_1\', \'C8_10\', \'C8_11\', \'C8_13\', \'C8_14\', \'C8_2\', \'C8_3\', \'C8_4\', \'C8_5\', \'C8_6\', \'C8_7\', \'C8_8\', \'C8_9\']

        >>> # c) get a subset of the Tables database using a list of names
        >>> tables_subset = tables[["C8_1", "C8_2", "C8_4", "C8_5", "C8_7"]]
        >>> tables_subset.names
        [\'C8_1\', \'C8_2\', \'C8_4\', \'C8_5\', \'C8_7\']
        '''
    def __hash__(self) -> int:
        '''
        Return a hash value for the current Tables database.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR, tables, Table
        >>> tables.load(f"{SAMPLE_DATA_DIR}/fun.tbl")
        >>> len(tables)
        46
        >>> original_hash = hash(tables)
        
        >>> # rename 1 table
        >>> tables.rename("ANAKNFF", "ANAKNFF_")
        >>> original_hash == hash(tables)
        False
        >>> tables.rename("ANAKNFF_", "ANAKNFF")  # revert the change
        >>> original_hash == hash(tables)
        True

        >>> # modify one table
        >>> tbl = tables["ANAKNFF"]
        >>> original_title = str(tbl[0])
        >>> tbl[0] = "New title"
        >>> tables["ANAKNFF"] = tbl
        >>> original_hash == hash(tables)
        False
        >>> tbl[0] = original_title  # revert the change
        >>> tables["ANAKNFF"] = tbl
        >>> original_hash == hash(tables)           # doctest: +SKIP
        True

        >>> # delete a table
        >>> original_table = tables["ANAKNFF"]
        >>> del tables["ANAKNFF"]
        >>> original_hash == hash(tables)
        False
        >>> tables["ANAKNFF"] = original_table
        >>> original_hash == hash(tables)           # doctest: +SKIP
        True

        >>> # add a table
        >>> tables["NEW"] = Table()
        >>> original_hash == hash(tables)
        False
        >>> del tables["NEW"]
        >>> original_hash == hash(tables)           # doctest: +SKIP
        True
        '''
    def __reduce__(self):
        """__reduce_cython__(self)"""
    def __setitem__(self, index, object) -> None:
        '''
        Update/add a (subset of) table(s) referenced by `key` from/to the Tables database.

        The `key` can represent a single object name (e.g. "ACAF") or a list of object names ("ACAF;ACAG;AOUC") 
        or a pattern (e.g. "A*") or a list of sub-patterns (e.g. "A*;*_").
        
        If the `key` represents a list of object names or of sub-patterns, each name or sub-pattern is separated 
        by a `separator` character which is either a whitespace ` `, or a comma `,`, or a semi-colon `;`, or a 
        tabulation `\\t`, or a newline `\\n`.

        A (sub-)`pattern` is a list of characters representing a group of object names. 
        It includes some special characters which have a special meaning:
        
            - `*` : any character sequence, even empty
            - `?` : any character (one and only one)
            - `@` : any alphanumerical char [A-Za-z0-9]
            - `&` : any non alphanumerical char
            - `|` : any alphanumeric character or none at the beginning and end of a string 
            - `!` : any non-alphanumeric character or none at the beginning and end of a string 
            - `\\` : escape the next character

        Note that the `key` can contain references to IODE lists which are prefixed with the symbol `$`.

        Parameters
        ----------
        key: str or list(str)
            (the list of) name(s) of the table(s) to update/add.
            The list of objects to update/add can be specified by a pattern or by a list of sub-patterns 
            (e.g. "A*;*_").
        value: int, tuple(...), dict(str, ...), Table or list of any of those
            If int, then it is interpreted as the number of columns to create a new empty table.
            If tuple or dictionary, it is forwarded to the Table constructor to create a new table.
            If Table, then it is used to update an existing table or to create a new table if it does not exist yet.

        See Also
        --------
        iode.Table

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import tables, Table, comments, TableGraphAxis
        >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
        >>> tables.load(f"{SAMPLE_DATA_DIR}/fun.tbl")
        
        >>> # a) -------- new table --------
        >>> # 1. specify list of line titles and list of LEC expressions
        >>> lines_titles = ["GOSG:", "YDTG:", "DTH:", "DTF:", "IT:", "YSSG+COTRES:", "RIDG:", "OCUG:"]
        >>> lines_lecs = ["GOSG", "YDTG", "DTH", "DTF", "IT", "YSSG+COTRES", "RIDG", "OCUG"]
        >>> tables["TABLE_CELL_LECS"] = {"nb_columns": 2, "table_title": "New Table", "lecs_or_vars": lines_lecs, 
        ...                              "lines_titles": lines_titles, "mode": True, "files": True, "date": True} 
        >>> tables["TABLE_CELL_LECS"]         # doctest: +NORMALIZE_WHITESPACE
        DIVIS | 1              |
        TITLE |         "New Table"
        ----- | ----------------------------
        CELL  | ""             |     "#S"
        ----- | ----------------------------
        CELL  | "GOSG:"        |        GOSG
        CELL  | "YDTG:"        |        YDTG
        CELL  | "DTH:"         |         DTH
        CELL  | "DTF:"         |         DTF
        CELL  | "IT:"          |          IT
        CELL  | "YSSG+COTRES:" | YSSG+COTRES
        CELL  | "RIDG:"        |        RIDG
        CELL  | "OCUG:"        |        OCUG
        ----- | ----------------------------
        MODE  |
        FILES |
        DATE  |
        <BLANKLINE>
        nb lines: 16
        nb columns: 2
        language: \'ENGLISH\'
        gridx: \'MAJOR\'
        gridy: \'MAJOR\'
        graph_axis: \'VALUES\'
        graph_alignment: \'LEFT\'
        <BLANKLINE>

        >>> # 2. specify list of variables
        >>> vars_list = ["GOSG", "YDTG", "DTH", "DTF", "IT", "YSSG", "COTRES", "RIDG", "OCUG", "$ENVI"]
        >>> tables["TABLE_VARS"] = {"nb_columns": 2, "table_title": "New Table", "lecs_or_vars": vars_list, 
        ...                         "mode": True, "files": True, "date": True}
        >>> tables["TABLE_VARS"]             # doctest: +NORMALIZE_WHITESPACE
        DIVIS | 1                                                                    |
        TITLE |                                  "New Table"
        ----- | ------------------------------------------------------------------------------
        CELL  | ""                                                                   |   "#S"
        ----- | ------------------------------------------------------------------------------
        CELL  | "Bruto exploitatie-overschot: overheid (= afschrijvingen)."          |    GOSG
        CELL  | "Overheid: geïnde indirecte belastingen."                            |    YDTG
        CELL  | "Totale overheid: directe belasting van de gezinnen."                |     DTH
        CELL  | "Totale overheid: directe vennootschapsbelasting."                   |     DTF
        CELL  | "Totale indirecte belastingen."                                      |      IT
        CELL  | "Globale overheid: ontvangen sociale zekerheidsbijdragen."           |    YSSG
        CELL  | "Cotisation de responsabilisation."                                  |  COTRES
        CELL  | "Overheid: inkomen uit vermogen."                                    |    RIDG
        CELL  | "Globale overheid: saldo van de ontvangen lopendeoverdrachten."      |    OCUG
        CELL  | "Wisselkoers van de USD t.o.v. de BEF (jaargemiddelde)."             |      EX
        CELL  | "Index wereldprijs - invoer van niet-energieprodukten, inUSD."       |   PWMAB
        CELL  | "Index wereldprijs - invoer van diensten, in USD."                   |    PWMS
        CELL  | "Index wereldprijs - uitvoer van niet-energieprodukten, inUSD."      |   PWXAB
        CELL  | "Index wereldprijs - uitvoer van diensten, in USD."                  |    PWXS
        CELL  | "Indicator van het volume van de wereldvraag naar goederen,1985=1."  |   QWXAB
        CELL  | "Indicator van het volume van de wereldvraag naar diensten,1985=1."  |    QWXS
        CELL  | "Brent olieprijs (USD per barrel)."                                  |    POIL
        CELL  | "Totale beroepsbevolking (jaargemiddelde)."                          |    NATY
        CELL  | "TFPFHP_"                                                            | TFPFHP_
        ----- | ------------------------------------------------------------------------------
        MODE  |
        FILES |
        DATE  |
        <BLANKLINE>
        nb lines: 27
        nb columns: 2
        language: \'ENGLISH\'
        gridx: \'MAJOR\'
        gridy: \'MAJOR\'
        graph_axis: \'VALUES\'
        graph_alignment: \'LEFT\'
        <BLANKLINE>

        >>> # b) -------- update table --------
        >>> tables["TABLE_CELL_LECS"]               # doctest: +NORMALIZE_WHITESPACE
        DIVIS | 1              |
        TITLE |         "New Table"
        ----- | ----------------------------
        CELL  | ""             |     "#S"
        ----- | ----------------------------
        CELL  | "GOSG:"        |        GOSG
        CELL  | "YDTG:"        |        YDTG
        CELL  | "DTH:"         |         DTH
        CELL  | "DTF:"         |         DTF
        CELL  | "IT:"          |          IT
        CELL  | "YSSG+COTRES:" | YSSG+COTRES
        CELL  | "RIDG:"        |        RIDG
        CELL  | "OCUG:"        |        OCUG
        ----- | ----------------------------
        MODE  |
        FILES |
        DATE  |
        <BLANKLINE>
        nb lines: 16
        nb columns: 2
        language: \'ENGLISH\'
        gridx: \'MAJOR\'
        gridy: \'MAJOR\'
        graph_axis: \'VALUES\'
        graph_alignment: \'LEFT\'
        <BLANKLINE>

        >>> # set graph axis type
        >>> tables["TABLE_CELL_LECS"].graph_axis = TableGraphAxis.SEMILOG
        >>> # print first line
        >>> tables["TABLE_CELL_LECS"][0]
        New Table
        >>> # print last line
        >>> tables["TABLE_CELL_LECS"][-1]
        <DATE>
        >>> # delete last line
        >>> del tables["TABLE_CELL_LECS"][-1]
        >>> # get index of line containing YSSG+COTRES
        >>> index = tables["TABLE_CELL_LECS"].index("YSSG+COTRES")
        >>> index
        9
        >>> tables["TABLE_CELL_LECS"][index]
        (\'"YSSG+COTRES:"\', \'YSSG+COTRES\')
        >>> # get line type
        >>> tables["TABLE_CELL_LECS"][index].line_type
        \'CELL\'
        >>> # get line graph type
        >>> tables["TABLE_CELL_LECS"][index].graph_type
        \'LINE\'
        >>> # know if axis is left
        >>> tables["TABLE_CELL_LECS"][index].axis_left
        True
        >>> # update cells
        >>> # double quotes "    -> STRING cell
        >>> # no double quotes   -> LEC cell
        >>> tables["TABLE_CELL_LECS"][index] = (\'"YSSG:"\', \'YSSG\')
        >>> tables["TABLE_CELL_LECS"][index]
        (\'"YSSG:"\', \'YSSG\')
        >>> # insert a new title line surrounded by two separator lines
        >>> tables["TABLE_CELL_LECS"].insert(index + 1, \'-\')
        >>> tables["TABLE_CELL_LECS"].insert(index + 2, "New Title")
        >>> tables["TABLE_CELL_LECS"].insert(index + 3, \'-\')
        >>> # append a new sepatator line
        >>> tables["TABLE_CELL_LECS"] += \'-\'

        >>> tables["TABLE_CELL_LECS"]                # doctest: +NORMALIZE_WHITESPACE
        DIVIS | 1       |
        TITLE |  "New Table"
        ----- | --------------
        CELL  | ""      | "#S"
        ----- | --------------
        CELL  | "GOSG:" | GOSG
        CELL  | "YDTG:" | YDTG
        CELL  | "DTH:"  |  DTH
        CELL  | "DTF:"  |  DTF
        CELL  | "IT:"   |   IT
        CELL  | "YSSG:" | YSSG
        ----- | --------------
        TITLE |  "New Title"
        ----- | --------------
        CELL  | "RIDG:" | RIDG
        CELL  | "OCUG:" | OCUG
        ----- | --------------
        MODE  |
        FILES |
        ----- | --------------
        <BLANKLINE>
        nb lines: 19
        nb columns: 2
        language: \'ENGLISH\'
        gridx: \'MAJOR\'
        gridy: \'MAJOR\'
        graph_axis: \'SEMILOG\'
        graph_alignment: \'LEFT\'
        <BLANKLINE>

        >>> # c) working on a subset
        >>> # 1) get subset
        >>> tables_subset = tables["C8_*"]
        >>> tables_subset.names
        [\'C8_1\', \'C8_10\', \'C8_11\', \'C8_13\', \'C8_14\', \'C8_2\', \'C8_3\', \'C8_4\', \'C8_5\', \'C8_6\', \'C8_7\', \'C8_8\', \'C8_9\']
        >>> # 2) add a table to the subset 
        >>> vars_list = ["XNATY", "XPOIL", "XPWMAB", "XPWXAB"]
        >>> tables_subset["X_GRT"] = {"nb_columns": 2, "table_title": "Croissance", "lecs_or_vars": vars_list, 
        ...                           "mode": True, "files": True, "date": True}  
        >>> tables_subset["X_GRT"]                      # doctest: +NORMALIZE_WHITESPACE
        DIVIS | 1                                                       |
        TITLE |                           "Croissance"
        ----- | ----------------------------------------------------------------
        CELL  | ""                                                      |  "#S"
        ----- | ----------------------------------------------------------------
        CELL  | "Croissance de la population active"                    |  XNATY
        CELL  | "Croissance du prix du pétrole"                         |  XPOIL
        CELL  | "Croissance des prix des biens importés"                | XPWMAB
        CELL  | "Croissance des prix des marchés pertinents à l\'export" | XPWXAB
        ----- | ----------------------------------------------------------------
        MODE  |
        FILES |
        DATE  |
        <BLANKLINE>
        nb lines: 12
        nb columns: 2
        language: \'ENGLISH\'
        gridx: \'MAJOR\'
        gridy: \'MAJOR\'
        graph_axis: \'VALUES\'
        graph_alignment: \'LEFT\'
        <BLANKLINE>
        >>> # --> new table also appears in the global workspace
        >>> "X_GRT" in tables
        True
        >>> tables["X_GRT"]                             # doctest: +NORMALIZE_WHITESPACE
        DIVIS | 1                                                       |
        TITLE |                           "Croissance"
        ----- | ----------------------------------------------------------------
        CELL  | ""                                                      |  "#S"
        ----- | ----------------------------------------------------------------
        CELL  | "Croissance de la population active"                    |  XNATY
        CELL  | "Croissance du prix du pétrole"                         |  XPOIL
        CELL  | "Croissance des prix des biens importés"                | XPWMAB
        CELL  | "Croissance des prix des marchés pertinents à l\'export" | XPWXAB
        ----- | ----------------------------------------------------------------
        MODE  |
        FILES |
        DATE  |
        <BLANKLINE>
        nb lines: 12
        nb columns: 2
        language: \'ENGLISH\'
        gridx: \'MAJOR\'
        gridy: \'MAJOR\'
        graph_axis: \'VALUES\'
        graph_alignment: \'LEFT\'
        <BLANKLINE>
        >>> # 3) update a table in the subset
        >>> table_x_grt = tables_subset["X_GRT"]
        >>> index = table_x_grt.index("XPWXAB")
        >>> table_x_grt.insert(index + 1, (f\'"{comments["XQWXSS"]}"\', "XQWXSS"))
        >>> tables_subset["X_GRT"]                      # doctest: +NORMALIZE_WHITESPACE
        DIVIS | 1                                                       |
        TITLE |                           "Croissance"
        ----- | ----------------------------------------------------------------
        CELL  | ""                                                      |  "#S"
        ----- | ----------------------------------------------------------------
        CELL  | "Croissance de la population active"                    |  XNATY
        CELL  | "Croissance du prix du pétrole"                         |  XPOIL
        CELL  | "Croissance des prix des biens importés"                | XPWMAB
        CELL  | "Croissance des prix des marchés pertinents à l\'export" | XPWXAB
        CELL  | "Croissance des marchés pertinents"                     | XQWXSS
        ----- | ----------------------------------------------------------------
        MODE  |
        FILES |
        DATE  |
        <BLANKLINE>
        nb lines: 13
        nb columns: 2
        language: \'ENGLISH\'
        gridx: \'MAJOR\'
        gridy: \'MAJOR\'
        graph_axis: \'VALUES\'
        graph_alignment: \'LEFT\'
        <BLANKLINE>
        >>> # --> table is also updated in the global workspace
        >>> tables["X_GRT"]                             # doctest: +NORMALIZE_WHITESPACE
        DIVIS | 1                                                       |
        TITLE |                           "Croissance"
        ----- | ----------------------------------------------------------------
        CELL  | ""                                                      |  "#S"
        ----- | ----------------------------------------------------------------
        CELL  | "Croissance de la population active"                    |  XNATY
        CELL  | "Croissance du prix du pétrole"                         |  XPOIL
        CELL  | "Croissance des prix des biens importés"                | XPWMAB
        CELL  | "Croissance des prix des marchés pertinents à l\'export" | XPWXAB
        CELL  | "Croissance des marchés pertinents"                     | XQWXSS
        ----- | ----------------------------------------------------------------
        MODE  |
        FILES |
        DATE  |
        <BLANKLINE>
        nb lines: 13
        nb columns: 2
        language: \'ENGLISH\'
        gridx: \'MAJOR\'
        gridy: \'MAJOR\'
        graph_axis: \'VALUES\'
        graph_alignment: \'LEFT\'
        <BLANKLINE>
        '''

class VarPositionalIndexer:
    def __init__(self, database) -> Any:
        """__init__(self, database)"""
    def __getitem__(self, index: int | tuple[int, int]) -> Any:
        """__getitem__(self, index: Union[int, Tuple[int, int]])"""
    def __setitem__(self, index: int | tuple[int, int], value) -> Any:
        """__setitem__(self, index: Union[int, Tuple[int, int]], value)"""

class Variables(_AbstractDatabase):
    __pyx_vtable__: ClassVar[PyCapsule] = ...
    df: DataFrame
    i: VarPositionalIndexer
    mode: str
    nb_periods: int
    periods: list[str]
    periods_as_float: list[float]
    sample: Sample
    def __init__(self, *args, **kwargs) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""
    def extrapolate(self, method: SimulationInitialization | str, from_period: str | Period = ..., to_period: str | Period = ..., variables_list: str | list[str] = ...) -> Any:
        '''extrapolate(self, method: Union[SimulationInitialization, str], from_period: Union[str, Period] = None, to_period: Union[str, Period] = None, variables_list: Union[str, List[str]] = None)

                Extrapolate variables using one the method described below, based on previous periods.

                The possible methods are as follows:

                  - :math:`Y := Y[-1], if Y null or NA` (TM1) : each null or NA endogen at the start takes the value of 
                    the previous period,
                  - :math:`Y := Y[-1], always` (TM1_A) : each endogen takes the value of the previous period at the start,
                  - :math:`Y := extrapolation, if Y null or NA` (EXTRA) : each null or NA endogen takes as value a linear 
                    extrapolation of the two previous periods,
                  - :math:`Y := extrapolation, always` (EXTRA_A) : each endogen takes as its value a linear extrapolation of 
                    the two preceding periods, whether or not it is zero at the start,
                  - :math:`Y unchanged` (ASIS): endogenous values are not initialized. They retain their value whether or 
                    not they are zero,
                  - :math:`Y := Y[-1], if Y = NA` (TM1_NA): each NA value takes the value of the previous period,
                  - :math:`Y := extrapolation, if Y = NA` (extra_na): each NA value takes the value of a linear extrapolation of 
                    the two previous periods.

                The sample over which the calculation is to be performed must be supplied. 
                This sample must be a sub-sample of the global sample defined in the current Variables database.

                In addition, the list of variables to be extrapolated can be specified. 
                If this list is not given, all variables will be extrapolated.

                Parameters
                ----------
                method: SimulationInitialization or str
                    initialization method. Possible values are:
                      - TM1: :math:`Y := Y[-1], if Y null or NA`
                      - TM1_A: :math:`Y := Y[-1], always`
                      - EXTRA: :math:`Y := extrapolation, if Y null or NA`
                      - EXTRA_A: :math:`Y := extrapolation, always`
                      - ASIS: :math:`Y unchanged`
                      - TM1_NA: :math:`Y := Y[-1], if Y = NA`
                      - EXTRA_NA: :math:`Y := extrapolation, if Y = NA`
                from_period: str or Period, optional
                    starting period to extrapolate variables.
                    Defaults to the starting period if the current sample.
                to_period: str or Period, optional
                    ending period to extrapolate variables.
                    Defaults to the ending period if the current sample.
                variables_list: str or list(str), optional
                    list of variables to extrapolate.
                    Defaults to None (all variables).

                Examples
                --------
                >>> from iode import variables, NA, SimulationInitialization
                >>> variables.clear()
                >>> variables.sample = "2000Y1:2020Y1"

                >>> def reset_ACAF():
                ...     variables["ACAF"] = "t"
                ...     variables["ACAF", ["2005Y1", "2007Y1"]] = NA
                >>> # create ACAF
                >>> reset_ACAF()
                >>> variables["ACAF", :"2010Y1"]
                [0.0, 1.0, 2.0, 3.0, 4.0, -2e+37, 6.0, -2e+37, 8.0, 9.0, 10.0]

                >>> # "TM1" (Y := Y[-1], if Y null or NA)
                >>> reset_ACAF()
                >>> variables.extrapolate(SimulationInitialization.TM1, "2005Y1", "2010Y1")
                >>> variables["ACAF", "2003Y1":"2009Y1"]
                [3.0, 4.0, 4.0, 6.0, 6.0, 8.0, 9.0]

                >>> # "TM1_A" (Y := Y[-1], always)
                >>> reset_ACAF()
                >>> variables.extrapolate(SimulationInitialization.TM1_A, "2005Y1", "2010Y1")
                >>> variables["ACAF", "2003Y1":"2009Y1"]
                [3.0, 4.0, 4.0, 4.0, 4.0, 4.0, 4.0]

                >>> # "EXTRA" (Y := extrapolation, if Y null or NA)
                >>> reset_ACAF()
                >>> variables.extrapolate(SimulationInitialization.EXTRA, "2005Y1", "2010Y1")
                >>> variables["ACAF", "2003Y1":"2009Y1"]
                [3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]

                >>> # "EXTRA_A" (Y := extrapolation, always)
                >>> reset_ACAF()
                >>> variables.extrapolate(SimulationInitialization.EXTRA_A, "2005Y1", "2010Y1")
                >>> variables["ACAF", "2003Y1":"2009Y1"]
                [3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]

                >>> # "ASIS" (Y unchanged)
                >>> reset_ACAF()
                >>> variables.extrapolate(SimulationInitialization.ASIS, "2005Y1", "2010Y1")
                >>> variables["ACAF", "2003Y1":"2009Y1"]
                [3.0, 4.0, -2e+37, 6.0, -2e+37, 8.0, 9.0]

                >>> # "TM1_NA" (Y := Y[-1], if Y = NA)
                >>> reset_ACAF()
                >>> variables.extrapolate(SimulationInitialization.TM1_NA, "2005Y1", "2010Y1")
                >>> variables["ACAF", "2003Y1":"2009Y1"]
                [3.0, 4.0, 4.0, 6.0, 6.0, 8.0, 9.0]

                >>> # "EXTRA_NA" (Y := extrapolation, if Y = NA)
                >>> reset_ACAF()
                >>> variables.extrapolate(SimulationInitialization.EXTRA_NA, "2005Y1", "2010Y1")
                >>> variables["ACAF", "2003Y1":"2009Y1"]
                [3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]
        '''
    @overload
    def from_array(self, array: Array, time_axis_name: str = ..., sep: str = ...) -> Any:
        '''from_array(self, array: Array, time_axis_name: str = u\'time\', sep: str = u\'_\')

                Copies the Array array into the IODE Variables database.
        
                If the passed array has more than two dimensions, the non-time axes are grouped 
                (with \'sep\' as separator) to form the Variables names.

                Parameters
                ----------
                array: Array
                    Input Array object.
                time_axis_name: str, optional
                    Name of the time axis in array.
                    Assumed to be \'time\' by default.
                sep: str, optional
                    If the \'array\' as more than two axes, the separator 
                    \'sep\' is used to group labels of the non-time axes.

                Warnings
                --------
                IODE and LArray don\'t use the same constant to represent NaN values.
                When loading a LArray Array into the Variables database, the LArray 
                NaN values (:math:`nan`) are converted to IODE NaN values (:math:`NA`).

                Examples
                --------
                >>> from iode import variables
                >>> import larray as la
                >>> import numpy as np
                >>> variables.clear()
                >>> len(variables)
                0

                >>> regions_axis = la.Axis("region=VLA,WAL,BXL")
                >>> code_axis = la.Axis("code=00..02")
                >>> periods_axis = la.Axis("time=1960Y1..1970Y1")
                >>> array = la.ndtest((regions_axis, code_axis, periods_axis), dtype=float)
                >>> array
                region  code\\time  1960Y1  1961Y1  1962Y1  ...  1968Y1  1969Y1  1970Y1
                   VLA         00     0.0     1.0     2.0  ...     8.0     9.0    10.0
                   VLA         01    11.0    12.0    13.0  ...    19.0    20.0    21.0
                   VLA         02    22.0    23.0    24.0  ...    30.0    31.0    32.0
                   WAL         00    33.0    34.0    35.0  ...    41.0    42.0    43.0
                   WAL         01    44.0    45.0    46.0  ...    52.0    53.0    54.0
                   WAL         02    55.0    56.0    57.0  ...    63.0    64.0    65.0
                   BXL         00    66.0    67.0    68.0  ...    74.0    75.0    76.0
                   BXL         01    77.0    78.0    79.0  ...    85.0    86.0    87.0
                   BXL         02    88.0    89.0    90.0  ...    96.0    97.0    98.0
        
                >>> # load the IODE Variables from the Array object
                >>> variables.from_array(array)
                >>> len(variables)
                9
                >>> variables.names
                [\'BXL_00\', \'BXL_01\', \'BXL_02\', \'VLA_00\', \'VLA_01\', \'VLA_02\', \'WAL_00\', \'WAL_01\', \'WAL_02\']
                >>> variables.sample
                Sample("1960Y1:1970Y1")
                >>> variables["VLA_00"]
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]
                >>> variables["BXL_02"]
                [88.0, 89.0, 90.0, 91.0, 92.0, 93.0, 94.0, 95.0, 96.0, 97.0, 98.0]
        '''
    @overload
    def from_array(cls, array) -> Any:
        '''from_array(self, array: Array, time_axis_name: str = u\'time\', sep: str = u\'_\')

                Copies the Array array into the IODE Variables database.
        
                If the passed array has more than two dimensions, the non-time axes are grouped 
                (with \'sep\' as separator) to form the Variables names.

                Parameters
                ----------
                array: Array
                    Input Array object.
                time_axis_name: str, optional
                    Name of the time axis in array.
                    Assumed to be \'time\' by default.
                sep: str, optional
                    If the \'array\' as more than two axes, the separator 
                    \'sep\' is used to group labels of the non-time axes.

                Warnings
                --------
                IODE and LArray don\'t use the same constant to represent NaN values.
                When loading a LArray Array into the Variables database, the LArray 
                NaN values (:math:`nan`) are converted to IODE NaN values (:math:`NA`).

                Examples
                --------
                >>> from iode import variables
                >>> import larray as la
                >>> import numpy as np
                >>> variables.clear()
                >>> len(variables)
                0

                >>> regions_axis = la.Axis("region=VLA,WAL,BXL")
                >>> code_axis = la.Axis("code=00..02")
                >>> periods_axis = la.Axis("time=1960Y1..1970Y1")
                >>> array = la.ndtest((regions_axis, code_axis, periods_axis), dtype=float)
                >>> array
                region  code\\time  1960Y1  1961Y1  1962Y1  ...  1968Y1  1969Y1  1970Y1
                   VLA         00     0.0     1.0     2.0  ...     8.0     9.0    10.0
                   VLA         01    11.0    12.0    13.0  ...    19.0    20.0    21.0
                   VLA         02    22.0    23.0    24.0  ...    30.0    31.0    32.0
                   WAL         00    33.0    34.0    35.0  ...    41.0    42.0    43.0
                   WAL         01    44.0    45.0    46.0  ...    52.0    53.0    54.0
                   WAL         02    55.0    56.0    57.0  ...    63.0    64.0    65.0
                   BXL         00    66.0    67.0    68.0  ...    74.0    75.0    76.0
                   BXL         01    77.0    78.0    79.0  ...    85.0    86.0    87.0
                   BXL         02    88.0    89.0    90.0  ...    96.0    97.0    98.0
        
                >>> # load the IODE Variables from the Array object
                >>> variables.from_array(array)
                >>> len(variables)
                9
                >>> variables.names
                [\'BXL_00\', \'BXL_01\', \'BXL_02\', \'VLA_00\', \'VLA_01\', \'VLA_02\', \'WAL_00\', \'WAL_01\', \'WAL_02\']
                >>> variables.sample
                Sample("1960Y1:1970Y1")
                >>> variables["VLA_00"]
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]
                >>> variables["BXL_02"]
                [88.0, 89.0, 90.0, 91.0, 92.0, 93.0, 94.0, 95.0, 96.0, 97.0, 98.0]
        '''
    @overload
    def from_frame(self, df: DataFrame) -> Any:
        '''from_frame(self, df: DataFrame)

                Copy the pandas DataFrame `df` into the IODE Variables database.
                The variable names to copy are deduced from the index of the DataFrame.
                The column names must match the sub-periods of the current Variables sample.

                Parameters
                ----------
                df: DataFrame
                    pandas DataFrame containing the variables to copy into the IODE Variables database.

                Notes
                -----
                The index of the passed DataFrame is sorted in alphabetical order before 
                copying to IODE Variables database.

                Warnings
                --------
                IODE and pandas don\'t use the same constant to represent NaN values.
                When loading a pandas DataFrame into the Variables database, the pandas 
                NaN values (:math:`nan`) are converted to IODE NaN values (:math:`NA`).

                See Also
                --------
                Variables.to_frame

                Examples
                --------
                >>> from iode import variables
                >>> import numpy as np
                >>> import pandas as pd
                >>> variables.clear()
                >>> len(variables)
                0

                >>> # create the pandas DataFrame
                >>> vars_names = [f"{region}_{code}" for region in ["VLA", "WAL", "BXL"] for code in ["00", "01", "02"]]
                >>> periods_list = [f"{i}Y1" for i in range(1960, 1971)]
                >>> data = np.arange(len(vars_names) * len(periods_list), dtype=float).reshape(len(vars_names), len(periods_list))
                >>> df = pd.DataFrame(index=vars_names, columns=periods_list, data=data)
                >>> # display the dataframe
                >>> df          # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                        1960Y1  1961Y1  1962Y1  1963Y1  ...  1967Y1  1968Y1  1969Y1  1970Y1
                VLA_00     0.0     1.0     2.0     3.0  ...     7.0     8.0     9.0    10.0
                VLA_01    11.0    12.0    13.0    14.0  ...    18.0    19.0    20.0    21.0
                VLA_02    22.0    23.0    24.0    25.0  ...    29.0    30.0    31.0    32.0
                WAL_00    33.0    34.0    35.0    36.0  ...    40.0    41.0    42.0    43.0
                WAL_01    44.0    45.0    46.0    47.0  ...    51.0    52.0    53.0    54.0
                WAL_02    55.0    56.0    57.0    58.0  ...    62.0    63.0    64.0    65.0
                BXL_00    66.0    67.0    68.0    69.0  ...    73.0    74.0    75.0    76.0
                BXL_01    77.0    78.0    79.0    80.0  ...    84.0    85.0    86.0    87.0
                BXL_02    88.0    89.0    90.0    91.0  ...    95.0    96.0    97.0    98.0
                <BLANKLINE>
                [9 rows x 11 columns]

                >>> # load into the IODE Variables database
                >>> variables.from_frame(df)
                >>> len(variables)
                9
                >>> variables.names
                [\'BXL_00\', \'BXL_01\', \'BXL_02\', \'VLA_00\', \'VLA_01\', \'VLA_02\', \'WAL_00\', \'WAL_01\', \'WAL_02\']
                >>> variables.sample
                Sample("1960Y1:1970Y1")
                >>> variables["VLA_00"]
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]
                >>> variables["BXL_02"]
                [88.0, 89.0, 90.0, 91.0, 92.0, 93.0, 94.0, 95.0, 96.0, 97.0, 98.0]
        '''
    @overload
    def from_frame(cls, df) -> Any:
        '''from_frame(self, df: DataFrame)

                Copy the pandas DataFrame `df` into the IODE Variables database.
                The variable names to copy are deduced from the index of the DataFrame.
                The column names must match the sub-periods of the current Variables sample.

                Parameters
                ----------
                df: DataFrame
                    pandas DataFrame containing the variables to copy into the IODE Variables database.

                Notes
                -----
                The index of the passed DataFrame is sorted in alphabetical order before 
                copying to IODE Variables database.

                Warnings
                --------
                IODE and pandas don\'t use the same constant to represent NaN values.
                When loading a pandas DataFrame into the Variables database, the pandas 
                NaN values (:math:`nan`) are converted to IODE NaN values (:math:`NA`).

                See Also
                --------
                Variables.to_frame

                Examples
                --------
                >>> from iode import variables
                >>> import numpy as np
                >>> import pandas as pd
                >>> variables.clear()
                >>> len(variables)
                0

                >>> # create the pandas DataFrame
                >>> vars_names = [f"{region}_{code}" for region in ["VLA", "WAL", "BXL"] for code in ["00", "01", "02"]]
                >>> periods_list = [f"{i}Y1" for i in range(1960, 1971)]
                >>> data = np.arange(len(vars_names) * len(periods_list), dtype=float).reshape(len(vars_names), len(periods_list))
                >>> df = pd.DataFrame(index=vars_names, columns=periods_list, data=data)
                >>> # display the dataframe
                >>> df          # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                        1960Y1  1961Y1  1962Y1  1963Y1  ...  1967Y1  1968Y1  1969Y1  1970Y1
                VLA_00     0.0     1.0     2.0     3.0  ...     7.0     8.0     9.0    10.0
                VLA_01    11.0    12.0    13.0    14.0  ...    18.0    19.0    20.0    21.0
                VLA_02    22.0    23.0    24.0    25.0  ...    29.0    30.0    31.0    32.0
                WAL_00    33.0    34.0    35.0    36.0  ...    40.0    41.0    42.0    43.0
                WAL_01    44.0    45.0    46.0    47.0  ...    51.0    52.0    53.0    54.0
                WAL_02    55.0    56.0    57.0    58.0  ...    62.0    63.0    64.0    65.0
                BXL_00    66.0    67.0    68.0    69.0  ...    73.0    74.0    75.0    76.0
                BXL_01    77.0    78.0    79.0    80.0  ...    84.0    85.0    86.0    87.0
                BXL_02    88.0    89.0    90.0    91.0  ...    95.0    96.0    97.0    98.0
                <BLANKLINE>
                [9 rows x 11 columns]

                >>> # load into the IODE Variables database
                >>> variables.from_frame(df)
                >>> len(variables)
                9
                >>> variables.names
                [\'BXL_00\', \'BXL_01\', \'BXL_02\', \'VLA_00\', \'VLA_01\', \'VLA_02\', \'WAL_00\', \'WAL_01\', \'WAL_02\']
                >>> variables.sample
                Sample("1960Y1:1970Y1")
                >>> variables["VLA_00"]
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]
                >>> variables["BXL_02"]
                [88.0, 89.0, 90.0, 91.0, 92.0, 93.0, 94.0, 95.0, 96.0, 97.0, 98.0]
        '''
    def high_to_low(self, type_of_series: HighToLowType | str, filepath: str | Path, var_list: str | list[str]) -> Any:
        '''high_to_low(self, type_of_series: Union[HighToLowType, str], filepath: Union[str, Path], var_list: Union[str, List[str]])

                Build series of lower periodicity by (summing the / taking the average of the / taking the 
                last observation of) sub-periods.

                The list of specified series (variables) from the input file are loaded into the current 
                Variables database and the periodicity of these series (variables) is modified simultaneously. 
                The new periodicity is the one currently defined in the current Variables database.

                The loaded series are added to or replace those (for existing names) in the current 
                Variables database.

                This procedure exists for the following cases:

                    - monthly to annual 
                      (annual observation = sum of 12 months / average of 12 months / December value) 
                    - quarterly to annual 
                      (annual observation = sum of 4 quarters / average of 4 quarters / last quarter value) 
                    - monthly to quarterly 
                      (quarterly observation = sum of 3 months / average of 3 months / value for the last month of the quarter) 

                Three types of series are available:

                    - SIM: addition of sub-periods data
                    - MEAN: average of sub-periods data
                    - LAST: last observation

                In the case of a non-existent value (:math:`NA`) for one of the sub-periods, the result is :math:`NA`.

                Parameters
                ----------
                type_of_series : HighToLowType or str
                    Three types of series are available: 
                        - SUM : addition of sub-period data
                        - MEAN : average of sub-period data
                        - LAST : last observation

                filepath : str or Path
                    Filepath to the source data file.

                var_list : str or list(str)
                    List of variables to include in the transformation.

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, variables, HighToLowType
                >>> variables.clear()
                >>> # define a yearly sample
                >>> variables.sample = "2000Y1:2020Y1"
                >>> # input filepath
                >>> filepath = f"{SAMPLE_DATA_DIR}/fun_q.var"
        
                Last Obs in year
        
                >>> variables.high_to_low(HighToLowType.LAST, filepath, ["ACAF", "ACAG"])
                >>> variables["ACAF", "2010Y1":"2014Y1"]
                [-37.82742883229439, -44.544792633543224, -55.55928981721873, -68.89465432262006, -83.34062511080091]
                >>> variables["ACAG", "2010Y1":"2014Y1"]
                [7.063482244552621, 7.321150091008727, 7.580990287577893, 7.842534702673841, 8.1050747072996]

                Mean of year
        
                >>> variables.high_to_low(HighToLowType.MEAN, filepath, ["ACAF", "ACAG"])
                >>> variables["ACAF", "2010Y1":"2014Y1"]
                [-37.82742883229439, -44.544792633543224, -55.55928981721873, -68.89465432262006, -83.34062511080091]
                >>> variables["ACAG", "2010Y1":"2014Y1"]
                [7.063482244552621, 7.321150091008727, 7.580990287577893, 7.842534702673841, 8.1050747072996]

                Sum
        
                >>> variables.high_to_low(HighToLowType.SUM, filepath, ["ACAF", "ACAG"])
                >>> variables["ACAF", "2010Y1":"2014Y1"]
                [-151.30971532917755, -178.1791705341729, -222.23715926887493, -275.5786172904802, -333.36250044320366]
                >>> variables["ACAG", "2010Y1":"2014Y1"]
                [28.253928978210485, 29.284600364034908, 30.323961150311572, 31.370138810695362, 32.4202988291984]
        '''
    def low_to_high(self, type_of_series: LowToHighType | str, method: LowToHighMethod | str, filepath: str | Path, var_list: str | list[str]) -> Any:
        '''low_to_high(self, type_of_series: Union[LowToHighType, str], method: Union[LowToHighMethod, str], filepath: Union[str, Path], var_list: Union[str, List[str]])

                Build series with higher periodicity for *stock data* (Unemployment, Debt, ...) or 
                *flow data* (GNP, Deficit, ...).

                The list of specified series (variables) from the input file are loaded into the current 
                Variables database and the periodicity of these series (variables) is modified simultaneously. 
                The new periodicity is the one currently defined in the current Variables database.

                The loaded series are added to or replace those (for existing names) in the current 
                Variables database.

                This procedure exists for the following cases:

                    - annual to monthly 
                    - annual to quarterly 
                    - quarterly to monthly 
        
                Two types of series are available, one for stocks (STOCK), the other for flows (FLOW).

                Three interpolation methods are available:

                    - linear (LINEAR): A[1980Q{1,2,3,4}] = A[1979Y1] + i * (A[1980Y1] - A[1979Y1])/4 i = 1,2,3,4 
                    - cubic splines (CUBIC_SPLINES): cubic interpolation 
                    - step (STEP) : A[1980Q{1,2,3,4}] = A[1980Y1] 

                Parameters
                ----------
                type_of_series : LowToHighType or str
                    Two types of series are considered: \'stock\' and \'flow\':
                        - STOCK: stock data (Unemployment, Debt, ...)
                        - FLOW: flow data (GNP, Deficit, ...)

                method : LowToHighMethod or str
                    Method to use for transformation. Three methods can be used:
                        - LINEAR (\'L\'): Linear interpolation
                        - CUBIC_SPLINES (\'C\'): Cubic Spliness
                        - STEP (\'S\'): Step

                filepath : str or Path
                    Filepath to the source data file.

                var_list : str or list(str)
                    List of variables to include in the transformation.

                Returns
                -------
                None

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR, variables, LowToHighType, LowToHighMethod
                >>> variables.clear()
                >>> # define a yearly sample
                >>> variables.sample = "2010Q1:2020Q4"
                >>> # input filepath
                >>> filepath = f"{SAMPLE_DATA_DIR}/fun.var"
        
                Linear interpolation / stock
        
                >>> # "stock" -> the result is a linear interpolation of the 2 surrounding source values.
                >>> variables.low_to_high(LowToHighType.STOCK, LowToHighMethod.LINEAR, filepath, ["ACAF", "ACAG"])
                >>> variables["ACAF", "2014Q1":"2014Q4"]
                [-72.50614701966526, -76.11763971671049, -79.7291324137557, -83.34062511080091]
                >>> variables["ACAG", "2014Q1":"2014Q4"]
                [31.63267881532112, 31.895218819946884, 32.15775882457264, 32.4202988291984]

                Linear interpolation / flow
        
                >>> # "flow" -> the result is the source value divided by the nb of sub-periods. 
                >>> variables.low_to_high(LowToHighType.FLOW, LowToHighMethod.LINEAR, filepath, ["ACAF", "ACAG"])
                >>> variables["ACAF", "2014Q1":"2014Q4"]
                [-20.83515627770023, -20.83515627770023, -20.83515627770023, -20.83515627770023]
                >>> variables["ACAG", "2014Q1":"2014Q4"]
                [8.1050747072996, 8.1050747072996, 8.1050747072996, 8.1050747072996]

                Cubic splines / stock
        
                >>> variables.low_to_high(LowToHighType.STOCK, LowToHighMethod.CUBIC_SPLINES, filepath, ["ACAF", "ACAG"])
                >>> variables["ACAF", "2012Q1":"2012Q4"]
                [-47.2984169294621, -50.052041225380975, -52.80566552129986, -55.55928981721873]
                >>> variables["ACAG", "2012Q1":"2012Q4"]
                [29.544440560604077, 29.804280757173238, 30.064120953742403, 30.323961150311572]

                Cubic splines / flow
        
                >>> variables.low_to_high(LowToHighType.FLOW, LowToHighMethod.CUBIC_SPLINES, filepath, ["ACAF", "ACAG"])
                >>> variables["ACAF", "2012Q1":"2012Q4"]
                [-12.748422687629207, -13.436828761608925, -14.270289043196508, -15.103749324784092]
                >>> variables["ACAG", "2012Q1":"2012Q4"]
                [7.483230672890683, 7.548190722032976, 7.6135768258069625, 7.678962929580949]

                Step / stock
        
                >>> # "stock" -> the result has the same value as the source
                >>> variables.low_to_high(LowToHighType.STOCK, LowToHighMethod.STEP, filepath, ["ACAF", "ACAG"])
                >>> variables["ACAF", "2014Q1":"2014Q4"]
                [-83.34062511080091, -83.34062511080091, -83.34062511080091, -83.34062511080091]
                >>> variables["ACAG", "2014Q1":"2014Q4"]
                [32.4202988291984, 32.4202988291984, 32.4202988291984, 32.4202988291984]

                Step / flow
        
                >>> # "flow" -> the result is the source value plus a portion of 
                >>> # the difference between the 2 surrounding values in the source
                >>> variables.low_to_high(LowToHighType.FLOW, LowToHighMethod.STEP, filepath, ["ACAF", "ACAG"])
                >>> variables["ACAF", "2014Q1":"2014Q4"]
                [-20.83515627770023, -20.83515627770023, -20.83515627770023, -20.83515627770023]
                >>> variables["ACAG", "2014Q1":"2014Q4"]
                [8.1050747072996, 8.1050747072996, 8.1050747072996, 8.1050747072996]
        '''
    def periods_subset(self, from_period: str | Period = ..., to_period: str | Period = ..., as_float: bool = ...) -> Any:
        '''periods_subset(self, from_period: Union[str, Period] = None, to_period: Union[str, Period] = None, as_float: bool = False)

                Return a subset of the periods from the current Variables sample.

                Parameters
                ----------
                from_period: str or Period, optional
                    first period of the subset of periods.
                    Defaults to the first period of the current Variables sample.
                to_period: str or Period, optional
                    last period of the subset of periods.
                    Defaults to the last period of the current Variables sample.
                as_float: bool, optional
                    whether or not to return periods as float.
                    Defaults to False.

                Returns
                -------
                list(str) or list(float)

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import variables
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> variables.sample
                Sample("1960Y1:2015Y1")
                >>> variables.periods_subset("1990Y1", "2000Y1")                    # doctest: +ELLIPSIS
                [\'1990Y1\', \'1991Y1\', ..., \'1999Y1\', \'2000Y1\']
                >>> variables.periods_subset("1990Y1", "2000Y1", as_float=True)     # doctest: +ELLIPSIS
                [1990.0, 1991.0, ..., 1999.0, 2000.0]
        '''
    def seasonal_adjustment(self, input_file: str, eps_test: float = ..., series: str | list[str] = ...) -> Any:
        """seasonal_adjustment(self, input_file: str, eps_test: float = 5.0, series: Union[str, List[str]] = None)

                Eliminate seasonal variations in monthly series (= variables).

                The method constructs deseasonalized series using the *Census XI method*, as well as cyclical 
                and stochastic trend components. Thus, In addition to the deseasonalized series (named after 
                the series in the input file), the result is two other series:

                    - the series containing the cyclical trend component named : `_C<name>` where *name* is the original name
                    - the series containing the stochastic component called : `_I<name>` where *name* is the original name

                Note that the deseasonalized series is the product of the other two.

                To do this, the list of *series* (= variables) are loaded from the specified file *input_file* 
                into the current IODE Variables database and simultaneously modifies the selected *series* if necessary.

                The new *series* are added to or replace (for existing names) those in the current IODE Variables database.

                In the case of non-existent value (NA) for one of the periods, the result is NA for the whole year.

                Parameters
                ----------
                input_file: str or Path
                    Filepath to the input file.
                eps_test: float, optional
                    Criterion verifying whether a seasonal influence is present in a series
                    Default to 5.0.
                series: str or list(str), optional
                    list of series on which to apply the seasonal adjustment.
                    Defaults to None (all variables).

                See Also
                --------
                Variables.trend_correction
        """
    @overload
    def to_array(self, vars_axis_name: str = ..., time_axis_name: str = ..., sample_as_floats: bool = ...) -> Array:
        '''to_array(self, vars_axis_name: str = u\'names\', time_axis_name: str = u\'time\', sample_as_floats: bool = False) -> Array

                Creates an Array from the current IODE Variables database.
        
                Parameters
                ----------
                vars_axis_name: str, optional
                    Name of the axis representing the Variables names.
                    Defaults to \'names\'.
                time_axis_name: str, optional
                    Name of the axis representing the periods.
                    Defaults to \'time\'.
                sample_as_floats: bool, optional
                    Whether or not to convert periods as float values.
                    Defaults to False (periods are exported as string values).

                Warnings
                --------
                IODE and LArray don\'t use the same constant to represent NaN values.
                When exporting IODE variables as a LArray Array, the IODE NaN values 
                (:math:`NA`) are converted to LArray NaN values (:math:`nan`).

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import variables
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> len(variables)
                394
                >>> variables.sample
                Sample("1960Y1:2015Y1")
                >>> variables.nb_periods
                56

                >>> # Export the IODE Variables database as an (larray) Array object
                >>> array = variables.to_array()
                >>> array.shape
                (394, 56)
                >>> array.axes.info
                394 x 56
                 names [394]: \'ACAF\' \'ACAG\' \'AOUC\' ... \'ZKFO\' \'ZX\' \'ZZF_\'
                 time [56]: \'1960Y1\' \'1961Y1\' \'1962Y1\' ... \'2013Y1\' \'2014Y1\' \'2015Y1\'
                >>> variables["AOUC"]               # doctest: +ELLIPSIS
                [nan, 0.24783191606766575, ..., 1.4237139558484628, 1.4608626117037322]
                >>> array["AOUC"]
                time  1960Y1  ...              2014Y1              2015Y1
                         nan  ...  1.4237139558484628  1.4608626117037322
                >>> variables["ZKFO"]               # doctest: +ELLIPSIS
                [1.0, 1.0, ..., 1.0159901, 1.0159901]
                >>> array["ZKFO"]
                time  1960Y1  1961Y1  1962Y1  ...     2012Y1     2013Y1     2014Y1     2015Y1
                         1.0     1.0     1.0  ...  1.0159901  1.0159901  1.0159901  1.0159901

                >>> # Export a subset of the IODE Variables database as an (larray) Array object
                >>> array = variables["A*;*_"].to_array()
                >>> array.shape
                (33, 56)
                >>> array.axes.info
                33 x 56
                 names [33]: \'ACAF\' \'ACAG\' \'AOUC\' ... \'WNF_\' \'YDH_\' \'ZZF_\'
                 time [56]: \'1960Y1\' \'1961Y1\' \'1962Y1\' ... \'2013Y1\' \'2014Y1\' \'2015Y1\'
                >>> variables["AOUC"]               # doctest: +ELLIPSIS
                [nan, 0.24783191606766575, ..., 1.4237139558484628, 1.4608626117037322]
                >>> array["AOUC"]
                time  1960Y1  ...              2014Y1              2015Y1
                         nan  ...  1.4237139558484628  1.4608626117037322
                >>> variables["ZZF_"]               # doctest: +ELLIPSIS
                [0.68840039, 0.68840039, ..., 0.68840039, 0.68840039]
                >>> array["ZZF_"]
                time      1960Y1      1961Y1  ...      2013Y1      2014Y1      2015Y1
                      0.68840039  0.68840039  ...  0.68840039  0.68840039  0.68840039
        '''
    @overload
    def to_array(cls) -> Any:
        '''to_array(self, vars_axis_name: str = u\'names\', time_axis_name: str = u\'time\', sample_as_floats: bool = False) -> Array

                Creates an Array from the current IODE Variables database.
        
                Parameters
                ----------
                vars_axis_name: str, optional
                    Name of the axis representing the Variables names.
                    Defaults to \'names\'.
                time_axis_name: str, optional
                    Name of the axis representing the periods.
                    Defaults to \'time\'.
                sample_as_floats: bool, optional
                    Whether or not to convert periods as float values.
                    Defaults to False (periods are exported as string values).

                Warnings
                --------
                IODE and LArray don\'t use the same constant to represent NaN values.
                When exporting IODE variables as a LArray Array, the IODE NaN values 
                (:math:`NA`) are converted to LArray NaN values (:math:`nan`).

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import variables
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> len(variables)
                394
                >>> variables.sample
                Sample("1960Y1:2015Y1")
                >>> variables.nb_periods
                56

                >>> # Export the IODE Variables database as an (larray) Array object
                >>> array = variables.to_array()
                >>> array.shape
                (394, 56)
                >>> array.axes.info
                394 x 56
                 names [394]: \'ACAF\' \'ACAG\' \'AOUC\' ... \'ZKFO\' \'ZX\' \'ZZF_\'
                 time [56]: \'1960Y1\' \'1961Y1\' \'1962Y1\' ... \'2013Y1\' \'2014Y1\' \'2015Y1\'
                >>> variables["AOUC"]               # doctest: +ELLIPSIS
                [nan, 0.24783191606766575, ..., 1.4237139558484628, 1.4608626117037322]
                >>> array["AOUC"]
                time  1960Y1  ...              2014Y1              2015Y1
                         nan  ...  1.4237139558484628  1.4608626117037322
                >>> variables["ZKFO"]               # doctest: +ELLIPSIS
                [1.0, 1.0, ..., 1.0159901, 1.0159901]
                >>> array["ZKFO"]
                time  1960Y1  1961Y1  1962Y1  ...     2012Y1     2013Y1     2014Y1     2015Y1
                         1.0     1.0     1.0  ...  1.0159901  1.0159901  1.0159901  1.0159901

                >>> # Export a subset of the IODE Variables database as an (larray) Array object
                >>> array = variables["A*;*_"].to_array()
                >>> array.shape
                (33, 56)
                >>> array.axes.info
                33 x 56
                 names [33]: \'ACAF\' \'ACAG\' \'AOUC\' ... \'WNF_\' \'YDH_\' \'ZZF_\'
                 time [56]: \'1960Y1\' \'1961Y1\' \'1962Y1\' ... \'2013Y1\' \'2014Y1\' \'2015Y1\'
                >>> variables["AOUC"]               # doctest: +ELLIPSIS
                [nan, 0.24783191606766575, ..., 1.4237139558484628, 1.4608626117037322]
                >>> array["AOUC"]
                time  1960Y1  ...              2014Y1              2015Y1
                         nan  ...  1.4237139558484628  1.4608626117037322
                >>> variables["ZZF_"]               # doctest: +ELLIPSIS
                [0.68840039, 0.68840039, ..., 0.68840039, 0.68840039]
                >>> array["ZZF_"]
                time      1960Y1      1961Y1  ...      2013Y1      2014Y1      2015Y1
                      0.68840039  0.68840039  ...  0.68840039  0.68840039  0.68840039
        '''
    @overload
    def to_array(cls) -> Any:
        '''to_array(self, vars_axis_name: str = u\'names\', time_axis_name: str = u\'time\', sample_as_floats: bool = False) -> Array

                Creates an Array from the current IODE Variables database.
        
                Parameters
                ----------
                vars_axis_name: str, optional
                    Name of the axis representing the Variables names.
                    Defaults to \'names\'.
                time_axis_name: str, optional
                    Name of the axis representing the periods.
                    Defaults to \'time\'.
                sample_as_floats: bool, optional
                    Whether or not to convert periods as float values.
                    Defaults to False (periods are exported as string values).

                Warnings
                --------
                IODE and LArray don\'t use the same constant to represent NaN values.
                When exporting IODE variables as a LArray Array, the IODE NaN values 
                (:math:`NA`) are converted to LArray NaN values (:math:`nan`).

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import variables
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> len(variables)
                394
                >>> variables.sample
                Sample("1960Y1:2015Y1")
                >>> variables.nb_periods
                56

                >>> # Export the IODE Variables database as an (larray) Array object
                >>> array = variables.to_array()
                >>> array.shape
                (394, 56)
                >>> array.axes.info
                394 x 56
                 names [394]: \'ACAF\' \'ACAG\' \'AOUC\' ... \'ZKFO\' \'ZX\' \'ZZF_\'
                 time [56]: \'1960Y1\' \'1961Y1\' \'1962Y1\' ... \'2013Y1\' \'2014Y1\' \'2015Y1\'
                >>> variables["AOUC"]               # doctest: +ELLIPSIS
                [nan, 0.24783191606766575, ..., 1.4237139558484628, 1.4608626117037322]
                >>> array["AOUC"]
                time  1960Y1  ...              2014Y1              2015Y1
                         nan  ...  1.4237139558484628  1.4608626117037322
                >>> variables["ZKFO"]               # doctest: +ELLIPSIS
                [1.0, 1.0, ..., 1.0159901, 1.0159901]
                >>> array["ZKFO"]
                time  1960Y1  1961Y1  1962Y1  ...     2012Y1     2013Y1     2014Y1     2015Y1
                         1.0     1.0     1.0  ...  1.0159901  1.0159901  1.0159901  1.0159901

                >>> # Export a subset of the IODE Variables database as an (larray) Array object
                >>> array = variables["A*;*_"].to_array()
                >>> array.shape
                (33, 56)
                >>> array.axes.info
                33 x 56
                 names [33]: \'ACAF\' \'ACAG\' \'AOUC\' ... \'WNF_\' \'YDH_\' \'ZZF_\'
                 time [56]: \'1960Y1\' \'1961Y1\' \'1962Y1\' ... \'2013Y1\' \'2014Y1\' \'2015Y1\'
                >>> variables["AOUC"]               # doctest: +ELLIPSIS
                [nan, 0.24783191606766575, ..., 1.4237139558484628, 1.4608626117037322]
                >>> array["AOUC"]
                time  1960Y1  ...              2014Y1              2015Y1
                         nan  ...  1.4237139558484628  1.4608626117037322
                >>> variables["ZZF_"]               # doctest: +ELLIPSIS
                [0.68840039, 0.68840039, ..., 0.68840039, 0.68840039]
                >>> array["ZZF_"]
                time      1960Y1      1961Y1  ...      2013Y1      2014Y1      2015Y1
                      0.68840039  0.68840039  ...  0.68840039  0.68840039  0.68840039
        '''
    @overload
    def to_frame(self, vars_axis_name: str = ..., time_axis_name: str = ..., sample_as_floats: bool = ...) -> DataFrame:
        '''to_frame(self, vars_axis_name: str = u\'names\', time_axis_name: str = u\'time\', sample_as_floats: bool = False) -> DataFrame

                Create a pandas DataFrame from the current Variables database.
                The index of the returned DataFrame is build from the Variables names 
                and the columns from the periods.

                Parameters
                ----------
                vars_axis_name: str, optional
                    Name of the axis representing the Variables names.
                    Defaults to \'names\'.
                time_axis_name: str, optional
                    Name of the axis representing the periods.
                    Defaults to \'time\'.
                sample_as_floats: bool, optional
                    Whether or not to export periods as string or float values.
                    Defaults to False (periods are exported as string values).

                Warnings
                --------
                IODE and pandas don\'t use the same constant to represent NaN values.
                When exporting IODE variables as a pandas DataFrame, the IODE NaN values 
                (:math:`NA`) are converted to pandas NaN values (:math:`nan`).

                See Also
                --------
                Variables.from_frame

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import variables
                >>> import pandas as pd
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> len(variables)
                394
                >>> variables.sample
                Sample("1960Y1:2015Y1")
                >>> variables.nb_periods
                56

                >>> # Export the IODE Variables database as a pandas DataFrame
                >>> df = variables.to_frame()
                >>> df.shape
                (394, 56)
                >>> df.index.to_list()              # doctest: +ELLIPSIS
                [\'ACAF\', \'ACAG\', \'AOUC\', ..., \'ZKFO\', \'ZX\', \'ZZF_\']
                >>> df.columns.to_list()            # doctest: +ELLIPSIS
                [\'1960Y1\', \'1961Y1\', ..., \'2014Y1\', \'2015Y1\']
                >>> variables["AOUC"]               # doctest: +ELLIPSIS
                [nan, 0.24783191606766575, ..., 1.4237139558484628, 1.4608626117037322]
                >>> df.loc["AOUC"]                  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                time
                1960Y1         NaN
                1961Y1    0.247832
                ...
                2014Y1    1.423714
                2015Y1    1.460863
                Name: AOUC, dtype: float64    
                >>> variables["ZKFO"]               # doctest: +ELLIPSIS
                [1.0, 1.0, ... 1.0159901, 1.0159901]
                >>> df.loc["ZKFO"]                  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                time
                1960Y1    1.00000
                1961Y1    1.00000
                ...
                2014Y1    1.01599
                2015Y1    1.01599
                Name: ZKFO, dtype: float64

                >>> # Export a subset of the IODE Variables database as a pandas DataFrame
                >>> df = variables["A*;*_"].to_frame()
                >>> df.shape
                (33, 56)
                >>> df.index.to_list()              # doctest: +ELLIPSIS
                [\'ACAF\', \'ACAG\', \'AOUC\', ..., \'WNF_\', \'YDH_\', \'ZZF_\']
                >>> df.columns.to_list()            # doctest: +ELLIPSIS
                [\'1960Y1\', \'1961Y1\', ..., \'2014Y1\', \'2015Y1\']
                >>> variables["AOUC"]               # doctest: +ELLIPSIS
                [nan, 0.24783191606766575, ..., 1.4237139558484628, 1.4608626117037322]
                >>> df.loc["AOUC"]                  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                time
                1960Y1         NaN
                1961Y1    0.247832
                ...
                2014Y1    1.423714
                2015Y1    1.460863
                Name: AOUC, dtype: float64    
                >>> variables["ZZF_"]               # doctest: +ELLIPSIS
                [0.68840039, 0.68840039, ..., 0.68840039, 0.68840039]
                >>> df.loc["ZZF_"]                  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                time
                1960Y1    0.6884
                1961Y1    0.6884
                ...
                2014Y1    0.6884
                2015Y1    0.6884
                Name: ZZF_, dtype: float64
        '''
    @overload
    def to_frame(cls) -> Any:
        '''to_frame(self, vars_axis_name: str = u\'names\', time_axis_name: str = u\'time\', sample_as_floats: bool = False) -> DataFrame

                Create a pandas DataFrame from the current Variables database.
                The index of the returned DataFrame is build from the Variables names 
                and the columns from the periods.

                Parameters
                ----------
                vars_axis_name: str, optional
                    Name of the axis representing the Variables names.
                    Defaults to \'names\'.
                time_axis_name: str, optional
                    Name of the axis representing the periods.
                    Defaults to \'time\'.
                sample_as_floats: bool, optional
                    Whether or not to export periods as string or float values.
                    Defaults to False (periods are exported as string values).

                Warnings
                --------
                IODE and pandas don\'t use the same constant to represent NaN values.
                When exporting IODE variables as a pandas DataFrame, the IODE NaN values 
                (:math:`NA`) are converted to pandas NaN values (:math:`nan`).

                See Also
                --------
                Variables.from_frame

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import variables
                >>> import pandas as pd
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> len(variables)
                394
                >>> variables.sample
                Sample("1960Y1:2015Y1")
                >>> variables.nb_periods
                56

                >>> # Export the IODE Variables database as a pandas DataFrame
                >>> df = variables.to_frame()
                >>> df.shape
                (394, 56)
                >>> df.index.to_list()              # doctest: +ELLIPSIS
                [\'ACAF\', \'ACAG\', \'AOUC\', ..., \'ZKFO\', \'ZX\', \'ZZF_\']
                >>> df.columns.to_list()            # doctest: +ELLIPSIS
                [\'1960Y1\', \'1961Y1\', ..., \'2014Y1\', \'2015Y1\']
                >>> variables["AOUC"]               # doctest: +ELLIPSIS
                [nan, 0.24783191606766575, ..., 1.4237139558484628, 1.4608626117037322]
                >>> df.loc["AOUC"]                  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                time
                1960Y1         NaN
                1961Y1    0.247832
                ...
                2014Y1    1.423714
                2015Y1    1.460863
                Name: AOUC, dtype: float64    
                >>> variables["ZKFO"]               # doctest: +ELLIPSIS
                [1.0, 1.0, ... 1.0159901, 1.0159901]
                >>> df.loc["ZKFO"]                  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                time
                1960Y1    1.00000
                1961Y1    1.00000
                ...
                2014Y1    1.01599
                2015Y1    1.01599
                Name: ZKFO, dtype: float64

                >>> # Export a subset of the IODE Variables database as a pandas DataFrame
                >>> df = variables["A*;*_"].to_frame()
                >>> df.shape
                (33, 56)
                >>> df.index.to_list()              # doctest: +ELLIPSIS
                [\'ACAF\', \'ACAG\', \'AOUC\', ..., \'WNF_\', \'YDH_\', \'ZZF_\']
                >>> df.columns.to_list()            # doctest: +ELLIPSIS
                [\'1960Y1\', \'1961Y1\', ..., \'2014Y1\', \'2015Y1\']
                >>> variables["AOUC"]               # doctest: +ELLIPSIS
                [nan, 0.24783191606766575, ..., 1.4237139558484628, 1.4608626117037322]
                >>> df.loc["AOUC"]                  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                time
                1960Y1         NaN
                1961Y1    0.247832
                ...
                2014Y1    1.423714
                2015Y1    1.460863
                Name: AOUC, dtype: float64    
                >>> variables["ZZF_"]               # doctest: +ELLIPSIS
                [0.68840039, 0.68840039, ..., 0.68840039, 0.68840039]
                >>> df.loc["ZZF_"]                  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                time
                1960Y1    0.6884
                1961Y1    0.6884
                ...
                2014Y1    0.6884
                2015Y1    0.6884
                Name: ZZF_, dtype: float64
        '''
    @overload
    def to_frame(cls) -> Any:
        '''to_frame(self, vars_axis_name: str = u\'names\', time_axis_name: str = u\'time\', sample_as_floats: bool = False) -> DataFrame

                Create a pandas DataFrame from the current Variables database.
                The index of the returned DataFrame is build from the Variables names 
                and the columns from the periods.

                Parameters
                ----------
                vars_axis_name: str, optional
                    Name of the axis representing the Variables names.
                    Defaults to \'names\'.
                time_axis_name: str, optional
                    Name of the axis representing the periods.
                    Defaults to \'time\'.
                sample_as_floats: bool, optional
                    Whether or not to export periods as string or float values.
                    Defaults to False (periods are exported as string values).

                Warnings
                --------
                IODE and pandas don\'t use the same constant to represent NaN values.
                When exporting IODE variables as a pandas DataFrame, the IODE NaN values 
                (:math:`NA`) are converted to pandas NaN values (:math:`nan`).

                See Also
                --------
                Variables.from_frame

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import variables
                >>> import pandas as pd
                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> len(variables)
                394
                >>> variables.sample
                Sample("1960Y1:2015Y1")
                >>> variables.nb_periods
                56

                >>> # Export the IODE Variables database as a pandas DataFrame
                >>> df = variables.to_frame()
                >>> df.shape
                (394, 56)
                >>> df.index.to_list()              # doctest: +ELLIPSIS
                [\'ACAF\', \'ACAG\', \'AOUC\', ..., \'ZKFO\', \'ZX\', \'ZZF_\']
                >>> df.columns.to_list()            # doctest: +ELLIPSIS
                [\'1960Y1\', \'1961Y1\', ..., \'2014Y1\', \'2015Y1\']
                >>> variables["AOUC"]               # doctest: +ELLIPSIS
                [nan, 0.24783191606766575, ..., 1.4237139558484628, 1.4608626117037322]
                >>> df.loc["AOUC"]                  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                time
                1960Y1         NaN
                1961Y1    0.247832
                ...
                2014Y1    1.423714
                2015Y1    1.460863
                Name: AOUC, dtype: float64    
                >>> variables["ZKFO"]               # doctest: +ELLIPSIS
                [1.0, 1.0, ... 1.0159901, 1.0159901]
                >>> df.loc["ZKFO"]                  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                time
                1960Y1    1.00000
                1961Y1    1.00000
                ...
                2014Y1    1.01599
                2015Y1    1.01599
                Name: ZKFO, dtype: float64

                >>> # Export a subset of the IODE Variables database as a pandas DataFrame
                >>> df = variables["A*;*_"].to_frame()
                >>> df.shape
                (33, 56)
                >>> df.index.to_list()              # doctest: +ELLIPSIS
                [\'ACAF\', \'ACAG\', \'AOUC\', ..., \'WNF_\', \'YDH_\', \'ZZF_\']
                >>> df.columns.to_list()            # doctest: +ELLIPSIS
                [\'1960Y1\', \'1961Y1\', ..., \'2014Y1\', \'2015Y1\']
                >>> variables["AOUC"]               # doctest: +ELLIPSIS
                [nan, 0.24783191606766575, ..., 1.4237139558484628, 1.4608626117037322]
                >>> df.loc["AOUC"]                  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                time
                1960Y1         NaN
                1961Y1    0.247832
                ...
                2014Y1    1.423714
                2015Y1    1.460863
                Name: AOUC, dtype: float64    
                >>> variables["ZZF_"]               # doctest: +ELLIPSIS
                [0.68840039, 0.68840039, ..., 0.68840039, 0.68840039]
                >>> df.loc["ZZF_"]                  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
                time
                1960Y1    0.6884
                1961Y1    0.6884
                ...
                2014Y1    0.6884
                2015Y1    0.6884
                Name: ZZF_, dtype: float64
        '''
    def trend_correction(self, input_file: str, lambda_: float, series: str | list[str] = ..., log: bool = ...) -> Any:
        """trend_correction(self, input_file: str, lambda_: float, series: Union[str, List[str]] = None, log: bool = False)

                Implementation of the *Hodrick-Prescott method* for trend series (= variables) construction. 
                The principle is the same as for deseasonalization: series read from a file are imported and transformed simultaneously.

                If the `log` flag is set, the series are log-transformed before calculation and exp-transformed after calculation. 
                The series values must therefore be strictly positive.

                Parameters
                ----------
                input_file: str
                    filepath to the input file.
                lambda_: float
                    Lambda parameter of the Hodrick-Prescott method.
                series: str or list(str), optional
                    list of series on which to apply the trend correction.
                    Defaults to None (all variables).
                log: bool, optional
                    Whether or not the series are log-transformed before calculation and exp-transformed 
                    after calculation. The series values must therefore be strictly positive.
                    Defaults to False.

                See Also
                --------
                Variables.seasonal_adjustment
        """
    def __delitem__(self, other) -> None:
        '''
        Remove the (subset of) variable(s) referenced by `key` from the Variables database.

        Parameters
        ----------
        key: str or list(str)
            (list of) name(s) of the variable(s) to be removed.
            The list of names can be given as a string pattern (e.g. "A*;*_").

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")

        >>> # a) delete one variable
        >>> variables.get_names("A*")
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> del variables["ACAF"]
        >>> variables.get_names("A*")
        [\'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']

        >>> # b) delete several variables at once using a pattern
        >>> del variables["A*"]
        >>> variables.get_names("A*")
        []

        >>> # c) delete several variables at once using a list of names
        >>> variables.get_names("B*")
        [\'BENEF\', \'BQY\', \'BRUGP\', \'BVY\']
        >>> del variables[["BENEF", "BQY"]]
        >>> variables.get_names("B*")
        [\'BRUGP\', \'BVY\']

        >>> # delete one variable from a subset of the global database
        >>> variables_subset = variables["D*"]
        >>> variables_subset.names
        [\'DEBT\', \'DPU\', \'DPUF\', \'DPUG\', \'DPUGO\', \'DPUH\', \'DPUHO\', \'DPUU\', \'DTF\', \'DTFX\', \'DTH\', \'DTH1\', \'DTH1C\', \'DTHX\']
        >>> del variables_subset["DPUGO"]
        >>> variables_subset.names
        [\'DEBT\', \'DPU\', \'DPUF\', \'DPUG\', \'DPUH\', \'DPUHO\', \'DPUU\', \'DTF\', \'DTFX\', \'DTH\', \'DTH1\', \'DTH1C\', \'DTHX\']
        >>> # NOTE: the variable has also been deleted from the global database
        >>> "DPUGO" in variables
        False
        >>> variables.get_names("D*")
        [\'DEBT\', \'DPU\', \'DPUF\', \'DPUG\', \'DPUH\', \'DPUHO\', \'DPUU\', \'DTF\', \'DTFX\', \'DTH\', \'DTH1\', \'DTH1C\', \'DTHX\']
        '''
    def __getitem__(self, index):
        '''
        Return the (subset of) variable(s) referenced by `key`.

        The `key` can represent a single object name (e.g. "ACAF") or a list of object names ("ACAF;ACAG;AOUC") 
        or a pattern (e.g. "A*") or a list of sub-patterns (e.g. "A*;*_").
        
        If the `key` represents a list of object names or of sub-patterns, each name or sub-pattern is separated 
        by a `separator` character which is either a whitespace ` `, or a comma `,`, or a semi-colon `;`, or a 
        tabulation `\\t`, or a newline `\\n`.

        A (sub-)`pattern` is a list of characters representing a group of object names. 
        It includes some special characters which have a special meaning:
        
            - `*` : any character sequence, even empty
            - `?` : any character (one and only one)
            - `@` : any alphanumerical char [A-Za-z0-9]
            - `&` : any non alphanumerical char
            - `|` : any alphanumeric character or none at the beginning and end of a string 
            - `!` : any non-alphanumeric character or none at the beginning and end of a string 
            - `\\` : escape the next character

        Note that the `key` can contain references to IODE lists which are prefixed with the symbol `$`.

        Parameters
        ----------
        key: str or list(str)
            (the list of) name(s) of the variable(s) to get.
            The list of objects to get can be specified by a pattern or by a list of sub-patterns (e.g. "A*;*_").

        Returns
        -------
        Single variable or a subset of the database.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables, NA
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
        >>> variables.sample
        Sample("1960Y1:2015Y1")

        >>> # -------- a) get one Variable --------
        >>> # get the variable values for the whole sample
        >>> variables["ACAF"]                       # doctest: +ELLIPSIS 
        [-2e+37, -2e+37, ..., -83.34062511080091, -96.41041982848331]
        >>> # get the variable value for a specific period
        >>> variables["ACAF", "1990Y1"]
        23.771
        >>> # get the variable values for range of periods (using a Python slice)
        >>> variables["ACAF", "1990Y1":"2000Y1"]    # doctest: +ELLIPSIS 
        [23.771, 26.240999, ..., 13.530404919696034, 10.046610792200543]
        >>> # same as above but with the colon \':\' inside the periods range string
        >>> variables["ACAF", "1990Y1:2000Y1"]      # doctest: +ELLIPSIS 
        [23.771, 26.240999, ..., 13.530404919696034, 10.046610792200543]

        >>> # b) -------- get a subset of the Variables database using a pattern --------
        >>> variables_subset = variables["A*"]
        >>> variables_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> # get the variable values for a specific period
        >>> variables["A*", "1990Y1"]
        [23.771, -28.1721855713507, 1.0, 0.9373591502749314, 1.0]
        >>> # get the variable values for range of periods (using a Python slice)
        >>> variables["A*", "1990Y1":"2000Y1"]      # doctest: +ELLIPSIS 
        [[23.771, 26.240999, 30.159, ..., 1.2031082, 1.3429699656745855, 1.3386028553645442]]

        >>> # c) -------- get a subset of the Variables database using a list of names --------
        >>> variables_subset = variables[["ACAF", "ACAG", "AQC", "BQY", "BVY"]]
        >>> variables_subset.names
        [\'ACAF\', \'ACAG\', \'AQC\', \'BQY\', \'BVY\']
        >>> # get the variable values for a specific period
        >>> variables[["ACAF", "ACAG", "AQC", "BQY", "BVY"], "1990Y1"]
        [23.771, -28.1721855713507, 1.0, -34.099998, -34.099997]
        >>> # get the variable values for range of periods (using a Python slice)
        >>> variables[["ACAF", "ACAG", "AQC", "BQY", "BVY"], "1990Y1":"2000Y1"]      # doctest: +ELLIPSIS 
        [[23.771, 26.240999, 30.159, ..., 140.73978, 144.8587818455608, 150.05335230584103]]
        '''
    def __hash__(self) -> int:
        '''
        Return a hash value for the current Variables database.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR, variables
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
        >>> len(variables)
        394
        >>> original_hash = hash(variables)
        
        >>> # rename 1 variable
        >>> variables.rename("ACAF", "ACAF_")
        >>> original_hash == hash(variables)
        False
        >>> variables.rename("ACAF_", "ACAF")  # revert the change
        >>> original_hash == hash(variables)
        True

        >>> # modify one variable
        >>> original_variable = variables["ACAF"]
        >>> variables["ACAF"] = 0.0
        >>> original_hash == hash(variables)
        False
        >>> variables["ACAF"] = original_variable  # revert the change
        >>> original_hash == hash(variables)
        True

        >>> # delete a variable
        >>> original_variable = variables["ACAF"]
        >>> del variables["ACAF"]
        >>> original_hash == hash(variables)
        False
        >>> variables["ACAF"] = original_variable
        >>> original_hash == hash(variables)
        True

        >>> # add a variable
        >>> variables["NEW"] = 0.0
        >>> original_hash == hash(variables)
        False
        >>> del variables["NEW"]
        >>> original_hash == hash(variables)
        True
        '''
    def __reduce__(self):
        """__reduce_cython__(self)"""
    def __setitem__(self, index, object) -> None:
        '''
        Update/add a (subset of) variable(s) referenced by `key` from/to the Variables database.

        The `key` can represent a single object name (e.g. "ACAF") or a list of object names ("ACAF;ACAG;AOUC") 
        or a pattern (e.g. "A*") or a list of sub-patterns (e.g. "A*;*_").
        
        If the `key` represents a list of object names or of sub-patterns, each name or sub-pattern is separated 
        by a `separator` character which is either a whitespace ` `, or a comma `,`, or a semi-colon `;`, or a 
        tabulation `\\t`, or a newline `\\n`.

        A (sub-)`pattern` is a list of characters representing a group of object names. 
        It includes some special characters which have a special meaning:
        
            - `*` : any character sequence, even empty
            - `?` : any character (one and only one)
            - `@` : any alphanumerical char [A-Za-z0-9]
            - `&` : any non alphanumerical char
            - `|` : any alphanumeric character or none at the beginning and end of a string 
            - `!` : any non-alphanumeric character or none at the beginning and end of a string 
            - `\\` : escape the next character

        Note that the `key` can contain references to IODE lists which are prefixed with the symbol `$`.

        Parameters
        ----------
        key: str or list(str)
            (the list of) name(s) of the variable(s) to update/add.
            The list of objects to update/add can be specified by a pattern or by a list of sub-patterns 
            (e.g. "A*;*_").
        value: int, float, tuple(int, float), list(int, float), str or list of any of those
            If int, the value is first converted to a float and then used for all periods.
            If float, the value is used for all periods.
            If tuple or list, the number of elements must be equal to the number of periods. 
            Each element is then used for the corresponding period.
            If str, the value is interpreted as a LEC expression and is evaluated for each period.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import variables, NA
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
        
        >>> # a) -------- add one variable --------
        >>> # 1) same value for all periods
        >>> variables["A0"] = NA
        >>> variables["A0"]                     # doctest: +ELLIPSIS 
        [-2e+37, -2e+37, ..., -2e+37, -2e+37]
        >>> # 2) vector (list) containing a specific value for each period
        >>> variables["A1"] = list(range(variables.nb_periods))
        >>> variables["A1"]                     # doctest: +ELLIPSIS 
        [0.0, 1.0, 2.0, ..., 53.0, 54.0, 55.0]
        >>> # 3) LEC expression
        >>> variables["A2"] = "t + 10"
        >>> variables["A2"]                     # doctest: +ELLIPSIS 
        [10.0, 11.0, 12.0, ..., 63.0, 64.0, 65.0]

        >>> # b) -------- update one variable --------
        >>> # 1) update all values of a Variable
        >>> variables["ACAF"]                   # doctest: +ELLIPSIS 
        [-2e+37, -2e+37, ..., -83.34062511080091, -96.41041982848331]
        >>> # 1.I) same value for all periods
        >>> variables["ACAF"] = NA
        >>> variables["ACAF"]                   # doctest: +ELLIPSIS 
        [-2e+37, -2e+37, ..., -2e+37, -2e+37]
        >>> # 1.II) vector (list) containing a specific value for each period
        >>> variables["ACAF"] = list(range(variables.nb_periods))
        >>> variables["ACAF"]                   # doctest: +ELLIPSIS 
        [0.0, 1.0, 2.0, ..., 53.0, 54.0, 55.0]
        >>> # 1.III) LEC expression
        >>> variables["ACAF"] = "t + 10"
        >>> variables["ACAF"]                   # doctest: +ELLIPSIS 
        [10.0, 11.0, 12.0, ..., 63.0, 64.0, 65.0]

        >>> # 2) set one value of a Variable for a specific period
        >>> variables["ACAG", "1990Y1"]
        -28.1721855713507
        >>> variables["ACAG", "1990Y1"] = -28.2
        >>> variables["ACAG", "1990Y1"]
        -28.2

        >>> # 3) set the variable values for range of periods 
        >>> # 3.I) using a Python slice
        >>> # 3.I.a) variable(periods) = same value for all periods
        >>> variables["ACAF", "1991Y1":"1995Y1"] = 0.0
        >>> variables["ACAF", "1991Y1":"1995Y1"]
        [0.0, 0.0, 0.0, 0.0, 0.0]
        >>> # 3.I.b) variable(periods) = vector (list) containing a specific value for each period
        >>> variables["ACAF", "1991Y1":"1995Y1"] = [0., 1., 2., 3., 4.]
        >>> variables["ACAF", "1991Y1":"1995Y1"]
        [0.0, 1.0, 2.0, 3.0, 4.0]
        >>> # 3.I.c) variable(periods) = LEC expression
        >>> variables["ACAF", "1991Y1":"1995Y1"] = "t + 10"
        >>> variables["ACAF", "1991Y1":"1995Y1"]
        [41.0, 42.0, 43.0, 44.0, 45.0]

        >>> # 3.II) same as above but with the colon \':\' inside the periods range string
        >>> # 3.II.a) variable(periods) = same value for all periods
        >>> variables["ACAF", "1991Y1:1995Y1"] = 0.0
        >>> variables["ACAF", "1991Y1:1995Y1"]
        [0.0, 0.0, 0.0, 0.0, 0.0]
        >>> # 3.II.b) variable(periods) = vector (list) containing a specific value for each period
        >>> variables["ACAF", "1991Y1:1995Y1"] = [0., -1., -2., -3., -4.]
        >>> variables["ACAF", "1991Y1":"1995Y1"]
        [0.0, -1.0, -2.0, -3.0, -4.0]
        >>> # 3.II.c) variable(periods) = LEC expression
        >>> variables["ACAF", "1991Y1:1995Y1"] = "t - 10"
        >>> variables["ACAF", "1991Y1:1995Y1"]
        [21.0, 22.0, 23.0, 24.0, 25.0]

        >>> # c) -------- working on a subset --------
        >>> # 1) get subset
        >>> variables_subset = variables["A*"]
        >>> variables_subset.names
        [\'A0\', \'A1\', \'A2\', \'ACAF\', \'ACAG\', \'AOUC\', \'AOUC_\', \'AQC\']
        >>> # 2) add a variable to the subset
        >>> variables_subset["A3"] = NA
        >>> variables_subset["A3"]              # doctest: +ELLIPSIS 
        [-2e+37, -2e+37, ..., -2e+37, -2e+37]
        >>> # --> new variable also appears in the global workspace
        >>> "A3" in variables
        True
        >>> variables["A3"]                     # doctest: +ELLIPSIS 
        [-2e+37, -2e+37, ..., -2e+37, -2e+37]
        >>> # 3) update a variable in the subset
        >>> variables_subset["A3"] = 0.0
        >>> variables_subset["A3"]              # doctest: +ELLIPSIS 
        [0.0, 0.0, ..., 0.0, 0.0]
        >>> # --> variable is also updated in the global workspace
        >>> variables["A3"]                     # doctest: +ELLIPSIS 
        [0.0, 0.0, ..., 0.0, 0.0]
        '''

class VarsMode(enum.IntEnum):
    __new__: ClassVar[Callable] = ...
    DIFF: ClassVar[VarsMode] = ...
    GROWTH_RATE: ClassVar[VarsMode] = ...
    LEVEL: ClassVar[VarsMode] = ...
    Y0Y_DIFF: ClassVar[VarsMode] = ...
    Y0Y_GROWTH_RATE: ClassVar[VarsMode] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[int]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    def __format__(self, *args, **kwargs) -> str: ...

class WriteFileExt(enum.IntEnum):
    __new__: ClassVar[Callable] = ...
    A2M: ClassVar[WriteFileExt] = ...
    CSV: ClassVar[WriteFileExt] = ...
    DUMMY: ClassVar[WriteFileExt] = ...
    GDI: ClassVar[WriteFileExt] = ...
    HTML: ClassVar[WriteFileExt] = ...
    MIF: ClassVar[WriteFileExt] = ...
    RTF: ClassVar[WriteFileExt] = ...
    _generate_next_value_: ClassVar[Callable] = ...
    _member_map_: ClassVar[dict] = ...
    _member_names_: ClassVar[list] = ...
    _member_type_: ClassVar[type[int]] = ...
    _unhashable_values_: ClassVar[list] = ...
    _use_args_: ClassVar[bool] = ...
    _value2member_map_: ClassVar[dict] = ...
    def __format__(self, *args, **kwargs) -> str: ...

class _AbstractDatabase:
    description: str
    filename: str
    i: Incomplete
    iode_type: IodeType
    names: list[str]
    def __init__(self, *args, **kwargs) -> None:
        """Create and return a new object.  See help(type) for accurate signature."""
    @overload
    def clear(self) -> Any:
        '''clear(self)

                Delete all objects from the current database.

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import comments
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> len(comments)
                317
                >>> comments.clear()
                >>> len(comments)
                0
        '''
    @overload
    def clear(cls) -> Any:
        '''clear(self)

                Delete all objects from the current database.

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import comments
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> len(comments)
                317
                >>> comments.clear()
                >>> len(comments)
                0
        '''
    @overload
    def copy(self) -> Self:
        '''copy(self) -> Self

                Create a *copy* of an IODE database. 
                Any change made on the *copied database* will **NOT** 
                modify the corresponding the global IODE database.

                Returns
                -------
                Database
        
                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import comments
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")

                Database subset

                >>> # without using copy(), any modification made on 
                >>> # the copy subset will also change the corresponding 
                >>> # IODE database
                >>> cmt_subset = comments["A*"]
                >>> cmt_subset.names
                [\'ACAF\', \'ACAG\', \'AOUC\', \'AQC\']
                >>> # a) add a comment
                >>> cmt_subset["A_NEW"] = "New comment"
                >>> "A_NEW" in cmt_subset
                True
                >>> "A_NEW" in comments
                True
                >>> comments["A_NEW"]
                \'New comment\'
                >>> # b) modify a comment
                >>> cmt_subset["ACAF"] = "Modified Comment"
                >>> cmt_subset["ACAF"]
                \'Modified Comment\'
                >>> comments["ACAF"]
                \'Modified Comment\'
                >>> # c) delete a comment
                >>> del cmt_subset["ACAG"]
                >>> "ACAG" in cmt_subset
                False
                >>> "ACAG" in comments
                False

                Copied database subset

                >>> cmt_subset_copy = comments["B*"].copy()
                >>> cmt_subset_copy.names
                [\'BENEF\', \'BENEF_\', \'BQY\', \'BVY\']
                >>> # by using copy(), any modification made on the copy subset 
                >>> # let the global workspace unchanged
                >>> # a) add a comment -> only added in the copied subset
                >>> cmt_subset_copy["B_NEW"] = "New Comment"
                >>> "B_NEW" in cmt_subset_copy
                True
                >>> "B_NEW" in comments
                False
                >>> # b) modify a comment -> only modified in the copied subset
                >>> cmt_subset_copy["BENEF"] = "Modified Comment"
                >>> cmt_subset_copy["BENEF"]
                \'Modified Comment\'
                >>> comments["BENEF"]
                \'Ondernemingen: niet-uitgekeerde winsten.\'
                >>> # c) delete a comment -> only deleted in the copied subset
                >>> del cmt_subset_copy["BENEF_"]
                >>> "BENEF_" in cmt_subset_copy
                False
                >>> "BENEF_" in comments
                True
        '''
    @overload
    def copy(cls) -> Any:
        '''copy(self) -> Self

                Create a *copy* of an IODE database. 
                Any change made on the *copied database* will **NOT** 
                modify the corresponding the global IODE database.

                Returns
                -------
                Database
        
                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import comments
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")

                Database subset

                >>> # without using copy(), any modification made on 
                >>> # the copy subset will also change the corresponding 
                >>> # IODE database
                >>> cmt_subset = comments["A*"]
                >>> cmt_subset.names
                [\'ACAF\', \'ACAG\', \'AOUC\', \'AQC\']
                >>> # a) add a comment
                >>> cmt_subset["A_NEW"] = "New comment"
                >>> "A_NEW" in cmt_subset
                True
                >>> "A_NEW" in comments
                True
                >>> comments["A_NEW"]
                \'New comment\'
                >>> # b) modify a comment
                >>> cmt_subset["ACAF"] = "Modified Comment"
                >>> cmt_subset["ACAF"]
                \'Modified Comment\'
                >>> comments["ACAF"]
                \'Modified Comment\'
                >>> # c) delete a comment
                >>> del cmt_subset["ACAG"]
                >>> "ACAG" in cmt_subset
                False
                >>> "ACAG" in comments
                False

                Copied database subset

                >>> cmt_subset_copy = comments["B*"].copy()
                >>> cmt_subset_copy.names
                [\'BENEF\', \'BENEF_\', \'BQY\', \'BVY\']
                >>> # by using copy(), any modification made on the copy subset 
                >>> # let the global workspace unchanged
                >>> # a) add a comment -> only added in the copied subset
                >>> cmt_subset_copy["B_NEW"] = "New Comment"
                >>> "B_NEW" in cmt_subset_copy
                True
                >>> "B_NEW" in comments
                False
                >>> # b) modify a comment -> only modified in the copied subset
                >>> cmt_subset_copy["BENEF"] = "Modified Comment"
                >>> cmt_subset_copy["BENEF"]
                \'Modified Comment\'
                >>> comments["BENEF"]
                \'Ondernemingen: niet-uitgekeerde winsten.\'
                >>> # c) delete a comment -> only deleted in the copied subset
                >>> del cmt_subset_copy["BENEF_"]
                >>> "BENEF_" in cmt_subset_copy
                False
                >>> "BENEF_" in comments
                True
        '''
    def copy_from(self, input_files: str | list[str], objects_names: str | list[str] = ...) -> Any:
        '''copy_from(self, input_files: Union[str, List[str]], objects_names: Union[str, List[str]] = u\'*\')

                Copy (a subset of) objects from the input file(s) \'input_files\' into the current database.

                Parameters
                ----------
                input_file: str or list(str)
                    file(s) from which the copied objects are read.
                objects_names: str or list(str)
                    list of objects to copy from the input file(s).
                    Defaults to load all objects from the input file(s). 
        
                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import comments
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> len(comments)
                317
                >>> # delete all comments with a name starting with \'A\'
                >>> comments.remove("A*")
                >>> comments.get_names("A*")
                []

                >>> # load all comments with a name starting with \'A\'
                >>> comments.copy_from(f"{SAMPLE_DATA_DIR}/fun.cmt", "A*")
                >>> comments.get_names("A*")
                [\'ACAF\', \'ACAG\', \'AOUC\', \'AQC\']

                >>> comments.clear()
                >>> # load all comments
                >>> comments.copy_from(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> len(comments)
                317
        '''
    def get_name(self, pos: int) -> str:
        '''get_name(self, pos: int) -> str

                Return the name of the IODE object at position `pos` in the database.

                Parameters
                ----------
                pos: int
                   Position of the object in the database.

                Returns
                -------
                str

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import comments
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> comments.get_name(0)
                \'ACAF\'
        '''
    def get_names(self, pattern: str | list[str] = ...) -> list[str]:
        '''get_names(self, pattern: Union[str, List[str]] = u\'\') -> List[str]

                Returns the list of objects names given a pattern.

                Parameters
                ----------
                pattern: str or list(str), optional
                    pattern to select a subset of objects. 
                    For example, \'A*;*_\' will select all objects for which the name starts 
                    with \'A\' or ends with \'_\'.
                    Return all names contained in the database by default.

                Returns
                -------
                list(str)
        
                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import comments
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> comments.get_names("A*;*_")         # doctest: +NORMALIZE_WHITESPACE
                [\'ACAF\', \'ACAG\', \'AOUC\', \'AQC\', \'BENEF_\', \'GOSH_\', 
                \'IDH_\', \'PAFF_\', \'PC_\', \'PFI_\', \'QAFF_\', \'QAF_\', 
                \'QAI_\', \'QAT_\', \'QBBPPOT_\', \'QC_\', \'QQMAB_\', \'QS_\', 
                \'Q_\', \'TFPHP_\', \'VAFF_\', \'VAI_\', \'VAT_\', \'VC_\', \'VS_\', 
                \'WBF_\', \'WBU_\', \'WCF_\', \'WCR1_\', \'WCR2_\', \'WIND_\', 
                \'WNF_\', \'YDH_\', \'ZZ_\']
                >>> # or equivalently
                >>> comments.get_names(["A*", "*_"])    # doctest: +NORMALIZE_WHITESPACE
                [\'ACAF\', \'ACAG\', \'AOUC\', \'AQC\', \'BENEF_\', \'GOSH_\', 
                \'IDH_\', \'PAFF_\', \'PC_\', \'PFI_\', \'QAFF_\', \'QAF_\', 
                \'QAI_\', \'QAT_\', \'QBBPPOT_\', \'QC_\', \'QQMAB_\', \'QS_\', 
                \'Q_\', \'TFPHP_\', \'VAFF_\', \'VAI_\', \'VAT_\', \'VC_\', \'VS_\', 
                \'WBF_\', \'WBU_\', \'WCF_\', \'WCR1_\', \'WCR2_\', \'WIND_\', 
                \'WNF_\', \'YDH_\', \'ZZ_\']
                >>> # get the list of all names
                >>> comments.names                # doctest: +ELLIPSIS
                [\'ACAF\', \'ACAG\', \'AOUC\', ..., \'ZKF\', \'ZX\', \'ZZ_\']
        '''
    def get_position(self, name: str) -> int:
        '''get_position(self, name: str) -> int

                Return the position of the IODE object with name `name` in the database.

                Parameters
                ----------
                name: str
                    Name of the IODE object to search for in the database.
        
                Returns
                -------
                int

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import comments
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> comments.get_position("ACAF")
                0
        '''
    @overload
    def is_copy_subset(self) -> bool:
        '''is_copy_subset(self) -> bool

                Whether or not the present object represents of a subset of a global IODE database 
                and if the IODE objects of the subset represent *deep copies* of the IODE objects 
                from the global IODE database.

                A *deep copy* subset means that any change made on an object of the subset will **NOT** 
                modify the corresponding object in the global IODE database.

                A *shallow copy* subset means that any change made on an object of the subset will also 
                modify the corresponding object in the global IODE database.

                By default, the :py:meth:`iode.CommentsDatabase.subset` method return a *shallow copy* subset.

                Returns
                -------
                bool

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import comments
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> comments.is_subset()
                False
                >>> cmt_subset = comments["A*"].copy()
                >>> cmt_subset.is_subset()
                True
                >>> cmt_subset.is_copy_subset()
                True
        '''
    @overload
    def is_copy_subset(cls) -> Any:
        '''is_copy_subset(self) -> bool

                Whether or not the present object represents of a subset of a global IODE database 
                and if the IODE objects of the subset represent *deep copies* of the IODE objects 
                from the global IODE database.

                A *deep copy* subset means that any change made on an object of the subset will **NOT** 
                modify the corresponding object in the global IODE database.

                A *shallow copy* subset means that any change made on an object of the subset will also 
                modify the corresponding object in the global IODE database.

                By default, the :py:meth:`iode.CommentsDatabase.subset` method return a *shallow copy* subset.

                Returns
                -------
                bool

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import comments
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> comments.is_subset()
                False
                >>> cmt_subset = comments["A*"].copy()
                >>> cmt_subset.is_subset()
                True
                >>> cmt_subset.is_copy_subset()
                True
        '''
    @overload
    def is_subset(self) -> bool:
        '''is_subset(self) -> bool

                Whether or not the present object represents a subset of a global IODE database.

                Returns
                -------
                bool

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import comments
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> comments.is_subset()
                False
                >>> cmt_subset = comments["A*"].copy()
                >>> cmt_subset.is_subset()
                True
        '''
    @overload
    def is_subset(cls) -> Any:
        '''is_subset(self) -> bool

                Whether or not the present object represents a subset of a global IODE database.

                Returns
                -------
                bool

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import comments
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> comments.is_subset()
                False
                >>> cmt_subset = comments["A*"].copy()
                >>> cmt_subset.is_subset()
                True
        '''
    @overload
    def is_subset(cls) -> Any:
        '''is_subset(self) -> bool

                Whether or not the present object represents a subset of a global IODE database.

                Returns
                -------
                bool

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import comments
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> comments.is_subset()
                False
                >>> cmt_subset = comments["A*"].copy()
                >>> cmt_subset.is_subset()
                True
        '''
    def load(self, filepath: str) -> Any:
        '''load(self, filepath: str)

                Load objects stored in file \'filepath\' into the current database.
                Erase the database before to load the file.

                Parameters
                ----------
                filepath: str
                    path to the file to load
        
                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import comments, variables
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> len(comments)
                317

                >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")
                >>> len(variables)
                394
        '''
    def merge(self, other: Self, overwrite: bool = ...) -> Any:
        """merge(self, other: Self, overwrite: bool = True)"""
    def merge_from(self, input_file: str) -> Any:
        '''merge_from(self, input_file: str)

                Merge all objects stored in the input file \'input_file\' into the current database.

                Parameters
                ----------
                input_file: str
                    file from which the objects to merge are read.
        
                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import comments
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> len(comments)
                317
                >>> # delete all comments
                >>> comments.clear()
                >>> len(comments)
                0

                >>> # reload all comments
                >>> comments.merge_from(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> len(comments)
                317
        '''
    def remove(self, names: str | list[str]) -> Any:
        '''remove(self, names: Union[str, List[str]])

                Delete the object(s) named \'names\' from the current database.
        
                Parameters
                ----------
                names: str
                    name(s) of object(s) to be deleted. 
                    It can be a pattern (e.g. "A*").

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import comments
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> comments.get_names("A*")
                [\'ACAF\', \'ACAG\', \'AOUC\', \'AQC\']

                >>> # remove one object
                >>> comments.remove("ACAF")
                >>> comments.get_names("A*")
                [\'ACAG\', \'AOUC\', \'AQC\']

                >>> # remove all objects with a name ending by \'_\'
                >>> comments.get_names("*_")            # doctest: +ELLIPSIS
                [\'BENEF_\', \'GOSH_\', \'IDH_\', ..., \'WNF_\', \'YDH_\', \'ZZ_\']
                >>> comments.remove("*_")
                >>> comments.get_names("*_")
                []
        '''
    def rename(self, old_name: str, new_name: str) -> Any:
        '''rename(self, old_name: str, new_name: str)

                Rename an object of the database.

                Parameters
                ----------
                old_name: str
                    current name in the database
                new_name: str
                    new name in the database

                Warning
                -------
                Renaming an Equation is not allowed.
        
                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import comments
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> comments["ACAF"]
                \'Ondernemingen: ontvangen kapitaaloverdrachten.\'

                >>> # rename comment \'ACAF\' as \'ACCAF\'
                >>> comments.rename("ACAF", "ACCAF")
                >>> "ACCAF" in comments
                True
                >>> comments["ACCAF"]
                \'Ondernemingen: ontvangen kapitaaloverdrachten.\'
        '''
    def save(self, filepath: str) -> Any:
        '''save(self, filepath: str)

                Save objects of the current database into the file \'filepath\'.

                Parameters
                ----------
                filepath: str
        
                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import comments
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> len(comments)
                317
                >>> comments.save(f"{SAMPLE_DATA_DIR}/fun2.cmt")
                >>> comments.clear()
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun2.cmt")
                >>> len(comments)
                317
        '''
    def search(self, pattern: str, word: bool = ..., case_sensitive: bool = ..., in_name: bool = ..., in_formula: bool = ..., in_text: bool = ..., list_result: str = ...) -> Any:
        '''search(self, pattern: str, word: bool = True, case_sensitive: bool = True, in_name: bool = True, in_formula: bool = True, in_text: bool = True, list_result: str = u\'_RES\')
        Return a list of all objects from the current database having a specific pattern in their names or LEC expression, comment...
          
                The following characters in *pattern* have a special meaning:
        
                    - `*` : any character sequence, even empty
                    - `?` : any character (one and only one)
                    - `@` : any alphanumerical char [A-Za-z0-9]
                    - `&` : any non alphanumerical char
                    - `|` : any alphanumeric character or none at the beginning and end of a string 
                    - `!` : any non-alphanumeric character or none at the beginning and end of a string 
                    - `\\` : escape the next character
        
                The Search method depends on the type of object:

                - Comments: the name and text of the comments are analyzed 
                - Equations: the name and LEC form of the equations are analyzed 
                - Identities: the name and LEC form of the identities are analyzed 
                - Lists: the name and text of the lists are analyzed 
                - Scalars: the name of the scalars is analyzed 
                - Tables: the table name, titles and LEC forms are analyzed 
                - Variables: the name of the variables is analyzed 

                Parameters
                ----------
                pattern: str     
                    string to search
                word: bool, optional
                    Whether or not the pattern to be searched for must be a whole word and not part of a word. 
                    Default to True.
                case_sensitive: bool, optional
                    Whether or not the search is case sensitive.
                    Default to True.
                in_name: bool, optional
                    Whether or not to also search in object names.
                    Default to True.
                in_formula: bool, optional
                    Whether or not to also search in LEC expressions (for Equations, Identities and Tables (LEC cells)).
                    Default to True.
                in_text: bool, optional
                    Whether or not to also search in texts (for Comments, lists, Equations (comment) and Tables (text cells)).
                    Default to True.
                list_result: str, optional
                    Name of the IODE list in which the resulting list of objects is saved.
                    Default to *_RES*.

                Returns
                -------
                list(str)

                Examples
                --------
                >>> from iode import SAMPLE_DATA_DIR
                >>> from iode import comments, equations, identities, lists, tables
                >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
                >>> equations.load(f"{SAMPLE_DATA_DIR}/fun.eqs")
                >>> identities.load(f"{SAMPLE_DATA_DIR}/fun.idt")
                >>> lists.load(f"{SAMPLE_DATA_DIR}/fun.lst")
                >>> tables.load(f"{SAMPLE_DATA_DIR}/fun.tbl")

                >>> # get list of comments containing \'Bruto\'
                >>> comments.search("Bruto")            # doctest: +ELLIPSIS 
                [\'GOSF\', \'GOSG\', \'GOSH\', \'GOSH_\', ..., \'VIF\', \'VIG\', \'YDH\', \'YDH_\']

                >>> # get list of equations containing \'AOUC\'
                >>> equations.search("AOUC")
                [\'AOUC\', \'PC\', \'PIF\', \'PXS\', \'QXAB\']

                >>> # get list of identities containing \'NDOMY\'
                >>> identities.search("NDOMY")
                [\'LCLASS\', \'LKEYN\', \'UCLASS\']

                >>> # get list of IODE lists containing \'AOUC\'
                >>> lists.search("AOUC")
                [\'COPY0\', \'ENDO0\', \'TOTAL0\']

                >>> # get list of IODE tables containing \'AOUC\' in cells
                >>> tables.search("AOUC")
                [\'ANAPRIX\', \'MULT1FR\', \'MULT1RESU\', \'T1\', \'T1NIVEAU\']
        '''
    def __contains__(self, other) -> bool:
        '''
        Test if the IODE object named `item` is present in the current database.

        Parameters
        ----------
        item: str
            name of the IODE object. 

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import comments
        >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
        >>> "ACAF" in comments
        True
        >>> "ZCAF" in comments
        False
        '''
    def __delitem__(self, other) -> None:
        """Delete self[key]."""
    def __getitem__(self, index):
        """Return self[key]."""
    def __iter__(self):
        '''
        Iterate over object names.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import comments
        >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
        >>> cmt_subset = comments["A*"]
        >>> cmt_subset.names
        [\'ACAF\', \'ACAG\', \'AOUC\', \'AQC\']
        >>> for name in cmt_subset:
        ...     print(name)
        ACAF
        ACAG
        AOUC
        AQC
        '''
    def __len__(self) -> int:
        '''
        Return the number of IODE objects in the current database.

        Returns
        -------
        int

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import comments
        >>> comments.load(f"{SAMPLE_DATA_DIR}/fun.cmt")
        >>> len(comments)
        317
        '''
    def __reduce__(self):
        """__reduce_cython__(self)"""
    def __setitem__(self, index, object) -> None:
        """Set self[key] to value."""

def __reduce_cython__(self) -> Any:
    """__reduce_cython__(self)"""
def __setstate_cython__(self, __pyx_state) -> Any:
    """__setstate_cython__(self, __pyx_state)"""
def add_error_msg(msg: str = ...) -> Any:
    '''add_error_msg(msg: str = u\'\')

        Add an error message to the stack of error messages.
   
        Parameters
        ----------
        msg: str
            string containing the text of the error
        
        Examples
        --------
        >>> from iode import add_error_msg, display_error_msgs
        >>> add_error_msg("Missing variables: QIG, PIG")
        >>> display_error_msgs() # doctest: +SKIP
    '''
def build_command_functions_list(group: int, gui: bool = ...) -> list[str]:
    """build_command_functions_list(group: int, gui: bool = False) -> List[str]

        Setup the IODE $/#-functions (i.e. IODE commands).

        Interactive commands (which open something in the GUI) starts with the character '#'.
        Non interactive commands starts with the character '$'.

        Parameters
        ----------
        group: int
            'group' is used to defined groups for the color highlighting.
            Possible values are either 0 or 1.
        gui: bool, optional
            False: for the console, True: for the GUI

        Notes
        -----
        See b_rep_syntax.c (C API) for the list

        Warnings
        --------
        Intended to be used only for building the GUI, not called by users

        Examples
        --------
        >>> from iode.reports import build_command_functions_list
        >>> build_command_functions_list(0, False)     # doctest: +ELLIPSIS
        ['$label', '$goto', '$ask', ..., '$next', '$procdef', '$procexec']
        >>> build_command_functions_list(0, True)      # doctest: +ELLIPSIS
        ['#label', '#goto', '#ask', ..., '#next', '#procdef', '#procexec']
        >>> build_command_functions_list(1, False)     # doctest: +ELLIPSIS
        ['$FileImportVar', '$FileImportCmt', '$FileDelete', ..., '$CsvDec', '$CsvNaN', '$CsvAxes']
        >>> build_command_functions_list(1, True)      # doctest: +ELLIPSIS
        ['#FileImportVar', '#FileImportCmt', '#WsSample', ..., '#ReportExec', '#ReportEdit', '#ReportPrompt']
    """
@overload
def build_lec_functions_list() -> list[str]:
    """build_lec_functions_list() -> List[str]

        Set the list of functions available in LEC expression.

        Notes
        -----
        See l_token.c (C API) for the list

        Warnings
        --------
        Intended to be used only for building the GUI, not called by users
        Examples
        --------
        >>> from iode.reports import build_lec_functions_list
        >>> build_lec_functions_list()         # doctest: +ELLIPSIS
        ['abs', 'acf', 'acos', 'and', ..., 'urandom', 'var', 'vmax', 'vmin']
    """
@overload
def build_lec_functions_list() -> Any:
    """build_lec_functions_list() -> List[str]

        Set the list of functions available in LEC expression.

        Notes
        -----
        See l_token.c (C API) for the list

        Warnings
        --------
        Intended to be used only for building the GUI, not called by users
        Examples
        --------
        >>> from iode.reports import build_lec_functions_list
        >>> build_lec_functions_list()         # doctest: +ELLIPSIS
        ['abs', 'acf', 'acos', 'and', ..., 'urandom', 'var', 'vmax', 'vmin']
    """
@overload
def build_report_functions_list() -> list[str]:
    """build_report_functions_list() -> List[str]

        Setup the @-functions (i.e. the IODE functions)

        Notes
        -----
        See b_rep_syntax.c (C API) for the list

        Warnings
        --------
        Intended to be used only for building the GUI, not called by users

        Examples
        --------
        >>> from iode.reports import build_report_functions_list
        >>> build_report_functions_list()          # doctest: +ELLIPSIS
        ['@upper', '@date', '@time', '@lower', ..., '@mkdir', '@rmdir', '@void', '@version']
    """
@overload
def build_report_functions_list() -> Any:
    """build_report_functions_list() -> List[str]

        Setup the @-functions (i.e. the IODE functions)

        Notes
        -----
        See b_rep_syntax.c (C API) for the list

        Warnings
        --------
        Intended to be used only for building the GUI, not called by users

        Examples
        --------
        >>> from iode.reports import build_report_functions_list
        >>> build_report_functions_list()          # doctest: +ELLIPSIS
        ['@upper', '@date', '@time', '@lower', ..., '@mkdir', '@rmdir', '@void', '@version']
    """
@overload
def clear_error_msgs() -> Any:
    '''clear_error_msgs()

        Empty the stack of error messages.
        clear_error_msgs() is called by display_error_msgs()
   
        Examples
        --------
        >>> from iode import clear_error_msgs, add_error_msg, display_error_msgs
        >>> # reset the stack of messages
        >>> clear_error_msgs() 
        >>> add_error_msg("Missing variables: QIG, PIG")
        >>> display_error_msgs()    # doctest: +SKIP
    '''
@overload
def clear_error_msgs() -> Any:
    '''clear_error_msgs()

        Empty the stack of error messages.
        clear_error_msgs() is called by display_error_msgs()
   
        Examples
        --------
        >>> from iode import clear_error_msgs, add_error_msg, display_error_msgs
        >>> # reset the stack of messages
        >>> clear_error_msgs() 
        >>> add_error_msg("Missing variables: QIG, PIG")
        >>> display_error_msgs()    # doctest: +SKIP
    '''
@overload
def clear_error_msgs() -> Any:
    '''clear_error_msgs()

        Empty the stack of error messages.
        clear_error_msgs() is called by display_error_msgs()
   
        Examples
        --------
        >>> from iode import clear_error_msgs, add_error_msg, display_error_msgs
        >>> # reset the stack of messages
        >>> clear_error_msgs() 
        >>> add_error_msg("Missing variables: QIG, PIG")
        >>> display_error_msgs()    # doctest: +SKIP
    '''
def dickey_fuller_test(lec: str, drift: bool, trend: bool, order: int) -> Scalars:
    '''dickey_fuller_test(lec: str, drift: bool, trend: bool, order: int) -> Scalars

        Dickey-Fuller tests.

        Tests are saved in scalars whose name is composed of the prefix *df_* and the name of the first series 
        appearing in the formula to be tested. For example, the test for the formula :math:`d(A0GR+A0GF)` is `df_a0gr`.

        Parameters
        ----------
        lec: str
            LEC form to be tested.
        drift: bool
            Whether or not the formula to be estimated must incorporate a constant term. 
        trend: bool
            Whether or not the formula to be estimated should incorporate a trend term.
        order: int 
            Order of the polynomial to be estimated to obtain the tests.
    
        Returns
        -------
        Scalars
            Scalars database containing the results of the Dickey-Fuller tests.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR, scalars, variables
        >>> from iode import dickey_fuller_test
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")

        >>> # dickey_fuller_test("ACAF", True, True, 3) estimates the equation:
        >>> #     d(ACAF) := df_ * ACAF[-1] +            
        >>> #     df_d +            (DRIFT)          
        >>> #     df_t * t +        (TREND)            
        >>> #     df1 * d(ACAF[-1]) + df2*d(ACAF[-2]) + df3*d(ACAF[-3])  (ORDER)
        >>> df_scalars = dickey_fuller_test("ACAF", True, True, 3)
        >>> df_scalars.get_names("df*")
        [\'df1\', \'df2\', \'df3\', \'df_\', \'df_d\', \'df_t\']
        >>> # note: the function dickey_fuller_test() returns a separated Scalars database.
        >>> #       The global database scalars is left unchaged
        >>> scalars.get_names("df*")
        []
        >>> # order 0
        >>> df_scalars["df_"]
        Scalar(0.0132523, 0.0845155, 0.0845155)
        >>> # drift
        >>> df_scalars["df_d"]
        Scalar(13.0806, 1, 6.78675)
        >>> # trend
        >>> df_scalars["df_t"]
        Scalar(-0.492697, 0.187978, 0.187978)
        >>> # order 1
        >>> df_scalars["df1"]
        Scalar(-0.120123, 0.180991, 0.180991)
        >>> # order 2
        >>> df_scalars["df2"]
        Scalar(-0.476959, 0.154505, 0.154505)
        >>> # order 3
        >>> df_scalars["df3"]
        Scalar(-0.211047, 0.170708, 0.170708)
    '''
@overload
def display_error_msgs() -> Any:
    '''display_error_msgs()

        Display all messages accumulated on to the stack of error messages.
        Calls clear_error_msgs().
        
        Examples
        --------
        >>> from iode import add_error_msg, display_error_msgs
        >>> add_error_msg("Missing variables: QIG, PIG")
        >>> display_error_msgs()     # doctest: +SKIP
    '''
@overload
def display_error_msgs() -> Any:
    '''display_error_msgs()

        Display all messages accumulated on to the stack of error messages.
        Calls clear_error_msgs().
        
        Examples
        --------
        >>> from iode import add_error_msg, display_error_msgs
        >>> add_error_msg("Missing variables: QIG, PIG")
        >>> display_error_msgs()     # doctest: +SKIP
    '''
def dynamic_adjustment(method: AdjustmentMethod | str, eqs: str, c1: str = ..., c2: str = ...) -> str:
    '''dynamic_adjustment(method: Union[AdjustmentMethod, str], eqs: str, c1: str = u\'c1\', c2: str = u\'c2\') -> str

        Transform a LEC equation to add a dynamic adjustment.
    
        Two methods can be used. Given the equation :math:`LHS = RHS`, we have:
    
            - **Partial Adjustment** (PARTIAL): :math:`d(LHS) = c1 * (RHS - (LHS)[-1])`
            - **Error Correction Model** (ERROR_CORRECTION): :math:`d(LHS) = c1 * d(RHS) + c2 * (RHS -LHS)[-1]`
    
        Parameters
        ----------
        method: AdjustmentMethod or str 
            Method used for the dynamic adjustment. 
            Possible values are PARTIAL or ERROR_CORRECTION.
        eqs: str
            LEC equation to dynamically adjust.
        c1: str, optional        
            Name of the first coefficient.
            Default to "c1".
        c2: str, optional    
            Name of the second coefficient. 
            Not used with the *Partial Adjustment* method.
            Default to "c2".

        Returns
        -------
        str
            Dynamically adjusted equation.

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR, equations, dynamic_adjustment, AdjustmentMethod
        >>> equations.load(f"{SAMPLE_DATA_DIR}/fun.eqs")
        >>> lec = equations["ACAF"].lec           # doctest: +NORMALIZE_WHITESPACE
        >>> lec
        \'(ACAF/VAF[-1]) :=acaf1+acaf2*GOSF[-1]+\\nacaf4*(TIME=1995)\'
    
        Partial Adjustment

        >>> partial_adjust_eq = dynamic_adjustment(AdjustmentMethod.PARTIAL, lec)
        >>> partial_adjust_eq
        \'d((ACAF/VAF[-1])) := c1 * (acaf1+acaf2*GOSF[-1]+\\nacaf4*(TIME=1995) -((ACAF/VAF[-1]))[-1])\'

        Error Correction Model

        >>> error_corr_adjust_eq = dynamic_adjustment(AdjustmentMethod.ERROR_CORRECTION, lec)
        >>> error_corr_adjust_eq
        \'d((ACAF/VAF[-1])) := c1 * d(acaf1+acaf2*GOSF[-1]+\\nacaf4*(TIME=1995)) + c2 * (acaf1+acaf2*GOSF[-1]+\\nacaf4*(TIME=1995) -(ACAF/VAF[-1]))[-1]\'
    '''
def execute_command(command: str | list[str]) -> Any:
    '''execute_command(command: Union[str, List[str]])

        Execute one or several IODE commands. 

        Parameters
        ----------
        command: str or list(str)
            IODE command(s) to execute. 
            If multiple commands are passed as one string, each command must end 
            with the special character *\\n*.

        Examples
        --------
        >>> from iode import execute_command

        >>> # execute IODE command one by one
        >>> execute_command("$WsClearVar")
        >>> execute_command("$WsSample 2000Y1 2005Y1")
        >>> execute_command("$DataCalcVar A t+1")
        >>> execute_command("$DataCalcVar B t-1")
        >>> execute_command("$DataCalcVar C A/B")
        >>> execute_command("$DataCalcVar D grt A")
        >>> execute_command("$WsSaveVar test_var.av")

        >>> # check content of file test_var.av
        >>> with open("test_var.av", "r") as f:         # doctest: +NORMALIZE_WHITESPACE
        ...     print(f.read())
        sample 2000Y1 2005Y1
        A 1 2 3 4 5 6
        B -1 0 1 2 3 4
        C -1 na 3 2 1.66666666666667 1.5
        D na 100 50 33.3333333333333 25 20
        <BLANKLINE>
    '''
@overload
def execute_lec(lec: str, period: str | int | Period = ...) -> float | list[float]:
    '''execute_lec(lec: str, period: Union[str, int, Period] = None) -> Union[float, List[float]]

        Compute a LEC formula using the current Variables and Scalars databases.
        The formula may be evaluate at a specific period or on the whole sample 
        (no value for *period* is passed).
    
        Parameters
        ----------
        lec: str
            LEC expression to be evaluated.
        period: str or int or Period, optional
            Period at which the LEC formula is evaluated.
            If the passed value is of type int, then it represents 
            the position of the period in the current sample.
            If the passed value is None, empty string or -1, the LEC expression 
            is evaluated over the whole sample.
            Default to None.
    
        Returns
        -------
        float or list(float)

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import execute_lec, equations, scalars, variables
        >>> equations.load(f"{SAMPLE_DATA_DIR}/fun.eqs")
        >>> scalars.load(f"{SAMPLE_DATA_DIR}/fun.scl")
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")

        >>> equations["ACAF"].lec       # doctest: +NORMALIZE_WHITESPACE
        \'(ACAF/VAF[-1]) :=acaf1+acaf2*GOSF[-1]+\\nacaf4*(TIME=1995)\'
        >>> variables["ACAF", "2000Y1"]
        10.046610792200543
        >>> lec = "(acaf1 + acaf2 * GOSF[-1] + acaf4*(TIME=1995)) * VAF[-1]"
        >>> # pass a period value as string
        >>> execute_lec(lec, "2000Y1")
        10.046610792200543
        >>> # pass a period position in the sample
        >>> t = variables.sample.index("2000Y1")
        >>> t
        40
        >>> execute_lec(lec, t)
        10.046610792200543
        >>> # compute the LEC formula over the whole sample
        >>> execute_lec(lec)        # doctest: +ELLIPSIS
        [-2e+37, 4.2884154594335815, 4.532163174288473, ..., -83.34062511080091, -96.41041982848331]
        >>> variables["ACAF"]       # doctest: +ELLIPSIS
        [-2e+37, -2e+37, -2e+37, ..., -83.34062511080091, -96.41041982848331]
    '''
@overload
def execute_lec(lec, t) -> Any:
    '''execute_lec(lec: str, period: Union[str, int, Period] = None) -> Union[float, List[float]]

        Compute a LEC formula using the current Variables and Scalars databases.
        The formula may be evaluate at a specific period or on the whole sample 
        (no value for *period* is passed).
    
        Parameters
        ----------
        lec: str
            LEC expression to be evaluated.
        period: str or int or Period, optional
            Period at which the LEC formula is evaluated.
            If the passed value is of type int, then it represents 
            the position of the period in the current sample.
            If the passed value is None, empty string or -1, the LEC expression 
            is evaluated over the whole sample.
            Default to None.
    
        Returns
        -------
        float or list(float)

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import execute_lec, equations, scalars, variables
        >>> equations.load(f"{SAMPLE_DATA_DIR}/fun.eqs")
        >>> scalars.load(f"{SAMPLE_DATA_DIR}/fun.scl")
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")

        >>> equations["ACAF"].lec       # doctest: +NORMALIZE_WHITESPACE
        \'(ACAF/VAF[-1]) :=acaf1+acaf2*GOSF[-1]+\\nacaf4*(TIME=1995)\'
        >>> variables["ACAF", "2000Y1"]
        10.046610792200543
        >>> lec = "(acaf1 + acaf2 * GOSF[-1] + acaf4*(TIME=1995)) * VAF[-1]"
        >>> # pass a period value as string
        >>> execute_lec(lec, "2000Y1")
        10.046610792200543
        >>> # pass a period position in the sample
        >>> t = variables.sample.index("2000Y1")
        >>> t
        40
        >>> execute_lec(lec, t)
        10.046610792200543
        >>> # compute the LEC formula over the whole sample
        >>> execute_lec(lec)        # doctest: +ELLIPSIS
        [-2e+37, 4.2884154594335815, 4.532163174288473, ..., -83.34062511080091, -96.41041982848331]
        >>> variables["ACAF"]       # doctest: +ELLIPSIS
        [-2e+37, -2e+37, -2e+37, ..., -83.34062511080091, -96.41041982848331]
    '''
@overload
def execute_lec(lec) -> Any:
    '''execute_lec(lec: str, period: Union[str, int, Period] = None) -> Union[float, List[float]]

        Compute a LEC formula using the current Variables and Scalars databases.
        The formula may be evaluate at a specific period or on the whole sample 
        (no value for *period* is passed).
    
        Parameters
        ----------
        lec: str
            LEC expression to be evaluated.
        period: str or int or Period, optional
            Period at which the LEC formula is evaluated.
            If the passed value is of type int, then it represents 
            the position of the period in the current sample.
            If the passed value is None, empty string or -1, the LEC expression 
            is evaluated over the whole sample.
            Default to None.
    
        Returns
        -------
        float or list(float)

        Examples
        --------
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import execute_lec, equations, scalars, variables
        >>> equations.load(f"{SAMPLE_DATA_DIR}/fun.eqs")
        >>> scalars.load(f"{SAMPLE_DATA_DIR}/fun.scl")
        >>> variables.load(f"{SAMPLE_DATA_DIR}/fun.var")

        >>> equations["ACAF"].lec       # doctest: +NORMALIZE_WHITESPACE
        \'(ACAF/VAF[-1]) :=acaf1+acaf2*GOSF[-1]+\\nacaf4*(TIME=1995)\'
        >>> variables["ACAF", "2000Y1"]
        10.046610792200543
        >>> lec = "(acaf1 + acaf2 * GOSF[-1] + acaf4*(TIME=1995)) * VAF[-1]"
        >>> # pass a period value as string
        >>> execute_lec(lec, "2000Y1")
        10.046610792200543
        >>> # pass a period position in the sample
        >>> t = variables.sample.index("2000Y1")
        >>> t
        40
        >>> execute_lec(lec, t)
        10.046610792200543
        >>> # compute the LEC formula over the whole sample
        >>> execute_lec(lec)        # doctest: +ELLIPSIS
        [-2e+37, 4.2884154594335815, 4.532163174288473, ..., -83.34062511080091, -96.41041982848331]
        >>> variables["ACAF"]       # doctest: +ELLIPSIS
        [-2e+37, -2e+37, -2e+37, ..., -83.34062511080091, -96.41041982848331]
    '''
def execute_report(filepath: str | Path, parameters: str | list[str] = ...) -> Any:
    '''execute_report(filepath: Union[str, Path], parameters: Union[str, List[str]] = None)

        Execute an IODE report.

        Parameters
        ----------
        filepath: str, Path
            Filepath of the IODE report to execute. 
            The extension of the file must be *.rep*.
        parameters: str or list(str), optional
            Parameter(s) passed to the report.
            If multiple parameters are passed in one string, they must be separated by a whitespace.
            Default to None.

        Notes
        -----
        Equivalent to the IODE command `$ReportExec`

        Examples
        --------
        >>> from iode import execute_report, variables

        >>> # create an dump report
        >>> with open("create_var.rep", "w") as f:
        ...     f.write("$WsClearVar\\n")
        ...     f.write("$WsSample 2000Y1 2005Y1\\n")
        ...     f.write("$DataCalcVar %1% t+1 \\n")
        ...     f.write("$DataCalcVar %2% t-1 \\n")
        ...     f.write("$DataCalcVar %3% %1%/%2%\\n")
        ...     f.write("$DataCalcVar %4% grt %1% \\n")
        ...     f.write("$WsSaveVar test_var.av\\n")
        12
        24
        22
        22
        25
        26
        23

        >>> # execute report
        >>> execute_report("create_var.rep", ["A", "B", "C", "D"])

        >>> # check content of file test_var.av
        >>> with open("test_var.av", "r") as f:         # doctest: +NORMALIZE_WHITESPACE
        ...     print(f.read())
        sample 2000Y1 2005Y1
        A 1 2 3 4 5 6
        B -1 0 1 2 3 4
        C -1 na 3 2 1.66666666666667 1.5
        D na 100 50 33.3333333333333 25 20
        <BLANKLINE>
    '''
@overload
def iode_msg_path() -> str:
    """iode_msg_path() -> str
    Return the path to the iode.msg file.
    
        Returns
        -------
        str

        Examples
        --------
        >>> from pathlib import Path
        >>> from iode.iode_python import iode_msg_path
        >>> iode_msg_path = Path(iode_msg_path())
        >>> iode_msg_path.exists()
        True
        >>> iode_msg_path.name
        'iode.msg'
    """
@overload
def iode_msg_path() -> Any:
    """iode_msg_path() -> str
    Return the path to the iode.msg file.
    
        Returns
        -------
        str

        Examples
        --------
        >>> from pathlib import Path
        >>> from iode.iode_python import iode_msg_path
        >>> iode_msg_path = Path(iode_msg_path())
        >>> iode_msg_path.exists()
        True
        >>> iode_msg_path.name
        'iode.msg'
    """
def iode_number_to_str(value: float) -> str:
    """iode_number_to_str(value: float) -> str"""
@overload
def is_NA(value: float) -> bool:
    """is_NA(value: float) -> bool

        Check whether a float value represents a valid IODE number or an IODE *Not Available* :math:`NA` value.

        Parameters
        ----------
        value: float

        Returns
        -------
        bool
            True if the float value represents an IODE *Not Available* :math:`NA` value.

        Examples
        --------
        >>> from iode import NA, is_NA
        >>> is_NA(1.0)
        False
        >>> is_NA(NA)
        True
    """
@overload
def is_NA(NA) -> Any:
    """is_NA(value: float) -> bool

        Check whether a float value represents a valid IODE number or an IODE *Not Available* :math:`NA` value.

        Parameters
        ----------
        value: float

        Returns
        -------
        bool
            True if the float value represents an IODE *Not Available* :math:`NA` value.

        Examples
        --------
        >>> from iode import NA, is_NA
        >>> is_NA(1.0)
        False
        >>> is_NA(NA)
        True
    """
@overload
def load_extra_files(extra_files: str | Path | list[str | Path]) -> list[Path]:
    '''load_extra_files(extra_files: Union[str, Path, List[Union[str, Path]]]) -> List[Path]

        Load extra file(s) referenced in the generalized sample.
        Maximum 4 files can be passed as argument.
        The file [1] always refers to the current workspace. 
        Extra files are numeroted from 2 to 5.

        Parmeters
        ---------
        extra_files: str or Path or list(str) or list(Path)

        Examples
        --------
        >>> from pathlib import Path
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import load_extra_files

        >>> sample_data_dir = Path(SAMPLE_DATA_DIR)
        >>> extra_files = [sample_data_dir / "ref.av", sample_data_dir / "fun.av", 
        ...                sample_data_dir / "fun2.av", sample_data_dir / "a.var"]
        >>> extra_files = load_extra_files(extra_files)
        >>> [Path(filepath).name for filepath in extra_files]
        [\'ref.av\', \'fun.av\', \'fun2.av\', \'a.var\']
    '''
@overload
def load_extra_files(extra_files) -> Any:
    '''load_extra_files(extra_files: Union[str, Path, List[Union[str, Path]]]) -> List[Path]

        Load extra file(s) referenced in the generalized sample.
        Maximum 4 files can be passed as argument.
        The file [1] always refers to the current workspace. 
        Extra files are numeroted from 2 to 5.

        Parmeters
        ---------
        extra_files: str or Path or list(str) or list(Path)

        Examples
        --------
        >>> from pathlib import Path
        >>> from iode import SAMPLE_DATA_DIR
        >>> from iode import load_extra_files

        >>> sample_data_dir = Path(SAMPLE_DATA_DIR)
        >>> extra_files = [sample_data_dir / "ref.av", sample_data_dir / "fun.av", 
        ...                sample_data_dir / "fun2.av", sample_data_dir / "a.var"]
        >>> extra_files = load_extra_files(extra_files)
        >>> [Path(filepath).name for filepath in extra_files]
        [\'ref.av\', \'fun.av\', \'fun2.av\', \'a.var\']
    '''
def print_error_msg(*args, **kwargs):
    """print_error_msg(error_code: int) -> str
    Print an error message corresponding to an error code.
           See error codes and error messages in the iode.msg file.

        Parameters
        ----------
        error_code: int
           error code

        Returns
        -------
        str

        Examples
        --------
        >>> from iode.iode_python import print_error_msg
        >>> print_error_msg(16)
        ' Sample modified'
    """
def enable_msgs() -> Any:
    """enable_msgs()
    Reset the normal output mechanism during an IODE session."""
def suppress_msgs() -> Any:
    """suppress_msgs()
    Suppress the output during an IODE session."""
def version() -> str:
    """version() -> str
    Return the Iode version."""
def w_close() -> Any:
    """w_close()
    End an output session."""
def w_dest(filename: str | Path = ..., dest: WriteFileExt | str = ...) -> Any:
    """w_dest(filename: Union[str, Path] = u'', dest: Union[WriteFileExt, str] = WriteFileExt.DUMMY)
    Initialise a new output session"""
def w_flush() -> Any:
    """w_flush()
    Flush the output session buffer."""
def w_print(txt: str = ...) -> Any:
    """w_print(txt: str = u'')
    Send a string into the output session buffer."""
def w_print_cmd(level: int = ..., text: str = ...) -> Any:
    """w_print_cmd(level: int = 1, text: str = u'')
    Start a code block of the given level or end the block if level < 0"""
def w_print_enum(level: int = ..., text: str = ...) -> Any:
    """w_print_enum(level: int = 1, text: str = u'')
    Print a bulleted paragraph of the given level"""
def w_print_par(level: int = ..., text: str = ...) -> Any:
    """w_print_par(level: int = 1, text: str = u'')
    Print a paragraph of the given level"""
def w_print_pg_footer(arg: str = ...) -> Any:
    """w_print_pg_footer(arg: str = u'')
    Define the page footer as from the current page"""
def w_print_pg_header(arg: str = ...) -> Any:
    """w_print_pg_header(arg: str = u'')
    Define the page header as from the current page"""
def w_print_tit(level: int = ..., title: str = ...) -> Any:
    """w_print_tit(level: int = 1, title: str = u'')
    Print a title of the given level"""
